/* Adam Rogalewicz
 *
 * finite up abstraction 
 * - based on minimalization of tree automaton stoped after n steps
 * - this code is hacked code originally from file minimize.c - MONA project
 *  */


#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include "gta.h"
#include "../Mem/mem.h"
#include "../BDD/bdd_external.h"
#include "collaps2.h"
#include "initial_partition.h"

#define MAXVARS 1000

#define NO_SORT_DESCRIPTORS

/* Following variables are defined in file minimize.c, and are used for 
 * comunication between functions */
extern SsId s, lSs, rSs;
extern GTA *orig, *res;
extern unsigned **block;
extern unsigned *numBlocks;
extern bdd_ptr *matrix, *transposed;
extern enum Candidate {
  cNEVER, 
  cHASBEEN, 
  cYES
} *candidate;

extern bdd_ptr **sorted;
extern unsigned *original;
extern bdd_ptr *qm;
extern unsigned qcols, maxSize;
extern unsigned *qb;

/* Following functions are declared in file minimize.c in Mona GTA Library */
int compare(unsigned i, unsigned j);
void swap(unsigned i, unsigned j);
void quicksort(int from, int to);
void sort(bdd_ptr *m, unsigned *b, unsigned rows, unsigned cols);
unsigned fn_block(unsigned p);

/*--------------------------------------------------------------*/

Guide guide;
extern BddNode *table;
extern bdd_manager *import_bddm;


/*****************************************************
 * folowing functions are define in module modify_shape.c
 * **************************************************************/
char is_inside(int value, int *list, int length);

int get_next_state_direct(GTA *G, int state1,int state2,int ss);

int next_state(GTA *G, int state1, int state2,char *bit_vector,int ss);

int get_reject_state(GTA *G);	

int *get_final_states(GTA *G);

GTA *abstract_fin_length2_ll(GTA *g,int length)
{
  int done;
  int counter=0;
  unsigned *block1, *aux, **block_tmp;
  int block1_size,k,l,tmp_no;
  State i, j;
  unsigned p;

  /* pre-sort block1 */
  block1_size=0;
  block1=initial_partition_of_ss1(g);
  block_renum(block1,g->ss[1].size);
  for(k=0;k<g->ss[1].size;k++) 
	  if (block1[k]>block1_size) block1_size=block1[k];
  
  if (guide.numSs!=3) {
	  printf("ERROR: abstraction: work only with 3 statespaces\n");
	  exit(0);
  }

  
  /* initialize */
  orig = g;
  res = gtaMake();
  maxSize = 0;
  block = (unsigned **) mem_alloc(sizeof(unsigned *)*guide.numSs);
  numBlocks = (unsigned *) mem_alloc(sizeof(unsigned)*guide.numSs);
  candidate = (enum Candidate *) mem_alloc(sizeof(enum Candidate)*guide.numSs);

  for (s = 0; s < guide.numSs; s++) {
    /* for statespace 1 it is done in function initial_partition_of_ss1 */
    if (s!=1) block[s] = (unsigned *) mem_alloc(sizeof(unsigned)*orig->ss[s].size);
    else block[s]=block1;
    if (orig->ss[s].size > maxSize)
      maxSize = orig->ss[s].size;

    /* set up initial partitions and candidates */
    if (s > 0) {
      if (s==1) {
	 numBlocks[s]=block1_size;
         candidate[s] = cYES; /* never processed yet */
      }
      else {
        candidate[s] = cNEVER; /* never processed yet */
      	for (i = 0; i < orig->ss[s].size; i++)
		block[s][i] = 0;
	numBlocks[s] = 1;
      }
    }
    else {
      int anyAccept = 0, anyReject = 0, anyDontcare = 0;
      numBlocks[0] = 0;

      for (i = 0; i < orig->ss[0].size; i++) {
	invariant(orig->final[i] >= -1 && orig->final[i] <= 1); 
	if (orig->final[i] == -1) 
	  anyReject = 1;
	if (orig->final[i] == 1) 
	  anyAccept = 1;
	if (orig->final[i] == 0) 
	  anyDontcare = 1;
      }
      if (anyReject)
	anyReject = numBlocks[0]++;
      if (anyAccept)
	anyAccept = numBlocks[0]++;
      if (anyDontcare) 
	 anyDontcare = numBlocks[0]++;
      /* AnyX is the index of the block containing all AnyX states (if any) */
      for (i = 0; i < orig->ss[0].size; i++)
	block[0][i] = (orig->final[i] == -1) ? anyReject  
	  : (orig->final[i] == 1)? anyAccept : anyDontcare;      
      if (numBlocks[0] == 1)
	candidate[0] = cNEVER; /* only one block */
      else
	candidate[0] = cYES; /* initial s.s. is a candidate for splitting */
    }
  }

  /*....optimize case when numBlocks[0]==1...*/

  sorted = (bdd_ptr **) mem_alloc(sizeof(bdd_ptr*)*maxSize);
  matrix = (bdd_ptr *) mem_alloc(sizeof(bdd_ptr)*maxSize*maxSize);
  transposed = (bdd_ptr *) mem_alloc(sizeof(bdd_ptr)*maxSize*maxSize);
  original = (unsigned *) mem_alloc(sizeof(unsigned)*maxSize);
  qb = (unsigned *) mem_alloc(sizeof(unsigned)*maxSize);

  /* Initialize of tmp_blocks */
  block_tmp=(unsigned **) mem_alloc(sizeof(unsigned *)*guide.numSs*2);
  for(k=0;k<(guide.numSs*2);k++)
	  block_tmp[k]=(unsigned *) mem_alloc (sizeof(unsigned)*maxSize);
  /* refine partitions until fixed point reached */
  do {
    done = 1;
    tmp_no=0; /* number of temporary blocks */

    for (s = 0; s < guide.numSs; s++) 
      if (candidate[s] == cYES) {
	candidate[s] = cHASBEEN;

	lSs = guide.muLeft[s];
	rSs = guide.muRight[s];
	
	/* find behaviour wrt. current partition */
	bdd_prepare_apply1(orig->ss[s].bddm);
	if (res->ss[s].bddm)
	  bdd_kill_manager(res->ss[s].bddm);
	res->ss[s].bddm = 
	  bdd_new_manager(orig->ss[s].bddm->table_elements, 0);

	for (i = 0; i < orig->ss[lSs].size; i++)
	  for (j = 0; j < orig->ss[rSs].size; j++)
	    matrix[i*maxSize + j] = transposed[j*maxSize + i] = 
	      bdd_apply1(orig->ss[s].bddm, 
			 BDD_ROOT(orig->ss[s].bddm,BEH(orig->ss[s], i, j)),
			 res->ss[s].bddm,
			 fn_block);
	      
	/* refine left state space */
	if (lSs==1) { aux=block_tmp[tmp_no]; tmp_no++;}
	else aux=block[lSs];
	
	/* original part of refining
	 * block[lSs] was rewriten to aux */
	sort(matrix, block[lSs], 
	     orig->ss[lSs].size, orig->ss[rSs].size);
	p = 0;
	for (i = 0; i < orig->ss[lSs].size - 1; i++) {
	  invariant(compare(i, i+1) <= 0);
	  if (compare(i, i+1) != 0)
	    aux[original[i]] = p++;
	  else
	    aux[original[i]] = p;
	}
	aux[original[i]] = p++;
	invariant(p >= numBlocks[lSs]);
	if (p > numBlocks[lSs]) {
	  candidate[lSs] = cYES;
	  done = 0;
	  if (lSs!=1) numBlocks[lSs] = p;
	}
	
	/* refine right state space */
	if (rSs==1) { aux=block_tmp[tmp_no]; tmp_no++;}
	else aux=block[rSs];
	
	/* original part of refining
	 * block[rSs] was rewriten to aux */
	sort(transposed, block[rSs],
	     orig->ss[rSs].size, orig->ss[lSs].size);
	p = 0;
	for (i = 0; i < orig->ss[rSs].size - 1; i++) {
	  invariant(compare(i, i+1) <= 0);
	  if (compare(i, i+1) != 0)
	    aux[original[i]] = p++;
	  else
	    aux[original[i]] = p;
	}
	aux[original[i]] = p++;
	invariant(p >= numBlocks[rSs]);
	if (p > numBlocks[rSs]) {
	  candidate[rSs] = cYES;
	  done = 0;
	  if(rSs!=1) numBlocks[rSs] = p;
	}
      }
    /* saved blocks of SS 1 must be strore inside the block[1] */
    if (tmp_no>0) {
	    for(k=0;k<tmp_no-1;k++) 
		    numBlocks[1]=block_merge(block_tmp[k+1],block_tmp[k],orig->ss[1].size); 
	/* compy containt of block_tmp[tmp_no-1] into block[1] */
	for(i=0;i<orig->ss[1].size;i++) block[1][i]=block_tmp[tmp_no-1][i];
    }
    counter++;
  } while ((!done)&&(counter<length));
  
  /* leave memory free 
   * !!! It is necessary to do this before colaps2, becaurse inside colaps2 function is called
   * minimalize. This global variables are shared with the minimalize function. */
  aux=block[1];
  for (s = 0; s < guide.numSs; s++)
	  if (s!=1) mem_free(block[s]);
  mem_free(block);
  mem_free(transposed);
  mem_free(numBlocks);
  mem_free(candidate);
  mem_free(sorted);
  mem_free(original);
  mem_free(qb);
  mem_free(matrix);
  //gtaFree(res);
  mem_free(res); /* !!!!!!! prozkoumat */
  /* build result automaton */
  res=(GTA *) colaps2(g,aux);
  /* clean up */
   mem_free(aux);
  return res;
}

GTA *abstract_fin_length2(GTA *G,int length) {
	GTA *res, *aux;
	aux=gtaMinimize(G);
	res=abstract_fin_length2_ll(aux,length);
	gtaFree(aux);
	return res;
}


/*
int main() {
  GTA* transd;
  char **free_vars;
  int *orders,*state_list;
  unsigned numvars;
  SSSet *statespaces;
  transd = gtaImport("test.gta", &free_vars, &orders, &statespaces, TRUE);
  if(!transd) {
      printf("error: Chyba nacitani !!!!\n");
      exit(1);
  }
  state_list=sort_states(transd);
	
}*/

