/* Adam Rogalewicz
 *
 * Initial partition of state space 1
 * - for pointer verification, we propose special encoding of program configuration into
 *   tree automata. This initial partition separates diferent type of nodes ->
 *        -> it is unpossible to collaps memory node with pointer descriptor, etc ...
 *   */

#include <stdlib.h>
#include <stdio.h>

#include "gta.h"
#include "../BDD/bdd.h"
#include "../Mem/mem.h"
#include "initial_partition.h"
#include "modify_shape_gta.h"


int *search_for_next_states(bdd_ptr in_ptr, bdd_manager *bddm, int states) {
	int *out,*out1;
	int i;
	bdd_ptr next;
	if (bdd_is_leaf(bddm, in_ptr)) {
		out=(int *) mem_alloc (sizeof(int)*states);
		for(i=0;i<states;i++) out[i]=0;
		i=(bdd_leaf_value(bddm, in_ptr));
		out[i]=1;
		return out;
	}
        next=bdd_then(bddm, in_ptr);	
	out=search_for_next_states(next,bddm,states);
        next=bdd_else(bddm, in_ptr);	
	out1=search_for_next_states(next,bddm,states);
	for (i=0;i<states;i++)
		if (out1[i]) out[i]=1;
	mem_free(out1);
	return out;
}


int *search_for_states(GTA *G,int state1, int state2, int states, int no, char *bitvector) {
	int i,*out;
	bdd_manager *bddm = G->ss[1].bddm;
	bdd_ptr my_bdd_ptr = BDD_ROOT(bddm, BEH(G->ss[1], state1, state2));
	for(i=0;i<no; i++) {
		if (bdd_is_leaf(bddm, my_bdd_ptr)) {
			out=(int *)mem_alloc(sizeof(int)*1);
			out[0]=-1;
			return out;
		}
		if (bdd_ifindex(bddm, my_bdd_ptr)!=(i*2)) continue;
		if (bitvector[i]) my_bdd_ptr=bdd_then(bddm, my_bdd_ptr);
		else my_bdd_ptr=bdd_else(bddm, my_bdd_ptr);
	}
	out=search_for_next_states(my_bdd_ptr,bddm,states);
	return out;
}

/* merge 2 results from function search for states 
 * if list1[0]==-1 && list2[0]==-1 return list with -1 on the begining
 * elseif list1[0]==-1, returns copy of list2
 * elseif list2[0]==-1, returns copy of list1
 * else ret[i]=list1[i] || list2[i] */
int *result_merge(int *list1,int *list2, int size) {
	int *res,i;
	res=(int *)mem_alloc(sizeof(int)*size);
	if ((list1[0]==-1)&&(list2[0]==-1)) {
		res[0]=-1;
		return res;
	}
	if (list1[0]==-1) {
		for(i=0;i<size;i++) res[i]=list2[i];
		return res;
	}
	if (list2[0]==-1) {
		for(i=0;i<size;i++) res[i]=list1[i];
		return res;
	}
	for(i=0;i<size;i++) 
		if ((list1[i])||(list2[i])) res[i]=1;
		else res[i]=0;
	return res;		
}

int block_merge(unsigned *block1, unsigned *block2, int size) {
/* Merge two block into fist one.
 * This is used in abstraction function - two block describing state division are merged into one
 *
 * !!! it work for blocks with maximum values 2^16. For bigger values, it will not work properly
 * */
	int i,j,mark[size];
	unsigned max=0,multiplicator=2,next=0;
	/* find max num in block1 */
	for(i=0;i<size;i++) if (block1[i]>max) max=block1[i];
	/* found  multiplicator for all elements in block2 */
	while(multiplicator<max) multiplicator=multiplicator*2;
	for(i=0;i<size;i++) 
		block1[i]=block1[i]+(block2[i]*multiplicator);
	for(i=0;i<size;i++) mark[i]=1;
	for(i=0;i<size;i++)
		if (mark[i]>0) { /* skip processed elements */
			mark[i]=-1; /* element i was processed */
			for(j=i+1;j<size;j++) {
				if ((mark[j]>0)&&(block1[i]==block1[j])) {
					block1[j]=next;
					mark[j]=-1;
				}
			}
			block1[i]=next;
			next++;
		}
	return next; /* return number of classes in resulting block1 */
}


int block_renum(unsigned *block1, int size) {
	/* remap numbers in block.
	 * At the end of this function, block will contains numbers from '0' */
	int i,j,mark[size];
	unsigned next=0;
	/* find max num in block1 */
	for(i=0;i<size;i++) mark[i]=1;
	for(i=0;i<size;i++)
		if (mark[i]>0) { /* skip processed elements */
			mark[i]=-1; /* element i was processed */
			for(j=i+1;j<size;j++) {
				if ((mark[j]>0)&&(block1[i]==block1[j])) {
					block1[j]=next;
					mark[j]=-1;
				}
			}
			block1[i]=next;
			next++;
		}
	return next; /* return number of classes in resulting block1 */
}


/**************************************************************************
* The initial partitioning
**************************************************************************/

// We classify states in the first state space wrt. their semantics in the coding of pointers: mem. nodes, next pointers, ...
//
unsigned *initial_partition_of_ss1(GTA *G) {
	int *final,*tmp,*tmp1,*tmp2,*undef,*nullp,*top,*memory;
	unsigned *state_list;
	int states,reject,i,j,k,init,init2,final_no,undef_no,nullp_no,top_no,memory_no;
	char bit_vector[4]; // Used for describing the bit prefix of a symbol when selecting a certain sub-BDD.
#ifdef SORT_NEXT
	int *descr, *descr_cls, *descr_leaf_cls, *descr_last_cls;
	int descr_no,descr_cls_no, descr_leaf_cls_no, descr_last_cls_no, order;
#endif

        // The number of states in the automaton (state space 1 -- the main part of the automaton)
	states=G->ss[SS1].size;
	state_list=(unsigned *)mem_alloc(sizeof(int)*states);
        
	/* Guess everything to a next pointer -- will be changed subsequently. */
	for(i=0;i<states;i++) state_list[i]=NEXTPTR_CLS;                                   
	/* The Reject state */
	reject=get_reject_state(G);
	state_list[reject]=REJECT_CLS;
	// printf("REJECT: %i\n",reject);
        
	/* Final states */
	final=get_final_states(G);
	for(final_no=0;final[final_no]!=-1;final_no++) state_list[final[final_no]]=FINAL_CLS;
	/* printf("FINAL: ");
	   for(i=0;final[i]!=-1;i++) printf("%i ",final[i]);
	   printf("\n");  */

	/* The initial state */
	init=G->ss[SS1].initial;
	state_list[init]=INITIAL_CLS;
	// printf("INIT: %i\n",init);
	/* Initial state 2 'init2'
	 * this is state reachable by rule init x init -- '00' --> init2
	 * it there is no such state: -1 is given */
	bit_vector[0]=0; bit_vector[1]=0;
	tmp=search_for_states(G,init,init,states,2,bit_vector);
	init2=-1;
	if (tmp[0]!=-1)
		for(i=0;i<states;i++)
			if ((tmp[i])&&(i!=init)) init2=i;
	mem_free(tmp);
	if (init2!=-1) state_list[init2]=INITIAL_CLS2;

        
        /******************************************
	 * undef pointers are put into a separate class
	 * *****************************************/	
	undef_no=0;
	undef=(int *)mem_alloc(sizeof(int)*(states+1));
	for (i=0;i<states;i++) {
		tmp1=search_for_states(G,i,init,states,0,bit_vector);
		if (init2!=-1) {
			tmp2=search_for_states(G,i,init2,states,0,bit_vector);
			tmp=result_merge(tmp1,tmp2,states);
			mem_free(tmp1);
			mem_free(tmp2);
		} else tmp=tmp1;
		if (tmp[0]==-1) {
			mem_free(tmp);
			continue;
		}
		for(j=0;final[j]!=-1;j++) 
			if (tmp[final[j]]) {                           // !!! Cannot we count the same node several times?
				undef[undef_no]=i;                     // ... when it reaches different final states?
				undef_no++;
			}
		mem_free(tmp);
	}
	undef[undef_no]=-1;
	/* printf("UNDEF: ");
	   for(i=0;i<undef_no;i++) printf("%i ",undef[i]);
	   printf("\n"); */
	for(i=0;undef[i]!=-1;i++) state_list[undef[i]]=UNDEF_CLS;

        /******************************************
	 * null pointers are put in a separate class 
	 * *****************************************/	
	nullp_no=0;
	nullp=(int *)mem_alloc(sizeof(int)*(states+1));
	bit_vector[0]=0; bit_vector[1]=0;
	for (i=0;i<states;i++) {                                       // !!! Is it necessary to repeat this search?
		tmp1=search_for_states(G,i,init,states,2,bit_vector);   // No, we can compute it once storing it to a matrix...
		if (init2!=-1) {
			tmp2=search_for_states(G,i,init2,states,2,bit_vector);
			tmp=result_merge(tmp1,tmp2,states);
			mem_free(tmp1);
			mem_free(tmp2);
		} else tmp=tmp1;

		if (tmp[0]==-1) { 
			mem_free(tmp);
			continue;
		}
		for(j=0;undef[j]!=-1;j++) 
			if (tmp[undef[j]]) {                           // !!! Cannot we count the same node several times?
				nullp[nullp_no]=i;                     // ... when it reaches different undef states?
				nullp_no++;
			}
		mem_free(tmp);
	}
	nullp[nullp_no]=-1;
	/* printf("NULL: ");
	   for(i=0;i<nullp_no;i++) printf("%i ",nullp[i]);
	   printf("\n"); */
	for(i=0;nullp[i]!=-1;i++) state_list[nullp[i]]=NULL_CLS;

	/************************************************
	 * top memory nodes are put separatelly
	 * *********************************************/
	top_no=0;
	top=(int *)mem_alloc(sizeof(int)*(states+1));
	bit_vector[0]=1; bit_vector[1]=1;
	for (i=0;i<states;i++) 
		for(j=0;j<states;j++) {
			tmp=search_for_states(G,i,j,states,2,bit_vector);
			if (tmp[0]==-1) { 
				mem_free(tmp);
				continue;
			}
			for(k=0;k<states;k++) 
				if ((tmp[k])&&(k!=reject)                            // !!! Why is this so complicated?
						&&(!is_inside(k,top,top_no))         // ... esp. why inside(...final..)???
						&&(!is_inside(k,final,final_no))) {
					top[top_no]=k;
					top_no++;
				}
			mem_free(tmp);
		}
	top[top_no]=-1;
	/* printf("TOP: ");
	   for(i=0;i<top_no;i++) printf("%i ",top[i]);
	   printf("\n"); */
	for(i=0;top[i]!=-1;i++) state_list[top[i]]=TOPMN_CLS;
	/*******************************************
	 * memory nodes are put separatelly
	 * *****************************************/
	memory_no=0;
	memory=(int *)mem_alloc(sizeof(int)*(states+1));
	bit_vector[0]=1; bit_vector[1]=0;
	for (i=0;i<states;i++) 
		for(j=0;j<states;j++) {
		tmp=search_for_states(G,i,j,states,2,bit_vector);
		if (tmp[0]==-1) {
			mem_free(tmp);
			continue; 
		}
		for(k=0;k<states;k++) 
			if ((tmp[k])&&(k!=reject)
					&&(!is_inside(k,memory,memory_no))
					&&(!is_inside(k,final,final_no))) {
				memory[memory_no]=k;
				memory_no++;
			}
		mem_free(tmp);
		}
	memory[memory_no]=-1;
	/*printf("MEMORY: ");
	for(i=0;i<memory_no;i++) printf("%i ",memory[i]);
	printf("\n");*/
	for(i=0;memory[i]!=-1;i++) state_list[memory[i]]=MEMN_CLS;
	/*******************************************
	 *******************************************
	 * NEXT POINTERS
	 ******************************************
	 ******************************************/
	/* next pointers are marked by NEXTPTR_CLS in state_list 
	 * if is defined SORT_NEXT -> next pointers are sorted into a several classes
	 * according to 'order' of next pointers, and position in a tree (leaf/inner) */
#ifdef SORT_NEXT
	/* first find next pointers */
	bit_vector[0]=0; bit_vector[1]=1;
	descr_no=0;
	descr=(int *)mem_alloc(sizeof(int)*(states+1));
	for(i=0;i<states;i++)
		if (state_list[i]==NEXTPTR_CLS) {
			descr[descr_no]=i;
			descr_no++;
		}
	descr_cls_no=0;
	descr_cls=(int *)mem_alloc(sizeof(int)*(descr_no+1));
	descr_leaf_cls_no=0;
	descr_leaf_cls=(int *)mem_alloc(sizeof(int)*(descr_no+1));
	descr_last_cls_no=0;
	descr_last_cls=(int *)mem_alloc(sizeof(int)*(descr_no+1));
	/* ***********************************************
	 * First 'most right' next pointers 
	 * ***********************************************/
	/* search for inner next pointers */
	for(i=0;memory[i]!=-1;i++){
		tmp1=search_for_states(G,memory[i],init,states,2,bit_vector);
		if (init2!=-1) {
			tmp2=search_for_states(G,memory[i],init2,states,2,bit_vector);
			tmp=result_merge(tmp1,tmp2,states);
			mem_free(tmp1);
			mem_free(tmp2);
		} else tmp=tmp1;
		if (tmp[0]==-1) { 
			mem_free(tmp);
			continue;
		}
		for(k=0;k<states;k++) 
			if ((tmp[k])&&(is_inside(k,descr,descr_no))
					&&(!is_inside(k,descr_cls,descr_cls_no))) {
				descr_cls[descr_cls_no]=k;
				descr_cls_no++;
			}
		mem_free(tmp);
	}
	/* search for 'leaf' next pointers -
	 * -there can be all combinbations of init & init2 */
	tmp1=search_for_states(G,init,init,states,2,bit_vector);
	if (init2!=-1) {
		tmp2=search_for_states(G,init2,init,states,2,bit_vector);
		tmp=result_merge(tmp1,tmp2,states);
		mem_free(tmp1);
		mem_free(tmp2);
		tmp1=search_for_states(G,init,init2,states,2,bit_vector);
		tmp2=result_merge(tmp,tmp1,states);
		mem_free(tmp);
		mem_free(tmp1);
		tmp1=search_for_states(G,init2,init2,states,2,bit_vector);
		tmp=result_merge(tmp1,tmp2,states);
		mem_free(tmp1);
		mem_free(tmp2);
	} else tmp=tmp1;
	for(k=0;k<states;k++) {
		if ((tmp[k])&&(k!=reject)) {
			if (is_inside(k,descr_cls,descr_cls_no))
				state_list[k]=FIRST_IDX;
			else state_list[k]=FIRST_IDX+1;
			descr_last_cls[descr_last_cls_no]=k;
			descr_last_cls_no++;

		} else {
			if (is_inside(k,descr_cls,descr_cls_no)) {
				state_list[k]=FIRST_IDX+2;
				descr_last_cls[descr_last_cls_no]=k;
				descr_last_cls_no++;
			}
		}
	}
	mem_free(tmp);
	descr_last_cls[descr_last_cls_no]=-1;
	/*******************************************
	 * other next pointers 
	 *******************************************/
	for(order=1;descr_last_cls[0]!=-1;order++) {
		/*printf("LAST_CLS: ");
		for(i=0;i<descr_last_cls_no;i++) printf("%i ",descr_last_cls[i]);
		printf("\n");*/
		descr_cls_no=0;
		descr_leaf_cls_no=0;
		/* search for inner descriptors */
		for(i=0;memory[i]!=-1;i++)
			for(j=0;descr_last_cls[j]!=-1;j++) {
				tmp=search_for_states(G,memory[i],descr_last_cls[j],states,2,bit_vector);
				if (tmp[0]==-1) { 
					mem_free(tmp);
					continue;
				}
				for(k=0;k<states;k++) { 
					if ((tmp[k])&&(is_inside(k,descr,descr_no))
							&&(!is_inside(k,descr_cls,descr_cls_no))) {
						descr_cls[descr_cls_no]=k;
						descr_cls_no++;
					}
				}
				mem_free(tmp);
			}
		/* search for 'leaf' descriptors */
		for(j=0;descr_last_cls[j]!=-1;j++) {
			tmp1=search_for_states(G,init,descr_last_cls[j],states,2,bit_vector);
			if (init2!=-1) {
				tmp2=search_for_states(G,init2,descr_last_cls[j],states,2,bit_vector);
				tmp=result_merge(tmp1,tmp2,states);
				mem_free(tmp1);
				mem_free(tmp2);
			} else tmp=tmp1;
			if (tmp[0]==-1) { 
				mem_free(tmp);
				continue;
			}
			for(k=0;k<states;k++) 
				if ((tmp[k])&&(is_inside(k,descr,descr_no))
						&&(!is_inside(k,descr_leaf_cls,descr_leaf_cls_no))) {
					descr_leaf_cls[descr_leaf_cls_no]=k;
					descr_leaf_cls_no++;
				}
			mem_free(tmp);
		}
		/* reset descr_last_cls */
		descr_last_cls_no=0;
		for(i=0;i<states;i++){ 
			if (is_inside(i,descr_cls,descr_cls_no)) {
				if (is_inside(i,descr_leaf_cls,descr_leaf_cls_no)) 
					state_list[i]=(order*3)+FIRST_IDX;
				else state_list[i]=(order*3)+FIRST_IDX+1;
				descr_last_cls[descr_last_cls_no]=i;
				descr_last_cls_no++;
			} else {
				if (is_inside(i,descr_leaf_cls,descr_leaf_cls_no)) {
					state_list[i]=(order*3)+FIRST_IDX+2;
					descr_last_cls[descr_last_cls_no]=i;
					descr_last_cls_no++;
				}
			}
		}
		descr_last_cls[descr_last_cls_no]=-1;
	}
#endif
	/* -------------------------------------------------- */
	mem_free(final);
	mem_free(undef);
	mem_free(nullp);
	mem_free(top);
	mem_free(memory);
#ifdef SORT_NEXT
	mem_free(descr);
	mem_free(descr_cls);
	mem_free(descr_leaf_cls);
	mem_free(descr_last_cls);
#endif
	/*for(i=0;i<states;i++) printf("%i ",state_list[i]);
	printf("\n"); */
	return state_list;
}

