#include "main-rot.h"
#include "gta.h"
#include "../Mem/mem.h"
#include <stdio.h>

#include "stack/stack.h"
#include "stack/history_stack.h"
#include "stack/gtastack.h"

GTA *init_aut;
GTA *empty_aut;

#ifdef HISTORY
extern hist_stack history;
extern char *changed;
#else
extern stack wait;
#endif

#ifdef PREDICATES
/* in case of predicate abstraction, this predicates will be stored in stack called predicates */
extern gtastack predicates;
#endif

#ifdef RESTRICT
extern gtastack var_restrict;
#endif

/* init descriptors contains initial descriptors
 * They are used in case of computation restart (after abstraction refinement) */
GTA **init_descriptors;

SSSet *statespaces;
char **free_vars;
int *orders; 

void init_structures(int lines,int descr) {
	unsigned i;
	GTA *a1;
	program=(program_type *) mem_alloc(sizeof(program_type)*lines);
#ifndef HISTORY
	/* initialize stack for todo list, and put there line number 0 
	 * - we start computation by executing command on line 0 */
	stack_init(&wait,lines+1);
	stack_push(&wait,0);
#else
	/* initialize stack for history
	 * - it is used in case of backward transition */
	histstack_init(&history,100);
	changed=(char *) mem_alloc(sizeof(char)*lines);
	for(i=0;i<lines;i++) changed[i]=0;
#endif
	/* in field state we store input states for given command lines */
	state=(GTA **) mem_alloc(sizeof(GTA *)*lines);
	/* pointer sedcriptors are stored in this structure */
	descriptors=(descriptor_type *) mem_alloc(sizeof(descriptor_type)*descr);
	init_descriptors=(GTA **) mem_alloc(sizeof(GTA *)*descr);
}

int read_checkers(int lines) {
	/* checkers are tree automata, which define bad states.
	 * Tests for empty intersection between checkers and reachable states are performed
	 *
	 * There 2 type of checkers:
	 * - HARD ERRORS: interupt computation
	 * - WARNINGS: display warning on terminal
	 *
	 * Checker can be defined for input states of all program lines, 
	 * or input state of given line 
	 ***********************************
	 * The difference between standard version and version with rotations is "auto-checkers"
	 * This auto_checkers are defined for all rotation commands, and check whether the rotation is possible
	 */
	FILE *file;
	int i,j,k,res;
	char name[20];
	GTA *aut;

	checkers_no=0;
	file=fopen(checkers_file,"r");
	if (!file) {
		printf("WARNING: file of checkers is not open - ??? no checkers ???\n");
		checkers_no=0;
	} else {
		res=fscanf(file,"%i",&i);
		if ((res==EOF)||(res==0)||(i<1)) {
			printf("WARNING: file of checkers contains no checkers - ??? no checkers???\n");
			checkers_no=0;
		} else checkers_no=i;
	}

	checkers=(checker_type *) mem_alloc(sizeof(checker_type)*(i+lines));

	for(i=0;i<checkers_no;i++) {
		res=fscanf(file,"%i %i",&j,&k);
		if (res==0) { checkers_no=i; break; }
		if (res==EOF) { checkers_no=i; break; }
		checkers[i].type=j;
		checkers[i].line=k;
		sprintf(name,"checker-%i.gta",i);
#ifdef VERBOSE
		printf("importing file %s - checker of type %i, line no %i\n",name,j,k);
#endif
		checkers[i].aut=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(!checkers[i].aut) {
			printf("ERROR: File %s\n",name);
			exit(1);
		}
	}
	/* close file, if it is open */
	if (file) fclose(file);
	printf("IMPORT: standart checkers OK\n");
	/****************************
	 * Now import auto-checkers 
	 ****************************/
	for(i=0;i<lines;i++) {
		sprintf(name,"auto_checker-%i.gta",i);
		checkers[checkers_no].aut=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(checkers[checkers_no].aut) {
			printf("Loaded %s\n",name);
			checkers[checkers_no].type=1; /* HARD ERROR */ 
			checkers[checkers_no].line=i; /* apply this checker only on line i */
			checkers_no++;
		}
	}
	
	printf("IMPORT: auto-checkers OK\n");
	
	return checkers_no;
}

#ifdef RESTRICT
void read_restrict() {
	int i;
	char name[20];
	GTA *aux;
	i=0;
	gtastack_init(&var_restrict,10);
	while(1) {
		sprintf(name,"restrict_var-%i.gta",i);
		aux=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!aux) break;
		gtastack_push(&var_restrict,aux);
		i++;
	}
	printf("Read restrict predicates \n");
}

#endif

#ifdef PREDICATES
void init_and_read_predicates() {
	int i;
	char name[20];
	GTA *aux;
	i=0;
	gtastack_init(&predicates,50);
	while(1) {
		sprintf(name,"predicate-%i.gta",i);
		aux=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!aux) break;
		gtastack_push(&predicates,aux);
		i++;
	}
	printf("Read %i initial predicates \n",i);
}
#endif

/* Read initial descriptors */
void read_initial_descriptors() {
	int i;
	char name[20];
	for(i=1;i<descr;i++) {
		sprintf(name,"descr-%i.gta",i);
		descriptors[i].descr=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(!(descriptors[i].descr)) {
			printf("ERROR: File %s not exists\n",name);
			exit(1);
		}
		/* store initial descriptors for case of computation restart */
		init_descriptors[i]=gtaCopy(descriptors[i].descr);
	}
}

/* reset descriptors to initial values */
void reset_descriptors() {
	int i;
	for(i=1;i<descr;i++) {
		gtaFree(descriptors[i].descr);
		descriptors[i].descr=gtaCopy(init_descriptors[i]);
	}
}


void reset_states() {
	int i;
	for (i=0;i<lines;i++) gtaFree(state[i]);
	state[0]=gtaCopy(init_aut);
	printf("O\n");
	/* other states are set to empty on the beginning */
	for (i=1;i<lines;i++) state[i]=gtaCopy(empty_aut); 
}

int read_input() {
	FILE *file;
	int i,j,k,l,m,res;
	char name[20];
	GTA *aut1;

	file=fopen(dir_file_rot,"r");
	if (!file) {
		printf("ERROR: file is not open, aborded\n");
		exit(1);
	}

	fscanf(file,"%i %i %u %u",&lines,&descr,&NODE_WIDTH,&pointer_num);
	printf("READ lines: %i\nDescriptors: %i\nNode width: %u\nPointer num:%u\n",lines,descr,NODE_WIDTH,pointer_num);
	if ((lines<=0)||(descr<0)) {
		printf("ERROR: lines/descriptors number\n");
		exit(1);
	}
	if (NODE_WIDTH<=2) {
		printf("ERROR: node width\n");
		exit(1);
	}
	init_structures(lines,descr);
	while (1) {
		res=fscanf(file,"%i %i %i %i %i ",&i,&l,&j,&k,&m);
		if (res==0) continue;
		if (res==EOF) break;
		if (res!=5) {
			printf("WARNING - input file may contain error on the end - line skiped\n");
			continue;
		}
#ifdef VERBOSE
		printf("line: %i %i %i %i %i\n",i,l,j,k,m);
#endif
		if ((i<0)||(i>=lines)||(j>=lines)||(l<0)||(l>11)) {
			printf("ERROR: line %i %i %i %i %iwas skiped\n",i,l,j,k,m);
			exit(1);
		}
		if ((l>0)&&((j<0)||(j>=lines))) {
			printf("ERROR: next line id defined incorectly\n");
			exit(1);
		}
		if ((l==2)&&((k<0)||(k>=lines))) {
			printf("ERROR: second next line ID defined incorectly\n");
			exit(1);
		}
		if((l==3)&&((k<0)||(k>=descr))) {
			printf("ERROR: Descriptor ID defined incorectly\n");
			exit(1);
		}
		program[i].t=j;
		program[i].f=k;
		program[i].type=l;
		program[i].mark=m;
		if (program[i].type==0) {
			continue;
		}
		/* read corresponding transducer */
		if (program[i].type<4) {
			sprintf(name,"transd-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
		}
		if (program[i].type==2) {
			sprintf(name,"transd-%i-2.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
		}
		/* x.next=y is forbiden command */
		if (program[i].type==3) {
			printf("ERROR: This type of coomand is forbiden in version with rotations\n");
			exit(1);
		}
		/* x=y.next */
		if (program[i].type==4) {
			sprintf(name,"special-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			sprintf(name,"ph2-%i.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			program[i].settoken=(GTA **) mem_alloc(sizeof(GTA *)*descr);
			for(j=1;j<descr;j++) {
				sprintf(name,"settoken-%i-%i.gta",i,j);
				program[i].settoken[j]=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
				if(!(program[i].settoken[j])) {
					printf("ERROR: File %s\n",name);
					exit(1);
				}
			}
		}
		/* x.next=new */
		if (program[i].type==5) {
			sprintf(name,"transd-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			sprintf(name,"transd-used-%i.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			/* now read transducer to repare structure after generic new 
			   - we need null next pointers assigned 
			   - we need defined UP pointer
			   We use settoken[0] to store this transducer
			 */  
			sprintf(name,"reparenew-%i.gta",i);
			program[i].settoken=(GTA **) mem_alloc(sizeof(GTA *)); /* just one automata is stored in thie field */
			printf("IMPORTING: %s\n",name);
			program[i].settoken[0]=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].settoken[0])) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
		}
		/* rotate commands
		   - there is only repare transducers */
		if((program[i].type==10)||(program[i].type==11)) {
			sprintf(name,"repare-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			sprintf(name,"repare-back-%i.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			/* we use settoken[0] to store transducer for backward rotation preparation */
			program[i].settoken=(GTA **) mem_alloc(sizeof(GTA *)); /* just one automata is stored in thie field */
			sprintf(name,"prepare_back_rot-%i.gta",i);
			program[i].settoken[0]=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].settoken[0])) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}

		}
	}
	/* Read empty automaton
	 * used for nondefined descriptors
	 * initial states */
	sprintf(name,"empty.gta");
	empty_aut=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
	if(!(empty_aut)) {
		printf("ERROR: File %s\n",name);
		exit(1);
	}
	
	/* Read initial descriptors */
	read_initial_descriptors();

	/* Read initial configuration */
	sprintf(name,"init.gta");
	init_aut=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
	if(!init_aut) {
		printf("ERROR: File %s\n",name);
		exit(1);
	}
	state[0]=gtaCopy(init_aut);
	/* other states are set to empty on the beginning */
	for(i=1;i<lines;i++) state[i]=gtaCopy(empty_aut);

	read_checkers(lines);
#ifdef RESTRICT
	read_restrict();
#endif
#ifdef PREDICATES
	init_and_read_predicates();
#endif
	//gtaFree(empty_aut);
	printf("Input read succesfully\n-----------------------------------------------\n");
	return 1;
}

void save_automaton(GTA *G,char *name) {
	gtaExport(G, name, 2, free_vars,orders, statespaces, 0);
}

