#include "main.h"
#include "gta.h"
#include "../Mem/mem.h"
#include <stdio.h>
#include "input.h"

#include "stack/stack.h"


GTA *init_aut;
GTA *empty_aut;


#ifdef PREDICATES
/* in case of predicate abstraction, this predicates will be stored in stack called predicates */
extern gtastack predicates;
#endif

#ifdef RESTRICT
extern gtastack var_restrict;
#endif

/* init descriptors contains initial descriptors
 * They are used in case of computation restart (after abstraction refinement) */
GTA **init_descriptors;

SSSet *statespaces;
char **free_vars;
int *orders; 

void init_structures(int lines,int descr) {
	unsigned i;
	GTA *a1;
	program=(program_type *) mem_alloc(sizeof(program_type)*lines);
	/* in field state we store input states for given command lines */
	state=(GTA **) mem_alloc(sizeof(GTA *)*lines);
	/* pointer sedcriptors are stored in this structure */
	descriptors=(descriptor_type *) mem_alloc(sizeof(descriptor_type)*descr);
	init_descriptors=(GTA **) mem_alloc(sizeof(GTA *)*descr);
}

int read_checkers() {
	/* checkers are tree automata, which define bad states.
	 * Tests for empty intersection between checkers and reachable states are performed
	 *
	 * There 2 type of checkers:
	 * - HARD ERRORS: interupt computation
	 * - WARNINGS: display warning on terminal
	 *
	 * Checker can be defined for input states of all program lines, 
	 * or input state of given line */
	FILE *file;
	int i,j,k,res;
	char name[40];
	GTA *aut;

	checkers_no=0;
	file=fopen(checkers_file,"r");
	if (!file) {
		printf("WARNING: file of checkers is not open - ??? no checkers ???\n");
		return 0;
	}
	res=fscanf(file,"%i",&i);
	if ((res==EOF)||(res==0)||(i<1)) {
		printf("WARNING: file of checkers contains no checkers - ??? no checkers???\n");
		return 0;
	}
	checkers_no=i;
	checkers=(checker_type *) mem_alloc(sizeof(checker_type)*i);

	for(i=0;i<checkers_no;i++) {
		res=fscanf(file,"%i %i",&j,&k);
		if (res==0) { checkers_no=i; break; }
		if (res==EOF) { checkers_no=i; break; }
		checkers[i].type=j;
		checkers[i].line=k;
		snprintf(name,40,"checker-%i.gta",i);
#ifdef VERBOSE
		printf("importing file %s - checker of type %i, line no %i\n",name,j,k);
#endif
		checkers[i].aut=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(!checkers[i].aut) {
			printf("ERROR: File %s\n",name);
			exit(1);
		}
	}
	printf("IMPORT: checkers OK\n");
	return checkers_no;
}

#ifdef RESTRICT
void read_restrict() {
	int i;
	char name[40];
	GTA *aux;
	i=0;
	gtastack_init(&var_restrict,10);
	while(1) {
		snprintf(name,40,"restrict_var-%i.gta",i);
		aux=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!aux) break;
		gtastack_push(&var_restrict,aux);
		i++;
	}
	printf("Read restrict predicates \n");
}

#endif

#ifdef PREDICATES
void init_and_read_predicates() {
	int i;
	char name[40];
	GTA *aux;
	i=0;
	gtastack_init(&predicates,50);
	while(1) {
		snprintf(name,40,"predicate-%i.gta",i);
		aux=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!aux) break;
		gtastack_push(&predicates,aux);
		i++;
	}
	printf("Read %i initial predicates \n",i);
}
#endif

/* Read initial descriptors */
void read_initial_descriptors() {
	int i;
	char name[40];
	for(i=1;i<descr;i++) {
		snprintf(name,40,"descr-%i.gta",i);
		descriptors[i].descr=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(!(descriptors[i].descr)) {
			printf("WARNING: File %s not exists - using empty descriptor\n",name);
			descriptors[i].descr=gtaCopy(empty_aut);
		}
		/* store initial descriptors for case of computation restart */
		init_descriptors[i]=gtaCopy(descriptors[i].descr);
		/* there exist 'from shape' transducer for each descriptor */
		snprintf(name,40,"from-shape-%i.gta",i);
		descriptors[i].from_shape=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(!(descriptors[i].from_shape)) {
			printf("ERROR: File %s\n",name);
			exit(1);
		}
	}
}

/* reset descriptors to initial values */
void reset_descriptors() {
	int i;
	for(i=1;i<descr;i++) {
		gtaFree(descriptors[i].descr);
		descriptors[i].descr=gtaCopy(init_descriptors[i]);
	}
}


void reset_states() {
	int i;
	for (i=0;i<lines;i++) gtaFree(state[i]);
	state[0]=gtaCopy(init_aut);
	printf("O\n");
	/* other states are set to empty on the beginning */
	for (i=1;i<lines;i++) state[i]=gtaCopy(empty_aut); 
}

/************************************************************************************************/
/* Support for variance analysis */

/* reed seeding transducers for variance analysis 
 * This function is called directly from the modul variance.c */
gtastack read_seed(char read_seed_desc) {
	int i;
	char name[40];
	GTA *aux;
	gtastack seed;
	gtastack_init(&seed,20);
	for(i=0;;i++) {
		snprintf(name,40,"seed_var-%i.gta",i);
		aux=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(!aux) break;
		gtastack_push(&seed,aux);
	}
	printf("INPUT: read seeding transducers for %i variables\n",i);
	for(i=0;;i++) {
		snprintf(name,40,"seed_data_track-%i.gta",i);
		aux=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(!aux) break;
		gtastack_push(&seed,aux);
	}
	printf("INPUT: read seeding transducers for data of length %i\n",i);
	if (read_seed_desc) {
		snprintf(name,40,"seed_descriptors.gta",i);
		aux=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if(aux) {
			gtastack_push(&seed,aux);
			printf("INPUT: read seeding transducers for descriptors");
		}
	}

	return seed;
}

/* read variance predicates */
void load_variance_predicates(varpred_stack *vpstack, char vd) {
	char name[40];
	int i;
	GTA *aux;
	var_pred p1;
	/* init the stack */
	varpredstack_init(vpstack,20);
	/* first read variance predicates for all pointer variables */
	i=1;
	while(1) {
		snprintf(name,40,"variance_pred-%i-decr.gta",i);
		p1.decr=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.decr) break;
		snprintf(name,40,"variance_pred-%i-eq.gta",i);
		p1.eq=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.eq) { 
			printf("INPUT ERROR: equality predicate for variable %i does not exists... stop\n",i);
			exit(1); 
		}

		snprintf(name,40,"variance_pred-%i-inc.gta",i);
		p1.inc=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.inc)  p1.inc=gtaCopy(empty_aut);
		p1.name=(char *) mem_alloc(sizeof(char)*30);
		snprintf(p1.name,40,"Variable %i moves",i);
		varpredstack_push(vpstack, p1);
		printf("INPUT: Read variance predicate: %s\n",p1.name);
		i++;
	}	
	/* read variance predicate for values data (if exists) */
	snprintf(name,40,"variance_pred-data_decr.gta");
	p1.decr=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
	if (p1.decr) {
		snprintf(name,40,"variance_pred-data_eq.gta");
		p1.eq=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.eq) { 
			printf("INPUT ERROR: equality predicate for data does not exists... stop\n",i);
			exit(1); 
		}
		snprintf(name,40,"variance_pred-data_inc.gta");
		p1.inc=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.inc) { 
			printf("INPUT ERROR: increase predicate for data does not exists... stop\n",i);
			exit(1); 
		}

		p1.name=(char *) mem_alloc(sizeof(char)*30);
		snprintf(p1.name,40,"Data values changes");
		varpredstack_push(vpstack, p1);
		printf("INPUT: Read variance predicate: %s\n",p1.name);
	}
	/* read variance predicate for disposed nodes (if exists) */
	snprintf(name,40,"variance_pred-deleted.gta");
	p1.decr=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
	if (p1.decr) {
		/* This variance predicate is allowed only in the case of no new statement in the program.py 
		 * -> if the number of deleted nodes is not increased, then it must be equal (no new nodes) */
		p1.eq=gtaCopy(p1.decr);
		gtaNegation(p1.eq); // equality is negation of decreasing
		p1.inc=gtaCopy(empty_aut);
		p1.name=(char *) mem_alloc(sizeof(char)*40);
		snprintf(p1.name,40,"Increased number of deleted nodes");
		varpredstack_push(vpstack, p1);
		printf("INPUT: Read variance predicate: %s\n",p1.name);
	}
	/* read variance predicates for descriptor fields (if required) */
	i=0;
	while(vd) {
		snprintf(name,40,"variance_pred-descr_%i_decr.gta",i);
		p1.decr=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.decr) break;
		snprintf(name,40,"variance_pred-descr_%i_eq.gta",i);
		p1.eq=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.eq) { 
			printf("INPUT ERROR: equality predicate for descriptor %i does not exists... stop\n",i);
			exit(1); 
		}

		snprintf(name,40,"variance_pred-descr_%i_inc.gta",i);
		p1.inc=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
		if (!p1.inc) { 
			printf("INPUT ERROR: increase predicate for descriptor %i does not exists... stop\n",i);
			exit(1); 
		}
		p1.name=(char *) mem_alloc(sizeof(char)*40);
		snprintf(p1.name,40,"Values in descriptor %i changes",i);
		varpredstack_push(vpstack, p1);
		printf("INPUT: Read variance predicate: %s\n",p1.name);
		i++;
	}	

}

/************************************************************************************************/

int read_input() {
	FILE *file;
	int i,j,k,l,m,res;
	char name[40];
	GTA *aut1;

	file=fopen(dir_file,"r");
	if (!file) {
		printf("ERROR: file is not open, aborded\n");
		exit(1);
	}

	fscanf(file,"%i %i %u",&lines,&descr,&NODE_WIDTH);
	printf("READ lines: %u\nDescriptors: %i\nNode width: %u\n",lines,descr,NODE_WIDTH);
	if ((lines<=0)||(descr<0)) {
		printf("ERROR: lines/descriptors number\n");
		exit(1);
	}
	if (NODE_WIDTH<=2) {
		printf("ERROR: node width\n");
		exit(1);
	}
	init_structures(lines,descr);
	while (1) {
		res=fscanf(file,"%i %i %i %i %i ",&i,&l,&j,&k,&m);
		if (res==0) continue;
		if (res==EOF) break;
		if (res!=5) {
			printf("WARNING - input file may contain error on the end - line skiped\n");
			continue;
		}
#ifdef VERBOSE
		printf("line: %i %i %i %i\n",i,l,j,k,m);
#endif
		if ((i<0)||(i>=lines)||(j>=lines)||(l<0)||(l>6)) {
			printf("WARNING1: line %i %i %i %i %iwas skiped\n",i,l,j,k,m);
			continue;
		}
		if ((l>0)&&((j<0)||(j>=lines))) {
			printf("WARNING: next line id defined incorectly\n");
			continue;
		}
		if ((l==2)&&((k<0)||(k>=lines))) {
			printf("WARNING: second next line ID defined incorectly\n");
			continue;
		}
		if((l==3)&&((k<0)||(k>=descr))) {
			printf("WARNING: Descriptor ID defined incorectly\n");
			continue;
		}
		program[i].t=j;
		program[i].f=k;
		program[i].type=l;
		program[i].mark=m;
		if (program[i].type==0) {
			continue;
		}
		/* read corresponding transducer */
		if (program[i].type<4) {
			snprintf(name,40,"transd-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
		}
		if (program[i].type==2) {
			snprintf(name,40,"transd-%i-2.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
		}
#ifdef UPDATE_DESCRIPTOR
		if (program[i].type==3) {
			/* This transducers are used in case of descriptor update */
			snprintf(name,40,"to_shape-%i.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
#ifdef COVER_CHECK
			program[i].settoken=(GTA **)mem_alloc(sizeof(GTA *)*2);
			snprintf(name,40,"prepare-cover-check-%i.gta",i);
			program[i].settoken[0]=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].settoken[0])) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			snprintf(name,40,"settoken-%i-%i.gta",i,program[i].f);
			program[i].settoken[1]=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].settoken[1])) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
#endif
		}
#endif
		if (program[i].type==4) {
			snprintf(name,40,"special-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			snprintf(name,40,"ph2-%i.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			program[i].settoken=(GTA **) mem_alloc(sizeof(GTA *)*descr);
			for(j=1;j<descr;j++) {
				snprintf(name,40,"settoken-%i-%i.gta",i,j);
				program[i].settoken[j]=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
				if(!(program[i].settoken[j])) {
					printf("ERROR: File %s\n",name);
					exit(1);
				}
			}
		}
		if (program[i].type==5) {
			snprintf(name,40,"transd-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			snprintf(name,40,"transd-used-%i.gta",i);
			program[i].transd2=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd2)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
		}
		/* type 6 = free(x) */
		if (program[i].type==6) {
			snprintf(name,40,"transd-%i.gta",i);
			program[i].transd=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
			if(!(program[i].transd)) {
				printf("ERROR: File %s\n",name);
				exit(1);
			}
			/* the field program[i].f is used to store the number of prepare_free automata */
			if (program[i].f>0) {
				program[i].settoken=(GTA **)mem_alloc(sizeof(GTA *)*program[i].f);			
				for(j=0;j<program[i].f;j++) {
					snprintf(name,40,"prepare_free-%i-%i.gta",i,j);
					program[i].settoken[j]=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
					if(!(program[i].settoken[j])) {
						printf("ERROR: File %s\n",name);
						exit(1);
					}
				}
			}
		}
	}
	/* Read empty automaton
	 * used for nondefined descriptors
	 * initial states */
	sprintf(name,"empty.gta");
	empty_aut=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
	if(!(empty_aut)) {
		printf("ERROR: File %s\n",name);
		exit(1);
	}
	
	/* Read initial descriptors */
	read_initial_descriptors();

	/* Read initial configuration */
	sprintf(name,"init.gta");
	init_aut=gtaImport(name, &free_vars, &orders, &statespaces, TRUE);
	if(!init_aut) {
		printf("ERROR: File %s\n",name);
		exit(1);
	}
	state[0]=gtaCopy(init_aut);
	/* other states are set to empty on the beginning */
	for(i=1;i<lines;i++) state[i]=gtaCopy(empty_aut);

	read_checkers();
#ifdef RESTRICT
	read_restrict();
#endif
#ifdef PREDICATES
	init_and_read_predicates();
#endif
	printf("Input read succesfully\n-----------------------------------------------\n");
	return 1;
}

void save_automaton(GTA *G,char *name) {
	gtaExport(G, name, 2, free_vars,orders, statespaces, 0);
}

