/* This file is originally part of mona project - gta minimalization 
 * Modified for my own purposes */

/*
 * MONA
 * Copyright (C) 1997-2002 BRICS.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include "../Mem/mem.h"
#include "gta.h"
#include "collaps2.h"

/* Following variables are defined in file minimize.c, and are used for 
 * comunication between functions */
extern SsId s, lSs, rSs;
extern GTA *orig, *res;
extern unsigned **block;
extern unsigned *numBlocks;
extern bdd_ptr *matrix, *transposed;
extern enum Candidate {
  cNEVER, 
  cHASBEEN, 
  cYES
} *candidate;

extern bdd_ptr **sorted;
extern unsigned *original;
extern bdd_ptr *qm;
extern unsigned qcols, maxSize;
extern unsigned *qb;

/* Following functions are declared in file minimize.c in Mona GTA Library */
int compare(unsigned i, unsigned j);
void swap(unsigned i, unsigned j);
void quicksort(int from, int to);
void sort(bdd_ptr *m, unsigned *b, unsigned rows, unsigned cols);
unsigned fn_block(unsigned p);

/* ---------------------------------------------------------------------
 * My new functions 
 * ---------------------------------------------------------------------*/

/* Test if a is included in b */
int is_included(GTA *a,GTA *b) {
/* If the intersection of a and co(b) is empty, then a is included in b */
	int ret;
        GTA *pp, *pp1;
        /* I dont want to make changes in original automaton */
        pp=gtaCopy(b);
        gtaNegation(pp);
        pp1=gtaProduct(pp,a,gtaAND);
        Tree *pp2;
        pp2=gtaMakeExample(pp1,1);
        if (pp2) ret=0; else ret=1;
	gtaFree(pp);
	gtaFree(pp1);
	mem_free(pp2);
	return ret;
}

/* Test if there is a intersection between a and b */
int intersection(GTA *a,GTA *b) {
        GTA* pp;
	int ret;
        pp=gtaProduct(a,b,gtaAND);
        Tree *pp2;
        pp2=gtaMakeExample(pp,1);
        if (pp2) ret=1; else ret=0;
	gtaFree(pp);
	mem_free(pp2);
	return ret;
}

/* Test for emptyness of automaton */
int empty(GTA *a) {
        Tree *pp2;
	int ret;
        pp2=gtaMakeExample(a,1);
        if (pp2) ret=0; else ret=1;
	mem_free(pp2);
	return ret;
}

/* ---------------------------------------------------------------------
 * Abstraction on finite length trees - minimalisation algorithm is stoped 
 * after n steps.
 * --------------------------------------------------------------------- */

GTA *abstract_fin_length_ll(GTA *g,int length)
{
  int done;
  int counter=0;
  State i, j;
  unsigned p, *aux;

  /* initialize */
  orig = g;
  res = gtaMake();
  maxSize = 0;
  block = (unsigned **) mem_alloc(sizeof(unsigned *)*guide.numSs);
  numBlocks = (unsigned *) mem_alloc(sizeof(unsigned)*guide.numSs);
  candidate = (enum Candidate *) mem_alloc(sizeof(enum Candidate)*guide.numSs);

  for (s = 0; s < guide.numSs; s++) {
    block[s] = (unsigned *) mem_alloc(sizeof(unsigned)*orig->ss[s].size);
    if (orig->ss[s].size > maxSize)
      maxSize = orig->ss[s].size;

    /* set up initial partitions and candidates */
    if (s > 0) {
      candidate[s] = cNEVER; /* never processed yet */
      for (i = 0; i < orig->ss[s].size; i++)
	block[s][i] = 0;
      numBlocks[s] = 1;
    }
    else {
      int anyAccept = 0, anyReject = 0, anyDontcare = 0;
      numBlocks[0] = 0;

      for (i = 0; i < orig->ss[0].size; i++) {
	invariant(orig->final[i] >= -1 && orig->final[i] <= 1); 
	if (orig->final[i] == -1) 
	  anyReject = 1;
	if (orig->final[i] == 1) 
	  anyAccept = 1;
	if (orig->final[i] == 0) 
	  anyDontcare = 1;
      }
      if (anyReject)
	anyReject = numBlocks[0]++;
      if (anyAccept)
	anyAccept = numBlocks[0]++;
      if (anyDontcare) 
	 anyDontcare = numBlocks[0]++;
      /* AnyX is the index of the block containing all AnyX states (if any) */
      for (i = 0; i < orig->ss[0].size; i++)
	block[0][i] = (orig->final[i] == -1) ? anyReject  
	  : (orig->final[i] == 1)? anyAccept : anyDontcare;      
      if (numBlocks[0] == 1)
	candidate[0] = cNEVER; /* only one block */
      else
	candidate[0] = cYES; /* initial s.s. is a candidate for splitting */
    }
  }

  /*....optimize case when numBlocks[0]==1...*/

  sorted = (bdd_ptr **) mem_alloc(sizeof(bdd_ptr*)*maxSize);
  matrix = (bdd_ptr *) mem_alloc(sizeof(bdd_ptr)*maxSize*maxSize);
  transposed = (bdd_ptr *) mem_alloc(sizeof(bdd_ptr)*maxSize*maxSize);
  original = (unsigned *) mem_alloc(sizeof(unsigned)*maxSize);
  qb = (unsigned *) mem_alloc(sizeof(unsigned)*maxSize);

  /* refine partitions until fixed point reached */
  do {
    done = 1;

    for (s = 0; s < guide.numSs; s++) 
      if (candidate[s] == cYES) {
	candidate[s] = cHASBEEN;

	lSs = guide.muLeft[s];
	rSs = guide.muRight[s];
	
	/* find behaviour wrt. current partition */
	bdd_prepare_apply1(orig->ss[s].bddm);
	if (res->ss[s].bddm)
	  bdd_kill_manager(res->ss[s].bddm);
	res->ss[s].bddm = 
	  bdd_new_manager(orig->ss[s].bddm->table_elements, 0);

	for (i = 0; i < orig->ss[lSs].size; i++)
	  for (j = 0; j < orig->ss[rSs].size; j++)
	    matrix[i*maxSize + j] = transposed[j*maxSize + i] = 
	      bdd_apply1(orig->ss[s].bddm, 
			 BDD_ROOT(orig->ss[s].bddm,BEH(orig->ss[s], i, j)),
			 res->ss[s].bddm,
			 fn_block);
	      
	/* refine left state space */
	sort(matrix, block[lSs], 
	     orig->ss[lSs].size, orig->ss[rSs].size);
	p = 0;
	for (i = 0; i < orig->ss[lSs].size - 1; i++) {
	  invariant(compare(i, i+1) <= 0);
	  if (compare(i, i+1) != 0)
	    block[lSs][original[i]] = p++;
	  else
	    block[lSs][original[i]] = p;
	}
	block[lSs][original[i]] = p++;
	invariant(p >= numBlocks[lSs]);
	if (p > numBlocks[lSs]) {
	  candidate[lSs] = cYES;
	  done = 0;
	  numBlocks[lSs] = p;
	}
	
	/* refine right state space */
	sort(transposed, block[rSs],
	     orig->ss[rSs].size, orig->ss[lSs].size);
	p = 0;
	for (i = 0; i < orig->ss[rSs].size - 1; i++) {
	  invariant(compare(i, i+1) <= 0);
	  if (compare(i, i+1) != 0)
	    block[rSs][original[i]] = p++;
	  else
	    block[rSs][original[i]] = p;
	}
	block[rSs][original[i]] = p++;
	invariant(p >= numBlocks[rSs]);
	if (p > numBlocks[rSs]) {
	  candidate[rSs] = cYES;
	  done = 0;
	  numBlocks[rSs] = p;
	}
      }
    counter++;
  } while ((!done)&&(counter<length));
  
  /* leave memory free 
   * !!! It is necessary to do this before colaps2, becaurse inside colaps2 function is called
   * minimalize. This global variables are shared with the minimalize function. */
  aux=block[1];
  for (s = 0; s < guide.numSs; s++)
	  if (s!=1) mem_free(block[s]);
  mem_free(block);
  mem_free(transposed);
  mem_free(numBlocks);
  mem_free(candidate);
  mem_free(sorted);
  mem_free(original);
  mem_free(qb);
  mem_free(matrix);
  //gtaFree(res);
  mem_free(res); /* !!!!!!! prozkoumat */
  
  /* build result automaton */
  res=(GTA *) colaps2(g,aux);
  /* clean up */
   mem_free(aux);
  return res;
}

GTA *abstract_fin_length(GTA *G,int length) {
	GTA *res, *aux;
	aux=gtaMinimize(G);
	res=abstract_fin_length_ll(aux,length);
	gtaFree(aux);
	return res;
}
	

