/*
 * MONA
 * Copyright (C) 1997-2002 BRICS.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include "gta.h"
#include "../Mem/mem.h"
#include "../BDD/bdd_external.h"

#define MAXVARS 1000

extern Guide guide;
extern BddNode *table;
extern bdd_manager *import_bddm;

Table **get_tables(GTA *G) {
	unsigned i, j, l, r;
	Table **result = (Table **) mem_alloc(sizeof(Table*)*guide.numSs);

	/* remove all marks G->bddm */
	for (i = 0; i < guide.numSs; i++)
		bdd_prepare_apply1(G->ss[i].bddm); 

	for (i = 0; i < guide.numSs; i++)
		result[i] = tableInit();


	/* build table of tuples (idx,lo,hi) */
	// Tady se tvori tabulka, zacne se vzdy u nejake kombinace vstupnich stavu a vsechny 
	// mozne pruchody BDD strukturou se pridaji do tabulky
	for (i = 0; i < guide.numSs; i++) 
		for (l = 0; l < G->ss[guide.muLeft[i]].size; l++) 
			for (r = 0; r < G->ss[guide.muRight[i]].size; r++) {
			  	export(G->ss[i].bddm, 
					BDD_ROOT(G->ss[i].bddm, BEH(G->ss[i], l, r)), 
        				result[i]);
			}

	for (i = 0; i < guide.numSs; i++) {
		for (j = 0; j < result[i]->noelems; j++) {
			if (result[i]->elms[j].idx != -1) {
				result[i]->elms[j].lo = bdd_mark(G->ss[i].bddm, 
      					 result[i]->elms[j].lo)-1;
				result[i]->elms[j].hi = bdd_mark(G->ss[i].bddm, 
      					 result[i]->elms[j].hi)-1;
			}
  		}
	}
	return result;
}

GTA *put_tables(Table **tables, GTA *G) {
	unsigned i, j, l, r, p1, p2;
	BddNode **tables2;
	tables2 = (BddNode **) mem_alloc(sizeof(BddNode *)*guide.numSs);

	/* nyni je potreb a prejit od typu Table -> **BddNode */
	for (i = 0; i < guide.numSs; i++) {
		tables2[i] = (BddNode *) mem_alloc(sizeof(BddNode)* tables[i]->noelems);    
		for (j=0; j < tables[i]->noelems; j++) {
			tables2[i][j].p=-1;
			tables2[i][j].idx=tables[i]->elms[j].idx;
			tables2[i][j].lo=tables[i]->elms[j].lo;
			tables2[i][j].hi=tables[i]->elms[j].hi;
		}
	}
	/* ************************************************************
	* kopie + zmena table 
	***************************************************************/
	GTA *res = gtaMake();
	res->final = (int *) mem_alloc(sizeof(int)*G->ss[0].size);

	for (i = 0; i < G->ss[0].size; i++)
		res->final[i] = G->final[i];
	for (i = 0; i < guide.numSs; i++) {
		StateSpace *ss = &res->ss[i];

		ss->initial = G->ss[i].initial;
		ss->size = G->ss[i].size;
		ss->ls = G->ss[i].ls;
		ss->rs = G->ss[i].rs;
		ss->behaviour = (bdd_handle *) mem_alloc(sizeof(bdd_handle)*ss->ls*ss->rs);
		
		for (l = 0; l < ss->ls; l++) 
			for (r = 0; r < ss->rs; r++) {
				BEH(res->ss[i], l, r) = bdd_mark(G->ss[i].bddm, 
			 		BDD_ROOT(G->ss[i].bddm, 
					BEH(G->ss[i], l, r)))-1;
			}
	}

	for (i = 0; i < guide.numSs; i++) {
		table = tables2[i];
		res->ss[i].bddm = 
			bdd_new_manager(8 * res->ss[i].size, res->ss[i].size*4); /* ?? */
		import_bddm = res->ss[i].bddm;
		for (l = 0; l < res->ss[guide.muLeft[i]].size; l++) 
			for (r = 0; r < res->ss[guide.muRight[i]].size; r++) {
				PUSH_SEQUENTIAL_LIST(import_bddm->roots, unsigned, 
			     		make_node(BEH(res->ss[i], l, r)));
				BEH(res->ss[i], l, r) = BDD_LAST_HANDLE(import_bddm);
			}
	}
	/* leave memory free */
	for (i = 0; i < guide.numSs; i++)
		mem_free(tables2[i]);
	mem_free(tables2);
	return res;
}

GTA *put_tables_remap_behaviour(Table **tables, GTA *G,unsigned **beh, unsigned beh_ss) {
	/* equal to function put_tables. In addition remap behaviour of state space ss
	 * by values given in beh */
	unsigned i, j, l, r, p1, p2;
	BddNode **tables2;
	tables2 = (BddNode **) mem_alloc(sizeof(BddNode *)*guide.numSs);

	/* nyni je potreb a prejit od typu Table -> **BddNode */
	for (i = 0; i < guide.numSs; i++) {
		tables2[i] = (BddNode *) mem_alloc(sizeof(BddNode)* tables[i]->noelems);    
		for (j=0; j < tables[i]->noelems; j++) {
			tables2[i][j].p=-1;
			tables2[i][j].idx=tables[i]->elms[j].idx;
			tables2[i][j].lo=tables[i]->elms[j].lo;
			tables2[i][j].hi=tables[i]->elms[j].hi;
		}
	}
	/* ************************************************************
	* kopie + zmena table 
	***************************************************************/
	GTA *res = gtaMake();
	res->final = (int *) mem_alloc(sizeof(int)*G->ss[0].size);

	for (i = 0; i < G->ss[0].size; i++)
		res->final[i] = G->final[i];

	for (i = 0; i < guide.numSs; i++) {
		StateSpace *ss = &res->ss[i];

		ss->initial = G->ss[i].initial;
		ss->size = G->ss[i].size;
		ss->ls = G->ss[i].ls;
		ss->rs = G->ss[i].rs;
		ss->behaviour = (bdd_handle *) mem_alloc(sizeof(bdd_handle)*ss->ls*ss->rs);
		
		if (i==beh_ss) { /* remap behaviour in this state space */
			for (l = 0; l < ss->ls; l++) 
				for (r = 0; r < ss->rs; r++) 
					BEH(res->ss[i], l, r) = beh[l][r]; 
		} else { /* normal */
			for (l = 0; l < ss->ls; l++) 
				for (r = 0; r < ss->rs; r++) {
					BEH(res->ss[i], l, r) = bdd_mark(G->ss[i].bddm, 
			 			BDD_ROOT(G->ss[i].bddm, 
						BEH(G->ss[i], l, r)))-1;
				}
		}
	}

	for (i = 0; i < guide.numSs; i++) {
		table = tables2[i];
		res->ss[i].bddm = 
			bdd_new_manager(8 * res->ss[i].size, res->ss[i].size*4); /* ?? */
		import_bddm = res->ss[i].bddm;
		for (l = 0; l < res->ss[guide.muLeft[i]].size; l++) 
			for (r = 0; r < res->ss[guide.muRight[i]].size; r++) {
				PUSH_SEQUENTIAL_LIST(import_bddm->roots, unsigned, 
			     		make_node(BEH(res->ss[i], l, r)));
				BEH(res->ss[i], l, r) = BDD_LAST_HANDLE(import_bddm);
			}
	}
	/* leave memory free */
	for (i = 0; i < guide.numSs; i++)
		mem_free(tables2[i]);
	mem_free(tables2);
	return res;
}

unsigned **get_behaviour(GTA *G, unsigned ss) {
	/* Get behaviur table for statespace ss */
	/* ??? First it is necessary to run function get_tables, to get proper result. I don't know why */
	unsigned l,r;
	unsigned **beh=(unsigned **) mem_alloc(sizeof(unsigned *)*G->ss[guide.muLeft[ss]].size);
	for (l = 0; l < G->ss[guide.muLeft[ss]].size; l++) {
		beh[l]=(unsigned *) mem_alloc(sizeof(unsigned)*G->ss[guide.muRight[ss]].size);
		for (r = 0; r < G->ss[guide.muRight[ss]].size; r++) {
			beh[l][r]=bdd_mark(G->ss[ss].bddm, 
				 BDD_ROOT(G->ss[ss].bddm, BEH(G->ss[ss], l, r)))-1;
		}
	}
	return beh;
}
