#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include "gta.h"
#include "main.h"

#include "../Mem/mem.h"

#include "stack.h"
#include "history_stack.h"

#include "modify_shape_gta.h"
#include "transducer.h"
#include "gta_artmc_lib.h"
#include "shape_to_tr.h"
#include "input.h"
#include "finite_up_abstr.h"
#include "neighbour-abstract.h"
#include "predicate_abstraction/predicate_abstr.h"
#include "collaps2.h"
#include "initial_partition.h"
#include "forward_run.h"
#include "variance_analysis/variance.h"

#include "refine.h"

#include "reverse_run/reverse_run.h"


// !!!!!! TOMAS
#include "finite_height_abstr.h"
// !!!!!! TOMAS

#define MAXVARS 1000

Guide guide;

int lines,descr,checkers_no;
unsigned NODE_WIDTH;

#ifdef STATISTICS
unsigned max_automata_size=0;
unsigned refinement_no=0;
unsigned max_var_size=0;
#endif

/* this variable is used for computation restart */
char restart;

/* abstr_mode:
 * 0 - abstraction is performed only at lines given by directions.txt
 * 1 - abstraction is performed on all lines */
int abstr_mode=0;

program_type *program;
GTA **state;
checker_type *checkers;
descriptor_type *descriptors;
#ifdef RESTRICT
gtastack var_restrict;
#endif

#ifdef PREDICATES
/* in case of predicate abstraction, this predicates will be stored in stack called predicates */
gtastack predicates;
#endif


int abstr_length;


void destroy_structure(){
	mem_free(program);
	mem_free(state);
	mem_free(descriptors);
	mem_free(checkers);
	/* not fully implemented - operating system destroy alocated structures after 
	 * program termination !!!*/
}


#ifdef RESTRICT
GTA *apply_restrict_automata(GTA *G) {
	GTA *aut1,*res;
	int i;
	res=gtaCopy(G);
	for(i=0;i<=var_restrict.no;i++) {
		aut1=gtaProduct(res,var_restrict.val[i],gtaAND);
		gtaFree(res);
		res=aut1;
	}
	return res;
}
#endif


/* If you want to change used abstraction, do it here,
 * This function is called from all places, whre it is necessary */
GTA *run_abstraction(GTA *G) {
	int i;
	GTA *res,*aut1,*orig;
	char done=0;
	int *classes,*classes2;
	
	/* !!! it is necessary to run minimalization before abstraction !!!
	 * otherwice seq fault :-( */
	orig=gtaMinimize(G);
#undef OUTSIDE_COLLAPS
#undef INNER_COLLAPS
	
#ifdef FIN_HEIGHT
#define OUTSIDE_COLLAPS
	classes=abstract_fin_height_follow_pointer_str(orig,abstr_length);
#endif
#ifdef FIN_HEIGHT_GENERIC
#define INNER_COLLAPS
	res=abstract_fin_height(orig,abstr_length);
#endif
#ifdef PREDICATES
#define OUTSIDE_COLLAPS
	classes=run_predicate_abtraction(orig, predicates,1); 
    #ifdef COMBINE_PRED_FH
	printf("Combined abstraction\n");    
	classes2=abstract_fin_height_follow_pointer_str(orig,abstr_length);
	block_merge(classes,classes2,orig->ss[1].size);
	mem_free(classes2);
    #endif
#endif
#ifdef NEIGHBOUR
#define OUTSIDE_COLLAPS
	classes=abstract_neighbour(orig);
#endif
#ifdef FIN_UP
#define INNER_COLLAPS
	res=abstract_fin_length2(orig,abstr_length);
#endif

#ifdef INNER_COLLAPS
#ifdef OUTSIDE_COLLAPS
	printf("ERROR: check #define in main.h, set more types of abstraction ...\n");
	exit(1);
#endif
#endif

#ifdef RESTRICT_PREDICATE
   #ifndef INNER_COLLAPS
	classes2=run_predicate_abtraction(orig, var_restrict,0); 
	block_merge(classes,classes2,orig->ss[1].size);
	mem_free(classes2);
   #else
	printf("WARNING: Sorry, unpossible to do restriction by predicates\n");
   #endif
#endif

#ifdef OUTSIDE_COLLAPS
	/* perform collapsing */
	//save_automaton(orig,"tmp.gta");
   #ifdef VERBOSE
	for(i=0;i<orig->ss[1].size;i++) printf("%i:%i ",i,classes[i]);
	printf("\n");
   #endif
	res=collaps2(orig,classes);
	mem_free(classes);
#endif
	
#ifdef RESTRICT
   #ifndef RESTRICT_PREDICATE
	aut1=apply_restrict_automata(res);
	gtaFree(res);
	res=aut1;
   #endif
#endif
	return res;
}

#ifdef FIRST_UNION
/* First union, than abstract */
GTA *abstract_and_union(GTA *new,GTA *orig,int new_line){
	/* new_line - line number of union-ed result 
	 * If second bit in program[new_line].mark is set - abstraction is performed 
	 * */
	GTA *a1,*a2;
	a1=gtaProduct(new,orig,gtaOR);
	if ((abstr_mode)|| (!(program[new_line].mark & 2))) {
		a2=run_abstraction(a1);
		gtaFree(a1);
	}
	else a2=a1;
	a1=gtaMinimize(a2);
	gtaFree(a2);
	return a1;
}
#else
/* First abstract, than union */
GTA *abstract_and_union(GTA *new,GTA *orig,int new_line){
	/* new_line - line number of union-ed result 
	 * If second bit in program[new_line].mark is set - abstraction is performed 
	 * */
	GTA *a1,*a2;
	if ((abstr_mode)|| (!(program[new_line].mark & 2))) {
		a1=run_abstraction(new);
	}
	else
		a1=gtaCopy(new);
	a2=gtaProduct(a1,orig,gtaOR);
	gtaFree(a1);
	a1=gtaMinimize(a2);
	gtaFree(a2);
	return a1;
}
#endif

GTA *apply_x_assign_ynext(int actual,GTA *in_aut) {
	int i;
	GTA *a1,*a2,*res;
	/* phase 1 - separatelly for all descriptors */
	for (i=1;i<descr;i++){
		a1=apply_transducer(in_aut,program[actual].settoken[i]);
		a2=apply_transducer(a1,descriptors[i].descr);
		gtaFree(a1);
		if (i==1) res=a2;
		else { 
			a1=gtaProduct(a2,res,gtaOR);
			gtaFree(a2);
			gtaFree(res);
			res=a1;
		}
	}
	/* phase2 */
	a1=(GTA *) apply_transducer(res,program[actual].transd2);
	gtaFree(res);
	/* special cases */
	a2=(GTA *) apply_transducer(in_aut, program[actual].transd);
	res=gtaProduct(a1,a2,gtaOR);
	gtaFree(a1);
	gtaFree(a2);
	return res;

}

GTA *apply_xnext_ass_new(int actual,GTA *in_aut, hist_stack *history, char reverse) {
/* reverse: 0 - reverser run disabled, 1 - reverse run enabled */
	GTA *aux,*bad;
	int res;
	hist_value hv;
	aux=(GTA *) apply_transducer(in_aut,program[actual].transd2);
	if (!empty(aux)) {
		printf("****************************************************\n");
		printf("ERROR: x.next=new: x.next used - NOT IMPLEMENTED\n");
		printf("****************************************************\n");
		gtaFree(aux);
#ifdef REMEMBER_CONFIGS
		if (reverse==0)
#endif
		{ 	/* unpossible to start reverse run */
			printf("Backward computation not enabled -> stoped\n");
			exit(1);
		}
		printf("Starting backward computation\n");
		aux=get_source_of_transducer(in_aut,program[actual].transd2);
		/* pop last stat from history stack - not interesting for us */
		hv=histstack_pop(history);
		gtaFree(hv.state);
		bad=reverse_run(aux,actual,history);
		if (!empty(bad)) {
			/* error in initial configuration reached */
			//gtaPrintVerbose(bad);
			save_automaton(bad,"error-configuration.gta");
			printf("--------------\nintersection with initial configuration, -> Safety check failed\nError state save as 'error-configuration.gta'\n-----------\n");
			exit(1);
		}
		refine_abstraction();
		restart=1;
		gtaFree(aux);
	} else gtaFree(aux);
	aux=(GTA *) apply_transducer(in_aut,program[actual].transd);
	return aux;
}

GTA *apply_free_x(int actual,GTA *in_aut) {
	GTA *aux,*aux2;
	int i;
	aux=gtaCopy(in_aut);
	for (i=0;i<program[actual].f;i++) {
		aux2=apply_transducer(aux,program[actual].settoken[i]);
		gtaFree(aux);
		aux=aux2;
	}
	aux2=apply_transducer(aux,program[actual].transd);
	gtaFree(aux);
	return aux2;
}

void update_descriptor(GTA *aut,int actual,int descr_no){
	GTA *a1,*a2,*a3;
	//char name[30];
#ifdef COVER_CHECK
	/*  Check necessarity of descriptor update */
	if (!empty(descriptors[descr_no].descr)) {
		printf("Preparing cover check ...\n");
		a3=apply_transducer(aut,program[actual].settoken[1]); /* Set token for cover check */
		a1=apply_transducer(a3, program[actual].settoken[0]); /* Prepare cover check */
		gtaFree(a3);
		/* apply descriptor, and then apply descriptor in reverse mode */
		a2=apply_transducer(a1,descriptors[descr_no].descr);
		a3=apply_transducer_reverse(a2,descriptors[descr_no].descr);
		gtaFree(a2);

		if (is_included(a1,a3)) {
			gtaFree(a1);
			gtaFree(a3);
			printf("Cover: Skiping descriptor update\n");
			//int oo;
			//scanf("%i",&oo);
			return;
		}
		gtaFree(a1);
		gtaFree(a3);
	} else printf("Cover: original descriptor EMPTY\n");
#endif
	/* start descriptor update */
	printf("Updating descriptor ...\n");
	a1= apply_transducer_noremap(aut,program[actual].transd2); /* create shape */
	printf("Modifying shape ...\n");
	a2= modify_shape(a1); /* modify shape */
	gtaFree(a1);
	printf("Creating descriptor ...\n");
	a1= apply_from_shape(a2,descriptors[descr_no].from_shape); /* create descriptor from shape */
	gtaFree(a2);
	a2=gtaProduct(a1,descriptors[descr_no].descr,gtaOR);
	gtaFree(descriptors[descr_no].descr);
	descriptors[descr_no].descr=a2;
}

int check_for_bad(int line,GTA *st_line, hist_stack *history, char reverse) {
	int i;
	GTA *aux,*bad;
	#ifdef VERBOSE
	printf("Checking for bad configurations on input states for line %i:\n",line);
	#endif
	for(i=0;i<checkers_no;i++) {
		if ((checkers[i].line==line)||(checkers[i].line==-1)) {
			if (intersection(checkers[i].aut,st_line)) {
				if (checkers[i].type) {
					printf("ERROR: Violation of checker %i\n",i);
#ifdef REMEMBER_CONFIGS
					if (reverse==0)
#endif
					{	/* unpossible to start reverse run */
						printf("Backward computation not enabled -> stoped\n");
						exit(1);
					}
					/* start reverse run */
					printf("Starting backward computation ...\n");
					aux=gtaProduct(checkers[i].aut,st_line,gtaAND);
					bad=reverse_run(aux,line,history);
					gtaFree(aux);
					if (!empty(bad)) {
						/* error in initial configuration reached */
						//gtaPrintVerbose(bad);
						save_automaton(bad,"error-configuration.gta");
						printf("--------------\nintersection with initial configuration => Safety check failed\nBad configuration saved as 'error-configuration.gta'\n-----------\n");
						exit(1);
					}
					refine_abstraction(bad);
					restart=1;
					return 1;
				}
				printf("WARNING: Violation of checker %i\n",i);
			}
		}
	}
	#ifdef VERBOSE
	printf("DONE\n------------------------------------\n");
	#endif
	return 0;
}

void find_next(int *next,char *next_order,char *changed, hist_stack *history, char remove_confs) {
	hist_value last;
	if (histstack_empty(*history)) {
		*next=-1;
		return;
	}
	last=histstack_pop(history);
	//printf("POP: line: %i, order: %i, type: %i, changed: %i",last.line,last.order,program[last.line].type,changed[last.line]);
#ifdef REMEMBER_CONFIGS
	/* we need not automaton reached after transduction */
	if (remove_confs) gtaFree(last.state);
#endif
	/* this line is not branching line - skip it*/
	if (program[last.line].type!=2){
		//printf("... skiped\n");
		find_next(next, next_order, changed, history,remove_confs);
		return;
	}
	/* last line is processed */
	if ((last.order>0)||(!changed[last.line])) {
		//printf("... skiped\n");
		find_next(next, next_order, changed, history, remove_confs);
		return;
	}
	*next=last.line;
	*next_order=1;
	//printf("... processing\n");
}

void forward_run(int start_line) {
	int actual;
	char order;
	char *changed;
	hist_value act_conf;
	GTA *apl_res,*aut1;
	hist_stack history;
	histstack_init(&history,100);		
	int i;
//	int debug;
	actual=start_line;
	order=0;
	restart=0;
	changed=(char *) mem_alloc(sizeof(char)*lines);
	for(i=0;i<lines;i++) changed[i]=0;
	changed[actual]=1;
	/* if actual is less than 0 -> end of computation */
	while (actual>=0) {
		printf("--------------------------------\nLine num: %i, order %i\n",actual,order);
#ifdef STATISTICS
		printf("Act. size: %i, Max. size: %i\n",state[actual]->ss[1].size,max_automata_size);
		if (max_automata_size<state[actual]->ss[1].size)
			max_automata_size=state[actual]->ss[1].size;
#endif
		/* push actual configuration into a stack */
		act_conf.line=actual;
		act_conf.order=order;
#ifdef REMEMBER_CONFIGS
		act_conf.state=gtaCopy(state[actual]);
#endif

		histstack_push(&history,act_conf);
		/*save_automaton(act_conf.state,"pokusXXX.gta");
		scanf("%i",&debug);*/

		if (program[actual].type==0) {
			/* This line is equal to exit() in C
			 * no transition is performed */
			#ifdef VERBOSE
			printf("Command exit() - no transduction: line skiped\n");
			#endif
			find_next(&actual, &order, changed, &history,1);
			continue;
		}
		/* apply transducer / series of transducers 
		 * dependentely on actual, and order */
		if ((program[actual].type<=3)&&(order==0))
			apl_res=apply_transducer(state[actual],program[actual].transd);
		if ((program[actual].type<=3)&&(order==1))
			apl_res=apply_transducer(state[actual],program[actual].transd2);
		if (program[actual].type==4) 
			/* x=y.next */
			apl_res=apply_x_assign_ynext(actual,state[actual]);
		if (program[actual].type==5)  
			/* x.next=new */
			apl_res=apply_xnext_ass_new(actual,state[actual],&history,1);
		if (program[actual].type==6)  
			/* x.next=new */
			apl_res=apply_free_x(actual,state[actual]);
		/* if variable restart is set -> stop main loop */
		if (restart) break;
#ifdef VERBOSE
		printf("Transducer applied\n");
#endif
		/* transducer(s) aplication finished */
		if (order==0) {
			if (!is_included(apl_res,state[program[actual].t])) {
				aut1=abstract_and_union(apl_res,state[program[actual].t],program[actual].t);
				gtaFree(state[program[actual].t]);
				state[program[actual].t]=aut1;
				if (check_for_bad(program[actual].t,state[program[actual].t],&history,1)) {
					/* in case of bad state, stop and restart computation */
					restart=1;
					gtaFree(apl_res);
					break; /* stop main loop */
				}
#ifdef UPDATE_DESCRIPTOR
				/* x.next=y -- descriptor creation 
				 * if last bit in program[actual].mark is set to 1
				 * no descriptor update is requested 
				 * */
				if ((program[actual].type==3)&&(!(program[actual].mark & 1)))
					update_descriptor(apl_res,actual,program[actual].f);
#endif
				gtaFree(apl_res);
				actual=program[actual].t;
				changed[actual]=1;
				continue;
			}
			/* nothing new added - go back with history stack */
			gtaFree(apl_res);
			find_next(&actual, &order, changed, &history,1);
			continue;
		}
		/* this point is reachable only with order>0 */
		if (program[actual].type!=2) {
			/* this point is normaly unreachable !!! -> exit */
			printf("ERROR: reached forbiden combination of actual and order\n");
			exit(1);
		}
		changed[actual]=0;
		if (!is_included(apl_res,state[program[actual].f])) {
			aut1=abstract_and_union(apl_res,state[program[actual].f],program[actual].f);
			gtaFree(state[program[actual].f]);
			gtaFree(apl_res);
			state[program[actual].f]=aut1;
			if (check_for_bad(program[actual].f,state[program[actual].f],&history,1)) {
				/* in case of bad state, stop and restart computation */
				restart=1;
				break; /* stop main loop */
			}
			actual=program[actual].f;
			order=0;
			changed[actual]=1;
			continue;
		}
		gtaFree(apl_res);
		/* nothing new added - go back with history stack */
		find_next(&actual, &order, changed, &history,1);
	}
	if (restart) {
#ifdef STATISTICS
		refinement_no++;
#endif
		printf("*********************\nrestarting computation ...\n");
		reset_states();
		reset_descriptors();
		histstack_reset(&history);
		forward_run(start_line); // mode must be 0
	}
	/* destroy the history stack */
	histstack_reset(&history);
	histstack_destroy(&history);
}

int main(int argc,char **argv) {
	int i;
	char name[40], save, run_variance, vabstr, vdesc;
#ifndef HISTORY
	printf("Compile with #define HISTORY in main.h\n");
	exit(1);
#endif
	/* Fisrt argument is allways abstraction height */
	if (argc<2) {
		printf("To few arguments\nartmc 1 [-ae] [-variance] [-vabstr] [-save]\n");
		exit(0);
	}
	sscanf(argv[1],"%i",&abstr_length);
	/* check, wheather to abstract everywhere */
	abstr_mode=0;
	for (i=1;i<argc;i++) {
		if (strcmp(argv[i],"-ae")==0) abstr_mode=1;
	}
	/* check, whether to run variance analysis after safety check */
	run_variance=0;
	vdesc=0; /* use descriptor nodes inside variance analysis */

	for (i=1;i<argc;i++) {
		if (strcmp(argv[i],"-variance")==0) run_variance=1;
		if (strcmp(argv[i],"-vd")==0) vdesc=1;
	}
	/* check, whether to use abstracton during creation of variants.
	 * This is needed in case of loop between two breakpoints */
	vabstr=0;
	for (i=1;i<argc;i++) {
		if (strcmp(argv[i],"-vabstr")==0) vabstr=1;
	}

	/* check, whether to save invariants, variants, predicates, ... */
	save=0;
	for (i=1;i<argc;i++) {
		if (strcmp(argv[i],"-save")==0) save=1;
	}
	/* read input files */
	read_input();

	/* run the safety check */
	forward_run(0); 

	printf("\n*****************\nSafety check finished\nRESULT: OK\n*****************\n");
	if (run_variance) {
		create_variants(vabstr,save,vdesc);
		printf("\n*****************\nFinished\nSafety check: OK\nRelational automaton saved in file rel-automaton.txt\n*****************\n");
	}

#ifdef STATISTICS
	printf("Maximal automaton size (in safety check phase): %i\n", max_automata_size);
	printf("Number of refinements (in safety check phase): %i\n",refinement_no);
	if (run_variance)
		printf("Maximal automaton size (in variance analysis): %i\n", max_var_size);
#endif
	/* save configurations, descriptors and predicates if argument -save is given*/
	
	if (save) {
		printf("Saving configurations ...\n");
		for(i=0;i<lines;i++) {
			snprintf(name,40,"output-%i.gta",i);
			save_automaton(state[i],name);
		}
		printf("Saving descriptors ...\n");
		for(i=1;i<descr;i++) {
			snprintf(name,40,"output-descriptor-%i.gta",i);
			save_automaton(descriptors[i].descr,name);
		}
#ifdef PREDICATES
		printf("Saving predicates ...\n");
		for(i=0;i<= predicates.no;i++) {
			snprintf(name,40,"save-predicate-%i.gta",i);
			save_automaton(predicates.val[i],name);
		}
#endif
	}
	destroy_structure();
	return 0;
}
