/* Adam Rogalewicz
 * - this file contains function modify_shape, it is used for automatic creation of 
 *   pointer descriptors
 * - this function cut out 'empty' (#1=0 #3=0) nodes under descriptor, and over descriptor 
 *   and replace it by loop on empty symbols (#1=0 #3=0) */

#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include "gta.h"
#include "../Mem/mem.h"
#include "../BDD/bdd_external.h"
#include "gta_changes.h"
#include "gta_artmc_lib.h"

#define MAXVARS 1000

Guide guide;
extern BddNode *table;
extern bdd_manager *import_bddm;

int get_next_state_direct(GTA *G, int state1,int state2,int ss) {
	/* If the next state is given directly without reading of a symbol in node, 
	 * it will be returned by this function
	 * otherwise it will be returnde -1 */
	int res;
    bdd_manager *bddm = G->ss[ss].bddm;
    bdd_ptr my_bdd_ptr = BDD_ROOT(bddm, BEH(G->ss[ss], state1, state2));
    if (bdd_is_leaf(bddm, my_bdd_ptr)) return bdd_leaf_value(bddm, my_bdd_ptr); 
    else return (-1);
}

int next_state(GTA *G, int state1, int state2,char *bit_vector,int ss) {
    bdd_manager *bddm = G->ss[ss].bddm;
    bdd_ptr my_bdd_ptr = BDD_ROOT(bddm, BEH(G->ss[ss], state1, state2));
    while (!bdd_is_leaf(bddm, my_bdd_ptr)) {
      my_bdd_ptr =  (bit_vector[bdd_ifindex(bddm, my_bdd_ptr)])?
        bdd_then(bddm, my_bdd_ptr):
        bdd_else(bddm, my_bdd_ptr);
    }
    return(bdd_leaf_value(bddm, my_bdd_ptr));
  }

char is_inside(int value, int *list, int length) {
	/* check, if the value is inside a 'list' of size 'length' */
	int i;
	for (i=0;i<length;i++) {
		if (value==list[i]) return 1;
	}
	return 0;
}

int get_reject_state(GTA *G) {
	int i,j,states;
	int reject=-1;
	/* no of states in statespace 1 */
	states=G->ss[1].size;
	for(i=0;i<states;i++) {
		reject=i;
		for(j=0;j<states;j++) {
			if (get_next_state_direct(G,i,j,1)!=i) {reject=-1; break; }
			if (get_next_state_direct(G,j,i,1)!=i) {reject=-1; break; }
		}
		if (reject!=-1) break;
	}
	return reject;
}
		
	

int *get_final_states(GTA *G) {
	int *res,aux,i;
	unsigned states,fin_no;
  	char bit_vector[4] = {0,0,0,0};
	fin_no=0;
	/* no of states in statespace 1 */
	states=G->ss[1].size;
	res=(int *) mem_alloc(sizeof(int)*(states+1));
	/* no of states in statespace 0 */
	if ((G->ss[0].ls)>(G->ss[0].rs)) states=G->ss[0].ls;
	else states=G->ss[0].rs;
	for (i=0;i<states;i++) {
		aux=next_state(G, i,0,bit_vector,0);
		if (G->final[aux]==1) {
			res[fin_no]=i;
			fin_no++;
		}
	}
	res[fin_no]=-1; /* mark end of array */
	return res;
}

int *eq_to_init(GTA *G) {
/* Search for a set of states in shape automaton equal to the initial one.
 * List will be terminated by -1 */
	int *initial,reject;
	unsigned init_no,added,states,i,j,aux;
	char bit_vector[4] = {0,0,0,0};
	/* number of states in state space[1] of G */
	states=G->ss[1].size;
	initial=(int *) mem_alloc(sizeof(int)*(states+1));
	initial[0]=G->ss[1].initial;
	reject=get_reject_state(G);
	init_no=1;
	added=1;
	while (added) { /* Check, if new states were founded */
		added=0;
		for(i=0;i<init_no;i++) {
			for(j=0;j<init_no;j++) {
				/* count, what is next state in G for states initial[i], initial[j], symbol '00' in read node */
				aux=next_state(G, initial[i],initial[j],bit_vector,1); 
				if ((!is_inside(aux,initial,init_no)) && (aux!=reject)) {
					/* founded new state */
					initial[init_no]=aux;
					added=1;
					init_no++;
				}
			}
		}

	}
	initial[init_no]=-1; /* put terminal mark */
	return initial;
} /* end of eq_to_init */
	
int *eq_to_final(GTA *H) {
/* Search for a set of states in shape automaton equal to the initial one 
 * List will be terminated by -1 */
	int *initial,*final;
	unsigned init_no,final_no,i,j,added,aux,states;
	char bit_vector[4] = {0,0,0,0};
	/* number of states in state space[1] of H */
	states=H->ss[1].size;
	/* get set of states equal to initial states */
	initial=eq_to_init(H);
	for (init_no=0;initial[init_no]!=-1;init_no++) ; /* count a number of initial states */
	final=get_final_states(H); /* allocation is performed in this call */
	/* final contains now all final states; last element of field final contains symbol -1 */
	for (final_no=0;final[final_no]!=-1;final_no++);
	/* number of states in state space[1] of H */
	if ((H->ss[1].ls)>(H->ss[1].rs)) states=H->ss[1].ls;
	else states=H->ss[1].rs;
	added=1;
	while (added) { /* Check, if new states were founded */
		added=0;
		for(i=0;i<states;i++) {
			for(j=0;j<init_no;j++) {
				aux=next_state(H, i,initial[j],bit_vector,1); 
				if ((!(is_inside(i,final,final_no))) && (is_inside(aux,final,final_no))) {
					/* founded new state */
					final[final_no]=i;
					added=1;
					final_no++;
				}
				aux=next_state(H, initial[j], i, bit_vector,1); 
				if ((!(is_inside(i,final,final_no))) && (is_inside(aux,final,final_no))) {
					/* founded new state */
					final[final_no]=i;
					added=1;
					final_no++;
				}
			}
		}

	}
	final[final_no]=-1;
	return final;
} /* end of eq_to_final */

GTA *modify_accept(GTA *G, int *states, int state_no) {
	/* This function collaps all states over a descriptor to the accepting one, and add a loop on this state */
	Table **tables;
	int *final,reject,*initial,init_no,reject_bdd_node;
	BddNode *e = (BddNode *) mem_alloc(sizeof(BddNode));
	unsigned **beh;
	unsigned i;
	GTA *H,*out;
	tables= (Table **) get_tables(G);
	beh=(unsigned **) get_behaviour(G,1); 
	final=get_final_states(G);
	initial=eq_to_init(G);
	for (init_no=0;initial[init_no]!=-1;init_no++); /* count a number of initial states */
	if (init_no!=1) printf("WARNING: modify_accept: No. of states equal to init is %i; Expected: 1\n",init_no);
	reject=get_reject_state(G);
	/* when occur a state from field 'states' in bdd table, I change this state by final[0] - accepting state */
	for (i = 0; i < tables[1]->noelems; i++) {
		if ((tables[1]->elms[i].idx==-1)&&
			(is_inside(tables[1]->elms[i].lo,states,state_no))) 
				tables[1]->elms[i].lo=final[0];
		if ((tables[1]->elms[i].idx==-1)&&(tables[1]->elms[i].lo==reject)) reject_bdd_node=i;

	}
	/* Add a loop on top */
	e->idx = 1;
	e->lo = tables[1]->noelems+1;
	e->hi = reject_bdd_node; 
	tableInsert(tables[1],e);
	e->idx = 3;
	e->lo = tables[1]->noelems+1;
	e->hi = reject_bdd_node; 
	tableInsert(tables[1],e);
	e->idx = -1;
	e->lo = final[0];
	e->hi = 0; 
	tableInsert(tables[1],e);
	for (i=0;i<init_no;i++) {
		beh[final[0]][initial[i]]=tables[1]->noelems-3; /* !!!!! pouzit pro jistotu vice initial */
		beh[initial[i]][final[0]]=tables[1]->noelems-3;
	}
	H=(GTA *) put_tables_remap_behaviour(tables, G, beh, 1); 
	out=gtaMinimize(H);
	/* leave memory free */
        for (i = 0; i < guide.numSs; i++)
                tableFree(tables[i]);
        mem_free(tables);
	gtaFree(H);
	mem_free(e);
	mem_free(final);
	return out;
} /* End of modify_accept */

GTA *modify_initial(GTA *G, int *states, int state_no) {
	int *next10,*next01,*next11,reject,*aux_f;
	unsigned i,j,aux,aux2, next01_no,next10_no,next11_no,max_states,add_tracks,starttrack,state1,state3,*map;
  	char bit_vector[4]; 
	Table **tables;
	BddNode *e = (BddNode *) mem_alloc(sizeof(BddNode));
	unsigned **beh; 
	GTA *H,*I;
	/* number of states in state space[1] of G */
	if ((G->ss[1].ls)>(G->ss[1].rs)) max_states=G->ss[1].ls;
	else max_states=G->ss[1].rs;
	/* reject state */
	reject=get_reject_state(G);
	/****************************************************************/
	/* search for set of states reachable by '01' from set of initial states */
	next01_no=0;
	next01=(int *) mem_alloc(sizeof(int)*(max_states));
	bit_vector[0]=0; bit_vector[1]=0; bit_vector[2]=0; bit_vector[3]=1; /* tract 1:'0' track 3:'1'; don't care about rest */
	for(i=0;i<state_no;i++) {
		for(j=0;j<state_no;j++) {
			aux=next_state(G, states[i],states[j],bit_vector,1); 
			if ((!is_inside(aux,next01,next01_no))&&(aux!=reject)) {
				next01[next01_no]=aux;
				next01_no++;
			}
		}
	}
	/****************************************************************/
	/* search for set of states reachable by '10' from set of initial states */
	next10_no=0;
	next10=(int *) mem_alloc(sizeof(int)*(max_states));
	bit_vector[0]=0; bit_vector[1]=1; bit_vector[2]=0; bit_vector[3]=0; /* tract 1:'1' track 3:'0'; don't care about rest */
	for(i=0;i<state_no;i++) {
		for(j=0;j<state_no;j++) {
			aux=next_state(G, states[i],states[j],bit_vector,1); 
			if ((!is_inside(aux,next10,next10_no))&&(aux!=reject)) {
				next10[next10_no]=aux;
				next10_no++;
			}
		}
	}
	/****************************************************************/
	/* search for set of states reachable by '11' from set of initial states */
	next11_no=0;
	next11=(int *) mem_alloc(sizeof(int)*(max_states));
	bit_vector[0]=0; bit_vector[1]=1; bit_vector[2]=0; bit_vector[3]=1; /* tract 1:'1' track 3:'1'; don't care about rest */
	for(i=0;i<state_no;i++) {
		for(j=0;j<state_no;j++) {
			aux=next_state(G, states[i],states[j],bit_vector,1); 
			if ((!is_inside(aux,next11,next11_no))&&(aux!=reject)) {
				next11[next11_no]=aux;
				next11_no++;
			}
		}
	}
	/****************************************************************/
	/* Now, we are going to collaps initial states, and add an initial loop */
	
	/* Remaping of tracks in automaton G */
	 state1=max_states;
	 state3=max_states+1;
	map=(unsigned *) mem_alloc(sizeof(unsigned)*(max_states+2));
	map[1]=state1;
	map[3]=state3;
	starttrack=max_states+2;
  	gtaReplaceIndices(G, map);

	tables= (Table **) get_tables(G);
	beh=(unsigned **) get_behaviour(G,1); 

	 /* first create a new set of bdd nodes for 'original' initial state - state[0] */
	aux=tables[1]->noelems; /* remember this node for future changing of 'hi' entry */
	e->idx = state1;
	e->lo = tables[1]->noelems+1;
	e->hi = -2; /* will be changed soon */
	tableInsert(tables[1],e);
	aux2=tables[1]->noelems;  /* remember this node for future changing of 'hi' entry */
	e->idx = state3;
	e->lo = tables[1]->noelems+1;
	e->hi = -2; /* will be changed soon */
	tableInsert(tables[1],e);
	/* 00  - stay in initial state*/
	e->idx = -1;
	e->lo = states[0];
	e->hi = 0; 
	tableInsert(tables[1],e);
	/* 01 */
	tables[1]->elms[aux2].hi=tables[1]->noelems; /* return to the remembered value */
	if (next01_no==0) { /* insert reject state - no next states for '01' */
		e->idx = -1;
		e->lo = reject;
		e->hi = 0; 
		tableInsert(tables[1],e);
	}
	for (i=0;i<next01_no;i++){
		if (i+1<next01_no) { /* add an nondeterministich choose - tract 4+i will be removed by projection */
			e->idx = starttrack+i;
			e->lo = tables[1]->noelems+1;
			e->hi = tables[1]->noelems+2; 
			tableInsert(tables[1],e);
		}
		e->idx = -1;
		e->lo = next01[i];
		e->hi = 0; 
		tableInsert(tables[1],e);
	}
	/* 1x */
	tables[1]->elms[aux].hi=tables[1]->noelems; /* return to the remembered value */
	aux2=tables[1]->noelems;
	e->idx = state3;
	e->lo = tables[1]->noelems+1;
	e->hi = -2; 
	tableInsert(tables[1],e);
	/* 10 */
	if (next10_no==0) {  /* insert reject state - no next states for '10' */
		e->idx = -1;
		e->lo = reject;
		e->hi = 0; 
		tableInsert(tables[1],e);
	}
	for (i=0;i<next10_no;i++){
		if (i+1<next10_no) { /* add an nondeterministich choose - tract 4+i will be removed by projection */
			e->idx = starttrack+i;
			e->lo = tables[1]->noelems+1;
			e->hi = tables[1]->noelems+2; 
			tableInsert(tables[1],e);
		}
		e->idx = -1;
		e->lo = next10[i];
		e->hi = 0; 
		tableInsert(tables[1],e);
	}
	/* 11 */
	tables[1]->elms[aux2].hi=tables[1]->noelems;/* return to the remembered value */
	if (next11_no==0) { /* insert reject state - no next states for '11' */
		e->idx = -1;
		e->lo = reject;
		e->hi = 0; 
		tableInsert(tables[1],e);
	}
	for (i=0;i<next11_no;i++){
		if (i+1<next11_no) { /* add an nondeterministich choose - tract 4+i will be removed by projection */
			e->idx = starttrack+i;
			e->lo = tables[1]->noelems+1;
			e->hi = tables[1]->noelems+2; 
			tableInsert(tables[1],e);
		}
		e->idx = -1;
		e->lo = next11[i];
		e->hi = 0; 
		tableInsert(tables[1],e);
	}
	/* *******************************************************************
	 * Changes of behaviour table\
	 * 
	 * First, change behaviour for init x init 
	 * ********************************************************************/
	beh[states[0]][states[0]]=aux;

	/* now, it is necessary to go through whole behaviour table, and find cases, where is combination of 
	 * one state from set 'states', and one state out of this set. If this combination allow accepting run, 
	 * it will be processed.
	 * 
	 * Cases, where more than one state from set 'states' is applicable together with an 'noninitial' state is allowed. 
	 * Nondeterministic choose of next bdd node is applied using dummy tracks, which are removed later by projection.
	 *
	 * It is necessary to do 2 cycles, one for lines, and second for colombs
	 */
	add_tracks=0;
	aux_f=(int *) mem_alloc(sizeof(int)*(max_states));
	aux=0;
	for(i=0;i<max_states;i++) { /* max_states - total no. of states in original automaton */
		if (i==states[0]) continue; /* ignore columb describing behaviour of initial state */
		aux=0;
		for(j=0;j<max_states ;j++) {
			if ((is_inside(j,states,state_no)) && 
				(!((tables[1]->elms[beh[i][j]].idx==-1)&&(tables[1]->elms[beh[i][j]].lo==reject)))) {
				if (!is_inside(beh[i][j],aux_f,aux)) {
					aux_f[aux]=beh[i][j];
					aux++;
				}
			}
		}
		if (aux>0) { /* behaviour will be changed only if there is at least 1 founded combination */
			if (aux==1) aux2=aux_f[0]; /* just one start bdd node => no necessarity of nondeterministich choose */
			else aux2=tables[1]->noelems; /* start nondeterministich choose HERE */
			for(j=0;j<aux-1;j++){
				if (j<aux-2) {
					e->idx = j;
					e->lo = aux_f[j];
					e->hi = tables[1]->noelems+1; 
					tableInsert(tables[1],e);
				} else { /* j==aux-1 */
					e->idx = j;
					e->lo = aux_f[j];
					e->hi = aux_f[j+1];
					tableInsert(tables[1],e);
				}
			}
			if (aux-1>add_tracks) add_tracks=aux-1;
			beh[i][states[0]]=aux2;
		}
	}
	
	aux=0;

	for(i=0;i<max_states;i++) { /* max_states - total no. of states in original automaton */
		if (i==states[0]) continue; /* ignore line describing behaviour of initial state */
		aux=0;
		for(j=0;j<max_states ;j++) {
			if ((is_inside(j,states,state_no)) && 
				(!((tables[1]->elms[beh[j][i]].idx==-1)&&(tables[1]->elms[beh[j][i]].lo==reject)))) {
				if (!is_inside(beh[j][i],aux_f,aux)) {
					aux_f[aux]=beh[j][i];
					aux++;
				}
			}
		}
		if (aux>0) { /* behaviour will be changed only if there is at least 1 founded combination */
			if (aux==1) aux2=aux_f[0]; /* just one start bdd node => no necessarity of nondeterministich choose */
			else aux2=tables[1]->noelems; /* start nondeterministich choose HERE */
			for(j=0;j<aux-1;j++){
				if (j<aux-2) {
					e->idx = j;
					e->lo = aux_f[j];
					e->hi = tables[1]->noelems+1; 
					tableInsert(tables[1],e);
				} else { /* j==aux-1 */
					e->idx = j;
					e->lo = aux_f[j];
					e->hi = aux_f[j+1];
					tableInsert(tables[1],e);
				}
			}
			if (aux-1>add_tracks) add_tracks=aux-1;
			beh[states[0]][i]=aux2;
		}
	}
	
	/*****************************************************************
	 * Store changed automaton into mona data structures 
	 * ***************************************************************/
	H=(GTA *) put_tables_remap_behaviour(tables, G, beh, 1); 

	/* projection on tracks 0,1,2,... 
	 * number of projection is given by add_tracks */
	for (i=0;i<add_tracks;i++) {
		I=gtaQuotientAndProject(H, i, 0);
		gtaFree(H);
		H=I;
	}
	
	/* projection on tracks starttrack+0, starttrack+1, starttrack+2, ... 
	 * number of projection is given by maximum of states_no, next01_no, next10_no, next11_no */
	add_tracks=0;
	if (add_tracks<state_no) add_tracks=state_no;
	if (add_tracks<next01_no) add_tracks=next01_no;
	if (add_tracks<next10_no) add_tracks=next10_no;
	if (add_tracks<next11_no) add_tracks=next11_no;
	for (i=0;i<add_tracks;i++) {
		I=gtaQuotientAndProject(H, starttrack+i, 0);
		gtaFree(H);
		H=I;
	}
	/* now we have in automaton just 2 tracks -> max_states & max_states+1 
	 * we remap them back to max_states -> 1; max_states+1 -> 3 */
	map[state1]=1;
	map[state3]=3;
  	gtaReplaceIndices(H,map);
	
	/* last few thinks to be done in this function */
	I=gtaMinimize(H);
	/* leave memory free */
        for (i = 0; i < guide.numSs; i++)
                tableFree(tables[i]);
        mem_free(tables);
	mem_free(aux_f);
	mem_free(next01);
	mem_free(next10);
	mem_free(next11);
	gtaFree(H);
	return I;
} /* end of modify_initial */
		

GTA *modify_shape(GTA *G) {
  /* This function perform shape automaton modifycation.
   * collaps all states under a pointer descriptor (into a initial one)
   * collaps all states over the descriptor (into a final one)
   * Add loops no initial and final state */  
  GTA *H,*I;
  unsigned init_no,final_no;
  int  *initial, *final;
  /* If input automaton is empty => nothing to do, just return copy of this automaton */
  if (empty(G)) {
	  H=gtaCopy(G);
	  return H;
  }
  /* --------------------------------------------------------------------------
   * states equal to the initial one
   * --------------------------------------------------------------------------*/
  /* Collaps all states in field initial + create loop on original initial state */
  initial=eq_to_init(G);
  for (init_no=0;initial[init_no]!=-1;init_no++); /* count a number of initial states */
  H=modify_initial(G,initial,init_no);
  /* --------------------------------------------------------------------------------
   * And now symbols on top
   * -------------------------------------------------------------------------------- */
  final=eq_to_final(H);
  for (final_no=0;final[final_no]!=-1;final_no++) ;
  I=modify_accept(H, final, final_no);  
  /* leave memory free */
  gtaFree(H);
  mem_free(initial);
  mem_free(final);
  return I;
}

