/* Adam Rogalewicz
 *
 * Predicate abstraction based on MONA GTA library 
 * */


#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include "gta.h"
#include "../Mem/mem.h"
#include "gtastack.h"
#include "../initial_partition.h"
#include "collaps2.h"

#define MAXVARS 1000

Guide guide;

/* ---------------------------------------------------------------------------- */

typedef struct {
	unsigned *orig,*pred;
	int no;
	int limit;
} todo_type;

/* in this stack will be touples, state(from original) x state (from predicate)
 * which have to be processed
 */
todo_type todo;

/* field pred_block:
 * lines are states in predicate automaton
 * rows are states in original automaton
 * pred_block[i][j] indicates, that state j satisfied predicate i (in predicate automaton)
 * */
unsigned **pred_block;

/* ---------------------------------------------------------------------------- */

int todo_pop(todo_type *s, int *orig, int *pred) {
	if (s->no<0) return -1;
	s->no--;
	*orig=s->orig[s->no+1];
	*pred=s->pred[s->no+1];
	return 1;
}

int todo_push(todo_type *s,int orig,int pred) {
	if (s->no>=s->limit-1) return 0;
	s->no++;
	s->orig[s->no]=orig;
	s->pred[s->no]=pred;
	return 1;
}

int todo_contains(todo_type s, int orig, int pred) {
	int i;
	for(i=0;i<=s.no;i++) {
		if ((s.orig[i]==orig)&&(s.pred[i]==pred)) return 1;
	}
	return 0;
}
	
int todo_empty(todo_type s) {
	if (s.no==-1) return 1;
	return 0;
}

void run_investigate(bdd_ptr orig,bdd_manager *bddm_orig,bdd_ptr pred,bdd_manager *bddm_pred) {
	/* make one step in both automata */
	unsigned s1,s2;
	/* both bdd pointer pointing to the leaf value -> exit */
	if ((bdd_is_leaf(bddm_orig, orig))&&(bdd_is_leaf(bddm_pred, pred))){
		s1=bdd_leaf_value(bddm_orig, orig);
		s2=bdd_leaf_value(bddm_pred, pred);
		if (!pred_block[s2][s1]){
			pred_block[s2][s1]=1;
			todo_push(&todo,s1,s2);
			//printf("adding %u %u\n",s1,s2);
		}
		return;
	}
	/* orig points to the leaf value -> make one step in pred automaton*/
	if(bdd_is_leaf(bddm_orig, orig)) {
		run_investigate(orig,bddm_orig,bdd_then(bddm_pred, pred),bddm_pred);
		run_investigate(orig,bddm_orig,bdd_else(bddm_pred, pred),bddm_pred);
		return;
	}
	/* pred points to the leaf value -> make one step in orig automaton*/
	if(bdd_is_leaf(bddm_orig, orig)) {
		run_investigate(bdd_then(bddm_orig,orig),bddm_orig,pred,bddm_pred);
		run_investigate(bdd_else(bddm_orig,orig),bddm_orig,pred,bddm_pred);
		return;
	}
	/* index in pred automaton in lower than in original -> make step in pred automaton */
	if(bdd_ifindex(bddm_orig, orig)>bdd_ifindex(bddm_pred, pred)) {
		run_investigate(orig,bddm_orig,bdd_then(bddm_pred, pred),bddm_pred);
		run_investigate(orig,bddm_orig,bdd_else(bddm_pred, pred),bddm_pred);
		return;
	}
	/* index in pred automaton in lower than in original -> make step in pred automaton */
	if(bdd_ifindex(bddm_orig, orig)<bdd_ifindex(bddm_pred, pred)) {
		run_investigate(bdd_then(bddm_orig,orig),bddm_orig,pred,bddm_pred);
		run_investigate(bdd_else(bddm_orig,orig),bddm_orig,pred,bddm_pred);
		return;
	}
	/* last case - run simultanous step in both automata */
	run_investigate(bdd_then(bddm_orig,orig),bddm_orig,bdd_then(bddm_pred, pred),bddm_pred);
	run_investigate(bdd_else(bddm_orig,orig),bddm_orig,bdd_else(bddm_pred, pred),bddm_pred);
	return;	
}

void investigate(GTA *orig, GTA *pred,unsigned state1,unsigned predstate1,unsigned state2,unsigned predstate2) {
	/* this function make the simulation run of automata orig and pred.
	 * if find next states in both automata
	 */
	/* initialization */
	bdd_manager *bddm_orig = orig->ss[1].bddm;
	bdd_manager *bddm_pred = pred->ss[1].bddm;
	bdd_ptr bdd_ptr_orig = BDD_ROOT(bddm_orig, BEH(orig->ss[1], state1, state2));
	bdd_ptr bdd_ptr_pred = BDD_ROOT(bddm_pred, BEH(pred->ss[1], predstate1, predstate2));
	run_investigate(bdd_ptr_orig,bddm_orig,bdd_ptr_pred,bddm_pred);
}


unsigned *map_classes(GTA *G, GTA *pred) {
	/* Make predicate abstraction - each state is predicate.
	 * two states (i & j) are equal => result[i]==result[j]
	 *
	 * !!!!! work only for s.s. 1 
	 * we suppose, that we work only with ss 1. Other statespaces are*/
	int i,j;
	unsigned orig,pr,*res;

	/* initialize pred_block */
	pred_block=(unsigned **)mem_alloc(sizeof(unsigned *)*pred->ss[1].size);
	/* initialize todo */
	todo.limit=G->ss[1].size*pred->ss[1].size;
	todo.no=-1;
	todo.orig=(unsigned *) mem_alloc (sizeof(unsigned)*todo.limit);
	todo.pred=(unsigned *) mem_alloc (sizeof(unsigned)*todo.limit);
	
	/*  fill pred_block with '0' values. */
	for(i=0;i<pred->ss[1].size;i++) {
		pred_block[i]=(unsigned *) mem_alloc(sizeof(unsigned)*G->ss[1].size);
		for(j=0;j<G->ss[1].size;j++) pred_block[i][j]=0;
	}
	/* initial state of original automaton satisfied initial state of predicate automaton:
	 * pred_block[init_G][init_pred]=1 
	 * put init_G x init_pred into 'todo' list*/
	pred_block[pred->ss[1].initial][G->ss[1].initial]=1;
	todo_push(&todo,G->ss[1].initial,pred->ss[1].initial);


	/* iterate till todo list is empty */
	while(!todo_empty(todo)) {
		todo_pop(&todo,&orig,&pr);
		//printf("processing: %u %u\n",orig,pr);
		for(i=0;i<G->ss[1].size;i++)
			for(j=0;j<pred->ss[1].size;j++) 
				/* process only if pred_block[j][i] is set */
				if (pred_block[j][i]) {
					/* orig x pred can be on the left side of rules */
					investigate(G,pred,orig,pr,i,j);
					/* orig x pred can be on the right side of rules */
					investigate(G,pred,i,j,orig,pr);
				}
		
	}
	/*
	for(i=0;i<(pred->ss[1].size);i++) {
		printf("%i: ",i);
  		for (j=0;j<G->ss[1].size;j++) printf("%i ",pred_block[i][j]);
		printf("\n");
	}*/
	/* merge all blocks */
	for(i=0;i<(pred->ss[1].size-1);i++) {
		block_merge(pred_block[i+1],pred_block[i],G->ss[1].size); 
		mem_free(pred_block[i]);
	}
	res=pred_block[pred->ss[1].size-1];
	
	/* leave memory free */
	mem_free(todo.orig);
	mem_free(todo.pred);
	mem_free(pred_block);
	return res;
}




/* Collapsing is not included. function returns list of equal states 
 *
 * !!! it is necessary to run minimalization before abstraction !!!
 * otherwice seq fault :-( */

int *run_predicate_abtraction(GTA *G, gtastack preds,char init_partition) {
/* G - input automaton
 * preds - set of predicates
 * init_part - 1: run initial partition, 0:don't run initial partition
 */
	unsigned *classes,*classes2,states;
	GTA *res;
	int i;
	states=G->ss[1].size;
	if (init_partition)
		/* perform 'standart' sorting - separate pointer descriptors, memory nodes, ... */
		classes=initial_partition_of_ss1(G);
	else {
		/* no initial partition */
		classes=(unsigned *) mem_alloc(sizeof(unsigned)*states);
		for(i=0;i<states;i++)classes[i]=0;
	}
	/* apply all predicates in stack predicates */
	for(i=0;i<=preds.no;i++) {
		classes2=map_classes(G,preds.val[i]);
		block_merge(classes,classes2,states);
		mem_free(classes2);
	}
	return classes;
}
