#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include "gta.h"
#include "../main.h"
#include "../transducer.h"
#include "history_stack.h"
#include "gta_artmc_lib.h"
#include "gtastack.h"
#include "../Mem/mem.h"


#define MAXVARS 1000


extern int lines;
extern GTA *empty_aut,*init_aut;

#ifdef PREDICATES
extern gtastack predicates;
#endif

#ifdef PREDICATES
int check_presence_of_predicate(GTA *pred) {
	int i;
	for(i=0;i<=predicates.no;i++) {
		if ((is_included(pred,predicates.val[i]))&&(is_included(predicates.val[i],pred)))
			return 1;
	}
	return 0;
}
#endif


GTA **back_state_init(GTA *bad, int line) {
	/* initialize array of potentionaly dangerous states 
	 * place bad configuration into the field corresponding to the line on which the bad conf was founded */
	GTA **back_state;
	int i;
	back_state=(GTA **) mem_alloc(sizeof(GTA *)*lines);
	for (i=0;i<lines;i++) back_state[i]=gtaCopy(empty_aut);
	back_state[line]=gtaCopy(bad);
	return back_state;
}

void destroy_back_state(GTA **back_state) {
	int i;
	for (i=0;i<lines;i++)  gtaFree(back_state[i]); 
	mem_free(back_state);
}

GTA *apply_x_assign_ynext_back(GTA *pre, int actual) {
	int i;
	GTA *spec,*ph2,*a1,*a2,*res;
	/* special cases -reverse */
	spec=(GTA *) apply_transducer_reverse(pre, program[actual].transd);

	/* phase2 - reverse */
	ph2=(GTA *) apply_transducer_reverse(pre,program[actual].transd2);
	/* phase 1 - separatelly for all descriptors */
	for (i=1;i<descr;i++){
		a1=(GTA *) apply_transducer_reverse(ph2,descriptors[i].descr);
		a2=(GTA *) apply_transducer_reverse(a1,program[actual].settoken[i]);
		gtaFree(a1);
		if (i==1) res=a2;
		else { 
			a1=gtaProduct(a2,res,gtaOR);
			gtaFree(a2);
			gtaFree(res);
			res=a1;
		}
	}
	a1=gtaProduct(spec,res,gtaOR);
	gtaFree(res);
	gtaFree(spec);	
	gtaFree(ph2);	
	return a1;
}

GTA *apply_free_x_back(GTA *pre, int actual) {
	GTA *aux,*aux2;
	int i;
	aux=apply_transducer_reverse(pre, program[actual].transd); 
	for (i=(program[actual].f-1);i>=0;i--) {
		aux2=apply_transducer_reverse(aux,program[actual].settoken[i]);
		gtaFree(aux);
		aux=aux2;
	}
	return aux;
}


GTA *reverse_run(GTA *bad, int line, hist_stack *history) {
	/* the function returns a set of real counterexamples 
	 * - in case of spurious conterexample, this set is empty 
	 * - in case of predicate based abstraction, new predicates are automatically created */
	int i;
	hist_value actual;
	GTA *apl_res,*aux,*aux2,**back_state,*lastseed;
	/* lastseed is an automaton containing last founded error configuration 
	 * initially lastseed=bad */
	lastseed=gtaCopy(bad);
	/* At the end of this procedure, the back_state[i] will contain states on line i, 
	 * from which a bad configuration is reachable.
	 * "back_state[0] \cap init_aut" is a set of real counterexamples */
	back_state=back_state_init(bad, line);

	printf("*********************\nStarting backward run\n*******************\n");
	while(!histstack_empty(*history)) {
		actual=histstack_pop(history);
		printf("----------------------------\nBackward process: Line num: %i\n",actual);
		if (program[actual.line].type==0) {
			/* This line is equal to exit() in C
			 * no transition is performed */
			printf("Command exit() - no transduction: line skiped\n");
			continue;
		}
		/* apply transducer / series of transducers 
		 * dependentely on actual, and order */
		if ((program[actual.line].type<=3)&&(actual.order==0))
			apl_res=apply_transducer_reverse(back_state[program[actual.line].t],program[actual.line].transd);
		if ((program[actual.line].type<=3)&&(actual.order==1))
			apl_res=apply_transducer_reverse(back_state[program[actual.line].f],program[actual.line].transd2);
		if (program[actual.line].type==4) 
			/* x=y.next */
			apl_res=apply_x_assign_ynext_back(back_state[program[actual.line].t], actual.line);
		if (program[actual.line].type==5)  
			/* x.next=new */
			apl_res=apply_transducer_reverse(back_state[program[actual.line].t],program[actual.line].transd);
		if (program[actual.line].type==6) 
			/* x=y.next */
			apl_res=apply_free_x_back(back_state[program[actual.line].t], actual.line);
#ifdef VERBOSE
		printf("Reverese transducer applied\n");
#endif
		aux=gtaProduct(apl_res,actual.state,gtaAND);
		gtaFree(actual.state);
		gtaFree(apl_res);
		if(empty(aux)) {
			gtaFree(aux);
			continue;
		}
		/* add newly computed dangerous configurations */
		aux2=gtaProduct(back_state[actual.line],aux,gtaOR);
		/* put aux as a "lastseed" of error */
		gtaFree(lastseed);
		lastseed=aux;
		gtaFree(back_state[actual.line]);
		aux=gtaMinimize(aux2);
		gtaFree(aux2);
		back_state[actual.line]=aux;
	}

#ifdef PREDICATES
	/* Report error, if we fonded existing predicate */
	if (check_presence_of_predicate(lastseed)) {
		printf("ERROR: no new predicate added during the reverse run\nSomething is odd :-(\n");
		exit(1);
	}
	/* add lastseed as a new predicate */
	gtastack_push(&predicates,lastseed);
	printf("Added new predicate\n");
#else
	/* we need not the lastseed to refine abstraction */
	gtaFree(lastseed);
#endif

	aux=gtaProduct(back_state[0],init_aut,gtaAND);
	/* clean memory */
	destroy_back_state(back_state);
	return aux;
}
