/*****************************************************************************
  Adam Rogalewicz
  
  Rotations on trees encoded in "bhrv06" style
  - supposed binary branching
  ****************************************************************************/

#include "../main.h"
#include "gta.h"
#include "../BDD/bdd.h"
#include "../Mem/mem.h"


#include "../initial_partition.h"
#include "../stack/stack.h"

unsigned ROT_act_new_state, ROT_actual, ROT_actual_new;
int ROT_reject;

/*************************************/
/* Functions to manipulate classes   */
/*************************************/

/* check, whether state is a memory node, or top memory node */
int is_mem_node(unsigned state, unsigned *classes) {
	if (classes[state]==MEMN_CLS) return 1;
	if (classes[state]==TOPMN_CLS) return 1;
	return 0;
}

/* check, whether state is a next pointer node */
int is_next_ptr(unsigned state, unsigned *classes) {
	if (classes[state]>=NEXTPTR_CLS) return 1;
	return 0;
}

/* check, whether state is a bottom */
int is_bot_node(unsigned state, unsigned *classes) {
	if (classes[state]==INITIAL_CLS) return 1;
	if (classes[state]==INITIAL_CLS2) return 1;
	return 0;
}

int ROT_get_reject(unsigned *classes, unsigned size) {
	unsigned i;
	for(i=0;i<size;i++) {
		if (classes[i]==REJECT_CLS) return i;
	}
	return -1;
	
}

int ROT_get_init(unsigned *classes, unsigned size) {

	unsigned i;
	for(i=0;i<size;i++) {
		if (classes[i]==INITIAL_CLS) return i;
	}
	return -1;
}

int ROT_get_init1(unsigned *classes, unsigned size) {

	unsigned i;
	for(i=0;i<size;i++) {
		if (classes[i]==INITIAL_CLS2) return i;
	}
	return -1;
}

/*************************************/
/* Stack operations                  */
/*************************************/

void push_triple(Ustack *st,unsigned a,unsigned b,unsigned c) {
	Ustack_push(st,a);
	Ustack_push(st,b);
	Ustack_push(st,c);
}

void push_4(Ustack *st,unsigned a,unsigned b,unsigned c,unsigned d) {
	Ustack_push(st,a);
	Ustack_push(st,b);
	Ustack_push(st,c);
	Ustack_push(st,d);
}

void push_5(Ustack *st,unsigned a,unsigned b,unsigned c,unsigned d,unsigned e) {
	Ustack_push(st,a);
	Ustack_push(st,b);
	Ustack_push(st,c);
	Ustack_push(st,d);
	Ustack_push(st,e);
}

void push_6(Ustack *st,unsigned a,unsigned b,unsigned c,unsigned d,unsigned e,unsigned f) {
	Ustack_push(st,a);
	Ustack_push(st,b);
	Ustack_push(st,c);
	Ustack_push(st,d);
	Ustack_push(st,e);
	Ustack_push(st,f);
}

void push_double(Ustack *st,unsigned a,unsigned b) {
	Ustack_push(st,a);
	Ustack_push(st,b);
}

void pop_double(Ustack *st,unsigned *a,unsigned *b) {
	*b=Ustack_pop(st);
	*a=Ustack_pop(st);
}

void pop_triple(Ustack *st,unsigned *a,unsigned *b,unsigned *c) {
	*c=Ustack_pop(st);
	*b=Ustack_pop(st);
	*a=Ustack_pop(st);
}
void pop_4(Ustack *st,unsigned *a,unsigned *b,unsigned *c,unsigned *d) {
	*d=Ustack_pop(st);
	*c=Ustack_pop(st);
	*b=Ustack_pop(st);
	*a=Ustack_pop(st);
}

void pop_5(Ustack *st,unsigned *a,unsigned *b,unsigned *c,unsigned *d,unsigned *e) {
	*e=Ustack_pop(st);
	*d=Ustack_pop(st);
	*c=Ustack_pop(st);
	*b=Ustack_pop(st);
	*a=Ustack_pop(st);
}

void pop_6(Ustack *st,unsigned *a,unsigned *b,unsigned *c,unsigned *d,unsigned *e,unsigned *f) {
	*f=Ustack_pop(st);
	*e=Ustack_pop(st);
	*d=Ustack_pop(st);
	*c=Ustack_pop(st);
	*b=Ustack_pop(st);
	*a=Ustack_pop(st);
}


/*************************************************************************/
/* Create special purpose BDD                                            */
/*************************************************************************/

bdd_manager *create_bdd_for_x(int track) {
	/* create small BDD (only 3 nodes)
	   - this BDD returns value of the track^th bit
	*/
	bdd_manager *bddm;
	bdd_ptr zero, one;
	bdd_handle handle;


	/* bdd manager with 10 nodes will be enough
	   (in case of overflow, the table is increased by 10) */
	bddm = bdd_new_manager(10,10);
	zero = bdd_find_leaf_hashed_add_root(bddm, 0);
	one =  bdd_find_leaf_hashed_add_root(bddm, 1);
	handle = bdd_handle_find_node_hashed_add_root(bddm, zero, one, track);
	/* Make sure, that it is created correctly - we will refer to this bdd only through handle no. '2' */
	//assert(handle == 2);
	if(handle!=2) {
		printf("ERROR: rotations: create_bdd_for_x\n");
		exit(1);
	}
	return bddm;
}

/*************************************************************************/
/* functions used inside BDD-apply commands */
/*************************************************************************/

unsigned mark_x_as_new_state(unsigned a, unsigned b) {
	if ((b)&&(a==ROT_actual)) return ROT_act_new_state;
	return ROT_reject;
}

unsigned set_x_as_reject(unsigned a, unsigned b) {
	if (b) return ROT_reject;
	return a;
}

unsigned mark_actual(unsigned a) {
	if (a==ROT_actual) return ROT_actual_new;
	return ROT_reject;
}

/*unsigned check_rule_exists(unsigned a) {
	if (a==ROT_actual) return 1;
	return 0;
	
}*/


/**************************************************************************************/
/**************************************************************************************/
/* RIGHT ROTATE                                                                        */
/**************************************************************************************/
/**************************************************************************************/

GTA *run_right_rotate(GTA *aut_in, unsigned track,unsigned *classes) {
	unsigned i,j,states;
	int init0,init1;
	GTA *aut;
	bdd_manager *aut_bddm,*xbddm;
	bdd_ptr res;
	bdd_handle tmp_handle,step2_handle,tmp_handle2;
	Ustack new_behaviour, step1_down, step2, step3, step4, step5; /* stacs of int */
	unsigned st_orig,st_new,top_state,orig_top_new,step2_post, used, used1;
	aut=gtaCopy(aut_in);

	aut_bddm=aut->ss[SS1].bddm;
	states=aut->ss[SS1].size;
	xbddm=create_bdd_for_x(track);

	/* the lowest unused state number */	
	ROT_act_new_state=states; 

	/*for(i=0;i<states;i++) printf("%i ",classes[i]);
	printf("\n");*/

	/* search for init, init1, and reject state in "classes" */
	ROT_reject=ROT_get_reject(classes,states);
	init0=ROT_get_init(classes,states);
	init1=ROT_get_init1(classes,states);
	if (ROT_reject==-1) { printf("ERROR: rotations: No reject state !!!\n"); exit(1);}
	if (init0==-1) { printf("ERROR: rotations: No initial state !!!\n"); exit(1);}
	if (init1==-1) { printf("WARNING: rotations: No initial state 2\n"); }

	/* Newly created behaviour will be temporary stored in this stac as triples */
	Ustack_init(&new_behaviour,100); /* size for 100 numbers, we use triples (left, right, bdd_handle) -> 33 triples */

	/************************************************************************/
	/* STEP1 */
	/************************************************************************/
	
	Ustack_init(&step1_down,10); 
	
	
	for(j=0;j<states;j++) {
		if (!is_next_ptr(j,classes)) continue;
		used=0;
		used1=0;
		for(i=0;i<states;i++) {
			/* just memory nodes can be processed */
			if (!is_mem_node(i,classes)) continue;
			ROT_actual=i;
			/* now is necessary to reinitialize cashe, because 
			   "mark_x_as_new_state" function depends on ROT_actual */
			bdd_kill_cache(aut_bddm);
			bdd_make_cache(aut_bddm, 100, 50);
			//printf(">>>> %i %i\n",i,j);
				/* check, wheather rules ...x...(j,init) -> i exists
				   if yes, 
				     - create new rules ...x...(new(j),init) -> new(i)
				     - place reference to (new(j), init) to behaviour table
				 */

				/* first for init0 */

		 	res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init0)),
                                                  xbddm,BDD_ROOT(xbddm,2),
                                                  aut_bddm,&mark_x_as_new_state);  /* set i to act_new_state */
						  
			if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
				/* NO transition via 'x' from (i,init0) - nothing to do with this bdd_handle */
                                //POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,BDD_LAST_HANDLE(aut_bddm));
                        	POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
			} else {
				/* founded corresponding rule */
				//printf("fonded: i:%i j:%i\n",i,j);
				/* new behaviour */
				push_triple(&new_behaviour,ROT_act_new_state+1,init0,BDD_LAST_HANDLE(aut_bddm));
				/* process down from this node */
				push_4(&step1_down,j,ROT_act_new_state+1,i,ROT_act_new_state);
				ROT_act_new_state+=2; /* next state number to be used */
				used=1;   
				
			}
				/* NOW for init1 */
			if (init1==-1) continue; /* nothing to do */
			res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],i,init1)),
                                                  xbddm,BDD_ROOT(xbddm,2),
                                                  aut_bddm,&mark_x_as_new_state);
                        if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
				        /* NO transition via 'x' from (i,init1) - nothing to do with this bdd_handle */
                                        //POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,BDD_LAST_HANDLE(aut_bddm));
                                POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
			} else {
				/* founded corresponding rule */
				/* new behaviour */
				push_triple(&new_behaviour,ROT_act_new_state+1,init1,BDD_LAST_HANDLE(aut_bddm));
				/* process down from this node */
				push_4(&step1_down,j,ROT_act_new_state+1,i,ROT_act_new_state);
				ROT_act_new_state+=2; /* next state number to be used */
				used1=1;
			}
		} /* for(i=0; ;i++) */
		/* REMOVE old rules ...x...(j,init0) -> i */
		/* !!!! We need to restart apply2 cache, becaurse we use here another function !!! */
		bdd_kill_cache(aut_bddm);
		bdd_make_cache(aut_bddm, 100, 50);
		if (used) {
			res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init0)),
                                      xbddm,BDD_ROOT(xbddm,2),
                                      aut_bddm,&set_x_as_reject); /* set i to act_new_state */
			push_triple(&new_behaviour,j,init0,BDD_LAST_HANDLE(aut_bddm));
		}
		/* REMOVE old rules ...x...(j,init1) -> i */
		if (used1) {
			res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init1)),
                                      xbddm,BDD_ROOT(xbddm,2),
                                      aut_bddm,&set_x_as_reject); /* set i to act_new_state */
			push_triple(&new_behaviour,j,init1,BDD_LAST_HANDLE(aut_bddm));
		}
	}  /* for(j=0; ;j++) */
	
	bdd_kill_cache(aut_bddm);

	
	/************************************************************************/
	/* STEP2 */
	/************************************************************************/
	//printf("step 2\n");
	Ustack_init(&step2,10);  
	while(!Ustack_empty(step1_down)) {
		pop_4(&step1_down,&st_orig,&st_new,&top_state,&orig_top_new);
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=st_new; /* for function mark_actual */
		///printf("POPED: %i %i %i %i\n",st_orig,st_new,top_state,orig_top_new);

		for (i=0;i<states;i++) {
			//used=0;
			/* just memory nodes can be on left */
			if (!is_mem_node(i,classes)) continue;
			for(j=0;j<states;j++) {
				if (!is_next_ptr(j,classes)) continue;
				/* we have on the left memory node, and on the right next_pointer node */
				/* we need to check, if there exists a rule ...(i,j) -> st_orig */
		  		bdd_prepare_apply1(aut_bddm);  
				
				res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],i,j)),
                                         aut_bddm,&mark_actual); 
						  
                                if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject))  {
                               		POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
				} else {
				        /* There exists a transition via from (i,j) -> st_new */
					push_5(&step2,i,j,BDD_LAST_HANDLE(aut_bddm),top_state,orig_top_new);
					//printf("step2: i:%i j:%i\n",i,j);
					//printf("step2: pushed: i:%i j:%i top:%i top_new:%i\n", i,j,top_state,orig_top_new);
				}

			}
			//if (used) ROT_act_new_state++;
		}

		
	}
	Ustack_destroy(&step1_down); 
	/************************************************************************/
	/* STEP3 */
	/************************************************************************/
	//printf("step 3\n");
	Ustack_init(&step3,10);  

	while(!Ustack_empty(step2)) {
		pop_5(&step2,&st_orig,&step2_post,&step2_handle,&top_state,&orig_top_new);
		//printf("Step3: POP: %i %i %i %i\n",st_orig,step2_post,top_state,orig_top_new);
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=top_state; /* for function mark_actual */

		for(j=0;j<states;j++) {
			if (!is_next_ptr(j,classes)) continue;
			/* first for init0 */
		  	bdd_prepare_apply1(aut_bddm);  
		 	res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init0)),
                                                aut_bddm,&mark_actual); /* set i to act_new_state */
                        if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
			        /* NO from ...(j,init0) -> i: nothing to do with this bdd_handle */
                                POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
		  	} else {
				push_triple(&new_behaviour,ROT_act_new_state,init0,BDD_LAST_HANDLE(aut_bddm));
				push_5(&step3,j,ROT_act_new_state,step2_post,step2_handle,orig_top_new);
				//printf("step3: push: j:%i, new:%i,step2_post:%i,orig_top:%i\n",j,ROT_act_new_state,step2_post,orig_top_new);
				ROT_act_new_state++;
			}
			/* now for init1 */
			if (init1==-1) continue;
		  	bdd_prepare_apply1(aut_bddm);  
		 	res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init1)),
                                                aut_bddm,&mark_actual); /* set i to act_new_state */
                        if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
			        /* NO from ...(j,init0) -> i: nothing to do with this bdd_handle */
                                POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
			} else {
				push_triple(&new_behaviour,ROT_act_new_state,init1,BDD_LAST_HANDLE(aut_bddm));
				push_5(&step3,j,ROT_act_new_state,step2_post,step2_handle,orig_top_new);
				//printf("step3: push: j:%i, new:%i,step2_post:%i,orig_top:%i\n",j,ROT_act_new_state,step2_post,orig_top_new);
				ROT_act_new_state++;
			}
		}
	}
	Ustack_destroy(&step2); 
	/************************************************************************/
	/* STEP4 */
	/************************************************************************/
	//printf("step 4\n");
	Ustack_init(&step4,10);  
	while(!Ustack_empty(step3)) {
		pop_5(&step3,&st_orig,&st_new,&step2_post,&step2_handle,&orig_top_new);
		//printf("Step4: POP: %i %i %i %i\n",st_orig,st_new,step2_post,orig_top_new);
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=st_new; /* for function mark_actual */

		for(i=0;i<states;i++) {
			used=0;
			if (!is_next_ptr(i,classes)) continue;
			for(j=0;j<states;j++) {
				if (!((is_mem_node(j,classes))||(is_bot_node(j,classes)))) continue;
		  		bdd_prepare_apply1(aut_bddm);  
				res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,i)),
                                         aut_bddm,&mark_actual); 
						  
                                if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject))  {
                                		POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
				} else {
					push_triple(&new_behaviour,j,ROT_act_new_state,BDD_LAST_HANDLE(aut_bddm));
					//printf("step4: i:%i j:%i\n",i,j);
					used=1;
				}
			}
			if (used) { 
				push_5(&step4,i,ROT_act_new_state,step2_post,step2_handle,orig_top_new);
				//printf("step4: push: j:%i, new:%i,step2_post:%i,orig_top:%i\n",i,ROT_act_new_state,step2_post,orig_top_new);
				ROT_act_new_state++; 
				used=0; 
			} 
				
		}
	}	
	Ustack_destroy(&step3); 
	/************************************************************************/
	/* STEP5 */
	/************************************************************************/
	//printf("step 5\n");
	Ustack_init(&step5,10);  
	while(!Ustack_empty(step4)) {
		pop_5(&step4,&st_orig,&st_new,&step2_post,&step2_handle,&orig_top_new);
		//printf("Step5: POP: %i %i %i %i\n",st_orig,st_new,step2_post,orig_top_new);
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=st_new; /* for function mark_actual */
		for(i=0;i<states;i++) {
			/* i must be a pointer node, or init */
			if (!((is_next_ptr(i,classes))||(is_bot_node(i,classes)))) continue;
			used=0;
			for(j=0;j<states;j++) {
				/* j must be a memory node, or bottom */
				if (!((is_mem_node(j,classes))||(is_bot_node(j,classes)))) continue;
		  		bdd_prepare_apply1(aut_bddm);  
			 	res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,i)),
                                                aut_bddm,&mark_actual); /* set i to act_new_state */
 	                	if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
			        /* NO from ...(j,init0) -> i: nothing to do with this bdd_handle */
        	                        POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
		  		} else {
					push_triple(&step5,j,step2_post,step2_handle);
					//printf("j:%i i:%i ROT_actual:%i\n",j,i,ROT_actual);
					if (used) {
						if (states<used) {
							printf("ERROR: rotations: step5: overflow\n"); 
							exit(1);
						}
						//printf("Nondeterministic choice: %i %i\n",used,states-used);
						tmp_handle2=bdd_handle_find_node_hashed_add_root(aut_bddm,
							BDD_ROOT(aut_bddm,BDD_LAST_HANDLE(aut_bddm)),
							BDD_ROOT(aut_bddm,tmp_handle2),
							states-used);
						used++;
					} else {
						tmp_handle2=BDD_LAST_HANDLE(aut_bddm);
						used=1;
					} 
					
				}
			} /* for(j=0;j<states;j++) */
			if (used) {
				push_triple(&new_behaviour,orig_top_new,i,tmp_handle2);
				used=0;
			}
		} /* for(i=0;i<states;i++) */
	} /* while(Ustack_empty(step4)) */
	Ustack_destroy(&step4); 
	/************************************************************************/
	/* second part of STEP 2 */
	/************************************************************************/
	while(!Ustack_empty(step5)) {
		pop_triple(&step5,&i,&j,&step2_handle);
		tmp_handle=bdd_handle_find_node_hashed_add_root(aut_bddm,
							BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],i,j)),
							BDD_ROOT(aut_bddm,step2_handle),
							0);
		push_triple(&new_behaviour,i,j,tmp_handle);
	}
	Ustack_destroy(&step5); 

	bdd_kill_manager(xbddm);

	/* now create the resulting automaton - create new behaviour for SS1*/
	
	StateSpace *ss = &aut->ss[1];
	/* first change size of statespace, and realloc the behaviour table */
	ss->size = ROT_act_new_state;
	ss->ls = ROT_act_new_state;
	ss->rs = ROT_act_new_state;
	ss->behaviour = (bdd_handle *) mem_resize(ss->behaviour,sizeof(bdd_handle)*ss->ls*ss->rs);
	/* now fill the BDD-table with old values, and "rejects" */
	for (i=0;i<ss->size;i++) {
		for (j=0;j<ss->size;j++) {
			if ((i<aut_in->ss[1].size)&&(j<aut_in->ss[1].size))
				BEH(aut->ss[1], i, j)=BEH(aut_in->ss[1], i, j);
			else
				BEH(aut->ss[1], i, j)=BEH(aut_in->ss[1], ROT_reject, ROT_reject);
		}
	}
	while(!Ustack_empty(new_behaviour)) {
		pop_triple(&new_behaviour,&i,&j,&tmp_handle);
		BEH(aut->ss[1], i, j)=tmp_handle;
	}
	Ustack_destroy(&new_behaviour);

	/* fill behaviour for ss[0] */
	ss = &aut->ss[0];
	ss->size = ROT_act_new_state;
	ss->ls = ROT_act_new_state;
	ss->rs = 1;
	ss->behaviour = (bdd_handle *) mem_resize(ss->behaviour,sizeof(bdd_handle)*ss->ls*ss->rs);
	for(i=0;i<ROT_act_new_state;i++) {
		if(i<aut_in->ss[0].size)
			BEH(aut->ss[0], i, 0)=BEH(aut_in->ss[0], i, 0);
		else
			BEH(aut->ss[0], i, 0)=BEH(aut_in->ss[0], ROT_reject, 0);
	}

	return aut;
}

/* Na zacatku je treba predelat cislovani tracku 0 -> 1,
   na konci project na '0' a nasledne prepsat track 1 -> 0 */

GTA *right_rotate(GTA *aut, int variable) {
	bdd_manager *bddm_forx;
	GTA *aut1,*aut2;
	int i;
	unsigned *map,*classes,states;
	/* For sure, we run minimization */
	//aut1=gtaMinimize(aut);
	aut1=gtaCopy(aut);
	states=aut->ss[1].size;
	
	/* search for classes before remaping */
	classes=initial_partition_of_ss1(aut1);
	
	/* make space for n (n-number of states)  nondeterministic choices */
	/* Remap track 0 -> n, 1-> n+1 */
	map=(unsigned *) mem_alloc(sizeof(unsigned)*((2*NODE_WIDTH)+states));
        for (i=0; i<2*NODE_WIDTH; i=i+2) {
		map[i]=states+i;               
	}
	gtaReplaceIndices(aut1,map);

	/* run the rotation */
	aut2=run_right_rotate(aut1,(variable*2)+4+states,classes);
	gtaFree(aut1);
	//return aut2;
	/* project tracks 0-(states-1) - nondeterministic choices created by step 2b, and 5 */
	for(i=0;i<states;i++) {
		aut1=gtaQuotientAndProject(aut2,i,0);
		gtaFree(aut2);
		aut2=aut1;
	}
	/* remap tract n->0, n+1->1, ... */
        for (i=states; i<(2*NODE_WIDTH)+states; i=i+2) {
		map[i]=i-states;               
	}
	
	gtaReplaceIndices(aut2,map);
	/* minimize the result */
	aut1=gtaMinimize(aut2);
	gtaFree(aut2);

	/* clean memory */
	mem_free(map);
	mem_free(classes);
	return aut1;

}


/**************************************************************************************/
/**************************************************************************************/
/* LEFT ROTATE                                                                        */
/**************************************************************************************/
/**************************************************************************************/


GTA *run_left_rotate(GTA *aut_in, unsigned track,unsigned *classes) {
	unsigned i,j,states;
	int init0,init1;
	GTA *aut;
	bdd_manager *aut_bddm,*xbddm;
	bdd_ptr res;
	bdd_handle tmp_handle,step3_handle,tmp_handle2;
	Ustack new_behaviour, step1_down, step2, step3, step4, step5; /* stacs of int */
	unsigned st_orig,st_new,top_state,orig_top_new,step3_post, used, used1;
	aut=gtaCopy(aut_in);

	aut_bddm=aut->ss[SS1].bddm;
	states=aut->ss[SS1].size;
	xbddm=create_bdd_for_x(track);

	/* the lowest unused state number */	
	ROT_act_new_state=states; 

	/*for(i=0;i<states;i++) printf("%i ",classes[i]);
	printf("\n");*/

	/* search for init, init1, and reject state in "classes" */
	ROT_reject=ROT_get_reject(classes,states);
	init0=ROT_get_init(classes,states);
	init1=ROT_get_init1(classes,states);
	if (ROT_reject==-1) { printf("ERROR: rotations: No reject state !!!\n"); exit(1);}
	if (init0==-1) { printf("ERROR: rotations: No initial state !!!\n"); exit(1);}
	if (init1==-1) { printf("WARNING: rotations: No initial state 2\n"); }

	/* Newly created behaviour will be temporary stored in this stac as triples */
	Ustack_init(&new_behaviour,100); /* size for 100 numbers, we use triples (left, right, bdd_handle) -> 33 triples */

	/************************************************************************/
	/* STEP1 */
	/************************************************************************/
	//printf("step1\n");
	
	Ustack_init(&step1_down,10); 
	
	
	for(j=0;j<states;j++) {
		if (!is_next_ptr(j,classes)) continue;
		used=0;
		used1=0;
		for(i=0;i<states;i++) {
			/* just memory nodes can be processed */
			if (!is_mem_node(i,classes)) continue;
			ROT_actual=i;
			/* now is necessary to reinitialize cashe, because 
			   "mark_x_as_new_state" function depends on ROT_actual */
			bdd_kill_cache(aut_bddm);
			bdd_make_cache(aut_bddm, 100, 50);
			//printf(">>>> %i %i\n",i,j);
				/* check, wheather rules ...x...(j,init) -> i exists
				   if yes, 
				     - create new rules ...x...(new(j),init) -> new(i)
				     - place reference to (new(j), init) to behaviour table
				 */

				/* first for init0 */

		 	res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init0)),
                                                  xbddm,BDD_ROOT(xbddm,2),
                                                  aut_bddm,&mark_x_as_new_state);  /* set i to act_new_state */
						  
			if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
				/* NO transition via 'x' from (i,init0) - nothing to do with this bdd_handle */
                                //POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,BDD_LAST_HANDLE(aut_bddm));
                        	POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
			} else {
				/* founded corresponding rule */
				//printf("fonded: i:%i j:%i\n",i,j);
				/* new behaviour */
				push_triple(&new_behaviour,ROT_act_new_state+1,init0,BDD_LAST_HANDLE(aut_bddm));
				/* process down from this node */
				push_4(&step1_down,j,ROT_act_new_state+1,i,ROT_act_new_state);
				ROT_act_new_state+=2; /* next state number to be used */
				used=1;   
				
			}
				/* NOW for init1 */
			if (init1==-1) continue; /* nothing to do */
			res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],i,init1)),
                                                  xbddm,BDD_ROOT(xbddm,2),
                                                  aut_bddm,&mark_x_as_new_state);
                        if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
				        /* NO transition via 'x' from (i,init1) - nothing to do with this bdd_handle */
                                        //POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,BDD_LAST_HANDLE(aut_bddm));
                                POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
			} else {
				/* founded corresponding rule */
				/* new behaviour */
				push_triple(&new_behaviour,ROT_act_new_state+1,init1,BDD_LAST_HANDLE(aut_bddm));
				/* process down from this node */
				push_4(&step1_down,j,ROT_act_new_state+1,i,ROT_act_new_state);
				ROT_act_new_state+=2; /* next state number to be used */
				used1=1;
			}
		} /* for(i=0; ;i++) */
		/* REMOVE old rules ...x...(j,init0) -> i */
		/* !!!! We need to restart apply2 cache, becaurse we use here another function !!! */
		bdd_kill_cache(aut_bddm);
		bdd_make_cache(aut_bddm, 100, 50);
		if (used) {
			res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init0)),
                                      xbddm,BDD_ROOT(xbddm,2),
                                      aut_bddm,&set_x_as_reject); /* set i to act_new_state */
			push_triple(&new_behaviour,j,init0,BDD_LAST_HANDLE(aut_bddm));
		}
		/* REMOVE old rules ...x...(j,init1) -> i */
		if (used1) {
			res=bdd_apply2_hashed(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init1)),
                                      xbddm,BDD_ROOT(xbddm,2),
                                      aut_bddm,&set_x_as_reject); /* set i to act_new_state */
			push_triple(&new_behaviour,j,init1,BDD_LAST_HANDLE(aut_bddm));
		}
	}  /* for(j=0; ;j++) */
	
	bdd_kill_cache(aut_bddm);
	/************************************************************************/
	/* STEP2 - equal to step4 in right rotate */
	/************************************************************************/
	//printf("step 2\n");
	Ustack_init(&step2,10);  
	while(!Ustack_empty(step1_down)) {
		//pop_4(&step3,&st_orig,&st_new,&step2_post,&step2_handle,&orig_top_new);
		pop_4(&step1_down,&st_orig,&st_new,&top_state,&orig_top_new);
		
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=st_new; /* for function mark_actual */

		for(i=0;i<states;i++) {
			used=0;
			if (!is_next_ptr(i,classes)) continue;
			for(j=0;j<states;j++) {
				if (!((is_mem_node(j,classes))||(is_bot_node(j,classes)))) continue;
		  		bdd_prepare_apply1(aut_bddm);  
				res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,i)),
                                         aut_bddm,&mark_actual); 
						  
                                if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject))  {
                                		POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
				} else {
					push_triple(&new_behaviour,j,ROT_act_new_state,BDD_LAST_HANDLE(aut_bddm));
					//printf("step2: i:%i j:%i\n",i,j);
					used=1;
				}
			}
			if (used) { 
				push_4(&step2,i,ROT_act_new_state,top_state,orig_top_new);
				//printf("step4: push: j:%i, new:%i,step2_post:%i,orig_top:%i\n",i,ROT_act_new_state,step3_post,orig_top_new);
				ROT_act_new_state++; 
				used=0; 
			} 
				
		}
	}	
	Ustack_destroy(&step1_down); 

	/************************************************************************/
	/* STEP3 - equal to step2 in right rotate */
	/************************************************************************/
	//printf("step 3\n");
	Ustack_init(&step3,10);  
	while(!Ustack_empty(step2)) {
		pop_4(&step2,&st_orig,&st_new,&top_state,&orig_top_new);
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=st_new; /* for function mark_actual */
		//printf("POPED: %i %i %i %i\n",st_orig,st_new,top_state,orig_top_new);

		for(i=0;i<states;i++) {
			/* just memory nodes can be on left */
			if (!is_mem_node(i,classes)) continue;
			for(j=0;j<states;j++) {
				if (!((is_next_ptr(j,classes))||(is_bot_node(j,classes)))) continue;
				/* we have on the left memory node, and on the right next_pointer node/or bottom */
				/* we need to check, if there exists a rule ...(i,j) -> st_orig */
		  		bdd_prepare_apply1(aut_bddm);  
				
				res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],i,j)),
                                         aut_bddm,&mark_actual); 
						  
                                if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject))  {
                               		POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
				} else {
				        /* There exists a transition via from (i,j) -> st_new */
					push_5(&step3,i,j,BDD_LAST_HANDLE(aut_bddm),top_state,orig_top_new);
					//printf("step3: pushed: i:%i j:%i top:%i top_new:%i\n", i,j,top_state,orig_top_new);
				}

			}
		}

		
	}
	Ustack_destroy(&step2); 

	/************************************************************************/
	/* STEP4 - equal to step3 in right rotate3 */
	/************************************************************************/
	//printf("step 4\n");
	Ustack_init(&step4,10);  

	while(!Ustack_empty(step3)) {
		pop_5(&step3,&st_orig,&step3_post,&step3_handle,&top_state,&orig_top_new);
		//printf("Step4: POP: %i %i %i %i\n",st_orig,step3_post,top_state,orig_top_new);
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=top_state; /* for function mark_actual */

		for(j=0;j<states;j++) {
			if (!is_next_ptr(j,classes)) continue;
			/* first for init0 */
		  	bdd_prepare_apply1(aut_bddm);  
		 	res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init0)),
                                                aut_bddm,&mark_actual); /* set i to act_new_state */
                        if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
			        /* NO from ...(j,init0) -> i: nothing to do with this bdd_handle */
                                POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
		  	} else {
				push_triple(&new_behaviour,ROT_act_new_state,init0,BDD_LAST_HANDLE(aut_bddm));
				push_5(&step4,j,ROT_act_new_state,step3_post,step3_handle,orig_top_new);
				//printf("step4: push: j:%i, new:%i,step3_post:%i,orig_top:%i\n",j,ROT_act_new_state,step3_post,orig_top_new);
				ROT_act_new_state++;
			}
			/* now for init1 */
			if (init1==-1) continue;
		  	bdd_prepare_apply1(aut_bddm);  
		 	res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,init1)),
                                                aut_bddm,&mark_actual); /* set i to act_new_state */
                        if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
			        /* NO from ...(j,init0) -> i: nothing to do with this bdd_handle */
                                POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
			} else {
				push_triple(&new_behaviour,ROT_act_new_state,init1,BDD_LAST_HANDLE(aut_bddm));
				push_5(&step4,j,ROT_act_new_state,step3_post,step3_handle,orig_top_new);
				//printf("step4: push: j:%i, new:%i,step3_post:%i,orig_top:%i\n",j,ROT_act_new_state,step3_post,orig_top_new);
				ROT_act_new_state++;
			}
		}
	}
	Ustack_destroy(&step3); 
	/************************************************************************/
	/* STEP5 - aqual as in right rotate */
	/************************************************************************/
	//printf("step 5\n");
	Ustack_init(&step5,10);  
	while(!Ustack_empty(step4)) {
		pop_5(&step4,&st_orig,&st_new,&step3_post,&step3_handle,&orig_top_new);
		//printf("Step5: POP: %i %i %i %i\n",st_orig,st_new,step3_post,orig_top_new);
		ROT_actual=st_orig; /* for function mark_actual */
		ROT_actual_new=st_new; /* for function mark_actual */
		for(i=0;i<states;i++) {
			/* i must be a pointer node, or init */
			if (!(is_next_ptr(i,classes))) continue;
			used=0;
			for(j=0;j<states;j++) {
				/* j must be a memory node, or bottom */
				if (!((is_mem_node(j,classes))||(is_bot_node(j,classes)))) continue;
		  		bdd_prepare_apply1(aut_bddm);  
			 	res=bdd_apply1(aut_bddm,BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],j,i)),
                                                aut_bddm,&mark_actual); /* set i to act_new_state */
 	                	if ((bdd_is_leaf(aut_bddm,res)) && (bdd_leaf_value(aut_bddm,res)==ROT_reject)) { 
			        /* NO from ...(j,init0) -> i: nothing to do with this bdd_handle */
        	                        POP_SEQUENTIAL_LIST(aut_bddm->roots,bdd_handle,tmp_handle);
		  		} else {
					push_triple(&step5,j,step3_post,step3_handle);
					//printf("j:%i i:%i ROT_actual:%i\n",j,i,ROT_actual);
					//printf("step5: push: j:%i step2_post:%i\n", j,step2_post);
					//push_triple(&new_behaviour,orig_top_new,i,BDD_LAST_HANDLE(aut_bddm));
					if (used) {
						if (states<used) {
							printf("ERROR: rotations: step5: overflow\n"); 
							exit(1);
						}
						//printf("Nondeterministic choice: %i %i\n",used,states-used);
						tmp_handle2=bdd_handle_find_node_hashed_add_root(aut_bddm,
							BDD_ROOT(aut_bddm,BDD_LAST_HANDLE(aut_bddm)),
							BDD_ROOT(aut_bddm,tmp_handle2),
							states-used);
						used++;
					} else {
						tmp_handle2=BDD_LAST_HANDLE(aut_bddm);
						used=1;
					} 
					//printf("tmp_handle: %i\n",tmp_handle2);
					
				}
			} /* for(j=0;j<states;j++) */
			if (used) {
				//printf("tmp_handle: %i\n",tmp_handle2);
				push_triple(&new_behaviour,orig_top_new,i,tmp_handle2);
				used=0;
			}
		} /* for(i=0;i<states;i++) */
	} /* while(Ustack_empty(step4)) */
	Ustack_destroy(&step4); 

	/************************************************************************/
	/* second part of STEP 3 */
	/************************************************************************/
	//printf("step 3b\n");
	while(!Ustack_empty(step5)) {
		pop_triple(&step5,&i,&j,&step3_handle);
		//printf("step3B: i:%i j:%i\n",i,j);
		tmp_handle=bdd_handle_find_node_hashed_add_root(aut_bddm,
							BDD_ROOT(aut_bddm,BEH(aut->ss[SS1],i,j)),
							BDD_ROOT(aut_bddm,step3_handle),
							0);
		push_triple(&new_behaviour,i,j,tmp_handle);
	}
	Ustack_destroy(&step5); 

	bdd_kill_manager(xbddm);

	/* now create the resulting automaton - create new behaviour for SS1*/
	
	StateSpace *ss = &aut->ss[1];
	/* first change size of statespace, and realloc the behaviour table */
	ss->size = ROT_act_new_state;
	ss->ls = ROT_act_new_state;
	ss->rs = ROT_act_new_state;
	ss->behaviour = (bdd_handle *) mem_resize(ss->behaviour,sizeof(bdd_handle)*ss->ls*ss->rs);
	/* now fill the BDD-table with old values, and "rejects" */
	for (i=0;i<ss->size;i++) {
		for (j=0;j<ss->size;j++) {
			if ((i<aut_in->ss[1].size)&&(j<aut_in->ss[1].size))
				BEH(aut->ss[1], i, j)=BEH(aut_in->ss[1], i, j);
			else
				BEH(aut->ss[1], i, j)=BEH(aut_in->ss[1], ROT_reject, ROT_reject);
		}
	}
	while(!Ustack_empty(new_behaviour)) {
		pop_triple(&new_behaviour,&i,&j,&tmp_handle);
		BEH(aut->ss[1], i, j)=tmp_handle;
	}
	Ustack_destroy(&new_behaviour);

	/* fill behaviour for ss[0] */
	ss = &aut->ss[0];
	ss->size = ROT_act_new_state;
	ss->ls = ROT_act_new_state;
	ss->rs = 1;
	ss->behaviour = (bdd_handle *) mem_resize(ss->behaviour,sizeof(bdd_handle)*ss->ls*ss->rs);
	for(i=0;i<ROT_act_new_state;i++) {
		if(i<aut_in->ss[0].size)
			BEH(aut->ss[0], i, 0)=BEH(aut_in->ss[0], i, 0);
		else
			BEH(aut->ss[0], i, 0)=BEH(aut_in->ss[0], ROT_reject, 0);
	}
	


	return aut;
}


GTA *left_rotate(GTA *aut, int variable) {
	bdd_manager *bddm_forx;
	GTA *aut1,*aut2;
	int i;
	unsigned *map,*classes,states;
	/* For sure, we run minimization */
	//aut1=gtaMinimize(aut);
	aut1=gtaCopy(aut);
	states=aut->ss[1].size;
	
	/* search for classes before remaping */
	classes=initial_partition_of_ss1(aut1);
	
	/* make space for n (n-number of states)  nondeterministic choices */
	/* Remap track 0 -> n, 1-> n+1 */
	map=(unsigned *) mem_alloc(sizeof(unsigned)*((2*NODE_WIDTH)+states));
        for (i=0; i<2*NODE_WIDTH; i=i+2) {
		map[i]=states+i;               
	}
	gtaReplaceIndices(aut1,map);

	/* run the rotation */
	aut2=run_left_rotate(aut1,(variable*2)+4+states,classes);
	gtaFree(aut1);
	//return aut2;
	/* project tracks 0-(states-1) - nondeterministic choices created by step 2b, and 5 */
	for(i=0;i<states;i++) {
		aut1=gtaQuotientAndProject(aut2,i,0);
		gtaFree(aut2);
		aut2=aut1;
	}
	/* remap tract n->0, n+1->1, ... */
        for (i=states; i<(2*NODE_WIDTH)+states; i=i+2) {
		map[i]=i-states;               
	}
	
	gtaReplaceIndices(aut2,map);

	/* minimize the result */
	aut1=gtaMinimize(aut2);
	gtaFree(aut2);

	/* clean memory */
	mem_free(map);
	mem_free(classes);
	return aut1;

}
