/* Adam Rogalewicz
 * Transducer operations (based on MONA GTA library) */

#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include "gta.h"
#include "../Mem/mem.h"

#define MAXVARS 1000

/* NODE_WIDTH includes the first two identifying bits */
extern unsigned NODE_WIDTH;

extern Guide guide;
extern char *my_global_variable;
extern bdd_manager *global_bddm;


GTA *apply_transducer_reverse(GTA *init,GTA *transd) {
  	GTA *pp,*pp1;
        int *map;
	int i;
	map=(int *) mem_alloc(sizeof(int)*2*NODE_WIDTH);
        for (i=0; i<2*NODE_WIDTH; i=i+2) {
		map[i]=i+1;               
	}
	pp=gtaCopy(init);
#ifdef VERBOSE
       	printf("Remapping...\n");
#endif
	gtaReplaceIndices(pp,map);
	pp1=gtaProduct(pp,transd,gtaAND);
	gtaFree(pp);
#ifdef VERBOSE
	printf("transducer applied ...\n");
#endif
	pp=gtaMinimize(pp1); 
	gtaFree(pp1);
	/* projection */
#ifdef VERBOSE
        printf("Projecting...\n");
#endif
	for (i=1; i<2*NODE_WIDTH; i=i+2) {
                pp1=gtaQuotientAndProject(pp,i,0);
                gtaFree(pp);
                /*pp=gtaMinimize(pp1);
		gtaFree(pp1);*/
		pp=pp1;
        }
	pp1=gtaMinimize(pp);
	gtaFree(pp);
	mem_free(map);
	return pp1;
}
/* ********************************************************************* */
GTA *projectAndRemapStore(GTA *a_orig, unsigned *map,int remap,int project_start) {
        unsigned i;
	GTA *a1, *a2;
	a1=gtaCopy(a_orig);
#ifdef VERBOSE
        printf("Projecting...\n");
#endif
	/* in normal transduction, project_start is equal to null */
        for (i=project_start; i<2*NODE_WIDTH; i=i+2) {
                a2=gtaQuotientAndProject(a1,i,0);
                gtaFree(a1);
                /*a1=gtaMinimize(a2);
                gtaFree(a2);*/
		a1=a2;
        }
	if (remap) {
#ifdef VERBOSE
        	printf("Remapping...\n");
#endif
        	gtaReplaceIndices(a1,map);
	}
	return a1;
}

/* ********************************************************************* */




GTA *apply_transducer_ll(GTA *init,GTA *transd,int remap,int project_start) {
	/* in normal transducer application, project_start==0 */
  	GTA *pp,*pp1;
        unsigned *map;
	unsigned i;
	map=(unsigned *) mem_alloc(sizeof(unsigned)*2*NODE_WIDTH);
        for (i=1; i<2*NODE_WIDTH; i=i+2) {
		map[i]=i-1;               
	}
	pp1=gtaProduct(init,transd,gtaAND);
#ifdef VERBOSE
	printf("transducer applied ...\n");
#endif
	pp=gtaMinimize(pp1); 
	gtaFree(pp1);
	pp1=projectAndRemapStore(pp,map,remap,project_start);
	gtaFree(pp);
	pp=gtaMinimize(pp1);
	gtaFree(pp1);
	mem_free(map);
	return pp;
}

GTA *apply_transducer(GTA *init,GTA *transd) {
	/* standart case */
	return apply_transducer_ll(init,transd,1,0);
}
GTA *apply_transducer_noremap(GTA *init,GTA *transd) {
	/* in this case, remaping function is not called */
	return apply_transducer_ll(init,transd,0,0);
}

GTA *get_source_of_transducer(GTA *init,GTA*transd) {
	return apply_transducer_ll(init,transd,0,1);
}

GTA *get_transducer_preimage(GTA *transd) {
	/* This function returns automaton describing all possible inputs 
	 * (i.e. input is tree for which exists an output) of the transducer 
	 *
	 * This is done by projecting on odd tracks */
	return projectAndRemapStore(transd, NULL, 0, 1);
}
