#include "gta.h"
#include "../Mem/mem.h"

#include "../main.h"
#include "../input.h"
#include "../transducer.h"
#include "history_stack.h"
#include "var_stack.h"
#include "gta_artmc_lib.h"
#include "../forward_run.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>




extern GTA *empty_aut;
extern char restart;

#ifdef STATISTICS
extern unsigned max_var_size;
#endif




GTA **var_state_init() {
	/* initialize array of potentionaly dangerous states 
	 * place bad configuration into the field corresponding to the line on which the bad conf was founded */
	GTA **var_state;
	int i;
	var_state=(GTA **) mem_alloc(sizeof(GTA *)*lines);
	for (i=0;i<lines;i++) var_state[i]=gtaCopy(empty_aut);
	return var_state;
}

void destroy_var_state(GTA **var_state) {
	int i;
	for (i=0;i<lines;i++)  gtaFree(var_state[i]); 
	mem_free(var_state);
}


GTA *prepare_seed(GTA *input,gtastack seed) {
	int i;
	GTA *aux,*aux2;
	aux=gtaCopy(input);
	for(i=0;i<=seed.no;i++) {
		aux2=apply_transducer(aux, seed.val[i]);
		gtaFree(aux);
		aux=aux2;
	}
	return aux;
}

void run_create_variants(GTA ** var_state, int start_line, GTA *seed, int amode) {
	int actual;
	char order;
	char *changed;
	hist_value act_conf;
	hist_stack var_history;
	GTA *apl_res, *aut1, *cfg;
	histstack_init(&var_history,100);	
	int i;
//	int debug;
	actual=start_line;
	order=0;
	restart=0;
	changed=(char *) mem_alloc(sizeof(char)*lines);
	for(i=0;i<lines;i++) changed[i]=0;
	changed[actual]=1;
	/* if actual is less than 0 -> end of computation */
	while (actual>=0) {
		printf("--------------------------------\nLine num: %i, order %i\n",actual,order);
		/* push actual configuration into a stack */
		act_conf.line=actual;
		act_conf.order=order;

		histstack_push(&var_history,act_conf);
		if (program[actual].type==0) {
			/* This line is equal to exit() in C
			 * no transition is performed */
			#ifdef VERBOSE
			printf("Command exit() - no transduction: line skiped\n");
			#endif
			find_next(&actual, &order, changed, &var_history,0);
			continue;
		}
		/* in the case of start_line, we apply the transducers on the seed. */
		if (actual==start_line) cfg=seed;
		else cfg=var_state[actual];

/*		save_automaton(cfg,"pokusXXX.gta");
		scanf("%i",&debug);*/
#ifdef STATISTICS
		printf("Act. var. size: %i, Max. var.size: %i\n",cfg->ss[1].size,max_var_size);
		if (max_var_size<cfg->ss[1].size)
			max_var_size=cfg->ss[1].size;
#endif

/*		save_automaton(cfg,"tmp_cfg.gta");
		scanf("%i",&debug);*/
		/* apply transducer / series of transducers 
		 * dependentely on actual, and order */
		if ((program[actual].type<=3)&&(order==0))
			apl_res=apply_transducer(cfg,program[actual].transd);
		if ((program[actual].type<=3)&&(order==1))
			apl_res=apply_transducer(cfg,program[actual].transd2);
		if (program[actual].type==4) 
			/* x=y.next */
			apl_res=apply_x_assign_ynext(actual,cfg);
		if (program[actual].type==5)  
			/* x.next=new */
			apl_res=apply_xnext_ass_new(actual,cfg,&var_history,0);
		if (program[actual].type==6)  
			/* x.next=new */
			apl_res=apply_free_x(actual,cfg);
		/* if variable restart is set -> stop main loop */
		if (restart) break;
#ifdef VERBOSE
		printf("Transducer applied\n");
#endif
		/* transducer(s) aplication finished */
		if (order==0) {
			if (!is_included(apl_res,var_state[program[actual].t])) {
				if (amode) aut1=abstract_and_union(apl_res,var_state[program[actual].t],program[actual].t);
				else aut1=gtaProduct(apl_res,var_state[program[actual].t],gtaOR);
				gtaFree(var_state[program[actual].t]);
				var_state[program[actual].t]=aut1;
				if (check_for_bad(program[actual].t,var_state[program[actual].t],&var_history,0 )) {
					/* in case of bad state, stop */
					restart=1;
					gtaFree(apl_res);
					break; /* stop main loop */
				}
				gtaFree(apl_res);
				actual=program[actual].t;
				changed[actual]=1;
				/* we do not continue from breakpoint, find some other place to continue */
				if (program[actual].mark & 4) find_next(&actual, &order, changed, &var_history,0);
				continue;
			}
			/* nothing new added - go back with history stack */
			gtaFree(apl_res);
			find_next(&actual, &order, changed, &var_history,0);
			continue;
		}
		/* this point is reachable only with order>0 */
		if (program[actual].type!=2) {
			/* this point is normaly unreachable !!! -> exit */
			printf("ERROR: reached forbiden combination of actual and order\n");
			exit(1);
		}
		changed[actual]=0;
		if (!is_included(apl_res,var_state[program[actual].f])) {
			if (amode) aut1=abstract_and_union(apl_res,var_state[program[actual].f],program[actual].f);
			else aut1=gtaProduct(apl_res,var_state[program[actual].f],gtaOR);
			gtaFree(var_state[program[actual].f]);
			gtaFree(apl_res);
			var_state[program[actual].f]=aut1;
			if (check_for_bad(program[actual].f,var_state[program[actual].f],&var_history,0)) {
				/* in case of bad state, stop  */
				restart=1;
				break; /* stop main loop */
			}
			actual=program[actual].f;
			order=0;
			changed[actual]=1;
			/* we do not continue from breakpoint, find some other place to continue */
			if (program[actual].mark & 4) find_next(&actual, &order, changed, &var_history,0);
			continue;
		}
		gtaFree(apl_res);
		/* nothing new added - go back with history stack */
		find_next(&actual, &order, changed, &var_history,0);
	}
	if (restart) {
		printf("ERROR: reached bad configuration within variance analysis ...\n...something is wrong, STOP");
		exit(1);
	}
	/* destroy the history stack */
	histstack_destroy(&var_history);

}

char *concat_strings(char *a,char *b) {
	char *out;
	out=(char *) mem_alloc (sizeof(char)*(strlen(a) + strlen(b) + 1) );
	strcpy(out,a);
	strcat(out,b);
	return out;
}

void check_variants_on_preds(GTA *var,  varpred_stack *vpreds, int pos, char *out, FILE *output) {
	GTA *aut,*aut1,*aut2,*tmp;
	char *str,str1[10];
	
	if (pos>vpreds->no) {
		printf("*******> %s\n",out);
		// all predicated applied
		fprintf(output,"%s\n",out);	
		return;
	}
	// decreasing
	tmp=gtaProduct(var, vpreds->val[pos].decr,gtaAND);
	aut=gtaMinimize(tmp);
	gtaFree(tmp);
	if(!empty(aut)) {
		snprintf(str1,10,"%i:> ",pos);
		str=concat_strings(out,str1);
		check_variants_on_preds(aut, vpreds, pos+1, str, output);
		mem_free(str);		
	}
	// equal
	tmp=gtaProduct(var, vpreds->val[pos].eq,gtaAND);
	aut1=gtaMinimize(tmp);
	gtaFree(tmp);
	if(!empty(aut1)) {
		snprintf(str1,10,"%i:= ",pos);
		str=concat_strings(out,str1);
		check_variants_on_preds(aut1, vpreds, pos+1, str, output);
		mem_free(str);		
	}
	// increase
	tmp=gtaProduct(var, vpreds->val[pos].inc,gtaAND);
	aut2=gtaMinimize(tmp);
	gtaFree(tmp);
	if(!empty(aut2)) {
		snprintf(str1,10,"%i:< ",pos);
		str=concat_strings(out,str1);
		check_variants_on_preds(aut2, vpreds, pos+1, str, output);
		mem_free(str);		
	}
	
	// not in relation
	gtaNegation(aut);
	gtaNegation(aut1);
	gtaNegation(aut2);
	tmp=gtaProduct(var, aut,gtaAND);
	gtaFree(aut);
	aut=gtaMinimize(tmp);
	gtaFree(tmp);
	tmp=gtaProduct(aut, aut1,gtaAND);
	gtaFree(aut);
	gtaFree(aut1);
	aut=gtaMinimize(tmp);
	gtaFree(tmp);
	tmp=gtaProduct(aut, aut2,gtaAND);
	gtaFree(aut);
	gtaFree(aut2);
	aut=gtaMinimize(tmp);
	gtaFree(tmp);
	if(!empty(aut)) {
		snprintf(str1,10,"%i:? ",pos);
		str=concat_strings(out,str1);
		check_variants_on_preds(aut, vpreds, pos+1, str, output);
		mem_free(str);		
	}
	gtaFree(aut);

}


void create_relational_automaton(var_stack *variants, varpred_stack *vpreds) {
	int i;
	FILE *output;
	variant var;
	output=fopen("rel-automaton.txt","w");
	printf("Creating relational automaton ...\n");
	for(i=0; i<=vpreds->no;i++) {
		fprintf(output,"Variance predicate no. %i: %s\n",i,vpreds->val[i].name);
	}

	for (i=0; i<=variants->no;i++) {
		var=variants->val[i];
		fprintf(output,"-------------------------\n%i -> %i\n",var.line_from,var.line_to);
		printf("Processing variant %i -> %i\n",var.line_from,var.line_to);
		check_variants_on_preds(var.var, vpreds, 0, "",output);
	}
	fclose(output);
	printf("...done\n");
}


void save_variants(var_stack *variants) {
	int i;
	char name[50];
	variant var;
	printf("Saving variants ...\n");
	for (i=0; i<=variants->no;i++) {
		var=variants->val[i];
		snprintf(name,50,"variant-%i-%i.gta",var.line_from,var.line_to);
		save_automaton(var.var,name);
	}
	printf("...done\n");
}


void create_variants(char amode,char save,char vdesc) {
	int i,j;
	GTA **var_state,*seed;
	var_stack variants;
	variant var;
	gtastack seed_aut;
	varpred_stack vpreds;
	seed_aut=read_seed(vdesc);
	load_variance_predicates(&vpreds,vdesc);

	varstack_init(&variants,50); // init the stack to store variants
	printf("Creating variants ...\n");

	for (i=0; i<lines;i++) {
		/* check whether cutpoint is set for line i */
		if (program[i].mark & 4) {
			printf("Starting to create variants on breakpoint %i ...\n",i);
			seed=prepare_seed(state[i],seed_aut);
			var_state=var_state_init();
			run_create_variants(var_state, i, seed, amode);
			// save the created variants into the stack
			for (j=0; j<lines;j++) 
				if ((program[j].mark & 4)&&(!empty(var_state[j]))) {
					var.line_from=i;
					var.line_to=j;
					var.var=gtaCopy(var_state[j]);
					printf("Created variant %i -> %i\n",i,j);
					varstack_push(&variants,var);
				}
			destroy_var_state(var_state);
			gtaFree(seed);
			printf("... done\n");
		}
	}

	// save variants (if needed)
	if (save) save_variants(&variants);

	// create the relational automaton according to the variance predicates
	create_relational_automaton(&variants, &vpreds);


	// destroy the stack of variants
	varstack_reset(&variants);
	varstack_destroy(&variants);
}
