import sys

def error_msg():
	print "An internal ERROR occured in the processing..."
	sys.exit()

def error_msg_expl(msg):
	print "ERROR: ", msg
	sys.exit()

#=================================================================================

# This function copies a part of the description of a node starting at first_idx and with the given length.
# (i*2) indexes 0,2,4,... -- the input bits that are copied to 1,3,5,...

def copy_n_elements(bdd,index,length,first_idx):
	for i in range(length):
		bdd.append(((i*2)+first_idx,index+1,index+2))
		bdd.append(((i*2)+first_idx+1,index+3,0))
		bdd.append(((i*2)+first_idx+1,0,index+3))
		index=index+3
	return (bdd,index)

#=================================================================================

# This function copies a part of a node starting at first_idx and with the length given by number
# apart from the value of x which is set to 0.

def copy_n_elements_unset_x(bdd,index,length,first_idx,x):
	for i in range(length):
		if i==x:
			bdd.append(((i*2)+first_idx+1,index+1,0)) # We may ignore the current value of x at idx.
			index=index+1
		else:
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+3,0))
			bdd.append(((i*2)+first_idx+1,0,index+3))
			index=index+3
	return (bdd,index)

#---------------------------------------------------------------------------

# This function copies a part of a node starting at first_idx and with the length given by number
# apart from the value of x which is set to 1.

def copy_n_elements_set_x(bdd,index,length,first_idx,x):
	for i in range(length):
		if i==x:
			bdd.append(((i*2)+first_idx+1,0,index+1)) # We may ignore the current value of x at idx.
			index=index+1
		else:
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+3,0))
			bdd.append(((i*2)+first_idx+1,0,index+3))
			index=index+3
	return (bdd,index)

#================================================================================
# This function copies a part of a node starting at first_idx and with the length given by number
# value of position x is checked for 'xval'

def copy_n_elements_check_x_to_val(bdd,index,length,first_idx,x,xval):
	for i in range(length):
		if i==x:
			if xval:
				bdd.append(((i*2)+first_idx,0,index+1)) # x must be 1
				bdd.append(((i*2)+first_idx+1,0,index+2))
			else:
				bdd.append(((i*2)+first_idx,index+1,0)) # x must be 1
				bdd.append(((i*2)+first_idx+1,index+2,0))
			index=index+2
		else:
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+3,0))
			bdd.append(((i*2)+first_idx+1,0,index+3))
			index=index+3
	return (bdd,index)

# This function copies a part of a node starting at first_idx and with the length given by number
# value of position x is checked for '1'

def copy_n_elements_check_x(bdd,index,length,first_idx,x):
	return copy_n_elements_check_x_to_val(bdd,index,length,first_idx,x,1)

#=================================================================================

# The function copies the node but remembers the value of x in the output state.
# The result contains lists of tuples (index,position) where to patch the continuation for x=0 and x=1.
# So, the result is: (bdd,index,patch_x_unset,patch_x_set).

def copy_n_elements_and_remember_x(bdd,index,len,first_idx,x):
        result=([],0,[],[]) # just some dummy value for the python's type-checker to be happy
	for i in range(len):
                if i<x: # just copy
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+3,0))
			bdd.append(((i*2)+first_idx+1,0,index+3))
			index=index+3
		elif (i==x) and (x==len-1): # copy and prepare the result 
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,-2,0)) # (note: -2 will later be changed)
			bdd.append(((i*2)+first_idx+1,0,-2)) # (note: -2 will later be changed)
			index=index+3
                        result=(bdd,index,[(index-2,1)],[(index-1,2)]) # the result contains the final triples for x==0 and x==1
		elif (i==x) and (x<len-1): # copy and fork according to the value of x
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+3,0))
			bdd.append(((i*2)+first_idx+1,0,index+6)) # now, we will always have 2x 3 triples (for x==0 and x==1)
			index=index+3
		else: # we have to make the copies twice
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+6,0)) # for i==len-1, the values will be re-defined later
			bdd.append(((i*2)+first_idx+1,0,index+6))
			index=index+3
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+6,0))
			bdd.append(((i*2)+first_idx+1,0,index+6))
			index=index+3
                        if i==len-1:
                                result=(bdd,index,[(index-5,1),(index-4,2)],[(index-2,1),(index-1,2)])
	return result
        # end of copy_n_elements_and_remember_x

#=================================================================================

# The function copies the node but remembers whether x=y, x<>y, or x & y are unset in the output state.
# The result contains lists of tuples (index,position) where to patch the continuation for the above cases.
# So, the result is: (bdd,index,patch_init,patch_xneqy,patch_xeqy).

def copy_n_elm_and_remember_x_eq_y(bdd,index,len,first_idx,x,y):
        if x==y:
               error_msg_expl("Trying to compare x with x? Not supported: leave it out...") 
        if x>y: # we want x<y
                z=x
                x=y
                y=z
        result=([],0,[],[],[]) # just some dummy value for the python's type-checker to be happy
	for i in range(len):
                if i<x: # just copy
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+3,0))
			bdd.append(((i*2)+first_idx+1,0,index+3))
			index=index+3
		elif i==x: # copy and fork according to the value of x
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+3,0))
			bdd.append(((i*2)+first_idx+1,0,index+6)) # now, we will always have 2x 3 triples (for x==0 and x==1)
			index=index+3
		elif (x<i) and (i<y): # we have to make the copies twice
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+6,0))
			bdd.append(((i*2)+first_idx+1,0,index+6))
			index=index+3
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+6,0))
			bdd.append(((i*2)+first_idx+1,0,index+6))
			index=index+3
                elif (i==y) and (y==len-1): # copy and prepare three results: both x and y unset, x<>y, x==y
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,-2,0)) # both x unset and y unset
			bdd.append(((i*2)+first_idx+1,0,-2)) # x<>y
			index=index+3
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,-2,0)) # x<>y 
			bdd.append(((i*2)+first_idx+1,0,-2)) # x==y
			index=index+3
                        result=(bdd,index,[(index-5,1)],[(index-4,2),(index-2,1)],[(index-1,2)])
                elif (i==y) and (y<len-1): # copy and fork to three branches: both x and y unset, x<>y, x==y
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+6,0)) # both x unset and y unset
			bdd.append(((i*2)+first_idx+1,0,index+9)) # x<>y
			index=index+3
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+6,0)) # x<>y
			bdd.append(((i*2)+first_idx+1,0,index+9)) # x==y
			index=index+3
		else: # we have to make the copies three times now
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+9,0)) # for i==len-1, the values will be re-defined later
			bdd.append(((i*2)+first_idx+1,0,index+9))
			index=index+3
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+9,0))
			bdd.append(((i*2)+first_idx+1,0,index+9))
			index=index+3
			bdd.append(((i*2)+first_idx,index+1,index+2))
			bdd.append(((i*2)+first_idx+1,index+9,0))
			bdd.append(((i*2)+first_idx+1,0,index+9))
			index=index+3
                        if i==len-1:
                                result=(bdd,index,[(index-8,1),(index-7,2)],[(index-5,1),(index-4,2)],[(index-2,1),(index-1,2)])
	return result
        # end of copy_n_elm_and_remember_x_eq_y

#=================================================================================

# This function copies the node and sets the position of y to value yval (used in nodes of undef pointers when y:=x).
# Then, the returning state depends on the original value of x.
# if y>=number -> nothing is set/unset
# If 0 is on position x -> ret_state0 is returned.
# If 1 is on position x -> ret_state1 is returned (typically this means an error: y:=undef...).
# Note: the situation, where x==y also works, but !!! BE CAREFULL with x==y !!!
# (We return the state acc. the original value that's may be changed now...)
# It can be used e.g. in the manual creation of pointer descriptors.


def copy_node_return_state_dep_on_x_low_level(bdd,index,number,idx,x,y,ret_state0,ret_state1,yval):
	if number:
		if number==x: # the computation is forked into the possibilities x==0 and x==1
			bdd.append((idx,index+1,-2))
			# we now consider x==0
			aux=index
			if (x==y)and(yval==1): bdd.append((idx+1,0,index+2)) #x==y && yval==1 -> we set x to value 1.
			else: bdd.append((idx+1,index+2,0)) # (x!=y) or (xval==0) -> preserve X's value
			index=index+2
			(bdd,index)=copy_node_return_state_dep_on_x_low_level(bdd,index, number-1, idx+2, x, y, ret_state0, ret_state1,yval)
			# we now consider x==1
			bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
			if (x==y)and(yval==0): bdd.append((idx+1,index+1,0)) #x==y && yval==0 -> we set x to value 0.
			else: bdd.append((idx+1,0,index+1)) # (x!=y) or (xval==1) -> preserve X's value
			index=index+1
			# x was set to 1 -> return state is changed to ret_state1 (and it will stay so as the position of x is unique).
			(bdd,index)=copy_node_return_state_dep_on_x_low_level(bdd,index, number-1, idx+2, x, y, ret_state1, ret_state0,yval)
		elif number==y: # y is set to yval (and so we don't add anything for idx and ignore its value).
			if yval: bdd.append((idx+1,0,index+1))
			else: bdd.append((idx+1,index+1,0))
			index=index+1
			(bdd,index)=copy_node_return_state_dep_on_x_low_level(bdd,index, number-1, idx+2, x, y, ret_state0, ret_state1,yval)
		else: # just copy
			bdd.append((idx,index+1,index+2))
			bdd.append((idx+1,index+3,0))
			bdd.append((idx+1,0,index+3))
			index=index+3
			(bdd,index)=copy_node_return_state_dep_on_x_low_level(bdd,index, number-1, idx+2, x, y, ret_state0, ret_state1,yval)
	else: # we finished the iteration
		bdd.append((-1,ret_state0,0))
		index=index+1
	return (bdd,index)

#---------------------------------------------------------------------------
# The first call sets y to '0', the second one sets it to '1' (this is NOT influenced by y!)
# The return state depends on the original value of x.

def copy_node_return_state_dep_on_x_unset_y(bdd,index,number,first_idx,x,y,ret_state0,ret_state1):
	return copy_node_return_state_dep_on_x_low_level(bdd,index, number, first_idx, number-x, number-y, ret_state0, ret_state1,0)
def copy_node_return_state_dep_on_x_set_y(bdd,index,number,first_idx,x,y,ret_state0,ret_state1):
	return copy_node_return_state_dep_on_x_low_level(bdd,index, number, first_idx, number-x, number-y, ret_state0, ret_state1,1)
        # number-x/y is used as the loop variable which is being decreased whereas idx is being increased.
	# for number=5, x=1 (i.e. 2nd position in 0,1,...), we reach 5-1=4 after one decrease which matches one increase from 0 in idx.
	
def copy_node_return_state_dep_on_x(bdd,index,number,first_idx,x,ret_state0,ret_state1):
	return copy_node_return_state_dep_on_x_low_level(bdd,index, number, first_idx, number-x, number+1, ret_state0, ret_state1,1)
	
#=================================================================================
# This function sets the position of 'x' to the value of the position of 'y'

def copy_n_elements_set_x_to_y_lowlevel(bdd,index,number,idx,x,y,ret_state,mode):
# mode: 0 - nothing is known
#	1 - x set to 0
#	2 - x set to 1
#	3 - y is 0
#	4 - y is 1
	if number:
		if number==x:
			# The value at position idx (the input position for x) can be ignored -- it'll be set wrt. y.
			if mode==0: # We guess whether the value of x will be 0 or 1 -- has to be checked later.
				bdd.append((idx+1,index+1,)) 
				# We guess x is going to be 0.
				aux=index
				index=index+1
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,1) # x set to 0
				# We guess x is going to be 1.
				bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,2) # x set to 1
			elif mode==3: # x has to be set to 0
				bdd.append((idx+1,index+1,0))
				index=index+1
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,0) 
			elif mode==4: # x has to be set to 1
				bdd.append((idx+1,0,index+1))
				index=index+1
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,0) 
			else: error_msg() # normally, this position is unreachable
		elif number==y:
			# We remember or check the guessed value of y.
			if mode==0:
				bdd.append((idx,index+1,))
				# We remember 0.
				aux=index
				bdd.append((idx+1,index+2,0))
				index=index+2
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,3) # y is 0
				# We remember 1.
				bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
				bdd.append((idx+1,0,index+1))
				index=index+1
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,4) # y is 1
			elif mode==1: # it is necessary to check that y is 0
				bdd.append((idx,index+1,0))
				bdd.append((idx+1,index+2,0))
				index=index+2
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,0) 
			elif mode==2: # it is necessary to check that y is 1
				bdd.append((idx,0,index+1))
				bdd.append((idx+1,0,index+2))
				index=index+2
				(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,0) 
			else: error_msg() # normally this position is unreachable
		else: # just copy
			bdd.append((idx,index+1,index+2))
			bdd.append((idx+1,index+3,0))
			bdd.append((idx+1,0,index+3))
			index=index+3
			(bdd,index)=copy_n_elements_set_x_to_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state,mode)
	else: # number has reached 0 -- we are finished: add the return state
		bdd.append((-1,ret_state,0))
		index=index+1
	return (bdd,index)

#---------------------------------------------------------------------------

def copy_n_elements_set_x_to_y(bdd,index,number,first_idx,x,y,ret_state):
	return copy_n_elements_set_x_to_y_lowlevel(bdd,index,number, first_idx, number-x,number-y, ret_state, 0)
        # number-x/y is used as the loop variable which is being decreased whereas idx is being increased.
	# for number=5, x=1 (i.e. 2nd position in 0,1,...), we reach 5-1=4 after one decrease which matches one increase from 0 in idx.


#=================================================================================


# This function copies the node and sets x to the value of y.
# if y>=number -> nothing is set/unset.
# If 0 is on position y -> ret_state0 is returned.
# If 1 is on position y -> ret_state1 is returned.

def copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number,idx,x,y,ret_st0,ret_st1,mode):
# mode: 0 - nothing is known
#	1 - x guessed to be 0
#	2 - x guessed to be 1
#	3 - y is 0
#	4 - y is 1
	if number:
                if (number==x) and (x==y):
			aux=index
			bdd.append((idx,index+1,-2))
			# We remember 0.
			bdd.append((idx+1,index+2,0))
			index=index+2
			(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) # y is 0
			# We remember 1.
			bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
			bdd.append((idx+1,0,index+1))
			index=index+1
			(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) # y is 1
		elif number==x:
			# The value at position idx (the input position for x) can be ignored -- it'll be set wrt. y.
			if mode==0: # We guess whether the value of x will be 0 or 1 -- has to be checked later.
				aux=index
				bdd.append((idx+1,index+1,-2)) 
				index=index+1
				# We guessed x is going to be 0.
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,1) # x set to 0
				# We guessed x is going to be 1.
				bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,2) # x set to 1
			elif mode==3: # x has to be set to 0
				bdd.append((idx+1,index+1,0))
				index=index+1
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) # y is 0
			elif mode==4: # x has to be set to 1
				bdd.append((idx+1,0,index+1))
				index=index+1
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) # y is 1
			else: error_msg() # normally, this position is unreachable
		elif number==y:
			# We remember or check the guessed value of y.
			if mode==0:
				aux=index
				bdd.append((idx,index+1,-2))
				# We remember 0.
				bdd.append((idx+1,index+2,0))
				index=index+2
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) # y is 0
				# We remember 1.
				bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
				bdd.append((idx+1,0,index+1))
				index=index+1
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) # y is 1
			elif mode==1: # it is necessary to check that y is 0
				bdd.append((idx,index+1,0))
				bdd.append((idx+1,index+2,0)) 
				index=index+2
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) # y is 0
			elif mode==2: # it is necessary to check that y is 1
				bdd.append((idx,0,index+1))
				bdd.append((idx+1,0,index+2))
				index=index+2
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) # y is 1
			else: error_msg() # normally this position is unreachable
		else: # just copy
			bdd.append((idx,index+1,index+2))
			bdd.append((idx+1,index+3,0))
			bdd.append((idx+1,0,index+3))
			index=index+3
			(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,mode)
	else: # number has reached 0 -- we are finished: add the return state
		if (mode==3):
                        bdd.append((-1,ret_st0,0)) # y was 0
                else:
                        bdd.append((-1,ret_st1,0)) # y was 1
		index=index+1
	return (bdd,index)
        # end of copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level

#---------------------------------------------------------------------------

def copy_node_Ret_st_dep_on_y_Set_x_to_y(bdd,index,number,first_idx,x,y,ret_st0,ret_st1):
	return copy_node_Ret_st_dep_on_y_Set_x_to_y_Low_level(bdd,index,number,first_idx,number-x,number-y,ret_st0,ret_st1,0)
        # number-x/y is used as the loop variable which is being decreased whereas idx is being increased.
	# for number=5, x=1 (i.e. 2nd position in 0,1,...), we reach 5-1=4 after one decrease which matches one increase from 0 in idx.


#=================================================================================

# This function copies the node, sets x to the value of y, and unsets y.
# if y>=number -> nothing is set/unset.
# If 0 is on position y -> ret_state0 is returned.
# If 1 is on position y -> ret_state1 is returned.
# Note: if x==y, the value is set to to the original value of y (not always 0).


def copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number,idx,x,y,ret_st0,ret_st1,mode):
# mode: 0 - nothing is known
#	1 - x guessed to be 0
#	2 - x guessed to be 1
#	3 - y is 0
#	4 - y is 1
	if number:
                if (number==x) and (x==y):
			aux=index
			bdd.append((idx,index+1,-2))
			# We remember 0.
			bdd.append((idx+1,index+2,0))
			index=index+2
			(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) # y is 0
			# We remember 1.
			bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
			bdd.append((idx+1,0,index+1))
			index=index+1
			(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) # y is 1
		elif number==x:
			# The value at position idx (the input position for x) can be ignored -- it'll be set wrt. y.
			if mode==0: # We guess whether the value of x will be 0 or 1 -- has to be checked later.
				aux=index
				bdd.append((idx+1,index+1,-2)) 
				index=index+1
				# We guessed x is going to be 0.
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,1) # x set to 0
				# We guessed x is going to be 1.
				bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,2) # x set to 1
			elif mode==3: # x has to be set to 0
				bdd.append((idx+1,index+1,0))
				index=index+1
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) 
			elif mode==4: # x has to be set to 1
				bdd.append((idx+1,0,index+1))
				index=index+1
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) 
			else: error_msg() # normally, this position is unreachable
		elif number==y:
			# We remember or check the guessed value of y and set y to zero.
			if mode==0:
				aux=index
				bdd.append((idx,index+1,-2))
				# We remember 0.
				bdd.append((idx+1,index+2,0)) # y is already zero
				index=index+2
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) # y is 0
				# We remember 1 and set y to 0.
				bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
				bdd.append((idx+1,index+1,0)) # set y to zero
				index=index+1
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) # y is 1
			elif mode==1: # it is necessary to check that y is 0
				bdd.append((idx,index+1,0))
				bdd.append((idx+1,index+2,0)) # y is already 0
				index=index+2
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,3) 
			elif mode==2: # it is necessary to check that y is 1 and set y to 0
				bdd.append((idx,0,index+1))
				bdd.append((idx+1,index+2,0)) # set y to zero
				index=index+2
				(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,4) 
			else: error_msg() # normally this position is unreachable
		else: # just copy
			bdd.append((idx,index+1,index+2))
			bdd.append((idx+1,index+3,0))
			bdd.append((idx+1,0,index+3))
			index=index+3
			(bdd,index)=copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number-1,idx+2,x,y,ret_st0,ret_st1,mode)
	else: # number has reached 0 -- we are finished: add the return state
		if (mode==3):
                        bdd.append((-1,ret_st0,0)) # y was 0
                else:
                        bdd.append((-1,ret_st1,0)) # y was 1
		index=index+1
	return (bdd,index)
        # end of copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level

#---------------------------------------------------------------------------

def copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y(bdd,index,number,first_idx,x,y,ret_st0,ret_st1):
	return copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y_Low_level(bdd,index,number,first_idx,number-x,number-y,ret_st0,ret_st1,0)
        # number-x/y is used as the loop variable which is being decreased whereas idx is being increased.
	# for number=5, x=1 (i.e. 2nd position in 0,1,...), we reach 5-1=4 after one decrease which matches one increase from 0 in idx.

#=================================================================================

# This function sets x wrt. y:
# if y==0 -> x will be set to 0
# if y==1 -> value of x will be nondeterministically chosen
#
# ret_state0 => x was set to 0
# ret_state1 -> x was set to 1

def copy_n_elements_guess_x_dep_on_y_lowlevel(bdd,index,number,idx,x,y,ret_state0, ret_state1,mode):
# mode: 0 - nothing is known
#	1 - x set to 0 -> y can be from {0,1}
#	2 - x set to 1 -> y has to be 1
	if number:
		if number==x:
			# The value at position idx (the input position for x) can be ignored -- it'll be set wrt. y.
			if mode==0: # We guess whether the value of x will be 0 or 1 -- has to be checked later.
				bdd.append((idx+1,index+1,-2)) 
				# We guess x is going to be 0.
				aux=index
				index=index+1
				(bdd,index)=copy_n_elements_guess_x_dep_on_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state0, ret_state1,1) # x set to 0
				# We guess x is going to be 1.
				bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
				(bdd,index)=copy_n_elements_guess_x_dep_on_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state0, ret_state1,2) # x set to 1
			else: error_msg()# this position is unreachable in case x < y
		elif number==y:
			# We remember  the guessed value of x.
			if mode==1: # y can be 0, or 1.
				bdd.append((idx,index+1,index+2))
				bdd.append((idx+1,index+3,0))
				bdd.append((idx+1,0,index+3))
				index=index+3
				(bdd,index)=copy_n_elements_guess_x_dep_on_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state0, ret_state1,0) 
			elif mode==2: # it is necessary to check that y is 1
				bdd.append((idx,0,index+1))
				bdd.append((idx+1,0,index+2))
				index=index+2
				# return state is ret_state1 in this case
				(bdd,index)=copy_n_elements_guess_x_dep_on_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state1, ret_state0,0) 
			else: error_msg() # this position is unreachable in case x < y
		else: # just copy
			bdd.append((idx,index+1,index+2))
			bdd.append((idx+1,index+3,0))
			bdd.append((idx+1,0,index+3))
			index=index+3
			(bdd,index)=copy_n_elements_guess_x_dep_on_y_lowlevel(bdd,index,number-1,idx+2,x,y,ret_state0, ret_state1,mode)
	else: # number has reached 0 -- we are finished: add the return state
		bdd.append((-1,ret_state0,0))
		index=index+1
	return (bdd,index)

def copy_n_elements_guess_x_dep_on_y(bdd,index,number,first_idx,x,y,ret_state0, ret_state1):
	return copy_n_elements_guess_x_dep_on_y_lowlevel(bdd,index,number, first_idx, number-x,number-y, ret_state0, ret_state1, 0)
        # number-x/y is used as the loop variable which is being decreased whereas idx is being increased.
	# for number=5, x=1 (i.e. 2nd position in 0,1,...), we reach 5-1=4 after one decrease which matches one increase from 0 in idx.


#=================================================================================

# Check whether we are at line_1 and change it to line_2

# !!! BE CAREFUL: Both of the line nubmers must be of the same length, and the length must be the maximum possible one.
# Should be improved in the future by providing in addition a constant giving the maximum length.

def change_line_num(bdd,index,line_1,line_2):
	for i in range(len(line_1)):
		if (line_1[i]=="1"):
			bdd.append((i*2,0,index+1))
		else:
			bdd.append((i*2,index+1,0))
		if (line_2[i]=="1"):
			bdd.append(((i*2)+1,0,index+2))
		else:
			bdd.append(((i*2)+1,index+2,0))
		index=index+2
	return (bdd,index)

#---------------------------------------------------------------------------

# Check whether we are at line_1 and change it to ERROR (111111....)

def change_line_num_to_err(bdd,index,line_1):
	for i in range(len(line_1)):
		if (line_1[i]=="1"):  
			bdd.append((i*2,0,index+1))
		else:
			bdd.append((i*2,index+1,0))
		bdd.append(((i*2)+1,0,index+2))
		index=index+2
	return (bdd,index)

#=================================================================================

def gen_header(size):
	"""Generates the header of a file --- size means the size of the state spaces 0 and 1"""
	out=[]
	out.append("MONA GTA\n")
	out.append("number of variables: 2\n")
	out.append("state spaces: 3\n")
	out.append("universes: 2\n")
	out.append("state space sizes: %i %i 1\n" % (size,size) )
	# results of state space 0:
	# 0 - reject; 1 - accept; 2-... don't care
	out.append("final: -1 1")
	for i in range(size-2):
		out.append(" 0")
	out.append("\n")
	out.append("guide:\n")
	out.append(" <hat> 1 2\n")
	out.append(" <univ> 1 1\n")
	out.append(" <dummy> 2 2\n")
	out.append("types: 0\n")
	out.append("universes:\n")
	out.append(" <univ> 0\n")
	out.append(" <dummy> 1\n")
	out.append("variable orders and state spaces:\n")
	out.append(" P 2: 1\n")
	out.append(" Q 2: 1\n")
	return out

#---------------------------------------------------------------------------

def gen_end():
	"""Generates the end of a file -- always the same"""
	out=[]
	out.append("\nstate space 2:\n")
	out.append(" initial state: 0\n")
	out.append(" bdd nodes: 1\n")
	out.append(" behaviour:\n")
	out.append("  0\n")
	out.append(" bdd:\n")
	out.append("  -1 0 0\n")
	out.append("\n")
	out.append("end\n")
	return out

#---------------------------------------------------------------------------

def gen_state_space_0(list):
	"""Generate state space 0 -- list contains numbers 0 - reject and 1 - accept,
	   it is recomended to set first item to 0 - reject, otherwise the empty tree will be accepted"""
	out=[]
	out.append("\nstate space 0:\n")
	out.append(" initial state: 0\n")
	out.append(" bdd nodes: 2\n")
	out.append(" behaviour:\n")
	for i in list:
		out.append("  %i\n" %i)
	out.append(" bdd:\n")
	out.append("  -1 0 0\n")
	out.append("  -1 1 0\n\n")
	return out

#---------------------------------------------------------------------------

# Returns an element of the table or default if the appropriate entry is undefined
def get_from_table(i,j,table, default):
	for k in table:
		if ((k[0]==i) and (k[1]==j)): return (" %i"% k[2])
	return " %s" % default

#---------------------------------------------------------------------------
	
def gen_gta_file(bdd,table, fin_states,filename):
# this function generates a gta file:
#   bdd - a list of bdd nodes
#   table - a list corresponding to the behaviour in a gta file (the basic transition table with links to the bdd)
#   fin_states - a list of final states
#   filename - the filename
	# first, it is necessary to find the number of states
	# we find the highest state number in bdd and table, and then increase it by 1 (states are numbered from 0)
	state_num=0
	for i in bdd:
		if i[0]==-1: state_num=max(state_num,i[1])
	for i in table:
		state_num=max(state_num, i[0], i[1])
	state_num=state_num+1
	# output is generated in list "out", each element corresponds to one line
	out=gen_header(state_num)
	# create state space 0
	states=[]
	for i in range(state_num):
		if i in fin_states: states.append(1)
		else: states.append(0)
	out.extend(gen_state_space_0(states))
	# run function for the creation of bdds and behaviour of state space 1
	out.append("\nstate space 1:\ninitial state: 0\n bdd nodes: %i\n behaviour:\n" % len(bdd)) 
	str=" "
	# create the behaviour section
	for i in range(state_num):
		for j in range(state_num):
			str=str+get_from_table(i,j,table,"0") # BDD node 0 represents error (used when no other outcome is defined)
		str=str+"\n"
	out.append(str)
	out.append(" bdd:\n")
	str=""
	# create the bdd nodes section
	for i in bdd:
		str=str+(" %i %i %i\n"% i)
	out.append(str)
	out.extend(gen_end())
	# everything is generated -> we write the file
	file=open(filename,"w")
	file.writelines(out)
	file.close()
	# end of gen_gta_file

#=================================================================================
