import functions
import sys


def descr_1up(node_width, pointer_num, marker):
	'''generate routing expression 1up'''

# node_width - maximal size of node
# pointer_num -- the number of pointer variables
# marker  -- marker required in destination


# We use the following states in the transducer:
#   0 - reading the init symbols
#   1 - normal processing of the main tree structure
#   2 - read and deleted source marker
#   3 - set destination marker - read rest of the tree (we don't care about contain of nodes)
#   4 - reject (malformed, not the right line number, etc.) 

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,4,0))#0
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#   - empty nodes (00) till the first pointer description
	#   - the first pointer descriptor (01)
	#****************************************
	table.append((0,0,1))
	bdd.append((0,2,0))#1 -- the first bit of the type of the input node
	bdd.append((1,3,0))#2 -- the first bit of the type of the output node
	bdd.append((2,4,6))#3 -- if 1, we found the 1st pointer descriptor => go to line 6
	bdd.append((3,5,0))#4
	bdd.append((-1,0,0))#5
	index=index+5
	# a pointer descriptor (01)
	bdd.append((3,0,index+1))#6
	index=index+1
	descr_bdd=index # in the below case a pointer descriptor being read, we continue from here
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,1,0))
	index=index+1

	#****************************************
	#   - a memory node / a top memory node (1x)
	#   - a right-most pointer descriptor (01)
	#****************************************
	table.append((1,1,index))	
	table.append((0,1,index))
	table.append((1,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	index=index+1
	# we don't care if it is tom/normal memory node
	# here, we are checking for 'from' marker -> first marker - on position after all pointers.
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-1,2,pointer_num+1,pointer_num+1,1,2)
	
	# ***************************************
	# left state=0/1, right state=2
	#     - a pointer descriptor
	# ***************************************
	table.append((1,2,index))
	table.append((0,2,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,index+4))
	index=index+4
	descr_bdd2=index # in the below case a pointer descriptor being read, we continue from here
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,2,0))
	index=index+1

	# ***************************************
	# left state=2, right state=0/1
	#     - a pointer descriptor
	# ***************************************
	table.append((2,1,index))
	table.append((2,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd2)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node = we are going to set marker 'to' - it is destionation of this pointer descriptor
	# it is necessary to check whether marker is set.
	bdd.append((1,0,index+1))
	index=index+1
	(bdd,index)=functions.copy_node_return_state_dep_on_x_set_y(bdd,index,node_width-1,2,pointer_num+1+marker,pointer_num+1,4,3)
	
	# ****************************************
	# left state=3, right state=0/1:
	# left state=0/1, right state=3:
	#   - read of a rest of the tree
	# ****************************************
	table.append((3,0,index))
	table.append((3,1,index))
	table.append((1,3,index))
	table.append((0,3,index))
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)	
	bdd.append((-1,3,0))
	index=index+1
	return(bdd, table, [3])


#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_gen_descr_1up():
	"""test x:=null"""
	(bdd,table,acc_l)=descr_1up(13,4,2)
	functions.gen_gta_file(bdd,table, acc_l,"descr_1up.gta")

#test_gen_descr_1up()


#=================================================================================

