import functions
import sys


def if_x_eq_null(node_width,pointer_num,x,line_orig,line_goto_1,line_goto_2):
	'''if (x==null) goto line1; else goto line2;'''

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# pointer_num -- the number of pointer variables
# x           -- the number of the pointer variable being tested

# line_orig, line_goto_1 and line_goto_2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
#   0 - not read any top memory node yet
#   1 - read a top memory node
#   2 - read the null section WITH x SET (NULL)
#   3 - read the null section WITH x NOT SET (NOT NULL)
#   4 - read the undef section for x NULL
#   5 - read the undef section for x NOT NULL WITH X NOT UNDEFINED
#   6 - read the undef section for x NOT NULL WITH X UNDEFINED
#   7 - accept: the line has been successfully changed (either to line 1 or 2 or to ERROR)
#   8 - reject: malformed, not the right line number, etc. 

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,8,0)) #0
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#   - non-top-memory nodes (~11) till the first top memory node
	#   - the first top memory node (11)
	#****************************************
	table.append((0,0,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
        index=index+3
        index_bak=index
	bdd.append((3,-2,0)) #4 - 00 / 10 (note: -2 will later be changed)
	bdd.append((3,0,-2)) #5 - 01 (note: -2 will later be changed)
	bdd.append((1,0,7)) #6
	bdd.append((2,4,8)) #7
	bdd.append((3,0,9)) #8
        index=index+5
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,1,0)) # we return state 1 as we saw a top memory node
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
	bdd[index_bak+1]=(bdd[index_bak+1][0],bdd[index_bak+1][1],index)
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,0,0)) # we return state 0 as we did not see a top memory node
        index=index+1

       	#****************************************
	# left state=0, right state=1: 
	#   - the top memory nodes (11)
	#****************************************
	table.append((0,1,index))
	bdd.append((0,0,index+1))
        bdd.append((1,0,index+2))
        bdd.append((2,0,index+3))
        bdd.append((3,0,index+4))
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,1,0)) # we return state 1 -- we are still among the top memory nodes
        index=index+1

       	#****************************************
	# left state=1, right state=0: 
	#   - the null node (00)
	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,0))
        bdd.append((1,index+2,0))
        bdd.append((2,index+3,0))
        bdd.append((3,index+4,0))
        index=index+4
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,x)
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,2,0)) # we return state 2 -- x is null
        index=index+1
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,3,0)) # we return state 3 -- x is not null (may still be undef)
        index=index+1

       	#****************************************
	# left state=2, right state=0: 
	#   - the undef node (00)
	#****************************************
	table.append((2,0,index))
	bdd.append((0,index+1,0))
        bdd.append((1,index+2,0))
        bdd.append((2,index+3,0))
        bdd.append((3,index+4,0))
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,4,0)) # we know that x is null
        index=index+1

       	#****************************************
	# left state=3, right state=0: 
	#   - the undef node (00)
	#****************************************
	table.append((3,0,index))
	bdd.append((0,index+1,0))
        bdd.append((1,index+2,0))
        bdd.append((2,index+3,0))
        bdd.append((3,index+4,0))
        index=index+4
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,x)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,5,0)) # we return state 5 -- x is defined and not null
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,6,0)) # we return state 6 -- x is undef
        index=index+1

	# ****************************************
	# left state=4, right state=0:
	#  - a line number change to line_goto_1
	# ****************************************	
	table.append((4,0,index))	
	(bdd,index)=functions.change_line_num(bdd,index,line_orig,line_goto_1)
	bdd.append((-1,7,0)) # accept
	index=index+1

	# ****************************************
	# left state=5, right state=0:
	#  - a line number change to line_goto_2
	# ****************************************	
	table.append((5,0,index))	
	(bdd,index)=functions.change_line_num(bdd,index,line_orig,line_goto_2)
	bdd.append((-1,7,0)) # accept
	index=index+1

	# ****************************************
	# left state=6, right state=0:
	#  - a line number change to ERROR (x is undefined)
	# ****************************************	
	table.append((6,0,index))	
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line_orig)
	bdd.append((-1,7,0)) # accept
	index=index+1

	return(bdd,table,[7])
	# end of if_x_eq_null

#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_if_x_eq_null():
	"""test if (x==null) goto line_1; else goto line_2"""
	(bdd,table,accept_l)=if_x_eq_null(8,3,0,"000", "001","010")
	functions.gen_gta_file(bdd,table,accept_l,"if_x==null.gta")

#test_if_x_eq_null()

#=================================================================================

