import functions
import sys


def if_x_eq_y(node_width,pointer_num,x,y,line_orig,line_goto_1,line_goto_2):
	'''if (x==y) goto line1; else goto line2;'''

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# pointer_num -- the number of pointer variables
# x,y         -- the numbers of the pointer variables being tested

# line_orig, line_goto_1 and line_goto_2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0            # not read a node with x or y set neither a top memory node yet
        s_x_eq_y = 1          # x==y, a top memory node not read yet
        s_x_eq_y_top = 2      # x==y, read a top memory node
        s_x_eq_y_null = 3     # x==y, read the null node
        s_x_eq_y_undef = 4    # x==y, read the undef node
        s_accept = 5          # all changes finished, accept
        s_x_neq_y = 6         # x<>y, a top memory node not read yet
        s_x_neq_y_top = 7     # x<>y, read a top memory node
        s_x_neq_y_null = 8    # x<>y, read the null node
        s_x_neq_y_undef = 9   # x<>y, read the undef node
        s_top = 10            # read a top node, but x and y not encountered yet
        s_null = 11           # read the null node, x or y is undefined
        s_undef = 12          # read the undef node, x or y is undefined
        s_reject = 13         # reject: malformed, not the right line number, etc.

# All states are thus: s_init s_x_eq_y s_x_eq_y_top s_x_eq_y_null s_x_eq_y_undef s_accept s_x_neq_y s_x_neq_y_top s_x_neq_y_null s_x_neq_y_undef s_top s_null s_undef s_reject

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


# The states that can be to the right of s_init in a transition rule: 
#  s_init, s_x_eq_y, s_x_eq_y_top, s_x_neq_y, s_x_neq_y_top, s_top

       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 01 -> s_init
        #   - 10 x && y -> s_x_eq_y 
        #     10 (x && !y) || (!x && y) -> s_x_neq_y
        #     10 !x && !y -> s_init
        #   - 11 x && y -> s_x_eq_y_top
        #     11 (x && !y) || (!x && y) -> s_x_neq_y_top
        #     11 !x && !y -> s_top
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,10,0)) #4 - 00
	bdd.append((3,0,10)) #5 - 01
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
        index=index+7
        index_bak_1=index
	bdd.append((3,-2,0)) #8 - 10 (note: -2 will later be changed)
        index=index+1
        index_bak_2=index
	bdd.append((3,0,-2)) #9 - 11 (note: -2 will later be changed)
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	bdd[index_bak_1]=(bdd[index_bak_1][0],index,bdd[index_bak_1][2]) #processing a memory node
        (bdd,index,patch_init,patch_xneqy,patch_xeqy)=functions.copy_n_elm_and_remember_x_eq_y(bdd,index,node_width-2,4,x,y)
        for (i,j) in patch_xeqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_x_eq_y,0)) 
        index=index+1
        for (i,j) in patch_xneqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_x_neq_y,0)) 
        index=index+1
        for (i,j) in patch_init:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_init,0)) 
        index=index+1
	bdd[index_bak_2]=(bdd[index_bak_2][0],bdd[index_bak_2][1],index) #processing a top memory node
        index_bak_proc_init_top=index # we will re-use this part for other input states too
        (bdd,index,patch_init,patch_xneqy,patch_xeqy)=functions.copy_n_elm_and_remember_x_eq_y(bdd,index,node_width-2,4,x,y)
        for (i,j) in patch_xeqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_x_eq_y_top,0)) 
        index=index+1
        for (i,j) in patch_xneqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_x_neq_y_top,0)) 
        index=index+1
        for (i,j) in patch_init:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_top,0)) 
        index=index+1

       	#****************************************
	# left state = s_init, right state = s_x_eq_y: 
	#   - 01, 10 -> s_x_eq_y
	#   - 11 -> s_x_eq_y_top
	#****************************************
	table.append((s_init,s_x_eq_y,index))
        index_bak_proc_init_xeqy=index # we will jump here for other combinations of states too
	bdd.append((0,index+1,index+4)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2 -- 00 is not possible now
	bdd.append((3,0,index+8)) #+3 -- 01
	bdd.append((1,0,index+5)) #+4
	bdd.append((2,index+6,index+7)) #+5
	bdd.append((3,index+8,0)) #+6 -- 10
        index=index+7
        index_bak=index
	bdd.append((3,0,-2)) #+7 -- 11 (-2 will be later re-defined)
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+8
        bdd.append((-1,s_x_eq_y,0)) # we finished processing 01/10 
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],bdd[index_bak][1],index) # we are going to process 11
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_x_eq_y_top,0)) # we finished processing 11 
        index=index+1

       	#****************************************
	# left state = s_init, right state = s_x_neq_y: 
	#   - 01, 10 -> s_x_neq_y
	#   - 11 -> s_x_neq_y_top
	#****************************************
	table.append((s_init,s_x_neq_y,index))
        index_bak_proc_init_xneqy=index # we will jump here for other combinations of states too
	bdd.append((0,index+1,index+4)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2 -- 00 is not possible now
	bdd.append((3,0,index+8)) #+3 -- 01
	bdd.append((1,0,index+5)) #+4
	bdd.append((2,index+6,index+7)) #+5
	bdd.append((3,index+8,0)) #+6 -- 10
        index=index+7
        index_bak=index
	bdd.append((3,0,-2)) #+7 -- 11 (-2 will be later re-defined)
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+8
        bdd.append((-1,s_x_neq_y,0)) # we finished processing 01/10 
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],bdd[index_bak][1],index) # we are going to process 11
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_x_neq_y_top,0)) # we finished processing 11 
        index=index+1

       	#****************************************
	# left state = s_init, right state = s_x_eq_y_top: 
	#   - 11 -> s_x_eq_y_top
	#****************************************
	table.append((s_init,s_x_eq_y_top,index))
        index_bak_proc_init_xeqytop=index # we will jump here for other combinations of states too
	bdd.append((0,0,index+1)) #index -- 00/01 is not possible now
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,0,index+3)) #+2 -- 10 is not possible now
	bdd.append((3,0,index+4)) #+3 -- 11
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_x_eq_y_top,0)) # we finished processing 11 
        index=index+1

       	#****************************************
	# left state = s_init, right state = s_x_neq_y_top: 
	#   - 11 -> s_x_neq_y_top
	#****************************************
	table.append((s_init,s_x_neq_y_top,index))
        index_bak_proc_init_xneqytop=index # we will jump here for other combinations of states too
	bdd.append((0,0,index+1)) #index -- 00/01 is not possible now
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,0,index+3)) #+2 -- 10 is not possible now
	bdd.append((3,0,index+4)) #+3 -- 11
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_x_neq_y_top,0)) # we finished processing 11 
        index=index+1

       	#****************************************
	# left state = s_init, right state = s_top: 
        #   - 11 x && y -> s_x_eq_y_top
        #     11 (x && !y) || (!x && y) -> s_x_neq_y_top
        #     11 !x && !y -> s_top
	#****************************************
	table.append((s_init,s_top,index))
	bdd.append((0,0,index+1)) #index -- 00/01 is not possible now
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,0,index+3)) #+2 -- 10 is not possible now
	bdd.append((3,0,index_bak_proc_init_top)) #+3 -- 11 - we go to the part described above
        index=index+4


# The states that can be to the right of s_x_eq_y in a transition rule: 
#   s_init s_top

       	#****************************************
	# left state = s_x_eq_y, right state = s_init: 
	#****************************************
	table.append((s_x_eq_y,s_init,index_bak_proc_init_xeqy))

       	#****************************************
	# left state = s_x_eq_y, right state = s_top: 
	#****************************************
	table.append((s_x_eq_y,s_top,index_bak_proc_init_xeqytop))


# The states that can be to the right of s_x_neq_y in a transition rule: 
#   s_init s_x_neq_y s_top s_x_neq_y_top

       	#****************************************
	# left state = s_x_neq_y, right state = s_init: 
	# left state = s_x_neq_y, right state = s_x_neq_y:
	#   Here, we OVERAPPROXIMATE: we do not remember which variable appeared where.
	#   (We even allow two appearances of the same variable!)
	#****************************************
	table.append((s_x_neq_y,s_init,index_bak_proc_init_xneqy))
	table.append((s_x_neq_y,s_x_neq_y,index_bak_proc_init_xneqy))

       	#****************************************
	# left state = s_x_neq_y, right state = s_top: 
	# left state = s_x_neq_y, right state = s_x_neq_y_top:
	#   Here, we OVERAPPROXIMATE: we do not remember which variable appeared where.
	#   (We even allow two appearances of the same variable!)
	#****************************************
	table.append((s_x_neq_y,s_top,index_bak_proc_init_xneqytop))
	table.append((s_x_neq_y,s_x_neq_y_top,index_bak_proc_init_xneqytop))


# The states that can be to the right of s_x_eq_y_top in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_x_eq_y_top, right state = s_init: 
	#   - 00 (the null node) -> s_x_eq_y_null
	#****************************************
	table.append((s_x_eq_y_top,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_x_eq_y_null,0))
        index=index+1


# The states that can be to the right of s_x_neq_y_top in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_x_neq_y_top, right state = s_init: 
	#   - 00 (the null node) -> s_x_neq_y_null
	#****************************************
	table.append((s_x_neq_y_top,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_x_neq_y_null,0))
        index=index+1


# The states that can be to the right of s_x_eq_y_null in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_x_eq_y_null, right state = s_init: 
	#   - 00 (the undef node) -> s_x_eq_y_undef
	#****************************************
	table.append((s_x_eq_y_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_x_eq_y_undef,0))
        index=index+1


# The states that can be to the right of s_x_neq_y_null in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_x_neq_y_null, right state = s_init: 
	#   - 00 (the undef node) & x, y not undefined -> s_x_neq_y_undef
        #     00 (the undef node) & x or y undefined -> s_undef
	#****************************************
	table.append((s_x_neq_y_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index,patch_init,patch_xneqy,patch_xeqy)=functions.copy_n_elm_and_remember_x_eq_y(bdd,index,node_width-2,4,x,y)
        for (i,j) in patch_xeqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_reject,0)) # malformed: at least one of x and y has already appeared 
        index=index+1
        for (i,j) in patch_xneqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_undef,0)) # x or y is undefined 
        index=index+1
        for (i,j) in patch_init:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_x_neq_y_undef,0)) 
        index=index+1

# The states that can be to the right of s_x_eq_y_undef in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_x_eq_y_undef, right state = s_init: 
	#   - 0 line_orig -> s_accept (& change to line_goto_1)
	#     0 ~line_orig -> s_reject
	#****************************************
	table.append((s_x_eq_y_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line_orig,line_goto_1)
	bdd.append((-1,s_accept,0)) 
	index=index+1


# The states that can be to the right of s_x_neq_y_undef in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_x_neq_y_undef, right state = s_init: 
	#   - 0 line_orig -> s_accept (& change to line_goto_2)
	#     0 ~line_orig -> s_reject
	#****************************************
	table.append((s_x_neq_y_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line_orig,line_goto_2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


# The states that can be to the right of s_top in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_top, right state = s_init: 
	#   - 00 (the null node) & (x==null and y==nul)-> s_x_eq_y_null
	#   - 00 (the null node) & ~(x==null and y==nul)-> s_null
	#****************************************
	table.append((s_top,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index,patch_init,patch_xneqy,patch_xeqy)=functions.copy_n_elm_and_remember_x_eq_y(bdd,index,node_width-2,4,x,y)
        for (i,j) in patch_xeqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_x_eq_y_null,0))
        index=index+1
        for (i,j) in patch_xneqy:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_null,0)) # we know that x or y is undefined
        index=index+1
        for (i,j) in patch_init:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_null,0)) # we know that x and y are undefined
        index=index+1


# The states that can be to the right of s_null in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_null, right state = s_init: 
	#   - 00 (the undef node) -> s_undef
	#****************************************
	table.append((s_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_undef,0))
        index=index+1


# The states that can be to the right of s_undef in a transition rule: 
#   s_init 
# Here, we a bit OVERAPPROXIMATE: normally, to the right there should only be 00 nodes...

       	#****************************************
	# left state = s_undef, right state = s_init: 
	#   - 0 line_orig -> s_accept (& change to ERROR)
	#     0 ~line_orig -> s_reject
	#****************************************
	table.append((s_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line_orig)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of if_x_eq_y


#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_if_x_eq_y():
	"""test if (x==y) goto line_1; else goto line_2"""
	(bdd,table,accept_l)=if_x_eq_y(8,3,0,1,"000", "001","010")
	functions.gen_gta_file(bdd,table,accept_l,"if_x==y.gta")

#test_if_x_eq_y()

#=================================================================================

