import functions
import sys


def if_data(node_width, pointer_num, descr_num, x, data, line_orig, line_goto, then_type):
	'''if(x.data="value")'''

# node_width -- maximal size of a node in a tree
# pointer_num -- the number of pointer variables
# descr_num   -- number of pointer descriptors
# x           -- the number of pointer variable
# line_orig, and line_next must be of the same length equal to the maximum length possible!
# then type: 1 - generating transducer for then, 0 - generating transducer for else

# We use the following states in the transducer:
#   0 - reading the init symbols
#   1 - normal processing of the main tree structure (x not yet processed)
#   2 - normal processing of the main tree structure (x processed)
#   3 - read a top symbol (x processed)
#   4 - read a top symbol (x not yet processed) => x null/undefined -- ERROR
#   5 - read the null pointers section (x processed)
#   6 - read the null pointer section (x null/undefined) -- ERROR
#   7 - read the undef pointer section (x processed)
#   8 - read the undef pointer section (x null/undefined) -- ERROR
#   9 - accept
#   10 - reject (malformed, not the right line number, etc.) 

	st_reject=10  # define reject state number

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,st_reject,0))#0
	index=1

	#****************************************
	# left state=0, right state=0: 
	#   - empty nodes (00) till the first pointer description
	#   - the first pointer descriptor (01)
	#****************************************
	table.append((0,0,1))
	bdd.append((0,2,0))#1 -- the first bit of the type of the input node
	bdd.append((1,3,0))#2 -- the first bit of the type of the output node
	bdd.append((2,4,9))#3 -- if 1, we found the 1st pointer descriptor => go to line 9
	bdd.append((3,5,0))#4
	bdd.append((-1,0,0))#5 -- 00(0,0) -> 00, finished / note that 0 is a state, 00 an input symbol
	index=index+5

	#****************************************
	# left state=1, right state=1 or left state=0, right state=1: 
	#   - a next pointer description (01)
	# __________ X not yet processed ____________
	#****************************************
	table.append((1,1,index))	
	table.append((0,1,index))	
	bdd.append((0,index+1,0))#6
	bdd.append((1,index+2,0))#7
	bdd.append((2,0,index+3))#8
	bdd.append((3,0,index+4))#9 -- 01(1,1 or 0,1) -> 01 where 01 is a pointer descriptor
	index=index+4
	descr_bdd=index # in the below case of _(1,0) and a pointer descriptor being read, we continue from here	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,1,0))
	index=index+1
	#****************************************
	# left state=2, right state=1 
	# left state=1, right state=2 
	# left state=0, right state=2 
	#   - a next pointer description (01)
	#  __________ X processed __________
	#****************************************
	table.append((2,1,index))	
	table.append((1,2,index))	
	table.append((0,2,index))	
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3))
	bdd.append((3,0,index+4))# 01(1 or 0,1) -> 01 where 01 is a pointer descriptor
	index=index+4
	descr_bdd_x_read=index 
	# in the below case of (2,0) and a pointer descriptor being read, we continue from here	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,2,0))
	index=index+1

	#****************************************
	# left state=1, right state=0:
	#   - a memory node (10)
	#   - a right-most top memory node (11)
	#   - a right-most pointer descriptor (01)
	# ________ X not processed _________
	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,-2)) # a top or normal memory node
	aux=index+1
	  # a normal memory node
	bdd.append((3,index+3,0)) 
	index=index+3
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num+descr_num,4,x)
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	act_track=(pointer_num+descr_num+2)*2
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(pointer_num+descr_num+2),act_track)
	bdd.append((-1,1,0))
	index=index+1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	# check, if data are set, or not
	# note, that len(data) can be less than data_size. In this case we just copy the rest up to the data_size
	data_size=node_width-(pointer_num+descr_num+2)
	act_track=(pointer_num+descr_num+2)*2
	if data_size:
		bdd.append((act_track,index+1,index+2))
		# lower index = data still according
		if len(data)==0: # there is no data to be checked, only copy
			bdd.append((act_track+1,index+3,0))
			bdd.append((act_track+1,0,index+3))
		elif data[0]=="X": # we dont care about value on this position
			bdd.append((act_track+1,index+3,0))
			bdd.append((act_track+1,0,index+3))
		elif data[0]=="1":
			bdd.append((act_track+1,index+4,0))
			bdd.append((act_track+1,0,index+3))
		else:
			bdd.append((act_track+1,index+3,0))
			bdd.append((act_track+1,0,index+4))
		index=index+3
		act_track=act_track+2
		for i in range(data_size-1):
			bdd.append((act_track+(2*i),index+2,index+3))
			bdd.append((act_track+(2*i),index+4,index+5))
			if len(data)<i+2: # there is no more data to be checkd till end of node -> only copy
				bdd.append((act_track+1+(2*i),index+6,0))
				bdd.append((act_track+1+(2*i),0,index+6))
			elif data[i+1]=="X": # we dont care about value on this position
				bdd.append((act_track+1+(2*i),index+6,0))
				bdd.append((act_track+1+(2*i),0,index+6))
			elif data[i+1]=="1":
				bdd.append((act_track+1+(2*i),index+7,0))
				bdd.append((act_track+1+(2*i),0,index+6))
			else:
				bdd.append((act_track+1+(2*i),index+6,0))
				bdd.append((act_track+1+(2*i),0,index+7))
			bdd.append((act_track+1+(2*i),index+7,0))
			bdd.append((act_track+1+(2*i),0,index+7))
			index=index+6
		# Here we distinguish then and else 
		if then_type: #generating then file
			bdd.append((-1,2,0))
			bdd.append((-1,st_reject,0)) # DATA not set => REJECT
		else: # generating else file
			bdd.append((-1,st_reject,0)) # DATA set => REJECT
			bdd.append((-1,2,0))
		index=index+2
	else:
		# Here we distinguish then and else 
		if then_type: #generating then file
			bdd.append((-1,2,0))  # NO data => answer YES
		else:
			bdd.append((-1,st_reject,0))  # NO data => reject
		index=index+1
				
	  # a top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index) 
	bdd.append((3,0,index+1)) 
	index=index+1
	top_bdd=index
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num+descr_num,4,x)
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	act_track=(pointer_num+descr_num+2)*2
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(pointer_num+descr_num+2),act_track)
	bdd.append((-1,4,0))
	index=index+1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	# check, if data are set, or not
	# note that len(data) can be less than data_size. In this case we just copy the rest up to the data_size
	data_size=node_width-(pointer_num+descr_num+2)
	act_track=(pointer_num+descr_num+2)*2
	if data_size:
		bdd.append((act_track,index+1,index+2))
		# lower index = data still according
		if len(data)==0: # there is no data to be checked, only copy
			bdd.append((act_track+1,index+3,0))
			bdd.append((act_track+1,0,index+3))
		elif data[0]=="X": # we dont care about value on this position
			bdd.append((act_track+1,index+3,0))
			bdd.append((act_track+1,0,index+3))
		elif data[0]=="1":
			bdd.append((act_track+1,index+4,0))
			bdd.append((act_track+1,0,index+3))
		else:
			bdd.append((act_track+1,index+3,0))
			bdd.append((act_track+1,0,index+4))
		index=index+3
		act_track=act_track+2
		for i in range(data_size-1):
			bdd.append((act_track+(2*i),index+2,index+3))
			bdd.append((act_track+(2*i),index+4,index+5))
			if len(data)<i+2: # there is no more data to be checkd till end of node -> only copy
				bdd.append((act_track+1+(2*i),index+6,0))
				bdd.append((act_track+1+(2*i),0,index+6))
			elif data[i+1]=="X": # we dont care about value on this position
				bdd.append((act_track+1+(2*i),index+6,0))
				bdd.append((act_track+1+(2*i),0,index+6))
			elif data[i+1]=="1":
				bdd.append((act_track+1+(2*i),index+7,0))
				bdd.append((act_track+1+(2*i),0,index+6))
			else:
				bdd.append((act_track+1+(2*i),index+6,0))
				bdd.append((act_track+1+(2*i),0,index+7))
			bdd.append((act_track+1+(2*i),index+7,0))
			bdd.append((act_track+1+(2*i),0,index+7))
			index=index+6
		# Here we distinguish then and else 
		if then_type: #generating then file
			bdd.append((-1,3,0))
			bdd.append((-1,st_reject,0)) # DATA not set => REJECT
		else: # generating else file
			bdd.append((-1,st_reject,0)) # DATA set => REJECT
			bdd.append((-1,3,0))
		index=index+2
	else:
		# Here we distinguish then and else 
		if then_type: #generating then file
			bdd.append((-1,3,0))  # NO data => answer YES
		else:
			bdd.append((-1,st_reject,0))  # NO data => reject
		index=index+1
	# out 3: X processed, 4 X unprocessed

	#****************************************
	# left state=1, right state=4:
	#   - a top memory node (11)
	#****************************************
	table.append((1,4,index))
	bdd.append((0,0,index+1))
	bdd.append((1,0,index+2))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,top_bdd)) 
	index=index+4
	
	#****************************************
	# left state=2, right state=0:
	#   - a memory node (10)
	#   - a right-most top memory node (11)
	#   - a right-most pointer descriptor (01)
	#  ____________X processed___________
	#****************************************
	table.append((2,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd_x_read)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,-2)) # a top or normal memory node
	aux=index+1
	  # a normal memory node
	bdd.append((3,index+3,0)) 
	index=index+3
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,2,0))
	index=index+1
	bdd[aux]=(2,bdd[aux][1],index)
	top_bdd_x_read=index
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,3,0))
	index=index+1

	#****************************************
	# left state=1, right state=3:
	# left state=2, right state=4:
	#   - a top memory node (11)
	#  _________ X processed ____________
	#****************************************
	table.append((1,3,index))
	table.append((2,4,index))
	bdd.append((0,0,index+1))
	bdd.append((1,0,index+2))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,top_bdd_x_read)) 
	index=index+4

	# ****************************************
	# left state=3, right state=0:
	#   - null pointers (00)
	# ****************************************
	table.append((3,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,5,0))
	index=index+1
	# out: 5
	# ****************************************
	# left state=4, right state=0:
	#   - null pointers (00)
	# !!! x must be null/undefined -- ERROR
	# ****************************************
	table.append((4,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,6,0))
	index=index+1
	
	# ****************************************
	# left state=5, right state=0:
	#   - undef pointers (00)
	# ****************************************
	table.append((5,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,7,0))
	index=index+1

	# ****************************************
	# left state=6, right state=0:
	#   - undef pointers (00)
	# !!!! x must be undefined -- ERROR
	# ****************************************
	table.append((6,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,8,0))
	index=index+1

	# ****************************************
	# left state=7, right state=0:
	#  - a line number change 
	# ****************************************	
	table.append((7,0,index))	
	(bdd,index)=functions.change_line_num(bdd,index,line_orig,line_goto)
	bdd.append((-1,9,0)) # accept
	index=index+1

	# ****************************************
	# left state=8, right state=0:
	#  - a line number change to ERROR (y is null/undefined)
	# ****************************************	
	table.append((8,0,index))	
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line_orig)
	bdd.append((-1,9,0)) # accept
	index=index+1


	return(bdd,table,[9])
	# end of x_assign_y

#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_ifdata():
	"""test if(x.data="value")"""
	#(bdd,table,acc_list)=if_data(13, 4, 5, 1, "11", "0000011", "0000010", 1)
	(bdd,table,acc_list)=if_data(14, 4, 5, 1, "00X", "0000100", "0000010", 1)
	functions.gen_gta_file(bdd,table,acc_list,"transd_then.gta")

	#(bdd,table,acc_list)=if_data(13, 4, 5, 1, "11", "0000011", "0000100", 0)
	(bdd,table,acc_list)=if_data(14, 4, 5, 1, "0X1", "0000100", "0000100", 0)
	functions.gen_gta_file(bdd,table,acc_list,"transd_else.gta")

#test_ifdata()

#=================================================================================

