import functions
import sys


def descr_star(node_width, pointer_num, marker):
	'''generate routing expression \sigma^*'''

# node_width - maximal size of node
# pointer_num -- the number of pointer variables
# marker  -- marker required in destination


# We use the following states in the transducer:
#   0 - reading the init symbols
#   1 - normal processing of the main tree structure
#   2 - read and deleted source marker
#   3 - set a destination marker
#   4 - 2+3  - read rest of the tree (we don't care about contain of nodes)
#   5 - reject (malformed, not the right line number, etc.) 

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,5,0))#0
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#   - empty nodes (00) till the first pointer description
	#   - the first pointer descriptor (01)
	#****************************************
	table.append((0,0,1))
	bdd.append((0,2,0))#1 -- the first bit of the type of the input node
	bdd.append((1,3,0))#2 -- the first bit of the type of the output node
	bdd.append((2,4,6))#3 -- if 1, we found the 1st pointer descriptor => go to line 6
	bdd.append((3,5,0))#4
	bdd.append((-1,0,0))#5
	index=index+5
	# a pointer descriptor (01)
	bdd.append((3,0,index+1))#6
	index=index+1
	descr_bdd=index # in the below case a pointer descriptor being read, we continue from here
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,1,0))
	index=index+1

	#****************************************
	#   - a memory node / a top memory node (1x)
	#   - a right-most pointer descriptor (01)
	#****************************************
	table.append((1,1,index))	
	table.append((0,1,index))
	table.append((1,0,index))
	bdd.append((0,index+1,index+4)) # pointer desc / memory node
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	index=index+1
	# we don't care if it is tom/normal memory node
	(bdd,index)=functions.copy_n_elements(bdd,index,pointer_num+1,2)
	#now check whether the token is set
	bdd.append((4+(2*pointer_num),index+1,-2))
	aux2=index
	index=index+1
	# token is not set => guess whether this is a destination
	bdd.append((5+(pointer_num*2),index+1,-2))
	aux=index
	index=index+1
	# NO it is not a destination => just copy the rest
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(pointer_num+3),6+(pointer_num*2))
	bdd.append((-1,1,0))
	index=index+1
	# YES, this is a destination => check for presence of marker
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
	(bdd,index)=functions.copy_n_elements_check_x(bdd,index,node_width-(pointer_num+3),6+(pointer_num*2),marker-1)
	bdd.append((-1,3,0))
	index=index+1
	#-------------------------------
	# the token is set in this node
	bdd[aux2]=(bdd[aux2][0],bdd[aux2][1],index)
	bdd.append((5+(pointer_num*2),index+1,-2))
	aux=index
	index=index+1
	# NO it is not a destination => just copy the rest
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(pointer_num+3),6+(pointer_num*2))
	bdd.append((-1,2,0))
	index=index+1
	# YES, this is a destination => check for presence of marker
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
	(bdd,index)=functions.copy_n_elements_check_x(bdd,index,node_width-(pointer_num+3),6+(pointer_num*2),marker-1)
	bdd.append((-1,4,0))
	index=index+1

	#****************************************
	#   - a memory node / a top memory node (1x)
	#   - a right-most pointer descriptor (01)
	#****************************************
	table.append((3,1,index))	
	table.append((1,3,index))	
	table.append((0,3,index))
	table.append((3,0,index))
	bdd.append((0,index+1,-2)) # pointer descriptor / memory node
	aux=index
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,index+4)) # goto processing of a pointer descriptor
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,3,0))
	index=index+1

	# a memory node or a top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
	bdd.append((1,0,index+1))
	index=index+1
	# we don't care if it is tom/normal memory node
	# here, we are checking for 'from' marker -> first marker - on position after all pointers.
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-1,2,pointer_num+1,pointer_num+1,3,4)

	# ***************************************
	# left state=2, right state=3
	# left state=3, right state=4
	#     - a pointer descriptor -> output 4
	#     - top memory node -> output 4
	# ***************************************
	table.append((2,3,index))
	table.append((3,2,index))
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)
	bdd.append((-1,4,0))
	index=index+1

	# ***************************************
	# left state=2, right state=0/1
	# left state=0/1, right state=2
	#     - a pointer descriptor (01) -> output 2
	#     - a memory node (1x) - set destination marker -> output 4
	#     - a memory node (1x) - not set destination marker -> output 2
	# ***************************************
	table.append((2,1,index))
	table.append((2,0,index))
	table.append((1,2,index))
	table.append((0,2,index))
	bdd.append((0,index+1,-2)) # pointer descriptor / memory node
	aux=index
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,index+4)) # goto processing of a pointer descriptor
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,2,0))
	index=index+1
	# a memory node or a top memory node = we are going to set marker 'to' - it is destionation of this pointer descriptor
	# it is necessary to check whether marker is set.
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
	bdd.append((1,0,index+1))
	index=index+1
	(bdd,index)=functions.copy_n_elements(bdd,index,pointer_num+1,2)
	# now guess token and check it
	bdd.append((5+(pointer_num*2),index+1,-2))
	aux=index
	index=index+1
	#just copy rest of the node
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(pointer_num+3),6+(pointer_num*2))
	bdd.append((-1,2,0))
	index=index+1
	# check presence of marker
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
	(bdd,index)=functions.copy_n_elements_check_x(bdd,index,node_width-(pointer_num+3),6+(pointer_num*2),marker-1)
	bdd.append((-1,4,0))
	index=index+1

	# ****************************************
	# left state=4, right state=0/1:
	# left state=0/1, right state=4:
	#   - read of a rest of the tree
	# ****************************************
	table.append((4,0,index))
	table.append((4,1,index))
	table.append((1,4,index))
	table.append((0,4,index))
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)	
	bdd.append((-1,4,0))
	index=index+1
	return(bdd, table, [4])


#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_gen_exp_star():
	#(bdd,table,acc_l)=descr_1up(24,4,2)
	(bdd,table,acc_l)=descr_star(24,6,7)
	functions.gen_gta_file(bdd,table, acc_l,"descr_star.gta")

#test_gen_exp_star()


#=================================================================================

