import functions
import sys

#=================================================================================

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# This version does not cover the possibility of new returning null when there is not enough memory.

#=================================================================================

def x_assign_new(node_width,ptr_num,mark_num,next_num,data_size,data,x,line1,line2):
	'''x=new (a new top memory is linked below the right-most one or -- initially -- under the null node)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num     -- the number of pointer variables including the deleted flag
# mark_num    -- the number of markes (including the token) -- should be the same as the number of descriptors
# next_num    -- the number of the next pointers
# data_size   -- the size of the data - it has influence only in case of variance analysis
# data        -- initial data value. Can be composed from 0,1,X. Empty string = random data.
# x           -- the number of the pointer variable being assigned

# line1 and line2 must be of the same length equal to the maximum length possible!

	if len(data)>data_size:
		print "ERROR: Gen_new: The length of the data exceed the limit"
		sys.exit(1)

# We use the following states in the transducer:
        s_init = 0             # reading the initial nodes
	s_not_init_top = 1     # reading existing memory nodes and ptr. descriptors
	s_new_ptr_desc = 10    # creating the new ptr. desc. (+0, +1, ..., + next_num-1)
	s_new_top = 2          # created the new top memory node
	s_top = 3              # reading the existing top memory nodes
	s_null = 4             # have read the null node
	s_undef = 5            # have read the undef node
        s_accept = 6           # the line has been successfully changed
        s_reject = 7           # reject: malformed, not the right line number, etc.


# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00 -> s_init / s_new_ptr_desc
        #   - 01/10 -> s_not_init_top
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,-2)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,-2,4)) #3
	bdd.append((3,0,5)) #4 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #5
        bdd.append((-1,s_not_init_top,0)) 
        index=index+1
	# processing a memory node -- NOT a top one
	bdd[1]=(bdd[1][0],bdd[1][1],index) 
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,0))
	bdd.append((3,index+3,0))
        index=index+3
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
        bdd.append((-1,s_not_init_top,0)) 
        index=index+1
	# processing an init node
	bdd[3]=(bdd[3][0],index,bdd[3][2]) 
	bdd.append((3,index+1,index+2)) # keep init or change to a new ptr. desc.?
        index=index+1
	# we keep the node as init
        bdd.append((-1,s_init,0)) 
        index=index+1
	# we create a new ptr. desc.
	bdd.append((5,index+1,0)) # set the ptr. desc. to UNDEF (01|011) -- we thus ignore bits 4,6,8
	bdd.append((7,0,index+2))
	bdd.append((9,0,index+3))
        bdd.append((-1,s_new_ptr_desc,0)) 
        index=index+4
	

       	#****************************************
	# left state = s_init, right state = s_not_init_top: 
	# left state = s_not_init_top, right state = s_init: 
	# left state = s_not_init_top, right state = s_not_init_top: 
        #   - 01/10 -> s_not_init_top
	#****************************************
	table.append((s_init,s_not_init_top,index))
	table.append((s_not_init_top,s_init,index))
	table.append((s_not_init_top,s_not_init_top,index))
	index_bak=index
	bdd.append((0,index+1,-2)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_not_init_top,0)) 
        index=index+1
	# processing a (not top) memory node
	bdd[index_bak]=(bdd[index_bak][0],bdd[index_bak][1],index) 
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,0)) # we read 10
	bdd.append((3,index+3,0))
        index=index+3
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
        bdd.append((-1,s_not_init_top,0)) 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_new_ptr_desc + i (i = +0, +1, ..., + next_num-2):
        #   - 00 -> s_new_ptr_desc + i+1 
	#****************************************
        for i in range(next_num - 1): # i.e. 0, 1, ..., next_num-2
		table.append((s_init,s_new_ptr_desc + i,index))
		bdd.append((0,index+1,0)) #index
		bdd.append((1,index+2,0)) #+1
		bdd.append((2,index+3,0)) #+2
		bdd.append((3,0,index+4)) #+3 - 00 -> 01
        	index=index+4
		bdd.append((5,index+1,0)) # set the ptr. desc. to UNDEF (01|011) -- we thus ignore bits 4,6,8
		bdd.append((7,0,index+2))
		bdd.append((9,0,index+3))
        	bdd.append((-1,s_new_ptr_desc + i+1,0)) 
        	index=index+4


       	#****************************************
	# left state = s_new_ptr_desc + next_num-1, right state = s_init:
        #   - 00 -> s_new_top
	#****************************************
	table.append((s_new_ptr_desc + next_num-1,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,0,index+4)) #+3 - 00 -> 11
       	index=index+4
	# we initialize the new node and ignore all even bits now:
	#   not deleted, pointed to by x, no markers, undefined data 
	for i in range(ptr_num + mark_num):
		if (i==x):
			bdd.append((5+2*i,0,index+i+1)) # pointed to by x
		else:
			bdd.append((5+2*i,index+i+1,0)) # not deleted, pointed by x only, not marked
       	index=index+ptr_num+mark_num
	# if data initialization is requested
	for i in range(len(data)):
		if data[i]=="0":
			bdd.append((5+((ptr_num+mark_num)*2)+(i*2),index+1,0))
			index=index+1
		if data[i]=="1":
			bdd.append((5+((ptr_num+mark_num)*2)+(i*2),0,index+1))
			index=index+1
		# in case of data[i]=="X", we skip this position
	# in case of enabled variance analysis, we need to set old values of the pointer variables
	# - i.e. there was no pointer variables pointing into this node before its creation
	#
	# in case of disbaled variance analysis node_width=2+ptr_num+mark_num+data_size and the loop 
	# will be not iterated
	for i in range (node_width-(2+ptr_num+mark_num+data_size+data_size)):
		# set the "old" position of all pointer variables 
		# it is a new node -> all "old position" are set to "false"
		bdd.append((5+((ptr_num+mark_num+data_size)*2)+(i*2),index+1,0))
		index=index+1
       	bdd.append((-1,s_new_top,0)) 
       	index=index+1


       	#****************************************
	# left state = s_not_init_top, right state = s_new_top:
	# left state = s_not_init_top, right state = s_top:
        #   - 11 -> s_top
	#****************************************
	table.append((s_not_init_top,s_new_top,index))
	table.append((s_not_init_top,s_top,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 11
       	index=index+4
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
        bdd.append((-1,s_top,0)) 
        index=index+1


       	#****************************************
	# left state = s_top, right state = s_init:
	# left state = s_new_top, right state = s_init:
        #   - 00 -> s_null
	#****************************************
	table.append((s_top,s_init,index))
	table.append((s_new_top,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00 
       	index=index+4
	# processing the null node
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
        bdd.append((-1,s_null,0)) 
        index=index+1


       	#****************************************
	# left state = s_null, right state = s_init:
        #   - 00 -> s_undef
	#****************************************
	table.append((s_null,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00 
       	index=index+4
	# processing the undef node
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
        bdd.append((-1,s_undef,0)) 
        index=index+1


       	#****************************************
	# left state = s_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line1,line2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_new


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_assign_new(node_width,ptr_num,mark_num,next_num,x,line1,line2):

def test_x_assign_new():
	"""test_x_assign_new"""
	(bdd,table,accept_l)=x_assign_new(9,4,3,4,1,"000000000","000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_new.gta")

#test_x_assign_new()


#=================================================================================

