# Adam Rogalewicz
#
# this generator create transducer, which place pointer variable into randomly set memory node

import functions
import sys


def x_assign_random(node_width, x, line_orig, line_goto):
	'''x:=random'''

# node_width  -- maximal size of a node
# x           -- the index of pointer variable 

# line_orig and line_goto must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
#   0 - reading the init symbols
#   1 - processing of a tree structure - x not set
#   2 - processing of a tree structure - x set
#   3 - read the null pointers section
#   4 - read the undef pointer section
#   5 - accept
#   6 - reject (malformed, not the right line number, etc.) 
#   7 - read top symbol - x not yet set
#   8 - read a top symbol - x set 


# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,6,0))#0
	index=1

	#****************************************
	# left state=0, right state=0: 
	#   - empty nodes (00) till the first pointer description
	#   - the first pointer descriptor (01)
	#****************************************
	table.append((0,0,1))
	bdd.append((0,2,0))#1 -- the first bit of the type of the input node
	bdd.append((1,3,0))#2 -- the first bit of the type of the output node
	bdd.append((2,4,9))#3 -- if 1, we found the 1st pointer descriptor => go to line 9
	bdd.append((3,5,0))#4
	bdd.append((-1,0,0))#5 -- 00(0,0) -> 00, finished / note that 0 is a state, 00 an input symbol
	index=index+5

	#****************************************
	# left state=1, right state=1 or left state=0, right state=1: 
	#   - a next pointer description (01)
	#****************************************
	table.append((1,1,index))	
	table.append((0,1,index))	
	bdd.append((0,index+1,0))#6
	bdd.append((1,index+2,0))#7
	bdd.append((2,0,index+3))#8
	bdd.append((3,0,index+4))#9 -- 01(1 or 0,1) -> 01 where 01 is a pointer descriptor
	index=index+4
	descr_bdd=index # in the below case of _(1,0) and a pointer descriptor being read, we continue from here	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,1,0))
	index=index+1

	#****************************************
	# left state=1, right state=0:
	#   - a memory node (10)
	#   - a right-most top memory node (11)
	#   - a right-most pointer descriptor (01)
	#   X not yet set
	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,-2)) # a top or normal memory node
	aux=index+1
	#----------------------------
	# a normal memory node - guess, if we need to set X or not
	bdd.append((3,index+3,0)) 
	index=index+3
	# If memory node is undefined (code: 1x1) -> unset x
	bdd.append((4,-2,index+1)) 
	aux3=index
	bdd.append((5,0,index+2)) 
	index=index+2
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-3,6,x-1)
	bdd.append((-1,1,0))
	index=index+1
	# this memory node is defined (code: 1x0)
	bdd[aux3]=(bdd[aux3][0],index,bdd[aux3][2])
	bdd.append((5,index+1,0)) 
	index=index+1
	(bdd,index)=functions.copy_n_elements(bdd,index,x-1,6)
	# guess X
	bdd.append((5+(x*2),index+1,-2))
	aux2=index
	index=index+1
	# x not set - copy the rest
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(x+3),6+(x*2))
	bdd.append((-1,1,0))
	index=index+1
	# x is set - copy the rest
	bdd[aux2]=(bdd[aux2][0],bdd[aux2][1],index) 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(x+3),6+(x*2))
	bdd.append((-1,2,0))
	index=index+1
	#-------------------------
	# a top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index) 
	bdd.append((3,0,index+1)) 
	index=index+1
	top_bdd=index
	# If memory node is undefined (code: 1x1) -> unset x
	bdd.append((4,-2,index+1)) 
	aux3=index
	bdd.append((5,0,index+2)) 
	index=index+2
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-3,6,x-1)
	bdd.append((-1,7,0))
	index=index+1
	# this memory node is defined (code: 1x0)
	bdd[aux3]=(bdd[aux3][0],index,bdd[aux3][2])
	bdd.append((5,index+1,0)) 
	index=index+1
	(bdd,index)=functions.copy_n_elements(bdd,index,x-1,6)
	# guess X
	bdd.append((5+(x*2),index+1,-2))
	aux2=index
	index=index+1
	# x not set - copy the rest
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(x+3),6+(x*2))
	bdd.append((-1,7,0))
	index=index+1
	# x is set - copy the rest
	bdd[aux2]=(bdd[aux2][0],bdd[aux2][1],index) 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-(x+3),6+(x*2))
	bdd.append((-1,8,0))
	index=index+1

	#****************************************
	# left state=1, right state=7:
	#   - a top memory node (11)
	#  X not set
	#****************************************
	table.append((1,7,index))
	bdd.append((0,0,index+1))
	bdd.append((1,0,index+2))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,top_bdd)) 
	index=index+4

	#****************************************
	# left state=1, right state=2:
	# left state=2, right state=1:
	# left state=0, right state=2:
	#   - a pointer descriptor (01)
	# X is set
	#****************************************
	table.append((1,2,index))	
	table.append((2,1,index))	
	table.append((0,2,index))	
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3))
	bdd.append((3,0,index+4))
	index=index+4
	descr_bdd_x_set=index # in the below case of _(2,0) and a pointer descriptor being read, we continue from here	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,2,0))
	index=index+1

	#****************************************
	# left state=2, right state=0:
	#   - a pointer descriptor (01)
	#   - a memory node (10)
	#   - a top memory node (11)
	#   X is set
	#****************************************
	table.append((2,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd_x_set)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,-2)) # a top or normal memory node
	aux=index+1
	  # a normal memory node - just unset possible X
	bdd.append((3,index+3,0)) 
	index=index+3
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
	bdd.append((-1,2,0))
	index=index+1

	  # a top memory node - just unset possible X
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index) 
	bdd.append((3,0,index+1)) 
	index=index+1
	top_bdd_x_set=index
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
	bdd.append((-1,8,0))
	index=index+1
	#****************************************
	# left state=1, right state=8:
	#   - a top memory node (11)
	#  X is set
	#****************************************
	table.append((1,8,index))
	bdd.append((0,0,index+1))
	bdd.append((1,0,index+2))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,top_bdd_x_set)) 
	index=index+4


	# ****************************************
	# left state=8, right state=0:
	#   - null pointers (00)
	# ****************************************
	table.append((8,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	# unset x
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
	bdd.append((-1,3,0))
	index=index+1
	
	# ****************************************
	# left state=3, right state=0:
	#   - undef pointers (00)
	# ****************************************
	table.append((3,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	# unset x
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
	bdd.append((-1,4,0))
	index=index+1

	# ****************************************
	# left state=4, right state=0:
	#  - a line number change 
	# ****************************************	
	table.append((4,0,index))	
	(bdd,index)=functions.change_line_num(bdd,index,line_orig,line_goto)
	bdd.append((-1,5,0)) # accept
	index=index+1


	return(bdd,table,[5])
	# end of x_assign_random

#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_x_assign_random():
	"""test x:=random"""
	(bdd,table,acc_l)=x_assign_random(8,3,2,"001", "011")
	functions.gen_gta_file(bdd,table,acc_l,"x_assign_random.gta")

#test_x_assign_random()

#=================================================================================

