import functions
import sys

#=================================================================================

# !!!!!!!!!!!!!!!!!!!!!!!!!!

# The use of these functions is not easy: multiple of them must be used, sometimes
# in parallel, sometimes sequentially. All the results are to be united then.

# More precisely, there are the following possibilities:
# 1.(for every descriptor:(ph1_mid_desc or ph1_lst_desc)) -> (ph2_reg or ph2del)
#       [[ There are functions for both the phases that cover all possibilities. ]]
# 2. y==null
# 3. y==undef
# 4. y->next==null
# 5. y->next==undef 

# !!!!!!!!!!!!!!!!!!!!!!!!!!
# A BIG OVERAPPROXIMATION???
#  - we allow in some places 10...(s_init,s_init) -> s_init, but
#    10/11 can only appear above a chain of ptr. desc. that is accepted via some special states...
#
#=================================================================================

def x_assign_y_next_ph1_mid_desc(node_width,ptr_num,next_num,x,y,next,line1,line2,descr_no):
	'''x=y->next (a middle next pointer, phase 1: pre-processing for using descriptors)'''

# Middle here means that it is not the right-most (bottom-most) ptr. descriptor node. 

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num     -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)
# descr_no    -- the number of the descriptor to be used (numbered from one)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0             # not read a useful pair of a ptr. desc. and the corresp. mem. node yet 
        s_bef_gvn_ptr_d = 10   # ptr. desc. nodes before the tested one (we use s_bef_gvn_ptr_d + 0, + 1, ...)
        s_gvn_ptr_d_y = 1      # the given ptr. desc. node with the given desc. set and above it
        s_gvn_ptr_d_n = 2      # the given ptr. desc. node with the given desc. NOT set and above it
        s_token_set_nnull = 3  # the ptr. moving token has been set, waiting for the null node
        s_token_set_null = 4   # the ptr. moving token has been set, the null node has been read
        s_token_set_undef = 5  # the ptr. moving token has been set, the undef node has been read
        s_accept = 6           # the ptr. moving token has been set, the right line number has been found
        s_reject = 7           # reject: malformed, not the right line number, etc.

	if ((next<0) or (next>=next_num-1)):
		functions.error_msg_expl("x_assign_y_next_ph1_mid_desc: use a different function for x=y->next, phase 1, desc.!");
        	# use x_assign_y_next_ph1_lst_desc

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 10, 11 -> s_init
        #   - 01 -> s_bef_gvn_ptr_d+0
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,10,0)) #4 - 00
	bdd.append((3,0,-2)) #5 - 01 (note: -2 will later be changed)
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	bdd[5]=(bdd[5][0],bdd[5][1],index) # processing the last (right-most) pointer descriptor
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_bef_gvn_ptr_d,0)) 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + i:
        #   - 01 -> s_bef_gvn_ptr_d + i+1 ... till just under processing the given ptr. desc. 
	#****************************************
        for i in range(next_num - next - 2):
                table.append((s_init,s_bef_gvn_ptr_d + i,index)) # we are so far looking for the place to put the token
	        bdd.append((0,index+1,0)) #index                 # the case with the token already set is below  
	        bdd.append((1,index+2,0)) #+1
	        bdd.append((2,0,index+3)) #+2
	        bdd.append((3,0,index+4)) #+3 - 01
                index=index+4
                (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
                bdd.append((-1,s_bef_gvn_ptr_d + i+1,0)) 
                index=index+1
                

       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_y / _n
	#****************************************
        table.append((s_init,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index_bak=index+4
	bdd.append((4,index+5,-2)) #+4
	bdd.append((5,index+6,0)) #+5 - 010 (i.e. null or undef: descr_no not set)
        index=index+6
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6) #+6
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a not null/undef ptr. descr. node
	bdd[index_bak]=(bdd[index_bak][0],bdd[index_bak][1],index) 
	bdd.append((5,0,index+1))
        index=index+1
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-3,6,descr_no-1)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #  (the case of the token already set is covered below)
        #   - 10/11 -> s_token_set_nnull provided y is set 
        #   - 10/11 -> s_init provided y is NOT set 
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_node_Ret_st_dep_on_y_Set_x_to_y(bdd,index,node_width-2,4,ptr_num,y,s_init,s_token_set_nnull)
                # ptr_num is the first bit in the section of markers, i.e. the descriptor token


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the appr. ptr. d. was not set
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_token_set_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_token_set_nnull, right state = s_bef_gvn_ptr_d + i (nnull is appearing in the middle of the ptr.d. chain):
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_token_set_nnull (nnull came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_y, right state = s_token_set_nnull (nnull coming from a lower top memory node):
	# left state = s_gvn_ptr_d_n, right state = s_token_set_nnull:
        #   - 01/10/11 -> s_token_set_nnull
        #   - 00 (the null node) -> s_token_set_null
	#****************************************
        table.append((s_token_set_nnull,s_init,index))
        for i in range(next_num - next - 1):
                table.append((s_token_set_nnull,s_bef_gvn_ptr_d + i,index))
        table.append((s_token_set_nnull,s_gvn_ptr_d_y,index))
        table.append((s_token_set_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_token_set_nnull,index))
        table.append((s_gvn_ptr_d_y,s_token_set_nnull,index))
        table.append((s_gvn_ptr_d_n,s_token_set_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_token_set_nnull,0)) # the null node has not been found yet
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_token_set_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_token_set_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_token_set_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_token_set_undef,0))  # the undef node has just been processed
        index=index+1

        
       	#****************************************
	# left state = s_token_set_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_token_set_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line1,line2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_ph1_mid_desc


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_assign_y_next_ph1_mid_desc(node_width,ptr_num,next_num,x,y,next,line1,line2,descr_no):

def test_x_assign_y_next_ph1_mid_desc():
	"""test_x_assign_y_next_ph1_mid_desc"""
	(bdd,table,accept_l)=x_assign_y_next_ph1_mid_desc(9,4,4,1,2,1,"000000000","000000001",1)
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph1_mid_desc_1.gta")
	(bdd,table,accept_l)=x_assign_y_next_ph1_mid_desc(9,4,4,1,2,1,"000000000","000000001",2)
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph1_mid_desc_2.gta")

# test_x_assign_y_next_ph1_mid_desc()


#----------------------------------------------------------------------------------


def x_assign_y_next_ph1_lst_desc(node_width,ptr_num,next_num,x,y,line1,line2,descr_no):
	'''x=y->next (the last next pointer, phase 1: pre-processing for using descriptors)'''

# The last here means the right-most, i.e. the bottom-most, one.

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num     -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)
# descr_no    -- the number of the descriptor to be used (numbered from one)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0             # not read a useful pair of a ptr. desc. and the corresp. mem. node yet 
        s_gvn_ptr_d_y = 1      # the given ptr. desc. node with the given desc. set and above it
        s_gvn_ptr_d_n = 2      # the given ptr. desc. node with the given desc. NOT set and above it
        s_token_set_nnull = 3  # the ptr. moving token has been set, waiting for the null node
        s_token_set_null = 4   # the ptr. moving token has been set, the null node has been read
        s_token_set_undef = 5  # the ptr. moving token has been set, the undef node has been read
        s_accept = 6           # the ptr. moving token has been set, the right line number has been found
        s_reject = 7           # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 10, 11 -> s_init
        #   - 01 & descr_no set -> s_gvn_ptr_d_y
        #   - 01 & descr_no NOT set -> s_gvn_ptr_d_n
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,10,0)) #4 - 00
	bdd.append((3,0,-2)) #5 - 01 (note: -2 will later be changed)
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	# processing the last (right-most) pointer descriptor
	bdd[5]=(bdd[5][0],bdd[5][1],index) 
	index_bak=index
	bdd.append((4,index+1,-2)) 
	bdd.append((5,index+2,0)) # 010 (i.e. null or undef: descr_no not set)
        index=index+2
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a not null/undef ptr. descr. node
	bdd[index_bak]=(bdd[index_bak][0],bdd[index_bak][1],index) 
	bdd.append((5,0,index+1)) 
        index=index+1
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-3,6,descr_no-1)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 -> s_token_set_nnull provided y is set 
        #   - 10/11 -> s_init provided y is NOT set 
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_node_Ret_st_dep_on_y_Set_x_to_y(bdd,index,node_width-2,4,ptr_num,y,s_init,s_token_set_nnull)
                # ptr_num is the first bit in the section of markers, i.e. the descriptor token


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the appr. ptr. d. was not set
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_token_set_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_token_set_nnull (nnull came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_y, right state = s_token_set_nnull (nnull coming from a lower top memory node):
	# left state = s_gvn_ptr_d_n, right state = s_token_set_nnull:
        #   - 01/10/11 -> s_token_set_nnull
        #   - 00 (the null node) -> s_token_set_null
	#****************************************
        table.append((s_token_set_nnull,s_init,index))
        table.append((s_token_set_nnull,s_gvn_ptr_d_y,index))
        table.append((s_token_set_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_token_set_nnull,index))
        table.append((s_gvn_ptr_d_y,s_token_set_nnull,index))
        table.append((s_gvn_ptr_d_n,s_token_set_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_token_set_nnull,0)) # the null node has not been found yet
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_token_set_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_token_set_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_token_set_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_token_set_undef,0))  # the undef node has just been processed
        index=index+1

        
       	#****************************************
	# left state = s_token_set_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_token_set_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line1,line2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_ph1_lst_desc


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------

# def x_assign_y_next_ph1_lst_desc(node_width,ptr_num,next_num,x,y,line1,line2,descr_no):

def test_x_assign_y_next_ph1_lst_desc():
	"""test_x_assign_y_next_ph1_lst_desc"""
	(bdd,table,accept_l)=x_assign_y_next_ph1_lst_desc(9,4,4,1,2,"000000000","000000001",1)
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph1_lst_desc_1.gta")
	(bdd,table,accept_l)=x_assign_y_next_ph1_lst_desc(9,4,4,1,2,"000000000","000000001",2)
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph1_lst_desc_2.gta")

# test_x_assign_y_next_ph1_lst_desc()


#----------------------------------------------------------------------------------


def x_assign_y_next_ph1_desc(node_width,ptr_num,desc_num,next_num,x,y,next,line1,line2,prefix):
	'''x=y->next (phase 1: pre-processing for using descriptors)'''

	for i in range(desc_num):
		if (i!=0): # 0 is the ptr. desc. type flag
			if (next==next_num-1):
				(bdd,table,accept_l)=x_assign_y_next_ph1_lst_desc(node_width,ptr_num,next_num,x,y,line1,line2,i)
				#functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph1_lst_desc_%i.gta"%i)
				functions.gen_gta_file(bdd,table,accept_l,"%s%i.gta"%(prefix,i))
			else:
				(bdd,table,accept_l)=x_assign_y_next_ph1_mid_desc(node_width,ptr_num,next_num,x,y,next,line1,line2,i)
				#functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph1_mid_desc_%i.gta"%i)
				functions.gen_gta_file(bdd,table,accept_l,"%s%i.gta"%(prefix,i))



#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------

def test_x_assign_y_next_ph1_desc():
	"""test_x_assign_y_next_ph1_desc"""
	x_assign_y_next_ph1_desc(9,4,3,4,1,2,1,"000000000","000000001","x_assign_y_next_ph1_desc_")
	x_assign_y_next_ph1_desc(9,4,3,4,1,2,3,"000000000","000000001","x_assign_y_next_ph1_desc_")

#test_x_assign_y_next_ph1_desc()


#=================================================================================


def x_assign_y_next_y_null(node_width,y,line1):
	'''x=y->next (y is null)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# y           -- the numbers of the pointer variable being used

# We use the following states in the transducer:
        s_init = 0             # not read a top node yet
        s_top = 1              # read a top node
        s_null = 2             # read the null node
        s_undef = 3            # read the undef node
        s_accept = 4           # accept: read line1 and changed it to the error notification
        s_reject = 5           # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 01, 10 -> s_init
        #   - 11 -> s_top
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,10,0)) #4 - 00
	bdd.append((3,0,10)) #5 - 01
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,-2)) #9 - 11 (note: -2 will later be changed)
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	# processing a top node
	bdd[9]=(bdd[9][0],bdd[9][1],index) 
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_top,0))
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_top:
        #   - 11 -> s_top
	#****************************************
        table.append((s_init,s_top,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 11
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_top,0)) 
        index=index+1


       	#****************************************
	# left state = s_top, right state = s_init:
        #   - 00 & (y is null) -> s_null
	#   - 00 & (y is not null) -> s_reject
	#****************************************
        table.append((s_top,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00
        index=index+4
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,y)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_reject,0)) # y is not null
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_null,0)) # y is really null
        index=index+1


       	#****************************************
	# left state = s_null, right state = s_init:
        #   - 00 -> s_undef
	#****************************************
        table.append((s_null,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_undef,0)) 
        index=index+1


       	#****************************************
	# left state = s_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to ERROR)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line1)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_y_null


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_assign_y_next_y_null(node_width,y,line1):

def test_x_assign_y_next_y_null():
	"""test_x_assign_y_next_y_null"""
	(bdd,table,accept_l)=x_assign_y_next_y_null(9,2,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_null.gta")

#test_x_assign_y_next_y_null()


#=================================================================================

def x_assign_y_next_y_undef(node_width,y,line1):
	'''x=y->next (y is undefined)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# y           -- the numbers of the pointer variable being used

# We use the following states in the transducer:
        s_init = 0             # not read a top node yet
        s_top = 1              # read a top node
        s_null = 2             # read the null node
        s_undef = 3            # read the undef node
        s_accept = 4           # accept: read line1 and changed it to the error notification
        s_reject = 5           # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 01, 10 -> s_init
        #   - 11 -> s_top
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,10,0)) #4 - 00
	bdd.append((3,0,10)) #5 - 01
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,-2)) #9 - 11 (note: -2 will later be changed)
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	# processing a top node
	bdd[9]=(bdd[9][0],bdd[9][1],index) 
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_top,0))
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_top:
        #   - 11 -> s_top
	#****************************************
        table.append((s_init,s_top,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 11
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_top,0)) 
        index=index+1


       	#****************************************
	# left state = s_top, right state = s_init:
        #   - 00 -> s_null
	#****************************************
        table.append((s_top,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_null,0)) 
        index=index+1

	
       	#****************************************
	# left state = s_null, right state = s_init:
        #   - 00 & (y is set) -> s_undef
	#   - 00 & (y is not set) -> s_reject
	#****************************************
        table.append((s_null,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00
        index=index+4
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,y)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_reject,0)) # y is not undefined
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_undef,0)) # y is really undefined
        index=index+1


       	#****************************************
	# left state = s_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to ERROR)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line1)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_y_undef


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_assign_y_next_y_undef(node_width,y,line1):

def test_x_assign_y_next_y_undef():
	"""test_x_assign_y_next_y_undef"""
	(bdd,table,accept_l)=x_assign_y_next_y_undef(9,2,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_undef.gta")

#test_x_assign_y_next_y_undef()


#=================================================================================


def x_assign_y_next_y_next_null_mid(node_width,ptr_num,next_num,x,y,next,line1,line2):
	'''x=y->next (a middle next pointer that is null)'''

# Middle here means that it is not the right-most (bottom-most) ptr. descriptor node. 

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num      -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0             # not read a useful pair of a ptr. desc. and the corresp. mem. node yet 
        s_bef_gvn_ptr_d = 10   # ptr. desc. nodes before the tested one (we use s_head_ptr_d + 0, + 1, ...)
        s_gvn_ptr_d_y = 1      # the given ptr. desc. node set to NULL and above it
        s_gvn_ptr_d_n = 2      # the given ptr. desc. node NOT set to NULL and above it
        s_null_dsc_nnull = 3   # the null descr. found, waiting for the null node
        s_null_dsc_null = 4    # the null descr. found, the null node has been read
        s_null_dsc_undef = 5   # the null descr. found, the undef node has been read
        s_accept = 6           # the null descr. found, the right line number has been found
        s_reject = 7           # reject: malformed, not the right line number, etc.

	if ((next<0) or (next>=next_num-1)):
		functions.error_msg_expl("x_assign_y_next_y_next_null_mid: use a different function for x=y->next, y->next==null.!");
        	# use x_assign_y_next_y_next_null_mid

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 10, 11 -> s_init
        #   - 01 -> s_bef_gvn_ptr_d+0
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,-2,0)) #4 - 00
	bdd.append((3,0,-2)) #5 - 01 (note: -2 will later be changed)
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
        index=index+9
	# a (top) memory node
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	# the init node
	bdd[4]=(bdd[4][0],index,bdd[4][2]) 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_init,0)) 
        index=index+1
	# processing the last (right-most) pointer descriptor
	bdd[5]=(bdd[5][0],bdd[5][1],index) 
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) 
        bdd.append((-1,s_bef_gvn_ptr_d,0)) 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + i:
        #   - 01 -> s_bef_gvn_ptr_d + i+1 ... till just under processing the given ptr. desc. 
	#****************************************
        for i in range(next_num - next - 2):
                table.append((s_init,s_bef_gvn_ptr_d + i,index))
	        bdd.append((0,index+1,0)) #index
	        bdd.append((1,index+2,0)) #+1
	        bdd.append((2,0,index+3)) #+2
	        bdd.append((3,0,index+4)) #+3 - 01
                index=index+4
                (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
                bdd.append((-1,s_bef_gvn_ptr_d + i+1,0)) 
                index=index+1
                

       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_y / _n
	#****************************************
        table.append((s_init,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index_bak_1=index+4
	bdd.append((4,index+5,-2)) #+4
	bdd.append((5,index+6,0)) #+5 - 01|0 (i.e. null or undef)
	index_bak_2=index+6
	bdd.append((6,-2,index+7)) #+6
	bdd.append((7,0,index+8)) #+7 - 01|01
	index_bak_3=index+8
	bdd.append((8,index+9,-2)) #+8
	bdd.append((9,index+10,0)) #+9 - 01|010 (i.e. null)
	# The given ptr. desc. is null
        index=index+10
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10) #+10
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y 
        index=index+1
	# processing a regular ptr. descr. node (011)
	bdd[index_bak_1]=(bdd[index_bak_1][0],bdd[index_bak_1][1],index) 
	bdd.append((5,0,index+1)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a special, but not null/undef, ptr. descr. node (0100)
	bdd[index_bak_2]=(bdd[index_bak_2][0],index,bdd[index_bak_2][2]) 
	bdd.append((7,index+1,0)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-4,8)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing an undef ptr. descr. node (01011)
	bdd[index_bak_3]=(bdd[index_bak_3][0],bdd[index_bak_3][1],index) 
	bdd.append((9,0,index+1)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	

       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 -> s_null_dsc_nnull provided y is set 
        #   - 10/11 -> s_init provided y is NOT set 
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-2,4,y,x,s_init,s_null_dsc_nnull)
                # The semantics is (1) to return a state wrt. x and (2) to unset y.


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the appr. ptr. d. was not set
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_token_set_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_token_set_nnull, right state = s_bef_gvn_ptr_d + i (nnull is appearing in the middle of the ptr.d. chain):
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_token_set_nnull (nnull came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_y, right state = s_token_set_nnull (nnull coming from a lower top memory node):
	# left state = s_gvn_ptr_d_n, right state = s_token_set_nnull:
        #   - 01/10/11 -> s_null_dsc_nnull
        #   - 00 (the null node) -> s_null_dsc_null
	#****************************************
        table.append((s_null_dsc_nnull,s_init,index))
        for i in range(next_num - next - 1):
                table.append((s_null_dsc_nnull,s_bef_gvn_ptr_d + i,index))
        table.append((s_null_dsc_nnull,s_gvn_ptr_d_y,index))
        table.append((s_null_dsc_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_null_dsc_nnull,index))
        table.append((s_gvn_ptr_d_y,s_null_dsc_nnull,index))
        table.append((s_gvn_ptr_d_n,s_null_dsc_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,-2)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
	# a (top) memory node
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+9
        bdd.append((-1,s_null_dsc_nnull,0)) # the null node has not been found yet
        index=index+1
	# a ptr. desc. node
	bdd[index_bak+1]=(bdd[index_bak+1][0],bdd[index_bak+1][1],index)
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_null_dsc_nnull,0)) # the null node has not been found yet
        index=index+1
	# the null node
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,node_width-2,4,x) # we set x to null
        bdd.append((-1,s_null_dsc_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_null_dsc_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_null_dsc_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+4
        bdd.append((-1,s_null_dsc_undef,0))  # the undef node has just been processed
        index=index+1
        

       	#****************************************
	# left state = s_null_dsc_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_null_dsc_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line1,line2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_y_next_null_mid


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------

# def x_assign_y_next_y_next_null_mid(node_width,ptr_num,next_num,x,y,next,line1,line2):

def test_x_assign_y_next_y_next_null_mid():
	"""test_x_assign_y_next_y_next_null_mid"""
	(bdd,table,accept_l)=x_assign_y_next_y_next_null_mid(9,4,4,1,2,0,"000000000","000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_null_mid.gta")

# test_x_assign_y_next_y_next_null_mid()


#----------------------------------------------------------------------------------


def x_assign_y_next_y_next_null_lst(node_width,ptr_num,next_num,x,y,line1,line2):
	'''x=y->next (the last next pointer that is null)'''

# The last here means the right-most, i.e. the bottom-most, one.

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num      -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0             # not read a useful pair of a ptr. desc. and the corresp. mem. node yet 
        s_gvn_ptr_d_y = 1      # the given ptr. desc. node that is null and above it
        s_gvn_ptr_d_n = 2      # the given ptr. desc. node that is NOT null and above it
 	s_null_dsc_nnull = 3   # the given descr. is null, waiting for the null node
        s_null_dsc_null = 4    # the given descr. is null, the null node has been read
        s_null_dsc_undef = 5   # the given descr. is null, the undef node has been read
        s_accept = 6           # the given descr. is null, the right line number has been found
        s_reject = 7           # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 10, 11 -> s_init
        #   - 01 & null -> s_gvn_ptr_d_y
        #   - 01 & NOT null -> s_gvn_ptr_d_n
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,-2,0)) #4 - 00
	bdd.append((3,0,-2)) #5 - 01 (note: -2 will later be changed)
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
	# a (top) memory node
        index=index+9
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
        bdd.append((-1,s_init,0)) 
        index=index+1
	# the init node
	bdd[4]=(bdd[4][0],index,bdd[4][2]) 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_init,0)) 
        index=index+1
	# processing the last (right-most) pointer descriptor
	bdd[5]=(bdd[5][0],bdd[5][1],index) 
        index_bak_1=index
	bdd.append((4,index+1,-2))
	bdd.append((5,index+2,0)) # 01|0 (i.e. null or undef)
	index_bak_2=index+2
	bdd.append((6,-2,index+3))
	bdd.append((7,0,index+4)) # 01|01
	index_bak_3=index+4
	bdd.append((8,index+5,-2))
	bdd.append((9,index+6,0)) # 01|010 (i.e. null)
	# The given ptr. desc. is null
        index=index+6
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10)
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y 
        index=index+1
	# processing a regular ptr. descr. node (011)
	bdd[index_bak_1]=(bdd[index_bak_1][0],bdd[index_bak_1][1],index) 
	bdd.append((5,0,index+1)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a special, but not null/undef, ptr. descr. node (0100)
	bdd[index_bak_2]=(bdd[index_bak_2][0],index,bdd[index_bak_2][2]) 
	bdd.append((7,index+1,0)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-4,8)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing an undef ptr. descr. node (01011)
	bdd[index_bak_3]=(bdd[index_bak_3][0],bdd[index_bak_3][1],index) 
	bdd.append((9,0,index+1)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 -> s_null_dsc_nnull provided y is set 
        #   - 10/11 -> s_init provided y is NOT set 
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-2,4,y,x,s_init,s_null_dsc_nnull)


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the appr. ptr. d. was not set
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_token_set_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_token_set_nnull (nnull came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_y, right state = s_token_set_nnull (nnull coming from a lower top memory node):
	# left state = s_gvn_ptr_d_n, right state = s_token_set_nnull:
        #   - 01/10/11 -> s_null_dsc_nnull
        #   - 00 (the null node) -> s_null_dsc_null
	#****************************************
        table.append((s_null_dsc_nnull,s_init,index))
        table.append((s_null_dsc_nnull,s_gvn_ptr_d_y,index))
        table.append((s_null_dsc_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_null_dsc_nnull,index))
        table.append((s_gvn_ptr_d_y,s_null_dsc_nnull,index))
        table.append((s_gvn_ptr_d_n,s_null_dsc_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,-2)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
	# a (top) memory node
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+9
        bdd.append((-1,s_null_dsc_nnull,0)) # the null node has not been found yet
        index=index+1
	# a ptr. desc. node
	bdd[index_bak+1]=(bdd[index_bak+1][0],bdd[index_bak+1][1],index)
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_null_dsc_nnull,0)) # the null node has not been found yet
        index=index+1
	# the null node
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,node_width-2,4,x) # we set x to null
        bdd.append((-1,s_null_dsc_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_null_dsc_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_null_dsc_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+4
        bdd.append((-1,s_null_dsc_undef,0))  # the undef node has just been processed
        index=index+1

        

       	#****************************************
	# left state = s_null_dsc_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_null_dsc_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line1,line2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_y_next_null_mid


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------

# def x_assign_y_next_y_next_null_lst(node_width,ptr_num,next_num,x,y,line1,line2):

def test_x_assign_y_next_y_next_null_lst():
	"""test_x_assign_y_next_y_next_null_lst"""
	(bdd,table,accept_l)=x_assign_y_next_y_next_null_lst(9,4,4,1,2,"000000000","000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_null_lst.gta")

# test_x_assign_y_next_y_next_null_lst()


#----------------------------------------------------------------------------------


def x_assign_y_next_y_next_null(node_width,ptr_num,desc_num,next_num,x,y,next,line1,line2,file_name):
	'''x=y->next (the next pointer is null)'''
        if (next==next_num-1):
		(bdd,table,accept_l)=x_assign_y_next_y_next_null_lst(node_width,ptr_num,next_num,x,y,line1,line2)
		#functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_null_lst.gta")
		functions.gen_gta_file(bdd,table,accept_l,file_name)
	else:
                (bdd,table,accept_l)=x_assign_y_next_y_next_null_mid(node_width,ptr_num,next_num,x,y,next,line1,line2)
		#functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_null_mid.gta")
		functions.gen_gta_file(bdd,table,accept_l,file_name)


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


def test_x_assign_y_next_y_next_null():
	"""test_x_assign_y_next_y_next_null"""
        x_assign_y_next_y_next_null(9,4,3,4,1,2,0,"000000000","000000001","x_assign_y_next_y_next_null_mid.gta")
        x_assign_y_next_y_next_null(9,4,3,4,1,2,3,"000000000","000000001","x_assign_y_next_y_next_null_lst.gta")

#test_x_assign_y_next_y_next_null()


#=================================================================================


def x_assign_y_next_y_next_undef_mid(node_width,ptr_num,next_num,x,y,next,line1):
	'''x=y->next (a middle next pointer that is undef)'''

# Middle here means that it is not the right-most (bottom-most) ptr. descriptor node. 

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num      -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0             # not read a useful pair of a ptr. desc. and the corresp. mem. node yet 
        s_bef_gvn_ptr_d = 10   # ptr. desc. nodes before the tested one (we use s_head_ptr_d + 0, + 1, ...)
        s_gvn_ptr_d_y = 1      # the given ptr. desc. node set to undef and above it
        s_gvn_ptr_d_n = 2      # the given ptr. desc. node NOT set to undef and above it
        s_undef_dsc_nnull = 3  # the undef descr. found, waiting for the null node
        s_undef_dsc_null = 4   # the undef descr. found, the null node has been read
        s_undef_dsc_undef = 5  # the undef descr. found, the undef node has been read
        s_accept = 6           # the undef descr. found, the right line number has been found
        s_reject = 7           # reject: malformed, not the right line number, etc.

	if ((next<0) or (next>=next_num-1)):
		functions.error_msg_expl("x_assign_y_next_y_next_null_mid: use a different function for x=y->next, y->next==null.!");
        	# use x_assign_y_next_y_next_null_mid

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 10, 11 -> s_init
        #   - 01 -> s_bef_gvn_ptr_d+0
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,10,0)) #4 - 00
	bdd.append((3,0,-2)) #5 - 01 (note: -2 will later be changed)
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
        index=index+9
	# init or a (top) memory node
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	# processing the last (right-most) pointer descriptor
	bdd[5]=(bdd[5][0],bdd[5][1],index) 
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) 
        bdd.append((-1,s_bef_gvn_ptr_d,0)) 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + i:
        #   - 01 -> s_bef_gvn_ptr_d + i+1 ... till just under processing the given ptr. desc. 
	#****************************************
        for i in range(next_num - next - 2):
                table.append((s_init,s_bef_gvn_ptr_d + i,index))
	        bdd.append((0,index+1,0)) #index
	        bdd.append((1,index+2,0)) #+1
	        bdd.append((2,0,index+3)) #+2
	        bdd.append((3,0,index+4)) #+3 - 01
                index=index+4
                (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
                bdd.append((-1,s_bef_gvn_ptr_d + i+1,0)) 
                index=index+1
                

       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_y / _n
	#****************************************
        table.append((s_init,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index_bak_1=index+4
	bdd.append((4,index+5,-2)) #+4
	bdd.append((5,index+6,0)) #+5 - 01|0 (i.e. null or undef)
	index_bak_2=index+6
	bdd.append((6,-2,index+7)) #+6
	bdd.append((7,0,index+8)) #+7 - 01|01
	index_bak_3=index+8
	bdd.append((8,-2,index+9)) #+8
	bdd.append((9,0,index+10)) #+9 - 01|011 (i.e. undef)
	# The given ptr. desc. is undef
        index=index+10
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10) #+10
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y 
        index=index+1
	# processing a regular ptr. descr. node (011)
	bdd[index_bak_1]=(bdd[index_bak_1][0],bdd[index_bak_1][1],index) 
	bdd.append((5,0,index+1)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a special, but not null/undef, ptr. descr. node (0100)
	bdd[index_bak_2]=(bdd[index_bak_2][0],index,bdd[index_bak_2][2]) 
	bdd.append((7,index+1,0)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-4,8)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a null ptr. descr. node (01010)
	bdd[index_bak_3]=(bdd[index_bak_3][0],index,bdd[index_bak_3][2]) 
	bdd.append((9,index+1,0)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	

       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 -> s_undef_dsc_nnull provided y is set 
        #   - 10/11 -> s_init provided y is NOT set 
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-2,4,y,ptr_num,s_init,s_undef_dsc_nnull)
		# we unset the token which is not set -- just for not having a suitable library function


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the appr. ptr. d. was not set
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_undef_dsc_nnull, right state = s_init:
	# left state = s_undef_dsc_nnull, right state = s_bef_gvn_ptr_d + i:
	# left state = s_undef_dsc_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_undef_dsc_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_undef_dsc_nnull:
	# left state = s_gvn_ptr_d_y, right state = s_undef_dsc_nnull:
	# left state = s_gvn_ptr_d_n, right state = s_undef_dsc_nnull:
        #   - 01/10/11 -> s_undef_dsc_nnull
        #   - 00 (the null node) -> s_undef_dsc_null
	#****************************************
        table.append((s_undef_dsc_nnull,s_init,index))
        for i in range(next_num - next - 1):
                table.append((s_undef_dsc_nnull,s_bef_gvn_ptr_d + i,index))
        table.append((s_undef_dsc_nnull,s_gvn_ptr_d_y,index))
        table.append((s_undef_dsc_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_undef_dsc_nnull,index))
        table.append((s_gvn_ptr_d_y,s_undef_dsc_nnull,index))
        table.append((s_gvn_ptr_d_n,s_undef_dsc_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
	# a (top) memory node or a ptr. descr.
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_undef_dsc_nnull,0)) # the null node has not been found yet
        index=index+1
	# the null node
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_undef_dsc_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_undef_dsc_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_undef_dsc_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_undef_dsc_undef,0))  # the undef node has just been processed
        index=index+1
        

       	#****************************************
	# left state = s_undef_dsc_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_undef_dsc_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line1)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_y_next_undef_mid


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------

# def x_assign_y_next_y_next_undef_mid(node_width,ptr_num,next_num,x,y,next,line1):

def test_x_assign_y_next_y_next_undef_mid():
	"""test_x_assign_y_next_y_next_undef_mid"""
	(bdd,table,accept_l)=x_assign_y_next_y_next_undef_mid(9,4,4,1,2,2,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_undef_mid.gta")

# test_x_assign_y_next_y_next_undef_mid()


#----------------------------------------------------------------------------------


def x_assign_y_next_y_next_undef_lst(node_width,ptr_num,next_num,x,y,line1):
	'''x=y->next (the last next pointer that is undef)'''

# The last here means the right-most, i.e. the bottom-most, one.

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num      -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0             # not read a useful pair of a ptr. desc. and the corresp. mem. node yet 
        s_gvn_ptr_d_y = 1      # the given ptr. desc. node that is undef and above it
        s_gvn_ptr_d_n = 2      # the given ptr. desc. node that is NOT undef and above it
 	s_undef_dsc_nnull = 3  # the given descr. is undef, waiting for the null node
        s_undef_dsc_null = 4   # the given descr. is undef, the null node has been read
        s_undef_dsc_undef = 5  # the given descr. is undef, the undef node has been read
        s_accept = 6           # the given descr. is undef, the right line number has been found
        s_reject = 7           # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 10, 11 -> s_init
        #   - 01 & null -> s_gvn_ptr_d_y
        #   - 01 & NOT null -> s_gvn_ptr_d_n
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,10,0)) #4 - 00
	bdd.append((3,0,-2)) #5 - 01 (note: -2 will later be changed)
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
	# init or a (top) memory node
        index=index+9
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #10
        bdd.append((-1,s_init,0)) 
        index=index+1
	# processing the last (right-most) pointer descriptor
	bdd[5]=(bdd[5][0],bdd[5][1],index) 
        index_bak_1=index
	bdd.append((4,index+1,-2))
	bdd.append((5,index+2,0)) # 01|0 (i.e. null or undef)
	index_bak_2=index+2
	bdd.append((6,-2,index+3))
	bdd.append((7,0,index+4)) # 01|01
	index_bak_3=index+4
	bdd.append((8,-2,index+5))
	bdd.append((9,0,index+6)) # 01|011 (i.e. undef)
	# The given ptr. desc. is undef
        index=index+6
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10)
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y 
        index=index+1
	# processing a regular ptr. descr. node (011)
	bdd[index_bak_1]=(bdd[index_bak_1][0],bdd[index_bak_1][1],index) 
	bdd.append((5,0,index+1)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a special, but not null/undef, ptr. descr. node (0100)
	bdd[index_bak_2]=(bdd[index_bak_2][0],index,bdd[index_bak_2][2]) 
	bdd.append((7,index+1,0)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-4,8)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1
	# processing a null ptr. descr. node (01010)
	bdd[index_bak_3]=(bdd[index_bak_3][0],index,bdd[index_bak_3][2]) 
	bdd.append((9,index+1,0)) 
        index=index+1
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-5,10)
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 -> s_undef_dsc_nnull provided y is set 
        #   - 10/11 -> s_init provided y is NOT set 
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-2,4,y,ptr_num,s_init,s_undef_dsc_nnull)
		# we unset the token which is not set -- just for not having a suitable library function


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the appr. ptr. d. was not set
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_token_set_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_token_set_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_token_set_nnull (nnull came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_y, right state = s_token_set_nnull (nnull coming from a lower top memory node):
	# left state = s_gvn_ptr_d_n, right state = s_token_set_nnull:
        #   - 01/10/11 -> s_undef_dsc_nnull
        #   - 00 (the null node) -> s_undef_dsc_null
	#****************************************
        table.append((s_undef_dsc_nnull,s_init,index))
        table.append((s_undef_dsc_nnull,s_gvn_ptr_d_y,index))
        table.append((s_undef_dsc_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_undef_dsc_nnull,index))
        table.append((s_gvn_ptr_d_y,s_undef_dsc_nnull,index))
        table.append((s_gvn_ptr_d_n,s_undef_dsc_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
	# a (top) memory node or a ptr. desc.
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_undef_dsc_nnull,0)) # the null node has not been found yet
        index=index+1
	# the null node
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_undef_dsc_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_undef_dsc_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_undef_dsc_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_undef_dsc_undef,0))  # the undef node has just been processed
        index=index+1

        
       	#****************************************
	# left state = s_undef_dsc_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_undef_dsc_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line1)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_y_next_undef_lst


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_assign_y_next_y_next_undef_lst(node_width,ptr_num,next_num,x,y,line1):

def test_x_assign_y_next_y_next_undef_lst():
	"""test_x_assign_y_next_y_next_undef_lst"""
	(bdd,table,accept_l)=x_assign_y_next_y_next_undef_lst(9,4,4,1,2,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_undef_lst.gta")

# test_x_assign_y_next_y_next_undef_lst()


#----------------------------------------------------------------------------------


def x_assign_y_next_y_next_undef(node_width,ptr_num,desc_num,next_num,x,y,next,line1,file_name):
	'''x=y->next (the next pointer is undefined)'''
        if (next==next_num-1):
		(bdd,table,accept_l)=x_assign_y_next_y_next_undef_lst(node_width,ptr_num,next_num,x,y,line1)
		#functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_undef_lst.gta")
		functions.gen_gta_file(bdd,table,accept_l,file_name)
	else:
                (bdd,table,accept_l)=x_assign_y_next_y_next_undef_mid(node_width,ptr_num,next_num,x,y,next,line1)
		#functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_y_next_undef_mid.gta")
		functions.gen_gta_file(bdd,table,accept_l,file_name)


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


def test_x_assign_y_next_y_next_undef():
	"""test_x_assign_y_next_y_next_undef"""
        x_assign_y_next_y_next_undef(9,4,3,4,1,2,2,"000000000","x_assign_y_next_y_next_undef_mid.gta")
        x_assign_y_next_y_next_undef(9,4,3,4,1,2,3,"000000000","x_assign_y_next_y_next_undef_lst.gta")

#test_x_assign_y_next_y_next_undef()


#=================================================================================


def x_assign_y_next_ph2_reg(node_width,ptr_num,x):
	'''x=y->next (phase 2 regular: removing the target token and really moving x)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num      -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)
# descr_no    -- the number of the descriptor to be used (numbered from one)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0       # not found the target token yet
        s_token = 1      # the token was found and x moved, waiting for the null node
        s_null = 2       # the null has been processed
        s_undef = 3      # the undef node has been processed
        s_accept = 4     # all changes done, the right line number has been found
        s_reject = 5     # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 01 -> s_init
        #   - 10/11 & token -> s_token
        #   - 10/11 & NOT token -> s_init
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,-2,0)) #4 - 00 (note: -2 will later be changed)
	bdd.append((3,0,-2)) #5 - 01 
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
	# a (top) memory node
	bdd.append((4,-2,11)) #10 - deleted or not?
	bdd.append((5,0,12)) #11 - deleted!
        index=index+11
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6) #12
        bdd.append((-1,s_init,0)) # we are interested in not deleted nodes only now
        index=index+1
        # a non-deleted (top) memory node
	bdd[10]=(bdd[10][0],index,bdd[10][2]) 
	bdd.append((5,index+1,0)) #11 - not deleted
        index=index+1
        (bdd,index)=functions.copy_node_Ret_st_dep_on_y_Set_x_to_y_Res_y(bdd,index,node_width-3,6,x-1,ptr_num-1,s_init,s_token)
        # an init or ptr. des. node
	bdd[4]=(bdd[4][0],index,bdd[4][2]) 
	bdd[5]=(bdd[5][0],bdd[5][1],index)
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_token, right state = s_init: 
	# left state = s_init, right state = s_token: 
	#   - 00 -> s_null
        #   - 01/10/11 -> s_token
	#****************************************
	table.append((s_token,s_init,index))
	table.append((s_init,s_token,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,-2)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        # 10/11 -- we just set x to zero
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+9
        bdd.append((-1,s_token,0)) # still waiting for the null node
        index=index+1
        # 01 -- we just copy
	bdd[index_bak+1]=(bdd[index_bak+1][0],bdd[index_bak+1][1],index) 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_token,0)) # still waiting for the null node
        index=index+1
        # 00 -- the null node
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2]) 
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x)
        bdd.append((-1,s_null,0))  # the null node has just been processed
        index=index+1


       	#****************************************
	# left state = s_null, right state = s_init: 
	#   - 00 -> s_undef
        #   - 01/10/11 -> s_reject
	#****************************************
	table.append((s_null,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00
        index=index+4
        # 00 -- the undef node
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,node_width-2,4,x) #+4
        bdd.append((-1,s_undef,0))  # the null node has just been processed
        index=index+1


       	#****************************************
	# left state = s_undef, right state = s_init: 
	#   - 0 -> s_accept
	#   - 1 -> s_reject
	#****************************************
	table.append((s_undef,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
        index=index+2
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-1,2) # we suppose the line to be already changed
        bdd.append((-1,s_accept,0))  # the null node has just been processed
        index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_ph2


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------

# def x_assign_y_next_ph2_reg(node_width,ptr_num,x):

def test_x_assign_y_next_ph2_reg():
	"""test_x_assign_y_next_ph2_reg"""
	(bdd,table,accept_l)=x_assign_y_next_ph2_reg(9,4,1)
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph2_reg.gta")

# test_x_assign_y_next_ph2_reg()


#=================================================================================


def x_assign_y_next_ph2_del(node_width,ptr_num,line2):
	'''x=y->next (phase 2 deleted: the target is a deleted node---we raise an error)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note: We SUPPOSE that deleted nodes do not have any variable set!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num      -- the number of pointer variables including the deleted flag
# next_num    -- the number of the next pointers
# x,y         -- the numbers of the pointer variables being used
# next        -- the number of the next pointer used (numbered from zero)
# descr_no    -- the number of the descriptor to be used (numbered from one)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0       # not found the target token yet
        s_token = 1      # the token was found and x moved, waiting for the null node
        s_null = 2       # the null has been processed
        s_undef = 3      # the undef node has been processed
        s_accept = 4     # all changes done, the right line number has been found
        s_reject = 5     # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00, 01 -> s_init
        #   - 10/11 & token -> s_token
        #   - 10/11 & NOT token -> s_init
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,6)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,4,5)) #3
	bdd.append((3,-2,0)) #4 - 00 (note: -2 will later be changed)
	bdd.append((3,0,-2)) #5 - 01 
	bdd.append((1,0,7)) #6
	bdd.append((2,8,9)) #7
	bdd.append((3,10,0)) #8 - 10
	bdd.append((3,0,10)) #9 - 11
	# a (top) memory node
	bdd.append((4,11,-2)) #10 - deleted or not?
	bdd.append((5,12,0)) #11 - not deleted!
        index=index+11
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-3,6) #12
        bdd.append((-1,s_init,0)) # we are interested in the deleted nodes only now
        index=index+1
        # a deleted (top) memory node
	bdd[10]=(bdd[10][0],bdd[10][1],index) 
	bdd.append((5,0,index+1)) #11 - not deleted
        index=index+1
        (bdd,index)=functions.copy_node_Ret_st_dep_on_y_Set_x_to_y(bdd,index,node_width-3,6,ptr_num-1,ptr_num-1,s_init,s_token)
          # unlike in the regular case, we preserve the token as debug info
        # an init or ptr. des. node
	bdd[4]=(bdd[4][0],index,bdd[4][2]) 
	bdd[5]=(bdd[5][0],bdd[5][1],index)
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_token, right state = s_init: 
	# left state = s_init, right state = s_token: 
	#   - 00 -> s_null
        #   - 01/10/11 -> s_token
	#****************************************
	table.append((s_token,s_init,index))
	table.append((s_init,s_token,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,-2)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        # 10/11 -- unlike in the regular case, we keep x as debug info 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_token,0)) # still waiting for the null node
        index=index+1
        # 01 -- we just copy
	bdd[index_bak+1]=(bdd[index_bak+1][0],bdd[index_bak+1][1],index) 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_token,0)) # still waiting for the null node
        index=index+1
        # 00 -- the null node
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2]) 
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_null,0))  # the null node has just been processed
        index=index+1


       	#****************************************
	# left state = s_null, right state = s_init: 
	#   - 00 -> s_undef
        #   - 01/10/11 -> s_reject
	#****************************************
	table.append((s_null,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00
        index=index+4
        # 00 -- the undef node
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_undef,0))  # the null node has just been processed
        index=index+1


       	#****************************************
	# left state = s_undef, right state = s_init: 
	#   - 0 -> s_accept
	#   - 1 -> s_reject
	#****************************************
	table.append((s_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line2)
          # NOTE: we have changed the line already in phase 1.
          #  -- may be better to change only now, but this requires changes in phase1 and regular phase2...
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_assign_y_next_ph2_del


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------

# def x_assign_y_next_ph2_del(node_width,ptr_num,line2):

def test_x_assign_y_next_ph2_del():
	"""test_x_assign_y_next_ph2_del"""
	(bdd,table,accept_l)=x_assign_y_next_ph2_del(9,4,"000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph2_del.gta")

# test_x_assign_y_next_ph2_del()


#----------------------------------------------------------------------------------


def x_assign_y_next_ph2(node_width,ptr_num,x,line2):
	'''x=y->next (phase 2: removing the token, setting x)'''
	(bdd,table,accept_l)=x_assign_y_next_ph2_reg(node_width,ptr_num,x)
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph2_reg.gta")
        (bdd,table,accept_l)=x_assign_y_next_ph2_del(node_width,ptr_num,line2)
	functions.gen_gta_file(bdd,table,accept_l,"x_assign_y_next_ph2_del.gta")


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


def test_x_assign_y_next_ph2():
	"""test_x_assign_y_next_ph2"""
        x_assign_y_next_ph2(9,4,1,"000000001")

#test_x_assign_y_next_ph2()


#=================================================================================

