import functions
import sys

import gen_x_ass_ynext 

#=================================================================================

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# This version does not cover the possibility of new returning null when there is not enough memory.

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# Be careful when using the below functions -- 3 usage scenarios arise here:
# 1. x_next_assign_new adds a new element when all the necessary conditions hold.
# 2. x_next_assign_new_used returns an error which, however, only indicates that the
#    new element cannot be added directly under x->next and should be added
#    using X=new, x->next=X where X is a special purpose (or currently dead) variable.
# 3. x_next_assign_new_x_null/undef yield a real error when x is null or undef.

# Further, one should provide x_next_assign_new with the right numbers of
# the corresponding descr./marker pair to be used. The appropriate pointer descriptor
# transducer should be generated for this purpose using function gen_descr_down.

#=================================================================================

def x_next_assign_new_mid(node_width,ptr_num,descr_num,mark_num,next_num,data_size,init_data,x,next,descr,marker,line1,line2):
	'''x->next=new (a new top memory is linked below the x pointed node to the given (middle) next pointer -- if it's free)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note 1: We SUPPOSE that deleted nodes do not have any variable set!

# Note 2: We SUPPOSE that the node available via next from x is free (init) -- if not, we generate an error!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num     -- the number of pointer variables including the deleted flag
# descr_num   -- the number of the descriptors (including the type bit)
# mark_num    -- the number of markes (including the token) -- should be the same as the number of descriptors
# next_num    -- the number of the next pointers
# data_size   -- the size of the data (used only in case of enabled variance analysis)
# x           -- the number of the pointer variable being assigned
# next        -- the number of the next pointer used (numbered from zero)
# descr       -- the number of the descriptor to be used (numbered from one)
# marker      -- the marker to be set in the new node (typically the same as descr)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0               # reading initial nodes, ptr. desc., and mem. nodes not to be modified
        s_bef_gvn_ptr_d = 10     # ptr. desc. nodes before the tested one (we use s_bef_gvn_ptr_d + 0, + 1, ...)
        s_gvn_ptr_d_y = 1        # the given ptr. desc. node with the new node linked below it and above this node 
        s_gvn_ptr_d_n = 2        # the given ptr. desc. node without a new node and above it
	s_new_ptr_desc = 20      # creating the new ptr. desc. (+0, +1, ..., + next_num-1)
	s_new_node = 3           # created the new memory node
        s_new_nd_lnkd_nnull = 4  # the new node has been linked, waiting for the null node
        s_new_nd_lnkd_null = 5   # the new node has been linked, the null node has been read
        s_new_nd_lnkd_undef = 6  # the new node has been linked, the undef node has been read
        s_accept = 7             # all the changes done, the right line number has been found
        s_reject = 8             # reject: malformed, not the right line number, etc.

	if ((next<0) or (next>=next_num-1)):
		functions.error_msg_expl("x_next_assign_new_mid: use a different function for x->next=new than mid!");
        	# use x_next_assign_new_lst

	if ((next_num<1) or (next_num>=10)):
		functions.error_msg_expl("x_next_assign_new_mid:too many next pointers!");
        	# increase the gaps between s_bef_gvn_ptr_d and s_new_ptr_desc 

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00 -> s_init / s_new_ptr_desc
        #   - 01 -> s_bef_gvn_ptr_d + 0
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,0)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,-2,4)) #3
	bdd.append((3,0,5)) #4 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #5
        bdd.append((-1,s_bef_gvn_ptr_d,0)) 
        index=index+1
	# processing an init node
	bdd[3]=(bdd[3][0],index,bdd[3][2]) 
	bdd.append((3,index+1,index+2)) # keep init or change to a new ptr. desc.?
        index=index+1
	# we keep the node as init
        bdd.append((-1,s_init,0)) 
        index=index+1
	# we create a new ptr. desc.
	bdd.append((5,index+1,0)) # set the ptr. desc. to UNDEF (01|011) -- we thus ignore bits 4,6,8
	bdd.append((7,0,index+2))
	bdd.append((9,0,index+3))
        bdd.append((-1,s_new_ptr_desc,0)) 
        index=index+4


       	#****************************************
	# left state = s_init, right state = s_new_ptr_desc + i (i = +0, +1, ..., + next_num-2):
        #   - 00 -> s_new_ptr_desc + i+1 
	#****************************************
        for i in range(next_num - 1): # i.e. 0, 1, ..., next_num-2
		table.append((s_init,s_new_ptr_desc + i,index))
		bdd.append((0,index+1,0)) #index
		bdd.append((1,index+2,0)) #+1
		bdd.append((2,index+3,0)) #+2
		bdd.append((3,0,index+4)) #+3 - 00 -> 01
        	index=index+4
		bdd.append((5,index+1,0)) # set the ptr. desc. to UNDEF (01|011) -- we thus ignore bits 4,6,8
		bdd.append((7,0,index+2))
		bdd.append((9,0,index+3))
        	bdd.append((-1,s_new_ptr_desc + i+1,0)) 
        	index=index+4


       	#****************************************
	# left state = s_new_ptr_desc + next_num-1, right state = s_init:
        #   - 00 -> s_new_node
	#****************************************
	table.append((s_new_ptr_desc + next_num-1,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00 -> 10
       	index=index+4
	# we initialize the new node and ignore all even bits now:
	#   not deleted, not directly pointed to, only the given marker set, undefined data 
	for i in range(ptr_num + mark_num):
		if (i==ptr_num+marker):
			bdd.append((5+2*i,0,index+i+1)) # the given marker is set
		else:
			bdd.append((5+2*i,index+i+1,0)) # not deleted, not directly pointed to, no other markers set
       	index=index+ptr_num+mark_num
	# if data initialization is requested
	for i in range(len(init_data)):
		if init_data[i]=="0":
			bdd.append((5+((ptr_num+mark_num)*2)+(i*2),index+1,0))
			index=index+1
		if init_data[i]=="1":
			bdd.append((5+((ptr_num+mark_num)*2)+(i*2),0,index+1))
			index=index+1
		# in case of data[i]=="X", we skip this position
	# in case of enabled variance analysis, we need to set old values of the pointer variables
	# - i.e. there was no pointer variables pointing into this node before its creation
	#
	# in case of disbaled variance analysis node_width=2+ptr_num+mark_num+data_size and the loop 
	# will be not iterated
	for i in range (node_width-(2+ptr_num+mark_num+data_size+data_size)):
		print "DEBUG mid: %i %i"%(i,5+((ptr_num+mark_num+data_size)*2)+(i*2))
		bdd.append((5+((ptr_num+mark_num+data_size)*2)+(i*2),index+1,0))
		index=index+1

       	bdd.append((-1,s_new_node,0)) 
       	index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + i:
        #   - 01 -> s_bef_gvn_ptr_d + i+1 ... till just under processing the given ptr. desc. 
	#****************************************
        for i in range(next_num - next - 2):
                table.append((s_init,s_bef_gvn_ptr_d + i,index)) # we are so far looking for the place to put the 
	        bdd.append((0,index+1,0)) #index                 # new element 
	        bdd.append((1,index+2,0)) #+1
	        bdd.append((2,0,index+3)) #+2
	        bdd.append((3,0,index+4)) #+3 - 01
                index=index+4
                (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
                bdd.append((-1,s_bef_gvn_ptr_d + i+1,0)) 
                index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_n (the new node has not been added)
	#****************************************
        table.append((s_init,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1


       	#****************************************
	# left state = s_new_node, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_y & set descr. bit to descr 
	#****************************************
        table.append((s_new_node,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        # change the descriptor bit to only descr. set (with type 1)
	for i in range(descr_num):
		if (i==0) or (i==descr):
			bdd.append((5+2*i,0,index+i+1)) # the given regular (type 1) descriptor is set
		else:
			bdd.append((5+2*i,index+i+1,0)) # all other descriptor bits are set to zero
       	index=index+descr_num
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 & x set -> s_new_nd_lnkd_nnull 
        #   - 10/11 & x not set -> s_reject (as a new node has already been added)
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,x) #+5
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_reject,0)) # a new element added into a wrong place (as x is not set) 
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_new_nd_lnkd_nnull,0)) # a new element added into the right place
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the new node was not added
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_new_nd_lnkd_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_new_nd_lnkd_nnull, right state = s_bef_gvn_ptr_d + i (nd is appearing in the middle of the ptr.d. chain):
	# left state = s_new_nd_lnkd_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_new_nd_lnkd_nnull (nd came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_n, right state = s_new_nd_lnkd_nnull (nd coming from a lower top memory node):
        #   - 01/10/11 -> s_new_nd_lnkd_nnull
        #   - 00 (the null node) -> s_new_nd_lnkd_null
	# NOTE that ..._d_y does not appear here as we cannot add two nodes...
	#****************************************
        table.append((s_new_nd_lnkd_nnull,s_init,index))
        for i in range(next_num - next - 1):
                table.append((s_new_nd_lnkd_nnull,s_bef_gvn_ptr_d + i,index))
        table.append((s_new_nd_lnkd_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_new_nd_lnkd_nnull,index))
        table.append((s_gvn_ptr_d_n,s_new_nd_lnkd_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_new_nd_lnkd_nnull,0)) # the null node has not been found yet
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_lnkd_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_new_nd_lnkd_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_new_nd_lnkd_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_lnkd_undef,0))  # the undef node has just been processed
        index=index+1


       	#****************************************
	# left state = s_new_nd_lnkd_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_new_nd_lnkd_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line1,line2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_next_assign_new_mid


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_next_assign_new_mid(node_width,ptr_num,descr_num,mark_num,next_num,data_size,x,next,descr,marker,line1,line2):

def test_x_next_assign_new_mid():
	"""test_x_next_assign_new_mid"""
	(bdd,table,accept_l)=x_next_assign_new_mid(9,4,3,3,4,2,1,1,1,1,"000000000","000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_mid_1.gta")
	(bdd,table,accept_l)=x_next_assign_new_mid(9,4,3,3,4,1,1,2,1,1,"000000000","000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_mid_2.gta")

#test_x_next_assign_new_mid()


#----------------------------------------------------------------------------------


def x_next_assign_new_lst(node_width,ptr_num,descr_num,mark_num,next_num,data_size,init_data,x,descr,marker,line1,line2):
	'''x->next=new (a new top memory is linked below the x pointed node to the given right_most next pointer -- if it's free)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note 1: We SUPPOSE that deleted nodes do not have any variable set!

# Note 2: We SUPPOSE that the node available via next from x is free (init) -- if not, we generate an error!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# ptr_num     -- the number of pointer variables including the deleted flag
# descr_num   -- the number of the descriptors (including the type bit)
# mark_num    -- the number of markes (including the token) -- should be the same as the number of descriptors
# next_num    -- the number of the next pointers
# data_size   -- the size of the data (used only in case of enabled variance analysis)
# x           -- the number of the pointer variable being assigned
# descr       -- the number of the descriptor to be used (numbered from one)
# marker      -- the marker to be set in the new node (typically the same as descr)

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0               # reading initial nodes, ptr. desc., and mem. nodes not to be modified
        s_gvn_ptr_d_y = 1        # the given ptr. desc. node with the new node linked below it and above this node 
        s_gvn_ptr_d_n = 2        # the given ptr. desc. node without a new node and above it
	s_new_ptr_desc = 10      # creating the new ptr. desc. (+0, +1, ..., + next_num-1)
	s_new_node = 3           # created the new memory node
        s_new_nd_lnkd_nnull = 4  # the new node has been linked, waiting for the null node
        s_new_nd_lnkd_null = 5   # the new node has been linked, the null node has been read
        s_new_nd_lnkd_undef = 6  # the new node has been linked, the undef node has been read
        s_accept = 7             # all the changes done, the right line number has been found
        s_reject = 8             # reject: malformed, not the right line number, etc.

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00 -> s_init / s_new_ptr_desc + 0
        #   - 01 -> s_gvn_ptr_d_n 
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,0)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,-2,4)) #3
	bdd.append((3,0,5)) #4 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #5
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1
	# processing an init node
	bdd[3]=(bdd[3][0],index,bdd[3][2]) 
	bdd.append((3,index+1,index+2)) # keep init or change to a new ptr. desc.?
        index=index+1
	# we keep the node as init
        bdd.append((-1,s_init,0)) 
        index=index+1
	# we create a new ptr. desc.
	bdd.append((5,index+1,0)) # set the ptr. desc. to UNDEF (01|011) -- we thus ignore bits 4,6,8
	bdd.append((7,0,index+2))
	bdd.append((9,0,index+3))
        bdd.append((-1,s_new_ptr_desc,0)) 
        index=index+4


       	#****************************************
	# left state = s_init, right state = s_new_ptr_desc + i (i = +0, +1, ..., + next_num-2):
        #   - 00 -> s_new_ptr_desc + i+1 
	#****************************************
        for i in range(next_num - 1): # i.e. 0, 1, ..., next_num-2
		table.append((s_init,s_new_ptr_desc + i,index))
		bdd.append((0,index+1,0)) #index
		bdd.append((1,index+2,0)) #+1
		bdd.append((2,index+3,0)) #+2
		bdd.append((3,0,index+4)) #+3 - 00 -> 01
        	index=index+4
		bdd.append((5,index+1,0)) # set the ptr. desc. to UNDEF (01|011) -- we thus ignore bits 4,6,8
		bdd.append((7,0,index+2))
		bdd.append((9,0,index+3))
        	bdd.append((-1,s_new_ptr_desc + i+1,0)) 
        	index=index+4


       	#****************************************
	# left state = s_new_ptr_desc + next_num-1, right state = s_init:
        #   - 00 -> s_new_node
	#****************************************
	table.append((s_new_ptr_desc + next_num-1,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,0)) #+2
	bdd.append((3,index+4,0)) #+3 - 00 -> 10
       	index=index+4
	# we initialize the new node and ignore all even bits now:
	#   not deleted, not directly pointed to, only the given marker set, undefined data 
	for i in range(ptr_num + mark_num):
		if (i==ptr_num+marker):
			bdd.append((5+2*i,0,index+i+1)) # the given marker is set
		else:
			bdd.append((5+2*i,index+i+1,0)) # not deleted, not directly pointed to, no other markers set
       	index=index+ptr_num+mark_num
	# if data initialization is requested
	for i in range(len(init_data)):
		if init_data[i]=="0":
			bdd.append((5+((ptr_num+mark_num)*2)+(i*2),index+1,0))
			index=index+1
		if init_data[i]=="1":
			bdd.append((5+((ptr_num+mark_num)*2)+(i*2),0,index+1))
			index=index+1
		# in case of data[i]=="X", we skip this position

	# in case of enabled variance analysis, we need to set old values of the pointer variables
	# - i.e. there was no pointer variables pointing into this node before its creation
	#
	# in case of disbaled variance analysis node_width=2+ptr_num+mark_num+data_size and the loop 
	# will be not iterated
	for i in range (node_width-(2+ptr_num+mark_num+data_size+data_size)):
		bdd.append((5+((ptr_num+mark_num+data_size)*2)+(i*2),index+1,0))
		index=index+1

       	bdd.append((-1,s_new_node,0)) 
       	index=index+1


       	#****************************************
	# left state = s_new_node, right state = s_init: 
        #   - 01 -> s_gvn_ptr_d_y & set descr. bit to descr 
	#****************************************
	table.append((s_new_node,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        # change the descriptor bit to only descr. set (with type 1)
	for i in range(descr_num):
		if (i==0) or (i==descr):
			bdd.append((5+2*i,0,index+i+1)) # the given regular (type 1) descriptor is set
		else:
			bdd.append((5+2*i,index+i+1,0)) # all other descriptor bits are set to zero
       	index=index+descr_num
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 & x set -> s_new_nd_lnkd_nnull 
        #   - 10/11 & x not set -> s_reject (as a new node has already been added)
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,x)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_reject,0)) # a new element added into a wrong place (as x is not set) 
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_new_nd_lnkd_nnull,0)) # a new element added into the right place
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init as the new node was not added
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_new_nd_lnkd_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_new_nd_lnkd_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_init, right state = s_new_nd_lnkd_nnull (nd came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_n, right state = s_new_nd_lnkd_nnull (nd coming from a lower top memory node):
        #   - 01/10/11 -> s_new_nd_lnkd_nnull
        #   - 00 (the null node) -> s_new_nd_lnkd_null
	# NOTE that ..._d_y does not appear here as we cannot add two nodes...
	#****************************************
        table.append((s_new_nd_lnkd_nnull,s_init,index))
        table.append((s_new_nd_lnkd_nnull,s_gvn_ptr_d_n,index))
        table.append((s_init,s_new_nd_lnkd_nnull,index))
        table.append((s_gvn_ptr_d_n,s_new_nd_lnkd_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_new_nd_lnkd_nnull,0)) # the null node has not been found yet
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_lnkd_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_new_nd_lnkd_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_new_nd_lnkd_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_lnkd_undef,0))  # the undef node has just been processed
        index=index+1


       	#****************************************
	# left state = s_new_nd_lnkd_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to line2)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_new_nd_lnkd_undef,s_init,index))
	(bdd,index)=functions.change_line_num(bdd,index,line1,line2)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_next_assign_new_lst


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------



# def x_next_assign_new_lst(node_width,ptr_num,descr_num,mark_num,next_num,data_size,x,descr,marker,line1,line2):

def test_x_next_assign_new_lst():
	"""test_x_next_assign_new_lst"""
	(bdd,table,accept_l)=x_next_assign_new_lst(9,4,3,3,4,2,1,1,1,"000000000","000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_lst_1.gta")
	(bdd,table,accept_l)=x_next_assign_new_lst(9,4,3,3,4,1,1,1,1,"000000000","000000001")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_lst_2.gta")

#test_x_next_assign_new_lst()


#=================================================================================


# A wrapper to generate x->next=new when everything works.

def x_next_assign_new(node_width,ptr_num,descr_num,mark_num,next_num,data_size,init_data,x,next,descr,marker,line1,line2,filename):
	# if the data initialization is required, check whether the data fits
	if len(init_data)>data_size:
		print "ERROR: Gen_x.next=new: data %s does not fit in the size %i"%(init_data,data_size)
		sys.exit(1)
        if (next==next_num-1):
                (bdd,table,accept_l)=x_next_assign_new_lst(node_width,ptr_num,descr_num,mark_num,next_num,data_size,init_data,x,descr,marker,line1,line2)
        else:
                (bdd,table,accept_l)=x_next_assign_new_mid(node_width,ptr_num,descr_num,mark_num,next_num,data_size,init_data,x,next,descr,marker,line1,line2)
	#functions.gen_gta_file(bdd,table,accept_l,"x_%i_next_%i_assign_new_marked_%i.gta" % (x,next,marker))
	functions.gen_gta_file(bdd,table,accept_l,filename)


#=================================================================================


def x_next_assign_new_mid_used(node_width,next_num,x,next,line1):
	'''x->next=new (a new top memory cannot be linked below the x pointed node to the given (middle) next pointer -- it's NOT free)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note 1: We SUPPOSE that deleted nodes do not have any variable set!

# Note 2: We SUPPOSE that the node available via next from x is free (init) -- if not, we generate an error!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# descr_num   -- the number of the descriptors (including the type bit)
# x           -- the number of the pointer variable being assigned
# next        -- the number of the next pointer used (numbered from zero)

# We use the following states in the transducer:
        s_init = 0               # reading initial nodes, ptr. desc., and mem. nodes not to be modified
        s_bef_gvn_ptr_d = 10     # ptr. desc. nodes before the tested one (we use s_bef_gvn_ptr_d + 0, + 1, ...)
        s_gvn_ptr_d_y = 1        # the given ptr. desc. node has a node linked below it
        s_gvn_ptr_d_n = 2        # the given ptr. desc. node does not have a node linked below it
        s_new_nd_nlnk_nnull = 3  # the new node could not be linked, waiting for the null node
        s_new_nd_nlnk_null = 4   # the new node could not be linked, the null node has been read
        s_new_nd_nlnk_undef = 5  # the new node could not be linked, the undef node has been read
        s_accept = 6             # all the changes done, the right line number has been found
        s_reject = 7             # reject: malformed, not the right line number, etc.
        s_mem_node = 8           # a non-top memory node has been read

	if ((next<0) or (next>=next_num-1)):
		functions.error_msg_expl("x_next_assign_new_mid: use a different function for x->next=new than mid!");
        	# use x_next_assign_new_lst

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00 -> s_init
        #   - 01 -> s_bef_gvn_ptr_d + 0
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,0)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,-2,4)) #3
	bdd.append((3,0,5)) #4 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #5
        bdd.append((-1,s_bef_gvn_ptr_d,0)) 
        index=index+1
	# processing an init node
	bdd[3]=(bdd[3][0],index,bdd[3][2]) 
	bdd.append((3,index+1,0)) # keep init
        index=index+1
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_mem_node, right state = s_init: 
        #   - 01 -> s_bef_gvn_ptr_d + 0
	#****************************************
	table.append((s_mem_node,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_bef_gvn_ptr_d,0)) 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + i:
	# left state = s_mem_node, right state = s_bef_gvn_ptr_d + i:
        #   - 01 -> s_bef_gvn_ptr_d + i+1 ... till just under processing the given ptr. desc. 
	#****************************************
        for i in range(next_num - next - 2):
                table.append((s_init,s_bef_gvn_ptr_d + i,index)) 
                table.append((s_mem_node,s_bef_gvn_ptr_d + i,index))
	        bdd.append((0,index+1,0)) #index
	        bdd.append((1,index+2,0)) #+1
	        bdd.append((2,0,index+3)) #+2
	        bdd.append((3,0,index+4)) #+3 - 01
                index=index+4
                (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
                bdd.append((-1,s_bef_gvn_ptr_d + i+1,0)) 
                index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_n (the new node could be added)
	#****************************************
        table.append((s_init,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1


       	#****************************************
	# left state = s_mem_node, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_y (the new node could not be added)
	#****************************************
        table.append((s_mem_node,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
	# left state = s_mem_node, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
        table.append((s_mem_node,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
	# left state = s_gvn_ptr_d_y, right state = s_mem_node:
        #   - 10/11 & x set -> s_new_nd_nlnk_nnull 
        #   - 10/11 & x not set -> s_mem_node
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
        table.append((s_gvn_ptr_d_y,s_mem_node,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,x)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_mem_node,0)) # the new element should not be added here
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_new_nd_nlnk_nnull,0)) # a new element should be added here but it's impossible 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
	# left state = s_mem_node, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
        table.append((s_mem_node,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
	# left state = s_gvn_ptr_d_n, right state = s_mem_node:
        #   - 10/11 -> s_mem_node as the new node could be added (regardless of whether x is here) 
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
        table.append((s_gvn_ptr_d_n,s_mem_node,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_mem_node,0)) 
        index=index+1


       	#****************************************
	# left state = s_new_nd_nlnk_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_new_nd_nlnk_nnull, right state = s_mem_node (under a top memory node): 
	# left state = s_new_nd_nlnk_nnull, right state = s_bef_gvn_ptr_d + i (nd is appearing in the middle of the ptr.d. chain):
	# left state = s_new_nd_nlnk_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_new_nd_nlnk_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_init, right state = s_new_nd_nlnk_nnull (nd came from the left somewhere lower in the ptr. d. chain):
	# left state = s_mem_node, right state = s_new_nd_nlnk_nnull (nd came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_n, right state = s_new_nd_nlnk_nnull: (nd coming from a lower top memory node):
	# left state = s_gvn_ptr_d_y, right state = s_new_nd_nlnk_nnull:
        #   - 01/10/11 -> s_new_nd_nlnk_nnull
        #   - 00 (the null node) -> s_new_nd_nlnk_null
	#****************************************
        table.append((s_new_nd_nlnk_nnull,s_init,index))
        table.append((s_new_nd_nlnk_nnull,s_mem_node,index))
        for i in range(next_num - next - 1):
                table.append((s_new_nd_nlnk_nnull,s_bef_gvn_ptr_d + i,index))
        table.append((s_new_nd_nlnk_nnull,s_gvn_ptr_d_n,index))
        table.append((s_new_nd_nlnk_nnull,s_gvn_ptr_d_y,index))
        table.append((s_init,s_new_nd_nlnk_nnull,index))
        table.append((s_mem_node,s_new_nd_nlnk_nnull,index))
        table.append((s_gvn_ptr_d_n,s_new_nd_nlnk_nnull,index))
        table.append((s_gvn_ptr_d_y,s_new_nd_nlnk_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_new_nd_nlnk_nnull,0)) # the null node has not been found yet
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_nlnk_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_new_nd_nlnk_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_new_nd_nlnk_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_nlnk_undef,0))  # the undef node has just been processed
        index=index+1


       	#****************************************
	# left state = s_new_nd_nlnk_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to error)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_new_nd_nlnk_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line1)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_next_assign_new_mid_used


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_next_assign_new_mid_used(node_width,next_num,x,next,line1):

def test_x_next_assign_new_mid_used():
	"""test_x_next_assign_new_mid_used"""
	(bdd,table,accept_l)=x_next_assign_new_mid_used(9,4,2,1,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_mid_used_1.gta")
	(bdd,table,accept_l)=x_next_assign_new_mid_used(9,4,1,2,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_mid_used_2.gta")

#test_x_next_assign_new_mid_used()


#----------------------------------------------------------------------------------


def x_next_assign_new_lst_used(node_width,x,line1):
	'''x->next=new (a new top memory cannot be linked below the x pointed node to the given right_most next pointer -- it's not free)'''

# Note: we do not check that the input structure is precisely correct, use an intersection
# with an invariant of the encoding if necessary!

# Note 1: We SUPPOSE that deleted nodes do not have any variable set!

# Note 2: We SUPPOSE that the node available via next from x is free (init) -- if not, we generate an error!

# node_width  -- the number of bits per input symbol (including everything, all flags, etc.) 
# x           -- the number of the pointer variable being assigned

# line1 and line2 must be of the same length equal to the maximum length possible!

# We use the following states in the transducer:
        s_init = 0               # reading initial nodes, ptr. desc., and mem. nodes not to be modified
        s_gvn_ptr_d_y = 1        # the given ptr. desc. node has a node linked below it
        s_gvn_ptr_d_n = 2        # the given ptr. desc. node does not have a node linked below it
        s_new_nd_nlnk_nnull = 3  # the new node could not be linked, waiting for the null node
        s_new_nd_nlnk_null = 4   # the new node could not be linked, the null node has been read
        s_new_nd_nlnk_undef = 5  # the new node could not be linked, the undef node has been read
        s_accept = 6             # all the changes done, the right line number has been found
        s_reject = 7             # reject: malformed, not the right line number, etc.
        s_mem_node = 8           # a non-top memory node has been read

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00 -> s_init
        #   - 01 -> s_gvn_ptr_d_n 
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,0)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,-2,4)) #3
	bdd.append((3,0,5)) #4 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #5
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1
	# processing an init node
	bdd[3]=(bdd[3][0],index,bdd[3][2]) 
	bdd.append((3,index+1,0)) # keep init 
        index=index+1
        bdd.append((-1,s_init,0)) 
        index=index+1


       	#****************************************
	# left state = s_mem_node, right state = s_init: 
        #   - 01 -> s_gvn_ptr_d_y 
	#****************************************
	table.append((s_mem_node,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
	# left state = s_mem_node, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
        table.append((s_mem_node,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
	# left state = s_gvn_ptr_d_y, right state = s_mem_node:
        #   - 10/11 & x set -> s_new_nd_nlnk_nnull 
        #   - 10/11 & x not set -> s_mem_node
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
        table.append((s_gvn_ptr_d_y,s_mem_node,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index,patch_l_0,patch_l_1)=functions.copy_n_elements_and_remember_x(bdd,index,node_width-2,4,x)
        for (i,j) in patch_l_0:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_mem_node,0)) # a new element should not be added here
        index=index+1
        for (i,j) in patch_l_1:
                if j==1:
                        bdd[i]=(bdd[i][0],index,bdd[i][2])
                else:
                        bdd[i]=(bdd[i][0],bdd[i][1],index)
        bdd.append((-1,s_new_nd_nlnk_nnull,0)) # a new element could not be added here
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
	# left state = s_mem_node, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
        table.append((s_mem_node,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
	# left state = s_gvn_ptr_d_n, right state = s_mem_node:
        #   - 10/11 -> s_mem_node as the new could be added
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
        table.append((s_gvn_ptr_d_n,s_mem_node,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+5
        bdd.append((-1,s_mem_node,0)) 
        index=index+1


       	#****************************************
	# left state = s_new_nd_nlnk_nnull, right state = s_init (under a (top) memory node, down in a chain of dscr., under null): 
	# left state = s_new_nd_nlnk_nnull, right state = s_mem_node (under a top memory node): 
	# left state = s_new_nd_nlnk_nnull, right state = s_gvn_ptr_d_n:
	# left state = s_new_nd_nlnk_nnull, right state = s_gvn_ptr_d_y:
	# left state = s_init, right state = s_new_nd_nlnk_nnull (nd came from the left somewhere lower in the ptr. d. chain):
	# left state = s_mem_node, right state = s_new_nd_nlnk_nnull (nd came from the left somewhere lower in the ptr. d. chain):
	# left state = s_gvn_ptr_d_n, right state = s_new_nd_nlnk_nnull:
	# left state = s_gvn_ptr_d_y, right state = s_new_nd_nlnk_nnull:
        #   - 01/10/11 -> s_new_nd_nlnk_nnull
        #   - 00 (the null node) -> s_new_nd_nlnk_null
	#****************************************
        table.append((s_new_nd_nlnk_nnull,s_init,index))
        table.append((s_new_nd_nlnk_nnull,s_mem_node,index))
        table.append((s_new_nd_nlnk_nnull,s_gvn_ptr_d_n,index))
        table.append((s_new_nd_nlnk_nnull,s_gvn_ptr_d_y,index))
        table.append((s_init,s_new_nd_nlnk_nnull,index))
        table.append((s_mem_node,s_new_nd_nlnk_nnull,index))
        table.append((s_gvn_ptr_d_n,s_new_nd_nlnk_nnull,index))
        table.append((s_gvn_ptr_d_y,s_new_nd_nlnk_nnull,index))
	bdd.append((0,index+1,index+5)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,index+4)) #+2
        index_bak=index+3
	bdd.append((3,-2,0)) #+3 - 00
	bdd.append((3,0,index+9)) #+4 - 01
	bdd.append((1,0,index+6)) #+5
	bdd.append((2,index+7,index+8)) #+6
	bdd.append((3,index+9,0)) #+7 - 10
	bdd.append((3,0,index+9)) #+8 - 11
        index=index+9
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+9
        bdd.append((-1,s_new_nd_nlnk_nnull,0)) # the null node has not been found yet
        index=index+1
	bdd[index_bak]=(bdd[index_bak][0],index,bdd[index_bak][2])
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_nlnk_null,0))  # the null node has just been found
        index=index+1


       	#****************************************
	# left state = s_new_nd_nlnk_null, right state = s_init: 
	#   - 00 (the undef node) -> 
	#****************************************
	table.append((s_new_nd_nlnk_null,s_init,index))
	bdd.append((0,index+1,0)) #index -- 10/11 is not possible now
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,index+3,0)) #+2 -- 01 is not possible now
	bdd.append((3,index+4,0)) #+3 -- 00
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
        bdd.append((-1,s_new_nd_nlnk_undef,0))  # the undef node has just been processed
        index=index+1


       	#****************************************
	# left state = s_new_nd_nlnk_undef, right state = s_init: 
	#   - 0 line1 -> s_accept (& change to error)
	#     0 ~line1 -> s_reject
	#****************************************
	table.append((s_new_nd_nlnk_undef,s_init,index))
	(bdd,index)=functions.change_line_num_to_err(bdd,index,line1)
	bdd.append((-1,s_accept,0)) 
	index=index+1


	return(bdd,table,[s_accept])
	# end of x_next_assign_new_lst_used


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def x_next_assign_new_lst_used(node_width,x,line1):


def test_x_next_assign_new_lst_used():
	"""test_x_next_assign_new_lst_used"""
	(bdd,table,accept_l)=x_next_assign_new_lst_used(9,2,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_lst_used_1.gta")
	(bdd,table,accept_l)=x_next_assign_new_lst_used(9,1,"000000000")
	functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_lst_used_2.gta")

#test_x_next_assign_new_lst_used()


#=================================================================================


# A wrapper to generate error when a new node cannot be linked under the given node via the given next pointer.

def x_next_assign_new_used(node_width,next_num,x,next,line1,filename):
        if (next==next_num-1):
                (bdd,table,accept_l)=x_next_assign_new_lst_used(node_width,x,line1)
        else:
                (bdd,table,accept_l)=x_next_assign_new_mid_used(node_width,next_num,x,next,line1)
	#functions.gen_gta_file(bdd,table,accept_l,"x_%i_next_%i_assign_new_used.gta" % (x,next))
	functions.gen_gta_file(bdd,table,accept_l,filename)


#=================================================================================


# The same as y being null in x=y->next.

def x_next_assign_new_x_null(node_width,x,line1,filename):
	""" x->next=new, x==null ---> error """
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_y_null(node_width,x,line1)
	#functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_x_null.gta")
	functions.gen_gta_file(bdd,table,accept_l,filename)


#=================================================================================


# The same as y being null in x=y->next.

def x_next_assign_new_x_undef(node_width,x,line1,filename):
	""" x->next=new, x==undef ---> error """
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_y_undef(node_width,x,line1)
	#functions.gen_gta_file(bdd,table,accept_l,"x_next_assign_new_x_undef.gta")
	functions.gen_gta_file(bdd,table,accept_l,filename)


#=================================================================================


def gen_descr_down_mid(node_width,ptr_num,next_num,next,marker):
	'''generates the ptr. descr. to go one node down via the middle next ptr. "next" to a place marked with "marker"'''

# node_width -- maximal size of node
# ptr_num    -- the number of pointer variables
# next_num   -- the number of the next pointers
# next       -- the next pointer to be followed
# marker     -- marker required in the destination

# We use the following states in the transducer:
        s_init = 0              # reading the init symbols and the (top) memory nodes away of the token trajectory
        s_token_set = 1         # the token has been set
        s_bef_gvn_ptr_d = 10    # ptr. desc. nodes before the tested one (we use s_bef_gvn_ptr_d + 0, + 1, ...)
        s_gvn_ptr_d_y = 2       # the token has been put below the given ptr. desc. node
        s_gvn_ptr_d_n = 3       # the token has NOT been put below the given ptr. desc. node
        s_token_removed = 4     # the token has been removed from its source position -- we accept whatever comes now
        s_reject = 5            # reject (the token has been moved to a wrong place, malformed, ...)

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	

	if ((next<0) or (next>=next_num-1)):
		functions.error_msg_expl("gen_descr_down_mid: use a different function for generating the descr. than ...mid!");
        	# use gen_descr_down_mid


	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00 -> s_init
        #   - 01 -> s_bef_gvn_ptr_d + 0
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,0)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,-2,4)) #3
	bdd.append((3,0,5)) #4 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #5
        bdd.append((-1,s_bef_gvn_ptr_d,0)) 
        index=index+1
	# processing an init node
	bdd[3]=(bdd[3][0],index,bdd[3][2]) 
	bdd.append((3,index+1,0))
        bdd.append((-1,s_init,0)) 
        index=index+2


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + i:
        #   - 01 -> s_bef_gvn_ptr_d + i+1 ... till just under processing the given ptr. desc. 
	#****************************************
        for i in range(next_num - next - 2):
                table.append((s_init,s_bef_gvn_ptr_d + i,index)) # we are so far looking for the place to put the token
	        bdd.append((0,index+1,0)) #index
	        bdd.append((1,index+2,0)) #+1
	        bdd.append((2,0,index+3)) #+2
	        bdd.append((3,0,index+4)) #+3 - 01
                index=index+4
                (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
                bdd.append((-1,s_bef_gvn_ptr_d + i+1,0)) 
                index=index+1


       	#****************************************
	# left state = s_init, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_n (the token has not been set)
	#****************************************
        table.append((s_init,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) # we return s_gvn_ptr_d_n 
        index=index+1


       	#****************************************
	# left state = s_token_set, right state = s_bef_gvn_ptr_d + _ (just under the given ptr. desc.):
        #   - 01 -> s_gvn_ptr_d_y (the token has been put to its target position)
	#****************************************
        table.append((s_token_set,s_bef_gvn_ptr_d + next_num-next-2,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) # we return s_gvn_ptr_d_y
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0))  # !!!!!!!!! s_gvn_ptr_d_n
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 & token set -> s_token_removed & remove the token
        #   - 10/11 & token not set -> s_reject (as the token has been put to a wrong place)
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-2,4,ptr_num,ptr_num,s_reject,s_token_removed)


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init / s_token_set & setting the token bit (chosen nondeterministically)
        #                       (for 11, this is not necessary -- no down descr. can lead there)
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	# we check the marker and if it set, we may (but need not---we have to guess) put the token here
	(bdd,index)=functions.copy_n_elements_guess_x_dep_on_y(bdd,index,node_width-2,4,ptr_num,ptr_num+marker,s_init,s_token_set)


	# ****************************************
	# left state=s_token_removed, right state=s_init/s_gvn_ptr_d_n:
	# left state=s_init/s_gvn_ptr_d_n, right state=s_token_removed:
	#   - 00/01/10/11 -> s_token_removed (reading the rest of the tree)
	# ****************************************
	table.append((s_token_removed,s_init,index))
	table.append((s_token_removed,s_gvn_ptr_d_n,index))
	table.append((s_init,s_token_removed,index))
	table.append((s_gvn_ptr_d_n,s_token_removed,index))
	for i in range(next_num - next - 1):
		table.append((s_token_removed,s_bef_gvn_ptr_d + i,index))
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)	
	bdd.append((-1,s_token_removed,0))
	index=index+1


	return(bdd,table,[s_token_removed])
	# end of gen_descr_down_mid


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def gen_descr_down_mid(node_width,ptr_num,next_num,next,marker):


def test_gen_descr_down_mid():
	"""test_gen_descr_down_mid"""
	(bdd,table,accept_l)=gen_descr_down_mid(9,4,4,0,1)
	functions.gen_gta_file(bdd,table,accept_l,"descr_down_mid_0.gta")
	(bdd,table,accept_l)=gen_descr_down_mid(9,4,4,1,1)
	functions.gen_gta_file(bdd,table,accept_l,"descr_down_mid_1.gta")
	(bdd,table,accept_l)=gen_descr_down_mid(9,4,4,2,1)
	functions.gen_gta_file(bdd,table,accept_l,"descr_down_mid_2.gta")


#test_gen_descr_down_mid()


#=================================================================================


def gen_descr_down_lst(node_width,ptr_num,marker):
	'''generates the ptr. descr. to go one node down via the last next ptr. to a place marked with "marker"'''

# node_width -- maximal size of node
# ptr_num    -- the number of pointer variables
# marker     -- marker required in the destination

# We use the following states in the transducer:
        s_init = 0              # reading the init symbols and the (top) memory nodes away of the token trajectory
        s_token_set = 1         # the token has been set
        s_gvn_ptr_d_y = 2       # the token has been put below the given ptr. desc. node
        s_gvn_ptr_d_n = 3       # the token has NOT been put below the given ptr. desc. node
        s_token_removed = 4     # the token has been removed from its source position -- we accept whatever comes now
        s_reject = 5            # reject (the token has been moved to a wrong place, malformed, ...)

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.


	#****************************************
	# bdd node 0 = the reject state
	#****************************************
	bdd.append((-1,s_reject,0)) #0
	index=1


       	#****************************************
	# left state = s_init, right state = s_init: 
	#   - 00 -> s_init
        #   - 01 -> s_gvn_ptr_d_n (the token has not been set)
	#****************************************
	table.append((s_init,s_init,index))
	bdd.append((0,2,0)) #1 -- the first bit of the type of the input node
	bdd.append((1,3,0)) #2 -- the first bit of the type of the output node
	bdd.append((2,-2,4)) #3
	bdd.append((3,0,5)) #4 - 01
        index=index+4
	# processing a ptr. desc.
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #5
        bdd.append((-1,s_gvn_ptr_d_n,0)) # the token has not moved this way
        index=index+1
	# processing an init node
	bdd[3]=(bdd[3][0],index,bdd[3][2]) 
	bdd.append((3,index+1,0))
        bdd.append((-1,s_init,0)) 
        index=index+2


       	#****************************************
	# left state = s_token_set, right state = s_init
        #   - 01 -> s_gvn_ptr_d_y (the token has been set)
	#****************************************
        table.append((s_token_set,s_init,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) # the token has moved down this way
        index=index+1


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_y:
        #   - 01 -> s_gvn_ptr_d_y (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_y,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_y,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_y, right state = s_init:
        #   - 10/11 & token set -> s_token_removed & remove the token
        #   - 10/11 & token not set -> s_reject (as the token has been put to a wrong place)
	#****************************************
        table.append((s_gvn_ptr_d_y,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	(bdd,index)=functions.copy_node_return_state_dep_on_x_unset_y(bdd,index,node_width-2,4,ptr_num,ptr_num,s_reject,s_token_removed)


       	#****************************************
	# left state = s_init, right state = s_gvn_ptr_d_n:
        #   - 01 -> s_gvn_ptr_d_n (an APPROXIMATION: we don't count the exact number of next pointers!)
	#****************************************
        table.append((s_init,s_gvn_ptr_d_n,index))
	bdd.append((0,index+1,0)) #index
	bdd.append((1,index+2,0)) #+1
	bdd.append((2,0,index+3)) #+2
	bdd.append((3,0,index+4)) #+3 - 01
        index=index+4
        (bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4) #+4
        bdd.append((-1,s_gvn_ptr_d_n,0)) 
        index=index+1


       	#****************************************
	# left state = s_gvn_ptr_d_n, right state = s_init:
        #   - 10/11 -> s_init / s_token_set & setting the token bit (chosen nondeterministically)
        #                       (for 11, this is not necessary -- no down descr. can lead there)
	#****************************************
        table.append((s_gvn_ptr_d_n,s_init,index))
	bdd.append((0,0,index+1)) #index
	bdd.append((1,0,index+2)) #+1
	bdd.append((2,index+3,index+4)) #+2
	bdd.append((3,index+5,0)) #+3 - 10
	bdd.append((3,0,index+5)) #+4 - 11
        index=index+5
	# we check the marker and if it set, we may (but need not---we have to guess) put the token here
	(bdd,index)=functions.copy_n_elements_guess_x_dep_on_y(bdd,index,node_width-2,4,ptr_num,ptr_num+marker,s_init,s_token_set)


	# ****************************************
	# left state=s_token_removed, right state=s_init/s_gvn_ptr_d_n:
	# left state=s_init/s_gvn_ptr_d_n, right state=s_token_removed:
	#   - 00/01/10/11 -> s_token_removed (reading the rest of the tree)
	# ****************************************
	table.append((s_token_removed,s_init,index))
	table.append((s_token_removed,s_gvn_ptr_d_n,index))
	table.append((s_init,s_token_removed,index))
	table.append((s_gvn_ptr_d_n,s_token_removed,index))
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)	
	bdd.append((-1,s_token_removed,0))
	index=index+1


	return(bdd,table,[s_token_removed])
	# end of gen_descr_down_lst


#----------------------------------------------------------------------------------
#              T E S T S
#----------------------------------------------------------------------------------


# def gen_descr_down_lst(node_width,ptr_num,marker):


def test_gen_descr_down_lst():
	"""test_gen_descr_down_lst"""
	(bdd,table,accept_l)=gen_descr_down_lst(9,4,1)
	functions.gen_gta_file(bdd,table,accept_l,"descr_down_lst.gta")


#test_gen_descr_down_lst()


#=================================================================================


# A wrapper to generate descriptors to go one level down via the given next pointer to a place marked by the given marker.

def gen_descr_down(node_width,ptr_num,next_num,next,marker,filename):
        if (next==next_num-1):
                (bdd,table,accept_l)=gen_descr_down_lst(node_width,ptr_num,marker)
        else:
                (bdd,table,accept_l)=gen_descr_down_mid(node_width,ptr_num,next_num,next,marker)
	#functions.gen_gta_file(bdd,table,accept_l,"descr_down_via_%i_to_%i.gta" % (next,marker))
	functions.gen_gta_file(bdd,table,accept_l,filename)


#=================================================================================









