# Adam Rogalewicz
#
# Main generator loop
# - This loop is guided by "program", and create all coresponding transducers
# - program is described in python module called 'program' -> read: DOC-program-rotations.txt
#
# Use script run_generator-rotations.sh to start generation of transducers
# - set path in this script according to your filesystem
#
# In this version:
# - only binary trees are supported
# - destructive updates are forbiden
# - leftrotates, and rightrotate is allowed

# !!!!!!!!!!!!!!!!!!!!!!!! repare after x.next=new !!!!!!!!!!!!!!!


import program
import sys
import os
import functions

import gen_goto
import gen_if_x_eq_y
import gen_if_x_eq_null
import gen_restrict
import gen_xnext_ass_y
import gen_xnext_ass_y_eq_null
import source_to_shape
import shape_to_descr
import gen_xnext_ass_null
import gen_x_ass_y
import gen_x_ass_null
import gen_x_ass_ynext
import gen_xnext_ass_new
import gen_restrict_variable
import gen_cover_check_preprocess
import gen_setdata
import gen_ifdata
import gen_x_ass_random

#rotations stuff
import ROT_auto_checkers
import ROT_repare
import ROT_set_token
import gen_descr_1up


# Each program line touple can contain 'mark' field
# This field is optional - it is automatically set to ''
# NOABSTR - don't perform abstraction if you addnew configurations to input of this line
# NOUPDATE - don't execute automatic descriptor update after this line (just for x.next=y lines)
mark_key={'' : 0, 'NOABSTR': 2, 'NOUPDATE': 1, 'NOABSTR+NOUPDATE': 3 }


def create_x_ass_ynext(index,i,prog,env):
	(node_width,pointer_num,desc_num,next_num,err_line,restrict_var)=env

	# generate phase 1
	gen_x_ass_ynext.x_assign_y_next_ph1_desc(node_width,pointer_num,desc_num,next_num,i[2],i[3],i[4],i[1],prog[i[5]][1],"settoken-%i-"%index)
	
	
	# generate phase 2
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_ph2_reg(node_width,pointer_num,i[2])
	functions.gen_gta_file(bdd,table,accept_l,"tmp1.gta")
        (bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_ph2_del(node_width,pointer_num,err_line)
	functions.gen_gta_file(bdd,table,accept_l,"tmp2.gta")
	# union of tmp1 && tmp2
	os.system("union tmp1.gta tmp2.gta ph2-%i.gta"%index)

	# special cases:
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_y_null(node_width,i[3],i[1])
	functions.gen_gta_file(bdd,table,accept_l,"tmp1.gta")
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_y_undef(node_width,i[3],i[1])
	functions.gen_gta_file(bdd,table,accept_l,"tmp2.gta")
	gen_x_ass_ynext.x_assign_y_next_y_next_null(node_width,pointer_num,desc_num,next_num,i[2],i[3],i[4],i[1],prog[i[5]][1],"tmp3.gta")
	gen_x_ass_ynext.x_assign_y_next_y_next_undef(node_width,pointer_num,desc_num,next_num,i[2],i[3],i[4],i[1],"tmp4.gta")
	os.system("union tmp1.gta tmp2.gta tmp3.gta tmp4.gta special-%i.gta"%index)
			
def create_xnext_ass_new(index,i,prog,env):
	(node_width,pointer_num,desc_num,next_num,err_line,restrict_var)=env
	# we used descriptor equal to next+1 for this command
	gen_xnext_ass_new.x_next_assign_new(node_width,pointer_num,desc_num,2,next_num,i[2],i[3],i[3]+1,1,i[1],prog[i[4]][1],"tmp1.gta")
	gen_xnext_ass_new.x_next_assign_new_x_null(node_width,i[2],i[1],"tmp2.gta")
	gen_xnext_ass_new.x_next_assign_new_x_undef(node_width,i[2],i[1],"tmp3.gta")
	os.system("union tmp1.gta tmp2.gta tmp3.gta transd-%i.gta"%index)
	gen_xnext_ass_new.x_next_assign_new_used(node_width,next_num,i[2],i[3],i[1],"transd-used-%i.gta"%index)
	# REPARE after new - change undef pointers under the new node to the null ones
	(bdd,table,accept_l)=ROT_repare.repare_after_new(node_width, i[2] ) 
	functions.gen_gta_file(bdd,table,accept_l,"reparenew-%i.gta"%index)
	
	

def gen_transducers():
	(prog,env_prg)=program.get_program()
	(pointer_num,err_line,restrict_var,data_size)=env_prg
	desc_num=4
	next_num=3
	# NODE_WIDTH: 2 (size of identificator) + number of pointers + token + marker + data
	node_width=2+pointer_num+2+data_size # <- there is only one marker, becaurse all descriptors are deterministic	
	env=(node_width,pointer_num,desc_num,next_num,err_line,restrict_var)

	out=[]
	# this line is different from the standard SAS case
	# we need number of pointers for backward rotations
	out.append("%i %i %i %i\n"%(len(prog),desc_num, node_width,pointer_num))
	for j in range(len(prog)):
		i=prog[j]
		print i
		#----------------------------
		if i[0]=="goto":
			(bdd,table)=gen_goto.goto(node_width, i[1], prog[i[2]][1])
			functions.gen_gta_file(bdd,table, [1],"transd-%i.gta"%j)
			# set mark
			if (len(i)>3): mark=mark_key[i[3]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[2],mark))
		#----------------------------
		elif i[0]=="if*":
			# nondeterministic if
			(bdd,table)=gen_goto.goto(node_width, i[1], prog[i[2]][1])
			functions.gen_gta_file(bdd,table, [1],"transd-%i.gta"%j)
			(bdd,table)=gen_goto.goto(node_width, i[1], prog[i[3]][1])
			functions.gen_gta_file(bdd,table, [1],"transd-%i-2.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[2],i[3],mark))
		#----------------------------
		elif i[0]=="exit":
			# set mark
			if (len(i)>2): mark=mark_key[i[2]]
			else: mark=0
			out.append("%i 0 -1 -1 %i\n"%(j,mark));
		#----------------------------
		elif i[0]=="ifx==y":
			(bdd,table,accept_l)=gen_if_x_eq_y.if_x_eq_y(node_width,pointer_num,i[2],i[3],i[1], prog[i[4]][1],prog[i[5]][1])
			functions.gen_gta_file(bdd,table,accept_l,"tmp.gta")
			# set mark
			if (len(i)>6): mark=mark_key[i[6]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[4],i[5],mark))
			(bdd,table)=gen_restrict.restrict(prog[i[5]][1])
			functions.gen_gta_file(bdd,table, [1],"tmp2.gta")
			os.system("intersect tmp.gta tmp2.gta transd-%i-2.gta"%j)
			os.system("invert tmp2.gta tmp3.gta")
			os.system("intersect tmp.gta tmp3.gta transd-%i.gta"%j)
		#----------------------------
		elif i[0]=="ifx==null":
			(bdd,table,accept_l)=gen_if_x_eq_null.if_x_eq_null(node_width, pointer_num, i[2], i[1], prog[i[3]][1],prog[i[4]][1])
			functions.gen_gta_file(bdd,table,accept_l,"tmp.gta")
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[3],i[4],mark))
			(bdd,table)=gen_restrict.restrict(prog[i[4]][1])
			functions.gen_gta_file(bdd,table, [1],"tmp2.gta")
			os.system("intersect tmp.gta tmp2.gta transd-%i-2.gta"%j)
			os.system("invert tmp2.gta tmp3.gta")
			os.system("intersect tmp.gta tmp3.gta transd-%i.gta"%j)
		#----------------------------
		elif i[0]=="x=y.next":
			create_x_ass_ynext(j,i,prog,env)
			# set mark
			if (len(i)>6): mark=mark_key[i[6]]
			else: mark=0
			out.append("%i 4 %i -1 %i\n"%(j,i[5],mark))
		#----------------------------
		elif i[0]=="x=y":
			(bdd,table)=gen_x_ass_y.x_assign_y(node_width, pointer_num, i[2], i[3], i[1], prog[i[4]][1])
			functions.gen_gta_file(bdd,table, [5],"transd-%i.gta"%j)
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[4],mark))
		#----------------------------
		elif i[0]=="x=null":
			(bdd,table)=gen_x_ass_null.x_assign_null(node_width, pointer_num, i[2], i[1], prog[i[3]][1])
			functions.gen_gta_file(bdd,table, [5],"transd-%i.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[3],mark))
		#----------------------------
		elif i[0]=="x.next=new":
			create_xnext_ass_new(j,i,prog,env)
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 5 %i -1 %i\n"%(j,i[4],mark))
		#----------------------------
		elif i[0]=="setdata":
			# there is only one marker => the parametr descr_num is "2"
			(bdd,table,acc_l)=gen_setdata.setdata(node_width,pointer_num,2,i[2],i[3],i[1],prog[i[4]][1])
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[4],mark))
		#----------------------------
		elif i[0]=="ifdata":
			# there is only one marker => the parametr descr_num is "2"
			# transducer for then
			(bdd,table,acc_l)=gen_ifdata.if_data(node_width, pointer_num, 2, i[2], i[3], i[1], prog[i[4]][1], 1)
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# transducer for else
			(bdd,table,acc_l)=gen_ifdata.if_data(node_width, pointer_num, 2, i[2], i[3], i[1], prog[i[5]][1], 0)
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i-2.gta"%j)
			
			# set mark
			if (len(i)>6): mark=mark_key[i[6]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[4],i[5],mark))
		#----------------------------
		elif i[0]=="x=random":
			(bdd,table,acc_l)=gen_x_ass_random.x_assign_random(node_width,i[2],i[1],prog[i[3]][1])
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[3],mark))
			
		#----------------------------
		elif i[0]=="rightrotate":
			# this checker is used before rotations
			(bdd,table,acc_l)=ROT_auto_checkers.gen_checker_right_rotate(i[2])
			functions.gen_gta_file(bdd,table, acc_l,"auto_checker-%i.gta"%j)
			# for backward rotate - we set token above "x", and do standard leftrotate
			# on this token
			(bdd,table,acc_l)=ROT_set_token.prepare_right_rot_back(node_width,pointer_num,i[2])
			functions.gen_gta_file(bdd,table, acc_l,"prepare_back_rot-%i.gta"%j)
			# After rotation, it is necessary to repare the tree, and change the line number
			(bdd,table,acc_l)=ROT_repare.gen_repare(node_width, pointer_num, i[1], prog[i[3]][1])
			functions.gen_gta_file(bdd,table, acc_l,"repare-%i.gta"%j)
			# after backward rotation, we need to change line number backward
			(bdd,table,acc_l)=ROT_repare.gen_repare(node_width, pointer_num, prog[i[3]][1],i[1])
			functions.gen_gta_file(bdd,table, acc_l,"repare-back-%i.gta"%j)
			
			
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 10 %i %i %i\n"%(j,i[3],i[2],mark))
		#----------------------------
		elif i[0]=="leftrotate":
			# this checker is used before rotations
			(bdd,table,acc_l)=ROT_auto_checkers.gen_checker_left_rotate(i[2])
			functions.gen_gta_file(bdd,table, acc_l,"auto_checker-%i.gta"%j)
			# for backward rotate - we set token above "x", and do standard rightrotate
			# on this token
			(bdd,table,acc_l)=ROT_set_token.prepare_left_rot_back(node_width,pointer_num,i[2])
			functions.gen_gta_file(bdd,table, acc_l,"prepare_back_rot-%i.gta"%j)
			# After rotation, it is necessary to repare the tree, and change the line number
			(bdd,table,acc_l)=ROT_repare.gen_repare(node_width, pointer_num, i[1], prog[i[3]][1])
			functions.gen_gta_file(bdd,table, acc_l,"repare-%i.gta"%j)
			# after backward rotation, we need to change line number backward
			(bdd,table,acc_l)=ROT_repare.gen_repare(node_width, pointer_num, prog[i[3]][1],i[1])
			functions.gen_gta_file(bdd,table, acc_l,"repare-back-%i.gta"%j)
			
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 11 %i %i %i\n"%(j,i[3],i[2],mark))
		#----------------------------
		else:
			print "ERROR: %s unknown command" %i[0]
			sys.exit()

	# generate initial descriptors - no more descriptors are used !!!!
	# we use only one marker
	#1left
	gen_xnext_ass_new.gen_descr_down(node_width,pointer_num,next_num,0,1,"descr-1.gta")
	#1right
	gen_xnext_ass_new.gen_descr_down(node_width,pointer_num,next_num,1,1,"descr-2.gta")
	#1up
	(bdd,table,acc_l)=gen_descr_1up.descr_1up(node_width,pointer_num,1)
	functions.gen_gta_file(bdd,table, acc_l,"descr-3.gta")

	file=open("directions-rotations.txt","w")
	file.writelines(out)
	file.close()
	
	# generate variable restrictions
	# for each pointer variable will be generated one automaton to ensure,
	# that pointer variable is present only at unique position
	if restrict_var:
		for i in range(pointer_num-1):
			(bdd,table)=gen_restrict_variable.restrict_variable(i+1)
			functions.gen_gta_file(bdd,table, [9],"restrict_var-%i.gta"%(i))

gen_transducers()
