# Adam Rogalewicz
#
# Main generator loop
# - This loop is guided by "program", and create all coresponding transducers
# - program is described in python module called 'program' -> read: DOC-program.txt
#
# Use script run_generator.sh to start generation of transducers
# - set path in this script according to your filesystem

import program
import sys
import os
import functions

import gen_goto
import gen_if_x_eq_y
import gen_if_x_eq_null
import gen_restrict
import gen_xnext_ass_y
import gen_xnext_ass_y_eq_null
import source_to_shape
import shape_to_descr
import gen_xnext_ass_null
import gen_x_ass_y
import gen_x_ass_null
import gen_x_ass_ynext
import gen_xnext_ass_new
import gen_restrict_variable
import gen_cover_check_preprocess
import gen_setdata
import gen_ifdata
import gen_x_ass_random
import gen_x_ass_new
import gen_free_x
import gen_checkers
import gen_empty
import seed
import variance_checkers
import gen_sigma_star

# Each program line touple can contain 'mark' field
# This field is optional - it is automatically set to ''
# NOABSTR - don't perform abstraction if you addnew configurations to input of this line
# NOUPDATE - don't execute automatic descriptor update after this line (just for x.next=y lines)
# BP - set breakpoints for variance analysis
mark_key={'' : 0, 'NOABSTR': 2, 'NOUPDATE': 1, 'NOABSTR+NOUPDATE': 3, 'CP': 4, 'CP+NOUPDATE' :5, 
	'CP+NOABSTR':6, 'CP+NOABSTR+NOUPDATE': 7}


def create_x_ass_ynext(index,i,prog,env):
	(node_width,pointer_num,desc_num,next_num,err_line,restrict_var)=env

	# generate phase 1
	gen_x_ass_ynext.x_assign_y_next_ph1_desc(node_width,pointer_num,desc_num,next_num,i[2],i[3],i[4],i[1],prog[i[5]][1],"settoken-%i-"%index)
	
	
	# generate phase 2
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_ph2_reg(node_width,pointer_num,i[2])
	functions.gen_gta_file(bdd,table,accept_l,"tmp1.gta")
        (bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_ph2_del(node_width,pointer_num,prog[i[5]][1])
	functions.gen_gta_file(bdd,table,accept_l,"tmp2.gta")
	# union of tmp1 && tmp2
	os.system("union tmp1.gta tmp2.gta ph2-%i.gta"%index)

	# special cases:
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_y_null(node_width,i[3],i[1])
	functions.gen_gta_file(bdd,table,accept_l,"tmp1.gta")
	(bdd,table,accept_l)=gen_x_ass_ynext.x_assign_y_next_y_undef(node_width,i[3],i[1])
	functions.gen_gta_file(bdd,table,accept_l,"tmp2.gta")
	gen_x_ass_ynext.x_assign_y_next_y_next_null(node_width,pointer_num,desc_num,next_num,i[2],i[3],i[4],i[1],prog[i[5]][1],"tmp3.gta")
	gen_x_ass_ynext.x_assign_y_next_y_next_undef(node_width,pointer_num,desc_num,next_num,i[2],i[3],i[4],i[1],"tmp4.gta")
	os.system("union tmp1.gta tmp2.gta tmp3.gta tmp4.gta special-%i.gta"%index)
			
def create_xnext_ass_new(index,i,prog,env,data_size,init_data):
	(node_width,pointer_num,desc_num,next_num,err_line,restrict_var)=env
	gen_xnext_ass_new.x_next_assign_new(node_width,pointer_num,desc_num,desc_num,next_num,data_size,init_data,i[2],i[3],i[5],i[5],i[1],prog[i[4]][1],"tmp1.gta")
	gen_xnext_ass_new.x_next_assign_new_x_null(node_width,i[2],i[1],"tmp2.gta")
	gen_xnext_ass_new.x_next_assign_new_x_undef(node_width,i[2],i[1],"tmp3.gta")
	os.system("union tmp1.gta tmp2.gta tmp3.gta transd-%i.gta"%index)
	gen_xnext_ass_new.x_next_assign_new_used(node_width,next_num,i[2],i[3],i[1],"transd-used-%i.gta"%index)
	#generate descriptor, if requested
	if i[6]:
		print "Creating descriptor no %i: one-down by next-%i"%(i[5],i[3])
		gen_xnext_ass_new.gen_descr_down(node_width,pointer_num,next_num,i[3],i[5],"descr-%i.gta"%i[5])
	
	

def gen_transducers(variance):
	(prog,env)=program.get_program()
	(node_width,pointer_num,desc_num,next_num,err_line,restrict_var)=env
	if variance:
		if (2*pointer_num)+desc_num+2>node_width:
			print "ERROR: node_width is to small for variance analysis"
			sys.exit()
		# if the variance analysis is enabled also on descriptors, then check whether we fit inside node_width
		if (variance==2) and ((2*desc_num)+2>node_width):
			print "ERROR: node_width is to small to enable variance analysis on descriptors"
			sys.exit()
		# in case of variance analysis, we have 2 copies of all pointer variables
		# and 2 copies od data inside a single memory node
		data_size=(node_width-((2*pointer_num)+desc_num+2))/2
	else:
		if (pointer_num+desc_num+2)>node_width:
			print "ERROR: node_width is to small"
			sys.exit()
		data_size=node_width-(pointer_num+desc_num+2)
	# in case of variance analysis, enlarge the node width by (pointer_num-1+data_size)
	# in case of variance analysis, 
	# we will make a space to remember the old positions of all pointer variables and old data value
	# in fact, we set node_width inside program.py
	# start generating ...
	new=0  
	out=[]
	out.append("%i %i %i\n"%(len(prog),desc_num, node_width))
	for j in range(len(prog)):
		i=prog[j]
		print i
		#----------------------------
		if i[0]=="goto":
			(bdd,table)=gen_goto.goto(node_width, i[1], prog[i[2]][1])
			functions.gen_gta_file(bdd,table, [1],"transd-%i.gta"%j)
			# set mark
			if (len(i)>3): mark=mark_key[i[3]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[2],mark))
		#----------------------------
		elif i[0]=="if*":
			# nondeterministic if
			(bdd,table)=gen_goto.goto(node_width, i[1], prog[i[2]][1])
			functions.gen_gta_file(bdd,table, [1],"transd-%i.gta"%j)
			(bdd,table)=gen_goto.goto(node_width, i[1], prog[i[3]][1])
			functions.gen_gta_file(bdd,table, [1],"transd-%i-2.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[2],i[3],mark))
		#----------------------------
		elif i[0]=="exit":
			# set mark
			if (len(i)>2): mark=mark_key[i[2]]
			else: mark=0
			out.append("%i 0 -1 -1 %i\n"%(j,mark));
		#----------------------------
		elif i[0]=="ifx==y":
			(bdd,table,accept_l)=gen_if_x_eq_y.if_x_eq_y(node_width,pointer_num,i[2],i[3],i[1], prog[i[4]][1],prog[i[5]][1])
			functions.gen_gta_file(bdd,table,accept_l,"tmp.gta")
			# set mark
			if (len(i)>6): mark=mark_key[i[6]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[4],i[5],mark))
			(bdd,table)=gen_restrict.restrict(prog[i[5]][1])
			functions.gen_gta_file(bdd,table, [1],"tmp2.gta")
			os.system("intersect tmp.gta tmp2.gta transd-%i-2.gta"%j)
			os.system("invert tmp2.gta tmp3.gta")
			os.system("intersect tmp.gta tmp3.gta transd-%i.gta"%j)
		#----------------------------
		elif i[0]=="ifx==null":
			(bdd,table,accept_l)=gen_if_x_eq_null.if_x_eq_null(node_width, pointer_num, i[2], i[1], prog[i[3]][1],prog[i[4]][1])
			functions.gen_gta_file(bdd,table,accept_l,"tmp.gta")
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[3],i[4],mark))
			(bdd,table)=gen_restrict.restrict(prog[i[4]][1])
			functions.gen_gta_file(bdd,table, [1],"tmp2.gta")
			os.system("intersect tmp.gta tmp2.gta transd-%i-2.gta"%j)
			os.system("invert tmp2.gta tmp3.gta")
			os.system("intersect tmp.gta tmp3.gta transd-%i.gta"%j)
		#----------------------------
		elif i[0]=="x.next=y":
			(bdd,table)=gen_xnext_ass_y.xnext_assign_y(node_width,pointer_num,desc_num,i[2],i[3],i[4],i[6],i[1],prog[i[5]][1])
			functions.gen_gta_file(bdd,table, [4],"tmp1.gta")
			(bdd,table)=gen_xnext_ass_y_eq_null.xnext_assign_y_null(node_width,i[2],i[3],i[4],desc_num,i[1],prog[i[5]][1])
			functions.gen_gta_file(bdd,table, [4],"tmp2.gta")
			# union of tmp1 && tmp2
			os.system("union tmp1.gta tmp2.gta tmp3.gta")
			# project tmp3 on track node_width*2 - nondeterministic choice in transducer
			os.system("project tmp3.gta %i transd-%i.gta"%(node_width*2,j))
			# set mark
			if (len(i)>7): mark=mark_key[i[7]]
			else: mark=0
			out.append("%i 3 %i %i %i\n"%(j,i[5],i[6],mark))
			# to shape
			(bdd,table,accept_l)=source_to_shape.source_to_shape(i[2],i[3])
			functions.gen_gta_file(bdd,table, accept_l,"to_shape-%i.gta"%j)
			# cover checking preprocess(not correctly implemented)
			(bdd,table,accept_l)=gen_cover_check_preprocess.cover_check_preprocess(node_width, pointer_num, i[3], i[6])
			functions.gen_gta_file(bdd,table, accept_l,"prepare-cover-check-%i.gta"%j)
			# we need settoken  transducer just for descriptor i[6], but it is 
			# easier to generate all settoken-s
			gen_x_ass_ynext.x_assign_y_next_ph1_desc(node_width,pointer_num,desc_num,next_num,i[3],i[2],i[4],prog[i[5]][1],prog[i[5]][1],"settoken-%i-"%j)
			# Overaproximate descriptor i[6] by \sigma^* if required (last parameter is "*")
			# disabled by default
			if (len(i)>8) and (i[8]=="*"):
				print "WARNING: Overapproximating descriptor %i by \sigma*"%i[6]
				(bdd,table,accept_l)=gen_sigma_star.descr_star(node_width,pointer_num,i[6])
				functions.gen_gta_file(bdd,table, accept_l,"descr-%i.gta"%i[6])
				
		#----------------------------
		elif i[0]=="x.next=null":
			(bdd,table)=gen_xnext_ass_null.xnext_assign_y_null(node_width,pointer_num,i[2],i[3],desc_num,i[1],prog[i[4]][1])
			functions.gen_gta_file(bdd,table, [4],"tmp.gta")
			# project tmp on track node_width*2 - nondeterministic choice in transducer
			os.system("project tmp.gta %i transd-%i.gta"%(node_width*2,j))
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[4],mark))
		#----------------------------
		elif i[0]=="x=y.next":
			create_x_ass_ynext(j,i,prog,env)
			# set mark
			if (len(i)>6): mark=mark_key[i[6]]
			else: mark=0
			out.append("%i 4 %i -1 %i\n"%(j,i[5],mark))
		#----------------------------
		elif i[0]=="x=y":
			(bdd,table)=gen_x_ass_y.x_assign_y(node_width, pointer_num, i[2], i[3], i[1], prog[i[4]][1])
			functions.gen_gta_file(bdd,table, [5],"transd-%i.gta"%j)
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[4],mark))
		#----------------------------
		elif i[0]=="x=null":
			(bdd,table)=gen_x_ass_null.x_assign_null(node_width, pointer_num, i[2], i[1], prog[i[3]][1])
			functions.gen_gta_file(bdd,table, [5],"transd-%i.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[3],mark))
		#----------------------------
		elif i[0]=="x.next=new":
			#check whether data initialization is required
			if (len(i)>8): init_data=i[8]
			else: init_data=""
			create_xnext_ass_new(j,i,prog,env,data_size,init_data)
			# set mark
			if (len(i)>7): mark=mark_key[i[7]]
			else: mark=0
			out.append("%i 5 %i -1 %i\n"%(j,i[4],mark))
			new=1
		#----------------------------
		elif i[0]=="setdata":
			if len(i[3])>data_size:
				print "ERROR: data of size %i can not fit into the given node_width"%len(i[3])
				sys.exit(1)
			(bdd,table,acc_l)=gen_setdata.setdata(node_width,pointer_num,desc_num,i[2],i[3],i[1],prog[i[4]][1])
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# set mark
			if (len(i)>5): mark=mark_key[i[5]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[4],mark))
		#----------------------------
		elif i[0]=="ifdata":
			if len(i[3])>data_size:
				print "ERROR: data of size %i can not fit into the given node_width"%len(i[3])
				sys.exit(1)
			# transducer for then
			(bdd,table,acc_l)=gen_ifdata.if_data(node_width, pointer_num, desc_num, i[2], i[3], i[1], prog[i[4]][1], 1)
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# transducer for else
			(bdd,table,acc_l)=gen_ifdata.if_data(node_width, pointer_num, desc_num, i[2], i[3], i[1], prog[i[5]][1], 0)
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i-2.gta"%j)
			
			# set mark
			if (len(i)>6): mark=mark_key[i[6]]
			else: mark=0
			out.append("%i 2 %i %i %i\n"%(j,i[4],i[5],mark))
		#----------------------------
		elif i[0]=="x=random":
			(bdd,table,acc_l)=gen_x_ass_random.x_assign_random(node_width,i[2],i[1],prog[i[3]][1])
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[3],mark))
			
		#----------------------------
		elif i[0]=="new":
			#check whether data initialization is required
			if (len(i)>5): init_data=i[5]
			else: init_data=""
			(bdd,table,acc_l)=gen_x_ass_new.x_assign_new(node_width,pointer_num,desc_num,next_num,data_size,init_data,i[2] ,i[1],prog[i[3]][1])
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			out.append("%i 1 %i -1 %i\n"%(j,i[3],mark))
			new=1
			
		#----------------------------
		elif i[0]=="free":
			(bdd,table,acc_l)=gen_free_x.free(node_width,i[2],i[1],prog[i[3]][1])
			functions.gen_gta_file(bdd,table, acc_l,"transd-%i.gta"%j)
			# set mark
			if (len(i)>4): mark=mark_key[i[4]]
			else: mark=0
			#create prepare_free automata
			# this automata are used befor the main transducers and they set all
			# pointer variables aliased to i[2] to undefined
			tmp=0
			for tmp2 in range(pointer_num):
				if (tmp2!=i[2]) and (tmp2>0):
					(bdd,table,acc_l)=gen_free_x.prepare_free(node_width,i[2],tmp2)
					functions.gen_gta_file(bdd,table, acc_l,"prepare_free-%i-%i.gta"%(j,tmp))
					tmp=tmp+1
			# the number of generated files is storen in the 4th argument 

			out.append("%i 6 %i %i %i\n"%(j,i[3],tmp,mark))
			
		#----------------------------
		else:
			print "ERROR: %s unknown command" %i[0]
			sys.exit()

	# genneration of shape -> descriptor transducers
	for i in range(desc_num-1):
		print "Generating shape -> descriptor no %i"%(i+1)
		(bdd,table)=shape_to_descr.shape_to_descr(node_width, pointer_num, i+1)
		functions.gen_gta_file(bdd,table, [0],"from-shape-%i.gta"%(i+1))
			

	file=open("directions.txt","w")
	file.writelines(out)
	file.close()
	
	# generate variable restrictions
	# for each pointer variable will be generated one automaton to ensure,
	# that pointer variable is present only at unique position
	if restrict_var:
		for i in range(pointer_num-1):
			(bdd,table)=gen_restrict_variable.restrict_variable(i+1)
			functions.gen_gta_file(bdd,table, [9],"restrict_var-%i.gta"%(i))

	#--------------------------------------------------------------
	# now we generate checkers used to detect error configurations
	print "Generating checkers to detect error configurations ..."
	checker_conf=[]
	ch_num=0
	#capture error lines -- line number 1xxxxx
	checker_conf.append("1 -1\n")
	gen_checkers.line_num_checker("1","checker-%i.gta"%ch_num)
	ch_num=ch_num+1
	# capture the states, where a pointer variable is alliased to a disposed node
	checker_conf.append("1 -1\n")
	gen_checkers.delete_node_checker(pointer_num,"checker-%i.gta"%ch_num)
	ch_num=ch_num+1
	#write checkers configuration file
	file=open("check.txt","w")
	file.writelines(["%i\n"%ch_num]+checker_conf)
	file.close()
	print "... done"
	# create automaton "empty" - describing an empty language in MONA input
	gen_empty.gen_empty("empty.gta")

	# in case of variance analysis, generate seeding transducers
	if variance:
		# generate seeding transducers
		print "Generating predicates for variance analysis ..."
		if new:
			print """Only restricted class of variance predicates can be used due to "new" statement in program.py"""
		for i in range(pointer_num):
			(bdd,table,acc_list)=seed.seed_var(node_width,pointer_num,desc_num,data_size,i)
			functions.gen_gta_file(bdd,table,acc_list,"seed_var-%i.gta"%i)
		for i in range(data_size):
			(bdd,table,acc_list)=seed.seed_data_track(node_width, pointer_num, desc_num, data_size,i)
			functions.gen_gta_file(bdd,table,acc_list,"seed_data_track-%i.gta"%i)
		if (variance==2):
			(bdd,table,acc_list)=seed.seed_descriptors(node_width, desc_num)
			functions.gen_gta_file(bdd,table,acc_list,"seed_descriptors.gta")
			
		# generate variance predicates
		for i in range(pointer_num-1):
			(bdd,table,acc_list)=variance_checkers.gen_variance_var_desc(pointer_num, i+1, desc_num, data_size)
			functions.gen_gta_file(bdd,table,acc_list,"variance_pred-%i-decr.gta"%(i+1))
			if not new:
				(bdd,table,acc_list)=variance_checkers.gen_variance_var_inc(pointer_num, i+1, desc_num, data_size)
				functions.gen_gta_file(bdd,table,acc_list,"variance_pred-%i-inc.gta"%(i+1))
			(bdd,table,acc_list)=variance_checkers.gen_variance_var_stay(pointer_num, i+1, desc_num, data_size)
			functions.gen_gta_file(bdd,table,acc_list,"variance_pred-%i-eq.gta"%(i+1))
		if (not new) and (data_size>0):
			(bdd,table,acc_list)=variance_checkers.gen_variance_data_change(pointer_num, desc_num, data_size, 0)
			functions.gen_gta_file(bdd,table,acc_list,"variance_pred-data_inc.gta")
			(bdd,table,acc_list)=variance_checkers.gen_variance_data_change(pointer_num, desc_num, data_size, 1)
			functions.gen_gta_file(bdd,table,acc_list,"variance_pred-data_decr.gta")
			(bdd,table,acc_list)=variance_checkers.gen_variance_data_stay(pointer_num, desc_num, data_size)
			functions.gen_gta_file(bdd,table,acc_list,"variance_pred-data_eq.gta")
		if not new:
			(bdd,table,acc_list)=variance_checkers.gen_variance_inc_deleted(pointer_num, desc_num, data_size)
			functions.gen_gta_file(bdd,table,acc_list,"variance_pred-deleted.gta")
		# variance predicates for descriptors
		if (not new) and (variance==2):
			for i in range(next_num):
				(bdd,table,acc_list)=variance_checkers.gen_variance_descr_i_change(desc_num, next_num, i, 0)
				functions.gen_gta_file(bdd,table,acc_list,"variance_pred-descr_%i_inc.gta"%i)
				(bdd,table,acc_list)=variance_checkers.gen_variance_descr_i_change(desc_num, next_num, i, 1)
				functions.gen_gta_file(bdd,table,acc_list,"variance_pred-descr_%i_decr.gta"%i)
	
				(bdd,table,acc_list)=variance_checkers.gen_variance_descr_i_stay(desc_num, next_num,i)
				functions.gen_gta_file(bdd,table,acc_list,"variance_pred-descr_%i_eq.gta"%i)


		print "... done"

			
			

#----------------------------------
# main call

print sys.argv

variance=0

try:
	sys.argv.index("-variance")
	print "Support for variance analysis enabled"
	variance=1
	sys.argv.index("-vd")
	variance=2
except ValueError:
	pass

print variance

gen_transducers(variance)

