import functions
import sys

def seed_var(node_width, pointer_num, descr_num, data_len, x):
	'''seeding for variance analysis - seed variable x'''

# pointer_l   -- pointer descriptor length (i.e. the description of next pointers)
# node_l      -- memory node descriptor length (i.e. the allocated memory nodes below which we have pointer descriptors)
# pointer_num -- the number of pointer variables
# descr_num   -- number of pointer descriptors
# x, y        -- the numbers of pointer variables being assigned ( x := y )

# We use the following states in the transducer:
#   0 - init
#   1 - normal processing
#   2 - read top
#   3 - read null
#   4 - read undef
#   5 - ACCEPT
#   6 - REJECT

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,6,0))#0
	index=1

	#****************************************
	# left state=0, right state=0: 
	#   - empty nodes (00) till the first pointer description
	#   - the first pointer descriptor (01)
	#****************************************
	table.append((0,0,1))
	bdd.append((0,2,0))#1 -- the first bit of the type of the input node
	bdd.append((1,3,0))#2 -- the first bit of the type of the output node
	bdd.append((2,4,9))#3 -- if 1, we found the 1st pointer descriptor => go to line 9
	bdd.append((3,5,0))#4
	bdd.append((-1,0,0))#5 -- 00(0,0) -> 00, finished / note that 0 is a state, 00 an input symbol
	index=index+5

	#****************************************
	# left state=1, right state=1 or left state=0, right state=1: 
	#   - a next pointer description (01)
	#****************************************
	table.append((1,1,index))	
	table.append((0,1,index))	
	bdd.append((0,index+1,0))#6
	bdd.append((1,index+2,0))#7
	bdd.append((2,0,index+3))#8
	bdd.append((3,0,index+4))#9 -- 01(1 or 0,1) -> 01 where 01 is a pointer descriptor
	index=index+4
	descr_bdd=index # in the below case of _(1,0) and a pointer descriptor being read, we continue from here	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,1,0))
	index=index+1

	#****************************************
	# left state=1, right state=0:
	#   - a memory node (10)
	#   - a right-most top memory node (11)
	#   - a right-most pointer descriptor (01)
	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,-2)) # a top or normal memory node
	aux=index+1
	  # a normal memory node
	bdd.append((3,index+3,0)) 
	index=index+3
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num+descr_num+data_len,4,x)
	# x=0
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,x)
	bdd.append((-1,1,0))
	index=index+1
	#x=1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,x)
	bdd.append((-1,1,0))
	index=index+1

	  # a top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index) 
	bdd.append((3,0,index+1)) 
	index=index+1
	top_bdd=index
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num+descr_num+data_len,4,x)
	# x=0
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,x)
	bdd.append((-1,2,0))
	index=index+1
	#x=1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,x)
	bdd.append((-1,2,0))
	index=index+1


	#****************************************
	# left state=1, right state=2:
	#   - a top memory node (11)
	#****************************************
	table.append((1,2,index))
	bdd.append((0,0,index+1))
	bdd.append((1,0,index+2))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,top_bdd)) 
	index=index+4
	
	# ****************************************
	# left state=2, right state=0:
	#   - null pointers (00)
	# ****************************************
	table.append((2,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num,4,x)
	# x=0
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,pointer_num,(2+pointer_num)*2,x)
	bdd.append((-1,3,0))
	index=index+1
	#x=1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,pointer_num,(2+pointer_num)*2,x)
	bdd.append((-1,3,0))
	index=index+1

	
	# ****************************************
	# left state=3, right state=0:
	#   - undef pointers (00)
	# ****************************************
	table.append((3,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num,4,x)
	# x=0
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,pointer_num,(2+pointer_num)*2,x)
	bdd.append((-1,4,0))
	index=index+1
	#x=1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,pointer_num,(2+pointer_num)*2,x)
	bdd.append((-1,4,0))
	index=index+1


	# ****************************************
	# left state=4, right state=0:
	#  - a line number change 
	# ****************************************	
	table.append((4,0,index))	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)
	bdd.append((-1,5,0)) # accept
	index=index+1


	return(bdd,table,[5])

#------------------------------------------------------------------

def seed_data_track(node_width, pointer_num, descr_num, data_len, data_track):
	'''seeding for variance analysis - seed n-th track of data value'''

# pointer_l   -- pointer descriptor length (i.e. the description of next pointers)
# node_l      -- memory node descriptor length (i.e. the allocated memory nodes below which we have pointer descriptors)
# pointer_num -- the number of pointer variables
# descr_num   -- number of pointer descriptors
# x, y        -- the numbers of pointer variables being assigned ( x := y )

# We use the following states in the transducer:
#   0 - init
#   1 - normal processing
#   2 - read top
#   3 - read null
#   4 - read undef
#   5 - ACCEPT
#   6 - REJECT

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,6,0))#0
	index=1

	#****************************************
	# left state=0, right state=0: 
	#   - empty nodes (00) till the first pointer description
	#   - the first pointer descriptor (01)
	#****************************************
	table.append((0,0,1))
	bdd.append((0,2,0))#1 -- the first bit of the type of the input node
	bdd.append((1,3,0))#2 -- the first bit of the type of the output node
	bdd.append((2,4,9))#3 -- if 1, we found the 1st pointer descriptor => go to line 9
	bdd.append((3,5,0))#4
	bdd.append((-1,0,0))#5 -- 00(0,0) -> 00, finished / note that 0 is a state, 00 an input symbol
	index=index+5

	#****************************************
	# left state=1, right state=1 or left state=0, right state=1: 
	#   - a next pointer description (01)
	#****************************************
	table.append((1,1,index))	
	table.append((0,1,index))	
	bdd.append((0,index+1,0))#6
	bdd.append((1,index+2,0))#7
	bdd.append((2,0,index+3))#8
	bdd.append((3,0,index+4))#9 -- 01(1 or 0,1) -> 01 where 01 is a pointer descriptor
	index=index+4
	descr_bdd=index # in the below case of _(1,0) and a pointer descriptor being read, we continue from here	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,1,0))
	index=index+1

	#****************************************
	# left state=1, right state=0:
	#   - a memory node (10)
	#   - a right-most top memory node (11)
	#   - a right-most pointer descriptor (01)
	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,-2)) # a top or normal memory node
	aux=index+1
	  # a normal memory node
	bdd.append((3,index+3,0)) 
	index=index+3
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num+descr_num+data_len,4,pointer_num+descr_num+data_track)
	# x=0
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,pointer_num+data_track)
	bdd.append((-1,1,0))
	index=index+1
	#x=1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,pointer_num+data_track)
	bdd.append((-1,1,0))
	index=index+1

	  # a top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index) 
	bdd.append((3,0,index+1)) 
	index=index+1
	top_bdd=index
	(bdd,index,x_unset,x_set)=functions.copy_n_elements_and_remember_x(bdd,index,pointer_num+descr_num+data_len,4,pointer_num+descr_num+data_track)
	# x=0
	for i in x_unset:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_unset_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,pointer_num+data_track)
	bdd.append((-1,2,0))
	index=index+1
	#x=1
	for i in x_set:
		if i[1]==1:
			bdd[i[0]]=(bdd[i[0]][0],index,bdd[i[0]][2])
		else:
			bdd[i[0]]=(bdd[i[0]][0],bdd[i[0]][1],index)
	(bdd,index)=functions.copy_n_elements_set_x(bdd,index,pointer_num+data_len,(2+pointer_num+descr_num+data_len)*2,pointer_num+data_track)
	bdd.append((-1,2,0))
	index=index+1


	#****************************************
	# left state=1, right state=2:
	#   - a top memory node (11)
	#****************************************
	table.append((1,2,index))
	bdd.append((0,0,index+1))
	bdd.append((1,0,index+2))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,top_bdd)) 
	index=index+4
	
	# ****************************************
	# left state=2, right state=0:
	#   - null pointers (00)
	# ****************************************
	table.append((2,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,3,0))
	index=index+1

	
	# ****************************************
	# left state=3, right state=0:
	#   - undef pointers (00)
	# ****************************************
	table.append((3,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,4,0))
	index=index+1


	# ****************************************
	# left state=4, right state=0:
	#  - a line number change 
	# ****************************************	
	table.append((4,0,index))	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)
	bdd.append((-1,5,0)) # accept
	index=index+1


	return(bdd,table,[5])

#------------------------------------------------------------------

def seed_descriptors(node_width, descr_num):
	'''seeding for variance analysis - seed n-th track of data value'''


# NOTE: the empty nodes (00) are ignored by this function

# We use the following states in the transducer:
#   0 - init
#   1 - normal processing
#   2 - read top
#   3 - read null
#   4 - read undef
#   5 - ACCEPT
#   6 - REJECT

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,6,0))#0
	index=1

	#****************************************
	# left state=0, right state=0: 
	#   - empty nodes (00) till the first pointer description
	#   - the first pointer descriptor (01)
	#****************************************
	table.append((0,0,1))
	bdd.append((0,2,0))#1 -- the first bit of the type of the input node
	bdd.append((1,3,0))#2 -- the first bit of the type of the output node
	bdd.append((2,4,9))#3 -- if 1, we found the 1st pointer descriptor => go to line 9
	bdd.append((3,5,0))#4
	bdd.append((-1,0,0))#5 -- 00(0,0) -> 00, finished / note that 0 is a state, 00 an input symbol
	index=index+5

	#****************************************
	# left state=1, right state=1 or left state=0, right state=1: 
	#   - a next pointer description (01)
	#****************************************
	table.append((1,1,index))	
	table.append((0,1,index))	
	bdd.append((0,index+1,0))#6
	bdd.append((1,index+2,0))#7
	bdd.append((2,0,index+3))#8
	bdd.append((3,0,index+4))#9 -- 01(1 or 0,1) -> 01 where 01 is a pointer descriptor
	index=index+4
	descr_bdd=index # in the below case of _(1,0) and a pointer descriptor being read, we continue from here	
	#-------------------------------------------------------
	# here we seed descriptors
	# first handle NULL (01010)/UNDEFINED (01011)
	bdd.append((4,index+1,-2))
	aux=index
	bdd.append((5,index+2,0))
	bdd.append((6,0,index+3))
	bdd.append((7,0,index+4))
	index=index+4
	bdd.append((8,index+1,-2))
	aux1=index
	#01010
	bdd.append((9,index+2,0))
	index=index+2
	#seed "010" (without "01" on the beginning)
	bdd.append((5+(2*descr_num),index+1,0))
	bdd.append((7+(2*descr_num),0,index+2))
	bdd.append((9+(2*descr_num),index+3,0))
	bdd.append((-1,1,0))
	index=index+4
	#01011
	bdd[aux1]=(bdd[aux1][0],bdd[aux1][1],index)
	bdd.append((9,0,index+1))
	index=index+1
	#seed "010" (without "01" on the beginning)
	bdd.append((5+(2*descr_num),index+1,0))
	bdd.append((7+(2*descr_num),0,index+2))
	bdd.append((9+(2*descr_num),0,index+3))
	bdd.append((-1,1,0))
	index=index+4
	#----------------------------------------------------
	# normal descriptors -> only one occurence of "1" 011(0*10*)
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
	bdd.append((5,0,index+1))
	index=index+1
	for i in range(descr_num-1):
		if i>0: 
			bdd[aux]=(bdd[aux][0],index,bdd[aux][2])
			bdd.append((5+(2*i),index+1,0))
			index=index+1
		bdd.append((6+(2*i),0,index+1)) #remember this position
		aux=index
		bdd.append((7+(2*i),0,index+2))
		index=index+2
		(bdd,index)=functions.copy_n_elements(bdd,index,descr_num-(2+i),8+(2*i))
		# create seed
		bdd.append((5+(2*descr_num),0,index+1))
		index=index+1
		for j in range(descr_num-1):
			if i==j:
				bdd.append((7+(2*j)+(2*descr_num),0,index+1))
			else:
				bdd.append((7+(2*j)+(2*descr_num),index+1,0))
			index=index+1
		bdd.append((-1,1,0))
		index=index+1
	# ------ DONE

	#****************************************
	# left state=1, right state=0:
	#   - a memory node (10)
	#   - a right-most top memory node (11)
	#   - a right-most pointer descriptor (01)
	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor
	bdd.append((1,index+2,0))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,descr_bdd)) # goto processing of a pointer descriptor
	index=index+4
	# a memory node or a top memory node
	bdd.append((1,0,index+1))
	bdd.append((2,index+2,-2)) # a top or normal memory node
	aux=index+1
	# a normal memory node
	bdd.append((3,index+3,0)) 
	index=index+3
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,1,0))
	index=index+1
	# a top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index) 
	bdd.append((3,0,index+1)) 
	index=index+1
	top_bdd=index
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,2,0))
	index=index+1


	#****************************************
	# left state=1, right state=2:
	#   - a top memory node (11)
	#****************************************
	table.append((1,2,index))
	bdd.append((0,0,index+1))
	bdd.append((1,0,index+2))
	bdd.append((2,0,index+3)) 
	bdd.append((3,0,top_bdd)) 
	index=index+4
	
	# ****************************************
	# left state=2, right state=0:
	#   - null pointers (00)
	# ****************************************
	table.append((2,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,3,0))
	index=index+1

	
	# ****************************************
	# left state=3, right state=0:
	#   - undef pointers (00)
	# ****************************************
	table.append((3,0,index))
	bdd.append((0,index+1,0))
	bdd.append((1,index+2,0))
	bdd.append((2,index+3,0)) 
	bdd.append((3,index+4,0)) 
	index=index+4
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width-2,4)
	bdd.append((-1,4,0))
	index=index+1


	# ****************************************
	# left state=4, right state=0:
	#  - a line number change 
	# ****************************************	
	table.append((4,0,index))	
	(bdd,index)=functions.copy_n_elements(bdd,index,node_width,0)
	bdd.append((-1,5,0)) # accept
	index=index+1


	return(bdd,table,[5])
#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_seed():
	(bdd,table,acc_list)=seed_var(21, 4, 7,2,0)
	functions.gen_gta_file(bdd,table,acc_list,"test.gta")

	(bdd,table,acc_list)=seed_data_track(21, 4, 7,2,1)
	functions.gen_gta_file(bdd,table,acc_list,"test1.gta")

	(bdd,table,acc_list)=seed_descriptors(16, 8)
	functions.gen_gta_file(bdd,table,acc_list,"test2.gta")
#test_seed()

#=================================================================================

