import functions
import sys


def source_to_shape(from_var, to_var):

# from_var - pointer variable number describing 'from' position
# to_var - pointer variable number describing 'to' position

# This function generates transducer for 'shape automaton'
# INPUT: automaton describing data structure (even tracks) and position of variables describing 'from' and 'to' positions
# Output: (after projection on even tracks) 'shape automaton' Contains just 2 tracks (1 & 3)
# 00 - nothing
# 10 - sorce
# 01 - destination
# 11 - sorce/destination in one node
#
# There is just one node in automaton with 1 in 1st track, and one node with 1 in 3rd track
# This automaton has similar shape like original one, but it is possible to handle its contains easily.


# We use the following states in the transducer:
#   0 - initial state - nothing interesting has read
#   1 - read 'from' node
#   2 - read 'to' node
#   3 - read 'form' and 'to' position - accepting state
#   4 - reject state

# These are the codes of the types of the tree nodes (the first two bits, for transducers four bits: 1.in,1.out,2.in,2.out)
#   00 - an empty node
#   01 - a pointer descriptor
#   10 - a memory node
#   11 - a top memory node (above which there are only other top nodes and then the null & undef. pointers and the line number)
	
	# It is necessary to count with code of tree nodes - tracks 0 & 2
	# and with fact that input is just in odd tracks

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

	from_pos=(from_var*2)+4 # mona track describing 'from' position 
	to_pos=(to_var*2)+4 # mona track describing 'to' position

	#****************************************
	# bdd node 0 = a reject state
	#****************************************
	bdd.append((-1,4,0))#0
	index=1

       	#***************************************************************
	# input state 0,0
	# - in memory nodes (1x), we check for variable 'from' and 'to'
	# - Other nodes can be directly rewriten to '00' (output state 0)
       	#***************************************************************
	table.append((0,0,1))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor - OUTPUT: '00'
	bdd.append((1,index+2,0))
	bdd.append((3,index+3,0))
	bdd.append((-1,0,0))#5
	index=index+4
	# memory node: output depends on contain of tracks 'from_pos' and 'to_pos'
	if (from_pos<to_pos):
		bdd.append((1,index+1,index+8))
		bdd.append((3,index+2,index+5))
		# OUTPUT: '00'
		bdd.append((from_pos,index+3,0))
		bdd.append((to_pos,index+4,0))
		bdd.append((-1,0,0))
		index=index+5
		# OUTPUT: '01'
		bdd.append((from_pos,index+1,0))
		bdd.append((to_pos,0,index+2))
		bdd.append((-1,2,0))
		index=index+3
		bdd.append((3,index+1,index+4))
		# OUTPUT: '10'
		bdd.append((from_pos,0,index+2))
		bdd.append((to_pos,index+3,0))
		bdd.append((-1,1,0))
		index=index+4
		# OUTPUT: '11'
		bdd.append((from_pos,0,index+1))
		bdd.append((to_pos,0,index+2))
		bdd.append((-1,3,0))
		index=index+3
	else:
		bdd.append((1,index+1,index+8))
		bdd.append((3,index+2,index+5))
		# OUTPUT: '00'
		bdd.append((to_pos,index+3,0))
		bdd.append((from_pos,index+4,0))
		bdd.append((-1,0,0))
		index=index+5
		# OUTPUT: '01'
		bdd.append((to_pos,0,index+1))
		bdd.append((from_pos,index+2,0))
		bdd.append((-1,2,0))
		index=index+3
		bdd.append((3,index+1,index+4))
		# OUTPUT: '10'
		bdd.append((to_pos,index+2,0))
		bdd.append((from_pos,0,index+3))
		bdd.append((-1,1,0))
		index=index+4
		# OUTPUT: '11'
		bdd.append((to_pos,0,index+1))
		bdd.append((from_pos,0,index+2))
		bdd.append((-1,3,0))
		index=index+3
	#********************************************
	# input state (0,1) & (1,0)
	# - in memory nodes (1x), we check for variable 'to'
	# - Other nodes can be directly rewriten to '00' (output state 1)
       	#***************************************************************
	table.append((1,0,index))
	table.append((0,1,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor - OUTPUT: '00'
	bdd.append((1,index+2,0))
	bdd.append((3,index+3,0))
	bdd.append((-1,1,0))#5
	index=index+4
	# memory node: output depends on contain of track 'to_pos'
	# it is not necessary to check for 'from_pos', we suppose correct form of automaton
	bdd.append((1,index+1,0))
	bdd.append((3,index+2,index+4))
	# OUTPUT: '00'
	bdd.append((to_pos,index+3,0))
	bdd.append((-1,1,0))
	index=index+4
	# OUTPUT: '01'
	bdd.append((to_pos,0,index+1))
	bdd.append((-1,3,0))
	index=index+2

	#********************************************
	# input state (0,2) & (2,0)
	# - in memory nodes (1x), we check for variable 'from'
	# - Other nodes can be directly rewriten to '00' (output state 1)
       	#***************************************************************
	table.append((2,0,index))
	table.append((0,2,index))
	bdd.append((0,index+1,index+4)) # pointer descriptor / memory node
	# pointer descriptor - OUTPUT: '00'
	bdd.append((1,index+2,0))
	bdd.append((3,index+3,0))
	bdd.append((-1,2,0))#5
	index=index+4
	# memory node: output depends on contain of track 'from_pos'
	# it is not necessary to check for 'to_pos', we suppose correct form of automaton
	bdd.append((1,index+1,index+4))
	bdd.append((3,index+2,0))
	# OUTPUT: '00'
	bdd.append((from_pos,index+3,0))
	bdd.append((-1,2,0))
	index=index+4
	# OUTPUT: '01'
	bdd.append((3,index+1,0))
	bdd.append((from_pos,0,index+2))
	bdd.append((-1,3,0))
	index=index+3
	#********************************************
	# imput state (3,0), (0,3), (1,2), (2,1)
	# read both symbols - 'from' and 'to' - just output '00' and state 3
	#*******************************************
	table.append((3,0,index))
	table.append((0,3,index))
	table.append((2,1,index))
	table.append((1,2,index))
	# OUTPUT: '00'
	bdd.append((1,index+1,0))
	bdd.append((3,index+2,0))
	bdd.append((-1,3,0))#5
	index=index+3

	# 3rd argument are accepting states
	return (bdd,table,[3])

#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_source_to_shape():
	(bdd,table,accept_l)=source_to_shape(3,4)
	functions.gen_gta_file(bdd,table, accept_l,"sorce_to_shape.gta")

#test_source_to_shape()


#=================================================================================

