# Adam Rogalewicz

# TODO: case where more top nodes is present (not handled by most of the generators
#      - variable checkers - check the case where "1" appear in tree label
#      - ok for descriptor stay
#      - ok for data stay
#      - ok for data increase


import functions
import sys

def gen_variance_var_desc(var_num, x, descr_num, data_len):

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init, x, and old(x) not yet read
#  1 - read x
#  2 - read old(x)
#  3 - x and old(x) are in relation - ACCEPT
#  4 - reject


	# BDD node 0 - reject state
	bdd.append((-1,4,0))
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> 0
	#      memory node (x) -> 1
	#      memory node old(x) -> 2
	#      memory node x + old(x) -> 4
	#****************************************
	table.append((0,0,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,0,0))
	#memory node
	bdd.append((4+(x*2),index+3,index+6))
	index=index+3
	# x not present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+1,index+2))
	bdd.append((-1,0,0))
	bdd.append((-1,2,0)) #old(x) present
	index=index+3
	# x present -> check not old(x)
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+1,0))
	bdd.append((-1,1,0)) # x present in this memory node
	index=index+2

       	#****************************************
	# left state=1, right state=0: 
	#      descriptor -> 1
	#      memory node, not old(x) -> 1
	#      memory node,  old(x) -> reject
       	#****************************************
	table.append((1,0,index))
	table.append((0,1,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,1,0))
	#memory node
	bdd.append((4+(x*2),index+3,0)) # x can not be present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+4,0)) # old(x) can not be present
	bdd.append((-1,1,0))
	index=index+5
       	#****************************************
	# left state=2, right state=0: 
	#      descriptor -> 2
	#      memory node, x -> accept
	#      memory node,  not(x) -> 2
       	#****************************************
	table.append((2,0,index))
	table.append((0,2,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,2,0))
	#memory node
	bdd.append((4+(x*2),index+3,index+5)) # x
	# x not present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+4,0)) # old(x) can not be present
	bdd.append((-1,2,0))
	index=index+5
	# x present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+1,0)) # old(x) can not be present
	bdd.append((-1,3,0))
	index=index+2

       	#****************************************
	# left state=1, right state=2: 
	#      descriptor -> accept
	#      TOP memory node -> accept
       	#****************************************
	table.append((1,2,index))
	bdd.append((0,index+1,0))
	bdd.append((-1,3,0))
	index=index+2
	# NOTE: the rule (2,1) -> reject is implicit, not necessary to define it

       	#****************************************
	# left state=3, right state=0: 
	# left state=0, right state=3: 
	#      -> accept
       	#****************************************
	table.append((3,0,index))
	table.append((0,3,index))
	bdd.append((-1,3,0))
	index=index+1
	
	return(bdd,table,[3])

#********************************************************************************
def gen_variance_var_inc(var_num, x, descr_num, data_len):

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init, x, and old(x) not yet read
#  1 - read x
#  2 - read old(x)
#  3 - x and old(x) are in relation - ACCEPT
#  4 - reject


	# BDD node 0 - reject state
	bdd.append((-1,4,0))
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> 0
	#      memory node (x) -> 1
	#      memory node old(x) -> 2
	#      memory node x + old(x) -> 4
	#****************************************
	table.append((0,0,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,0,0))
	#memory node
	bdd.append((4+(x*2),index+3,index+6))
	index=index+3
	# x not present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+1,index+2))
	bdd.append((-1,0,0))
	bdd.append((-1,2,0)) #old(x) present
	index=index+3
	# x present -> check not old(x)
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+1,0))
	bdd.append((-1,1,0)) # x present in this memory node
	index=index+2

       	#****************************************
	# left state=1, right state=0: 
	#      descriptor -> 1
	#      memory node, not old(x) -> 1
	#      memory node,  old(x) -> accept
       	#****************************************
	table.append((1,0,index))
	table.append((0,1,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,1,0))
	#memory node
	bdd.append((4+(x*2),index+3,0)) # x can not be present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+4,index+5)) # old(x) 
	bdd.append((-1,1,0))
	bdd.append((-1,3,0)) # old(x) present -> accept
	index=index+6
       	#****************************************
	# left state=2, right state=0: 
	#      descriptor -> 2
	#      memory node, x -> reject
	#      memory node,  not(x) -> 2
       	#****************************************
	table.append((2,0,index))
	table.append((0,2,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,2,0))
	#memory node
	bdd.append((4+(x*2),index+3,0)) # x
	# x not present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+4,0)) # old(x) can not be present
	bdd.append((-1,2,0))
	index=index+5

       	#****************************************
	# left state=2, right state=1: 
	#      descriptor -> accept
       	#****************************************
	table.append((2,1,index))
	bdd.append((0,index+1,0))
	bdd.append((-1,3,0))
	index=index+2
	# NOTE: the rule (1,2) -> reject is implicit, not necessary to define it

       	#****************************************
	# left state=3, right state=0: 
	# left state=0, right state=3: 
	#      -> accept
       	#****************************************
	table.append((3,0,index))
	table.append((0,3,index))
	bdd.append((-1,3,0))
	index=index+1
	
	return(bdd,table,[3])

#********************************************************************************
def gen_variance_var_stay(var_num, x, descr_num, data_len):

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init, x, and old(x) not yet read

#  3 - x and old(x) are in the same node -ACCEPT
#  4 - reject


	# BDD node 0 - reject state
	bdd.append((-1,4,0))
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> 0
	#      memory node x + old(x) -> 3
	#****************************************
	table.append((0,0,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,0,0))
	#memory node
	bdd.append((4+(x*2),index+3,index+5))
	index=index+3
	# x not present => old(x) not present
	bdd.append(((2+var_num+descr_num+data_len+x)*2,index+1,0))
	bdd.append((-1,0,0))
	index=index+2
	# x present => old(x) must be presented
	bdd.append(((2+var_num+descr_num+data_len+x)*2,0,index+1))
	bdd.append((-1,3,0)) # x present in this memory node
	index=index+2

       	#****************************************
	# left state=3, right state=0: 
	# left state=0, right state=3: 
	#      -> accept
       	#****************************************
	table.append((3,0,index))
	table.append((0,3,index))
	bdd.append((-1,3,0))
	index=index+1
	
	return(bdd,table,[3])

#********************************************************************************
def gen_variance_inc_deleted(var_num, descr_num, data_len):

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init no increase
#  1 - increased number of deleted nodes
#  2 - increased number - read top memory node
#  3 - read null
#  4 - read undef
#  5 - read line number - ACCEPT
#  6 - reject

	# BDD node 0 - reject state
	bdd.append((-1,6,0))
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> 0
	#      memory node - deleted + orig_not_del -> 1 
	#      memory node - othervice ->0
	#      TOP memory node - deleted + orig_not_del -> 2
	#      TOP memory node - othervice -> reject
	#****************************************
	table.append((0,0,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,0,0))
	index=index+2
	#memory node
	bdd.append((2,index+1,index+7)) #normal/top memory node
	bdd.append((4,index+2,index+4))
	bdd.append((2*(2+var_num+descr_num+data_len),index+3,0)) # the node can not be resurected ...
	bdd.append((-1,0,0))
	index=index+4
	# deleted node
	bdd.append((2*(2+var_num+descr_num+data_len),index+1,index+2)) # check whether the node was deleted before
	bdd.append((-1,1,0)) #founded newly deleted node
	bdd.append((-1,0,0))
	index=index+3
	#TOP memory node
	bdd.append((4,0,index+1))
	bdd.append((2*(2+var_num+descr_num+data_len),index+2,0)) # the node can not be resurected ...
	bdd.append((-1,2,0)) #founded newly deleted node + TOP
	index=index+3
       	#****************************************
	# left state=1, right state=0 
	#      descriptor -> 1
	#      memory node - RESURECTED = unpossible -> reject 
	#      memory node - othervice -> 1
	#      TOP memory node - RESURECTED = unpossible -> reject
	#      TOP memory node - othervice -> 2
       	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,1,0))
	index=index+2
	#memory node
	bdd.append((2,index+1,index+4)) #normal/top memory node
	bdd.append((4,index+2,index+3))
	bdd.append((2*(2+var_num+descr_num+data_len),index+3,0)) # the node can not be resurected ...
	bdd.append((-1,1,0))
	index=index+4
	#TOP memory node
	bdd.append((4,index+1,index+2))
	bdd.append((2*(2+var_num+descr_num+data_len),index+2,0)) # the node can not be resurected ...
	bdd.append((-1,2,0))
	index=index+3

       	#****************************************
	# left state=1, right state=1 
	# left state=0, right state=1 
	#      descriptor -> 1
       	#****************************************
	table.append((0,1,index))
	table.append((1,1,index))
	bdd.append((0,index+1,0))
	bdd.append((-1,1,0))
	index=index+2

       	#****************************************
	# left state=2, right state=0 
       	#****************************************
	table.append((2,0,index))
	bdd.append((-1,3,0))
	index=index+1
       	#****************************************
	# left state=3, right state=0 
       	#****************************************
	table.append((3,0,index))
	bdd.append((-1,4,0))
	index=index+1
       	#****************************************
	# left state=4, right state=0 
       	#****************************************
	table.append((4,0,index))
	bdd.append((-1,5,0))
	index=index+1

	return(bdd,table,[5])

#********************************************************************************

def  gen_variance_inc_data_check_value(bdd,index,track,data,bdd_node_bigger, bdd_node_smaller,bdd_eq):
	if data=="": 
		bdd.append((-1,bdd_eq,0))
		index=index+1
		return (bdd,index)
	if data[0]=="1":
		bdd.append((track,bdd_node_bigger,index+1))
		return gen_variance_inc_data_check_value(bdd,index+1,track+2,data[1:],bdd_node_bigger, bdd_node_smaller,bdd_eq)
	else:
		bdd.append((track,index+1,bdd_node_smaller))
		return gen_variance_inc_data_check_value(bdd,index+1,track+2,data[1:],bdd_node_bigger, bdd_node_smaller,bdd_eq)

def gen_variance_inc_data_scan_value(bdd,index,track_scan,scan_len,track_check,data,bdd_node_bigger, bdd_node_smaller,bdd_eq):
	if scan_len==0:
		if data=="":
			print "ERROR: no data - unpossible to create variance checker"
			sys.exit()
		else:
			return gen_variance_inc_data_check_value(bdd,index,track_check,data,bdd_node_bigger, bdd_node_smaller,bdd_eq)
	bdd.append((track_scan,index+1,-2))
	
	aux1=index
	(bdd,index)=gen_variance_inc_data_scan_value(bdd,index+1,track_scan+2,scan_len-1,track_check,data+"0",bdd_node_bigger, bdd_node_smaller,bdd_eq)
	bdd[aux1]=(bdd[aux1][0],bdd[aux1][1],index)
	(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,scan_len-1,track_check,data+"1",bdd_node_bigger, bdd_node_smaller,bdd_eq)
	return (bdd,index)


def gen_variance_data_change(var_num, descr_num, data_len, dir):
# this function create an automaton, which accepts all trees, where
#	- in all nodes: the new data value is >= the old one
#       - exists a node: the new data value > the old one
# the parameter "dir" set to "1" changes the relation ">" into "<"

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init no increase
#  1 - increased data
#  2 - read top + increase
#  3 - read null
#  4 - read undef
#  5 - read line num
#  6 - reject
#  7 - read top + data stay

	track_scan=(2+var_num+descr_num)*2
	track_check=(2+var_num+descr_num+data_len+var_num)*2

	# BDD node 0 - reject state
	bdd.append((-1,6,0))
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> 0
	#      memory node - data increased -> 1 
	#      memory node - data not changed -> 0
	#      TOP memory node - data increased -> 2
	#****************************************
	table.append((0,0,index))
	table.append((0,7,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,0,0))
	index=index+2
	#memory node
	bdd.append((2,index+2,-2)) # top/normal
	aux=index
	bdd.append((-1,1,0))
	index=index+2
	if dir:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",0, aux+1, 0)
	else:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",aux+1, 0,0)
	# top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index+1)
	bdd.append((-1,2,0))
	aux=index
	index=index+1
	if dir:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",0,aux,7)
	else:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",aux, 0,7)

       	#****************************************
	# left state=1, right state=0 
	#      descriptor -> 1
	#      memory node - data equal or increased -> 1 
	#      TOP memory node - data equal or increased -> 2 
       	#****************************************
	table.append((1,0,index))
	table.append((1,7,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,1,0))
	index=index+2
	#memory node
	bdd.append((2,index+2,-2)) #top/normal
	aux=index
	bdd.append((-1,1,0))
	index=index+2
	if dir:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",0,aux+1, 1)
	else:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",aux+1, 0,1)
	# top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index+1)
	bdd.append((-1,2,0))
	aux=index
	index=index+1
	if dir:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",0, aux, 2)
	else:
		(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",aux, 0,2)

       	#****************************************
	# left state=1, right state=1 
	# left state=0, right state=1 
	#      descriptor -> 1
       	#****************************************
	table.append((0,1,index))
	table.append((1,1,index))
	bdd.append((0,index+1,0))
	bdd.append((-1,1,0))
	index=index+2

       	#****************************************
	# left state=2, right state=0 
       	#****************************************
	table.append((2,0,index))
	bdd.append((-1,3,0))
	index=index+1
       	#****************************************
	# left state=3, right state=0 
       	#****************************************
	table.append((3,0,index))
	bdd.append((-1,4,0))
	index=index+1
       	#****************************************
	# left state=4, right state=0 
       	#****************************************
	table.append((4,0,index))
	bdd.append((-1,5,0))
	index=index+1

	return(bdd,table,[5])

#********************************************************************************
def gen_variance_data_stay(var_num, descr_num, data_len):
# this function create an automaton, which accepts all trees, where the old and new data values are equal

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init + OK.
#  2 - read top + OK.
#  3 - read null 
#  4 - read undef
#  5 - read line num
#  6 - reject

	track_scan=(2+var_num+descr_num)*2
	track_check=(2+var_num+descr_num+data_len+var_num)*2

	# BDD node 0 - reject state
	bdd.append((-1,6,0))
	index=1

       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> 0
	#      memory node - data not changed -> 0
	#      TOP memory node - data not changed -> 2
	#****************************************
	table.append((0,0,index))
	bdd.append((0,index+1,index+2))
	bdd.append((-1,0,0))
	index=index+2
	#memory node
	bdd.append((2,index+1,-2)) # top/normal
	aux=index
	index=index+1
	(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",0, 0, 0)
	# top memory node
	bdd[aux]=(bdd[aux][0],bdd[aux][1],index)
	(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",0, 0,2)

       	#****************************************
	# left state=0, right state=2: 
	#      TOP memory node - data not changed -> 2
	#****************************************
	table.append((0,2,index))
	bdd.append((0,0,index+1))
	bdd.append((2,0,index+2)) # top
	index=index+2
	(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan,data_len,track_check,"",0, 0,2)


       	#****************************************
	# left state=2, right state=0 
       	#****************************************
	table.append((2,0,index))
	bdd.append((-1,3,0))
	index=index+1
       	#****************************************
	# left state=3, right state=0 
       	#****************************************
	table.append((3,0,index))
	bdd.append((-1,4,0))
	index=index+1
       	#****************************************
	# left state=4, right state=0 
       	#****************************************
	table.append((4,0,index))
	bdd.append((-1,5,0))
	index=index+1

	return(bdd,table,[5])


#***************************************************************************************************
# Variance checkers for descriptor fields
# checks values on the i^{th} descriptor field of all memory nodes

def gen_variance_descr_i_change(descr_num, next_num, next, dir):
# this function create an automaton, which accepts all trees, where
#	- in all descriptor nodes corresponding to the i^th next, the descriptor value is >= the old one
#       - exists a node: the new descriptor value > the old one
# the parameter "dir" set to "1" changes the relation ">" into "<"

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init no increase
#  1 - increased descr
#  2 - read top + increase
#  3 - read null
#  4 - read undef
#  5 - read line num
#  6 - reject
#  7 - read top + descr stay


	track_scan=4
	track_check=4+(descr_num*2)

	# BDD node 0 - reject state
	bdd.append((-1,6,0))
	index=1

	#----------------------------------------
	# descriptor handling
	#----------------------------------------

	# handle next nodes 
	for i in range(next_num):
		if i==next:
			# no descr node with increased value in the subtree
			if (i<next_num-1):
				table.append((0,8+(i*2),index))
				bdd.append((0,index+1,0))
				bdd.append((2,0,index+3))
				index=index+2
			else: descr_bdd=index+1
			if i==0: 
				bdd.append((-1,1,0))
				aux_leaf=index
				aux=0
			else: 
				bdd.append((-1,7+(i*2),0))
				aux_leaf=index
				aux=6+(i*2)
			index=index+1
			# distinguish null+undef /  normal descriptors
			bdd.append((4,-2,index+1))
			bdd_rem=index
			index=index+1
			if dir:
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,descr_num-1,track_check,"1",0,aux_leaf,aux )
				# return to null/undef
				bdd[bdd_rem]=(bdd[bdd_rem][0],index,bdd[bdd_rem][2])
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,2,track_check,"0",0,aux_leaf,aux )
			else:
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,descr_num-1,track_check,"1",aux_leaf, 0,aux)
				# return to null/undef
				bdd[bdd_rem]=(bdd[bdd_rem][0],index,bdd[bdd_rem][2])
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,2,track_check,"0",aux_leaf, 0,aux)

			# some descr node with increased value in the subtree
			if (i<next_num-1):
				table.append((1,8+(i*2),index))
				table.append((0,9+(i*2),index))
				table.append((1,9+(i*2),index))
				bdd.append((0,index+1,0))
				bdd.append((2,0,index+3))
				index=index+2
			else: descr_bdd1=index+1
			if i==0: 
				bdd.append((-1,1,0))
				aux_leaf=index
				aux=1
			else: 
				bdd.append((-1,7+(i*2),0))
				aux_leaf=index
				aux=6+(i*2)
			index=index+1
			# distinguish null+undef /  normal descriptors
			bdd.append((4,-2,index+1))
			bdd_rem=index
			index=index+1
			if dir:
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,descr_num-1,track_check,"1",0,aux_leaf,aux)
				# return to null/undef
				bdd[bdd_rem]=(bdd[bdd_rem][0],index,bdd[bdd_rem][2])
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,2,track_check,"0",0,aux_leaf,aux)
			else:
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,descr_num-1,track_check,"1",aux_leaf, 0,aux)
				# return to null/undef
				bdd[bdd_rem]=(bdd[bdd_rem][0],index,bdd[bdd_rem][2])
				(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,2,track_check,"0",aux_leaf, 0,aux)

		#---------------------------------
		# handle descriptor nodes, which are not interesting for this variance checker
		else:
			# no descr node with increased value in the subtree
			if (i<next_num-1):
				table.append((0,8+(i*2),index))
				bdd.append((0,index+1,0))
				bdd.append((2,0,index+2))
				index=index+2
			else: descr_bdd=index
			if i==0: bdd.append((-1,0,0))
			else: bdd.append((-1,6+(i*2),0))
			index=index+1
			# some descr node with increased value in the subtree
			if (i<next_num-1):
				table.append((1,8+(i*2),index))
				table.append((0,9+(i*2),index))
				table.append((1,9+(i*2),index))
				bdd.append((0,index+1,0))
				bdd.append((2,0,index+2))
				index=index+2
			else: descr_bdd1=index
			if i==0: bdd.append((-1,1,0))
			else: bdd.append((-1,7+(i*2),0))
			index=index+1
		
		


       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> check value of i^th descriptor
	#      memory node -> 0 
	#      TOP memory node -> 7
	#****************************************
	table.append((0,0,index))
	bdd.append((0,index+1,index+3))
	bdd.append((2,index+2,descr_bdd))
	bdd.append((-1,0,0))
	index=index+3

	#memory node
	bdd.append((2,index+1,index+2)) # top memory 
	bdd.append((-1,0,0))
	bdd.append((-1,7,0))
	index=index+3

       	#****************************************
	# left state=1, right state=0 
	#      descriptor -> check value of the i^th descriptor
	#      memory node -> 1 
	#      TOP memory node -> 2 
       	#****************************************
	table.append((1,0,index))
	bdd.append((0,index+1,index+2))
	bdd.append((2,0,descr_bdd1))
	index=index+2
	#memory node
	bdd.append((2,index+1,index+2)) # top/normal
	bdd.append((-1,1,0))
	bdd.append((-1,2,0))
	index=index+3

       	#****************************************
	# left state=0, right state=2: 
	# left state=1, right state=2: 
	#      TOP memory node -> 2
	#****************************************
	table.append((0,2,index))
	table.append((1,2,index))
	table.append((1,7,index))
	bdd.append((-1,2,0))
	index=index+1

       	#****************************************
	# left state=0, right state=2: 
	# left state=1, right state=2: 
	#      TOP memory node -> 2
	#****************************************
	table.append((0,7,index))
	bdd.append((-1,7,0))
	index=index+1

       	#****************************************
	# left state=2, right state=0 
       	#****************************************
	table.append((2,0,index))
	bdd.append((-1,3,0))
	index=index+1
       	#****************************************
	# left state=3, right state=0 
       	#****************************************
	table.append((3,0,index))
	bdd.append((-1,4,0))
	index=index+1
       	#****************************************
	# left state=4, right state=0 
       	#****************************************
	table.append((4,0,index))
	bdd.append((-1,5,0))
	index=index+1

	return(bdd,table,[5])

def gen_variance_descr_i_stay(descr_num, next_num, next):
# this function create an automaton, which accepts all trees, where
#	- the value in all descriptor nodes corresponding to the i^th next is preserved

	table=[] # The main transition table States x States --> BDD nodes
	bdd=[]   # Further looking up of the next state for the given situation.
                 # The triple is (the bit of input, the next BDD node for 0, the next BDD node for 1).
		 # (-1,x,_) then means that the target state has been found and it's x.

# We use the following states in the automaton:
#  0 - init no increase
#  2 - read top + increase
#  3 - read null
#  4 - read undef
#  5 - read line num
#  6 - reject


	track_scan=4
	track_check=4+(descr_num*2)

	# BDD node 0 - reject state
	bdd.append((-1,6,0))
	index=1

	#----------------------------------------
	# descriptor handling
	#----------------------------------------

	# handle next nodes 
	for i in range(next_num):
		if i==next:
			# check whether the value is preserved
			if (i<next_num-1):
				table.append((0,7+i,index))
				bdd.append((0,index+1,0))
				bdd.append((2,0,index+2))
				index=index+2
			else: descr_bdd=index
			if i==0: 
				aux=0
			else: 
				aux=6+i
			bdd.append((4,-2,index+1))
			bdd_rem=index
			index=index+1
			(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,descr_num-1,track_check,"1",0,0,aux )
			# return to null/undef
			bdd[bdd_rem]=(bdd[bdd_rem][0],index,bdd[bdd_rem][2])
			(bdd,index)=gen_variance_inc_data_scan_value(bdd,index,track_scan+2,2,track_check,"0",0,0,aux )

		#---------------------------------
		# handle descriptor nodes, which are not interesting for this variance checker
		else:
			# no descr node with increased value in the subtree
			if (i<next_num-1):
				table.append((0,7+i,index))
				bdd.append((0,index+1,0))
				bdd.append((2,0,index+2))
				index=index+2
			else: descr_bdd=index
			if i==0: bdd.append((-1,0,0))
			else: bdd.append((-1,6+i,0))
			index=index+1
		
		


       	#****************************************
	# left state=0, right state=0: 
	#      descriptor -> check value
	#      memory node - -> 0 
	#      TOP memory node -> 2
	#****************************************
	table.append((0,0,index))
	bdd.append((0,index+1,index+3))
	bdd.append((2,index+2,descr_bdd))
	bdd.append((-1,0,0))
	index=index+3

	#memory node
	bdd.append((2,index+1,index+2)) # normal/top
	bdd.append((-1,0,0))
	bdd.append((-1,2,0))
	index=index+3

       	#****************************************
	# left state=0, right state=2: 
	#      TOP memory node -> 2
	#****************************************
	table.append((0,2,index))
	bdd.append((-1,2,0))
	index=index+1

       	#****************************************
	# left state=2, right state=0 
       	#****************************************
	table.append((2,0,index))
	bdd.append((-1,3,0))
	index=index+1
       	#****************************************
	# left state=3, right state=0 
       	#****************************************
	table.append((3,0,index))
	bdd.append((-1,4,0))
	index=index+1
       	#****************************************
	# left state=4, right state=0 
       	#****************************************
	table.append((4,0,index))
	bdd.append((-1,5,0))
	index=index+1

	return(bdd,table,[5])


#--------------------------------------------------------------------------------
#              T E S T S
#--------------------------------------------------------------------------------

def test_variance():
	(bdd,table,acc_list)=gen_variance_var_desc(3, 1, 4, 1)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-x-desc.gta")

	(bdd,table,acc_list)=gen_variance_var_inc(3, 1, 4, 1)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-x-inc.gta")

	(bdd,table,acc_list)=gen_variance_var_stay(3, 1, 4, 1)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-x-stay.gta")

	(bdd,table,acc_list)=gen_variance_inc_deleted(3, 4, 1)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-deleted.gta")


	(bdd,table,acc_list)=gen_variance_data_change(3, 4, 2, 0)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-data_vals_incr.gta")

	(bdd,table,acc_list)=gen_variance_data_change(3, 4, 2, 1)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-data_vals_desc.gta")

	(bdd,table,acc_list)=gen_variance_data_stay(3, 4, 2)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-data_vals_eq.gta")

	(bdd,table,acc_list)=gen_variance_descr_i_change(3, 3, 1, 0)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-descr_vals_incr.gta")
	
	(bdd,table,acc_list)=gen_variance_descr_i_stay(4, 4, 0)
	functions.gen_gta_file(bdd,table,acc_list,"variance_checker-descr_vals_stay.gta")
#test_variance()
