/* Generated By:JavaCC: Do not edit this line. ProgramParser.java */
package parser ;

import java.io.* ;
import java.util.* ;


/**
 * The order parser class.
 *
 * @author Pierre Salagnac
 */
public class ProgramParser implements ProgramParserConstants {

  // comment prefixes
  private final static String VARIABLE_ORDER_PREFIX = "pointer variables are :" ;
  private final static String MEMBER_ORDER_PREFIX   = "next pointers are :" ;

  // orders
  private String variableOrder ;
  private String memberOrder ;
  private OrderParser variableParser ;
  private OrderParser memberParser ;

  // parsed values
  private Map<String,String> values ;

  // program lines
  private List<String> lineBinaries ;


  /**
   * Start the parse process.
   *
   * @exception ParseException The input cannot be parsed.
   */
  public void parse() throws ParseException {

    init() ;

    doParse() ;

    parseOrders() ;

  }


  /**
   * Initialize the parser before starting the parse process.
   */
  private void init() {

    values = new HashMap<String,String>() ;

    lineBinaries = new ArrayList<String>() ;

  }


  /**
   * Start parse orders in comments.
   *
   * @exception ParseException The order definition is wrong.
   */
  private void parseOrders() throws ParseException {

    StringReader memberReader   = new StringReader(memberOrder) ;
    StringReader variableReader = new StringReader(variableOrder) ;

    memberParser   = new OrderParser(memberReader) ;
    variableParser = new OrderParser(variableReader) ;

    memberParser.parse() ;
    variableParser.parse(-1) ;
  }


  /**
   * Return the definition of variables order. This order is parsed in a
   * comment.
   * 
   * @return The variables order.
   */
  public String getVariableOrder() {

    return variableOrder ;

  }


  /**
   * Return the definition of members order. This order is parsed in a
   * comment.
   *
   * @return The members order.
   */
  public String getMemberOrder() {

    return memberOrder ;

  }


  /**
   * Return the binary code of the specified program line.
   *
   * @param index The line index.
   *
   * @return The line binary code.
   */
  public String getLineBinary(int index) {

    return lineBinaries.get(index) ;

  }


  /**
   * Return the maximum with alowed for a tree node.
   *
   * @return The node with, or -1 if it is not specified.
   */
  public int getNodeWidth() {

    String key = "node_width" ;

    if (values.containsKey(key)) {
      String str = values.get(key) ;
      int width = Integer.parseInt(str) ;
      return width ;
    }

    else
      return -1 ;


  }


  /**
   * Return the size allocated for the data in a type 3 node.
   *
   * @return The data length.
   */
  public int getDataWidth() {

    int width = getNodeWidth() ;

    // remove fixed width
    width -= 4 ;

    // remove variables
    width -= getVariableCount() ;

    // remove descriptors
    width -= getDescriptorCount() ;

    return width ;

  }


  /**
   * Return the number of descriptors that must be put in type 3 nodes.
   *
   * @return The descriptor count, or -1 if it is not specified.
   */
  public int getDescriptorCount() {

    String key = "desc_num" ;

    if (values.containsKey(key)) {
      String str = values.get(key) ;
      int count = Integer.parseInt(str) ;
      return count ;
    }

    else
      return 1 ;

  }


  /**
   * Return the number of variables declared in the program.
   *
   * @return The variable count.
   */
  public int getVariableCount() {
   return variableParser.getInputCount() ;
  }


  /**
   * Return the number of member declared in the program.
   *
   * @return The member count.
   */
  public int getMemberCount() {
    return memberParser.getInputCount() ;
  }


  /**
   * Return the index of the specified variable.
   *
   * @param name The variable name.
   */
  public int getVariableIndex(String name) {
    return variableParser.getInputIndex(name) ;
  }


  /**
   * Return the index of the specified member.
   *
   * @param name The member name.
   */
  public int getMemberIndex(String name) {
    return memberParser.getInputIndex(name) ;
  }


  /**
   * Return the variable associated with this index.
   *
   * @param index The variable index.
   *
   * @return The corresponding variable name.
   */
  public String getVariable(int index) {
    return variableParser.getInput(index) ;
  }


  /**
   * Return the member associated with this index.
   *
   * @param index The pointer index.
   *
   * @return The corresponding variable name.
   */
  public String getMember(int index) {
    return memberParser.getInput(index) ;
  }


  /**
   * Return if the specified variable is declared or not.
   *
   * @param name The variable name.
   *
   * @return true if the variable is declared.
   */
  public boolean containsVariable(String name) {
    return variableParser.containsInput(name) ;
  }


  /**
   * Return if the specified member is declared.
   *
   * @param name The member name.
   *
   * @return true if the member is declared.
   */
  public boolean containsMember(String name) {
    return memberParser.containsInput(name) ;
  }

  final private void doParse() throws ParseException {
    label_1:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case COMMENT:
        ;
        break;
      default:
        jj_la1[0] = jj_gen;
        break label_1;
      }
      parseComment();
    }
    jj_consume_token(DEF);
    jj_consume_token(GET_PROGRAM);
    jj_consume_token(OPENING_PAR);
    jj_consume_token(CLOSING_PAR);
    jj_consume_token(COLON);
    label_2:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case COMMENT:
        ;
        break;
      default:
        jj_la1[1] = jj_gen;
        break label_2;
      }
      parseComment();
    }
    parseIdentifier();
    jj_consume_token(EQUAL);
    parseProgram();
    label_3:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case IDENTIFIER:
      case COMMENT:
        ;
        break;
      default:
        jj_la1[2] = jj_gen;
        break label_3;
      }
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case COMMENT:
        parseComment();
        break;
      case IDENTIFIER:
        parseAssignment();
        break;
      default:
        jj_la1[3] = jj_gen;
        jj_consume_token(-1);
        throw new ParseException();
      }
    }
    jj_consume_token(RETURN);
    parseExpression();
    jj_consume_token(0);
  }

  final private void parseProgram() throws ParseException {
  String binary ;
    jj_consume_token(OPENING_BOX);
    label_4:
    while (true) {
      jj_consume_token(OPENING_PAR);
      parseString();
      jj_consume_token(COMMA);
      // binary code
          binary = parseString();
      lineBinaries.add(binary) ;
      label_5:
      while (true) {
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case COMMA:
          ;
          break;
        default:
          jj_la1[4] = jj_gen;
          break label_5;
        }
        jj_consume_token(COMMA);
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case INTEGER:
          parseInteger();
          break;
        case STRING:
          parseString();
          break;
        default:
          jj_la1[5] = jj_gen;
          jj_consume_token(-1);
          throw new ParseException();
        }
      }
      jj_consume_token(CLOSING_PAR);
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case COMMA:
        jj_consume_token(COMMA);
        break;
      default:
        jj_la1[6] = jj_gen;
        ;
      }
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case OPENING_PAR:
        ;
        break;
      default:
        jj_la1[7] = jj_gen;
        break label_4;
      }
    }
    jj_consume_token(CLOSING_BOX);
  }

  final private void parseAssignment() throws ParseException {
  String name ;
  String value ;
    name = jj_consume_token(IDENTIFIER).image;
    jj_consume_token(EQUAL);
    value = parseExpression();
    values.put(name,value) ;
  }

  final private void parseComment() throws ParseException {
  Token token ;
  String comment ;
    token = jj_consume_token(COMMENT);
    comment = token.image ;
    comment = comment.substring(1).trim() ;

    if      (comment.startsWith(VARIABLE_ORDER_PREFIX)) {
      comment = comment.substring(VARIABLE_ORDER_PREFIX.length()) ;
      variableOrder = comment.trim() ;
    }

    else if (comment.startsWith(MEMBER_ORDER_PREFIX)) {
      comment = comment.substring(MEMBER_ORDER_PREFIX.length()) ;
      memberOrder = comment.trim() ;
    }
  }

  final private String parseExpression() throws ParseException {
  Token token ;
  String string ;
  StringBuffer buffer = new StringBuffer() ;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case INTEGER:
      token = jj_consume_token(INTEGER);
      {if (true) return token.image ;}
      break;
    case STRING:
      string = parseString();
      {if (true) return string ;}
      break;
    case IDENTIFIER:
      string = parseIdentifier();
      {if (true) return string ;}
      break;
    case OPENING_PAR:
      jj_consume_token(OPENING_PAR);
      string = parseExpression();
      buffer.append('(') ;
      buffer.append(string) ;
      label_6:
      while (true) {
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case COMMA:
          ;
          break;
        default:
          jj_la1[8] = jj_gen;
          break label_6;
        }
        jj_consume_token(COMMA);
        string = parseExpression();
        buffer.append(',') ;
        buffer.append(string) ;
      }
      jj_consume_token(CLOSING_PAR);
      buffer.append(')') ;
      {if (true) return buffer.toString() ;}
      break;
    default:
      jj_la1[9] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
    throw new Error("Missing return statement in function");
  }

  final private String parseIdentifier() throws ParseException {
  Token token ;
    token = jj_consume_token(IDENTIFIER);
    {if (true) return token.image ;}
    throw new Error("Missing return statement in function");
  }

  final private int parseInteger() throws ParseException {
  Token token ;
    token = jj_consume_token(INTEGER);
    {if (true) return Integer.parseInt(token.image) ;}
    throw new Error("Missing return statement in function");
  }

  final private String parseString() throws ParseException {
  Token token ;
    token = jj_consume_token(STRING);
    int length = token.image.length() ;
    {if (true) return token.image.substring(1,length-1) ;}
    throw new Error("Missing return statement in function");
  }

  public ProgramParserTokenManager token_source;
  SimpleCharStream jj_input_stream;
  public Token token, jj_nt;
  private int jj_ntk;
  private int jj_gen;
  final private int[] jj_la1 = new int[10];
  static private int[] jj_la1_0;
  static {
      jj_la1_0();
   }
   private static void jj_la1_0() {
      jj_la1_0 = new int[] {0x200000,0x200000,0x220000,0x220000,0x40,0x50000,0x40,0x200,0x40,0x70200,};
   }

  public ProgramParser(java.io.InputStream stream) {
     this(stream, null);
  }
  public ProgramParser(java.io.InputStream stream, String encoding) {
    try { jj_input_stream = new SimpleCharStream(stream, encoding, 1, 1); } catch(java.io.UnsupportedEncodingException e) { throw new RuntimeException(e); }
    token_source = new ProgramParserTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 10; i++) jj_la1[i] = -1;
  }

  public void ReInit(java.io.InputStream stream) {
     ReInit(stream, null);
  }
  public void ReInit(java.io.InputStream stream, String encoding) {
    try { jj_input_stream.ReInit(stream, encoding, 1, 1); } catch(java.io.UnsupportedEncodingException e) { throw new RuntimeException(e); }
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 10; i++) jj_la1[i] = -1;
  }

  public ProgramParser(java.io.Reader stream) {
    jj_input_stream = new SimpleCharStream(stream, 1, 1);
    token_source = new ProgramParserTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 10; i++) jj_la1[i] = -1;
  }

  public void ReInit(java.io.Reader stream) {
    jj_input_stream.ReInit(stream, 1, 1);
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 10; i++) jj_la1[i] = -1;
  }

  public ProgramParser(ProgramParserTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 10; i++) jj_la1[i] = -1;
  }

  public void ReInit(ProgramParserTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 10; i++) jj_la1[i] = -1;
  }

  final private Token jj_consume_token(int kind) throws ParseException {
    Token oldToken;
    if ((oldToken = token).next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    if (token.kind == kind) {
      jj_gen++;
      return token;
    }
    token = oldToken;
    jj_kind = kind;
    throw generateParseException();
  }

  final public Token getNextToken() {
    if (token.next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    jj_gen++;
    return token;
  }

  final public Token getToken(int index) {
    Token t = token;
    for (int i = 0; i < index; i++) {
      if (t.next != null) t = t.next;
      else t = t.next = token_source.getNextToken();
    }
    return t;
  }

  final private int jj_ntk() {
    if ((jj_nt=token.next) == null)
      return (jj_ntk = (token.next=token_source.getNextToken()).kind);
    else
      return (jj_ntk = jj_nt.kind);
  }

  private java.util.Vector jj_expentries = new java.util.Vector();
  private int[] jj_expentry;
  private int jj_kind = -1;

  public ParseException generateParseException() {
    jj_expentries.removeAllElements();
    boolean[] la1tokens = new boolean[22];
    for (int i = 0; i < 22; i++) {
      la1tokens[i] = false;
    }
    if (jj_kind >= 0) {
      la1tokens[jj_kind] = true;
      jj_kind = -1;
    }
    for (int i = 0; i < 10; i++) {
      if (jj_la1[i] == jj_gen) {
        for (int j = 0; j < 32; j++) {
          if ((jj_la1_0[i] & (1<<j)) != 0) {
            la1tokens[j] = true;
          }
        }
      }
    }
    for (int i = 0; i < 22; i++) {
      if (la1tokens[i]) {
        jj_expentry = new int[1];
        jj_expentry[0] = i;
        jj_expentries.addElement(jj_expentry);
      }
    }
    int[][] exptokseq = new int[jj_expentries.size()][];
    for (int i = 0; i < jj_expentries.size(); i++) {
      exptokseq[i] = (int[])jj_expentries.elementAt(i);
    }
    return new ParseException(token, exptokseq, tokenImage);
  }

  final public void enable_tracing() {
  }

  final public void disable_tracing() {
  }

}
