/*
 * TransducerRule.java
 */
package automata ;

/**
 * An automata rule with output.
 * 
 * @author Pierre Salagnac
 */
public class TransducerRule extends Rule {

  // output state
  private String output ;
  
  
  /**
   * Creates a new instance.
   * 
   * @param input The binary input.
   * @param output The binary output.
   * @param leftState The left state of the automata.
   * @param rightState The right state of the automata.
   * @param state The output state.
   */
  public TransducerRule(String input,String output,String leftState,
      String rightState,String state) {
    
    super(input,leftState,rightState,state) ;
    
    this.output = output ;
    
  }
  
  
  /**
   * Returns the output of the transducer rule.
   * 
   * @return The binary output.
   */
  public String getOutput() {
    
    return output ;
    
  }
  
  
  /**
   * Returns a string describing this rule.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    buffer.append("left=") ;
    buffer.append(getLeftState()) ;
    
    buffer.append(",right=") ;
    buffer.append(getRightState()) ;
    
    buffer.append(",input=") ;
    buffer.append(getInput()) ;
    
    buffer.append(",output=") ;
    buffer.append(output) ;
    
    buffer.append(",state=") ;
    buffer.append(getState()) ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }
  
}
