/*
 * Bdd.java
 */
package bdd;

import java.util.* ;


/**
 * The complete BDD.
 * 
 * @author Pierre Salagnac
 */
public class Bdd {

  // all nodes
  private List<BddNode> nodes ;
  
  // all trees
  private List<BddTree> trees ;
  
  
  /**
   * Create a new instance.
   * 
   * @param trees All bdd trees.
   * @param nodes Some nodes that not in the trees.
   */
  Bdd(BddTree[] trees,BddNode[] nodes) {
    
    this.nodes = new ArrayList<BddNode>() ;
    this.trees = new ArrayList<BddTree>() ;
    
    
    for (BddTree tree : trees) {
      this.trees.add(tree) ;
      addNode(tree) ;
    }
    
    for (BddNode node : nodes)
      addNode(node) ;
    
  }
  
  
  /**
   * Return the max depth of the BDD.
   */
  public int getMaxBitIndex() {
    
    int max = 0 ;
    
    for (BddTree tree : trees)
      max = Math.max(max,tree.getMaxBitIndex()) ;
    
    for (BddNode node : nodes)
      max = Math.max(max,node.getMaxBitIndex()) ;
    
    
    return max ;
  }
  
  
  /**
   * Build the array with all nodes.
   * 
   * @param node The node to add.
   */
  private void addNode(BddNode node) {
    
    if (nodes.contains(node))
      return ;
    
    nodes.add(node) ;
    
    BddNode trueNode = node.getChild(true) ;
    if (trueNode != null)
      addNode(trueNode) ;
    
    BddNode falseNode = node.getChild(false) ;
    if (falseNode != null)
      addNode(falseNode) ;
  }
  
  
  /**
   * Return the nomber of node.
   * 
   * @return The number of nodes.
   */
  public int getNodeCount() {
    
    return nodes.size() ;
    
  }
  
  
  /**
   * Return the tree with the specified input states if exists.
   * 
   * @param left The left state to search.
   * @param right The right state to search.
   * 
   * @return The tree with these inputs, or null.
   */
  public BddTree getTree(String left,String right) {
    
    for (BddTree tree : trees) {
      
      if (left.equals(tree.getLeftState())
        &&right.equals(tree.getRightState())) {
        return tree ;
      }
    }
    
    return null ;
    
  }
  
  
  /**
   * Return the node with the specified index.
   * 
   * @param index The index node to seach.
   * 
   * @return The node with this index, or null.
   */
  public BddNode getNode(int index) {
    
    for (BddNode node : nodes) {
      if (node.getIndex() == index)
        return node ;
    }
    
    return null ;
  }
  

}
