/*
 * BddTree.java
 */
package bdd ;


/**
 * BDD tree. A tree is a root instance of BddNode.
 * 
 * @author Pierre Salagnac
 */
public class BddTree extends BddNode {

  // initial states
  private String leftState ;
  private String rightState ;
  
  
  /**
   * Create a new instance.
   * 
   * @param leftState The left input state.
   * @param rightState The right input state.
   */
  BddTree(String leftState,String rightState)  {
    
    // root looks at the first bit.
    super(0) ;
    
    this.leftState  = leftState ;
    this.rightState = rightState ;
    
  }
  
  
  /**
   * Return the left input state.
   * 
   * @return The left state.
   */
  public String getLeftState() {
    return leftState ;
  }
  
  
  /**
   * Return the right input state.
   * 
   * @return The right state.
   */
  public String getRightState() {
    return rightState ;
  }

  
////////////////////////////////////////////////////////////////////////////////
//                               EXPAND                                       //
////////////////////////////////////////////////////////////////////////////////


  /**
   * Expand the BDD node to have real binary tree.
   * 
   * @param nextBit The next available bit index.
   * 
   * @exception BddException We are unable to expand the nodes.
   */
  void expand(int nextBit) throws BddException {

    expandNotLeaf(nextBit) ;

    expandMultiple(nextBit) ;

  }
  
}
