/*
 * TypedefBddBuilder.java
 */
package bdd;

import automata.* ;


/**
 * Creates a BDD for the typedef rules.
 * 
 * @author Pierre Salagnac
 */
public class TypedefBddBuilder extends BddBuilder<Rule>{
  
  
  /**
   * Create a new instance.
   * 
   * @param nodeWidth The max length of a node binary.
   */
  public TypedefBddBuilder(int nodeWidth) {
    
    super(nodeWidth) ;
    
  }
  
  
  /**
   * Build the BDD nodes.
   * 
   * @param rules The automata rules.
   */
  void buildNodes(Rule[] rules) throws BddException {
    
    int nodeWidth = getNodeWidth() ;
    StateIndexer states = getStateIndexer() ;
    
    // error node
    BddNode error = new BddNode(0) ;
    error.addState(states.getErrorState()) ;
    addNode(error) ;
    
    for (Rule rule : rules) {
    
      String left   = rule.getLeftState() ;
      String right  = rule.getRightState() ; 
      String binary = rule.getInput() ;
      String state  = rule.getState() ;
    
      if (binary.length() > nodeWidth)
        throw new BddException("rule input too large, max:"+nodeWidth+
            ", found:"+binary.length()+"  "+rule) ;
      
      BddTree tree  = getTree(left,right) ;

      
      int stateIndex = states.getStateIndex(state) ;
      buildNodes(tree,binary,stateIndex) ;
    }
  }
  
  
  /**
   * Build nodes accepting a binary input.
   * 
   * @param node The root node.
   * @param binary The binary input.
   * @param state The output state.
   */
  private void buildNodes(BddNode node,String binary,int state) {
    
    // end
    if (node.getBitIndex()/2 == binary.length()) {
      node.addState(state) ;
    }
    
    // in the bdd
    else {
      char c = binary.charAt(node.getBitIndex()/2) ;
      boolean side = (c == '1') ;
      
      BddNode child = node.getChild(side) ;
      if (child == null) {
        child = new BddNode(node.getBitIndex()+2) ;
        node.setChild(side,child) ;
      }
      
      buildNodes(child,binary,state) ;
    }
    
  }
  
}
