/*
 * Move.java
 */
package parser;

/**
 * Atomic move in a path.
 * 
 * @author Pierre Salagnac
 */
public class Move extends Path {

  // primitive move
  private String name ;
  private boolean reverse ;
  
  
  /**
   * Creates a new instance.
   * 
   * @param start The starting point, or null if the move is relative.
   * @param name The primitive move to use.
   * @param reverse Use the opposite direction of the primitive move.
   * @param multiple true if the move is a multiple one.
   * @param end The ending point, or null.
   */
  public Move(String start,String name,boolean reverse,boolean multiple,String end) {
    
    this.name     = name ;
    this.reverse  = reverse ;
    
    setStartPoint(start) ;
    setEndPoint(end) ;
    setMultiple(multiple) ;
    
  }
  
  
  /**
   * Returns the name of the primitive move.
   * 
   * @return The primitive name.
   */
  public String getName() {
    return name ;
  }
  
  
  /**
   * Return the side of the primitive link.
   * 
   * @return true if the primitive move is reversed.
   */
  public boolean isReversed() {
    return reverse ;
  }
  
  
  /**
   * Clones this single move.
   * 
   * @return An equvalent single move.
   */
  @Override
  public Move clone() {
    
    return new Move(
        getStartPoint(),
        name,
        reverse,
        isMultiple(),
        getEndPoint()
    ) ;
    
  }
  
  
  /**
   * Returns a string describing this path.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    buffer.append("name=") ;
    buffer.append(name) ;
    
    if (reverse)
      buffer.append(",reverse") ;
    
    if (isMultiple())
      buffer.append(",multiple") ;
    
    if (getStartPoint() != null) {
      buffer.append(",start=") ;
      buffer.append(getStartPoint()) ;
    }
    
    if (getEndPoint() != null) {
      buffer.append(",end=") ;
      buffer.append(getEndPoint()) ;
    }
    
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }
  
  
}
