/*
 * Typedef.java 
 */
package parser;

import java.util.* ;


/**
 * This class is the result of a typedef parse. It contais all the sub-elements
 * of the delcaration.
 * 
 * @author Pierre Salagnac
 */
public class Typedef {
  
  // name of the type
  private String name ;
  
  // contents
  private List<MemberList> contents ;
  
  // root variables
  private List<String> rootVariables ;
  
  
  /**
   * Create a new instance with the specified name.
   * 
   * @param name The name of the type.
   */
  Typedef(String name) {
    
    this.name = name ;
    
    contents      = new ArrayList<MemberList>() ;
    rootVariables = new ArrayList<String>() ;
  }
  
  
  /**
   * Return the name of the type.
   * 
   * @return The type name.
   */
  public String getName() {
    return name ;
  }
  
  
  /**
   * Return the names of variable pointing roots.
   * 
   * @return Variables pointing to the roots.
   */
  public String[] getRootVariables() {
    
    String[] array = new String[rootVariables.size()] ;
    rootVariables.toArray(array) ;
    
    return array ;
  }
  
  
  /**
   * Return the differents contents of the structure.
   */
  public MemberList[] getContents() {
    
    MemberList[] array = new MemberList[contents.size()] ;
    
    contents.toArray(array) ;
    
    return array ;
    
  }
  
  
  /**
   * Add a variable, pointing to the root of this type.
   * 
   * @param variable The variable name.
   */
  void addRootVariable(String variable) {
    
    rootVariables.add(variable) ;
    
  }
  
  
  /**
   * Add a new content definition to the type.
   * 
   * @param content The content of the structure.
   */
  public void addContent(MemberList content) {
    
    contents.add(content) ;
    
  }
  
  
  
  
  /**
   * Return a string describing this typedef.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    buffer.append("name=") ;
    buffer.append(name) ;
    buffer.append(',') ;
    
    buffer.append("root=") ;
    buffer.append('{') ;
    for (int i=0;i<rootVariables.size();i++) {
      if (i != 0)
        buffer.append(',') ;
      buffer.append(rootVariables.get(i)) ;
    }
    buffer.append('}') ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }
  
}
