/*
 * Input.java
 */
package tree ;


/**
 * An input of a tree pattern.
 * 
 * @author Pierre Salagnac
 */
public class Input {

  // data type
  private String type ;
  
  // name of corresponding member
  private String name ;
  
  // route
  private int route ;
  private boolean primitive ;
  
  // this is a null input
  private boolean null_ ;
  
  // this is an undifined input
  private boolean undefined ;
  
  
  /**
   * Private constructor.
   * 
   * @param name The member name.
   */
  private Input(String name) {
    
    this.name = name ;
    
  }
  
  
  /**
   * Return the name of this input.
   * 
   * @return The member name.
   */
  public String getName() {
    return name ;
  }
  
  
  /**
   * Return the data type of this input.
   * 
   * @return The input type, ou null if the input is undefined.
   */
  public String getType() {
    return type ;
  }
  
  /**
   * Return if the associated move is primitive.
   * 
   * @return true if the input is primitive.
   */
  public boolean isPrimitive() {
    return primitive ;
  }
  
  
  /**
   * Return if the input is null.
   * 
   * @return true if the input is null.
   */
  public boolean isNull() {
    return null_ ;
  }
  
  
  /**
   * Return if the input is undefined.
   * 
   * @return true if the input is undefined.
   */
  public boolean isUndefined() {
    return undefined ;
  }
  
  
  /**
   * Return the index of the route in this input.
   * 
   * @return The route index, or -1 is the input has no route.
   */
  public int getRoute() {
    return route ;
  }
  
  
////////////////////////////////////////////////////////////////////////////////
//                              FACTORY                                       //
////////////////////////////////////////////////////////////////////////////////
  
  /**
   * Create a new input for a primitive member.
   * 
   * @param name The member name.
   * @param type The data type of the input.
   * @param route The route index.
   */
  public static Input createPrimitiveInput(String name,String type,int route) {
    
    Input input = new Input(name) ;
    
    input.type      = type ;
    input.route     = route ;
    input.primitive = true ;
    
    return input ;
  }
  
  
  /**
   * Create a new input for a routed member.
   * 
   * @param name The member name.
   * @param type The data type of the input.
   * @param route The route index.
   */
  public static Input createRoutedInput(String name,String type,int route) {
    
    Input input = new Input(name) ;
    
    input.type  = type ;
    input.route = route ;
    
    return input ;
  }
  
  
  /**
   * Create a new instance for a null input.
   * 
   * @param name The member name.
   * @param type The data type of the input.
   */
  public static Input createNullInput(String name,String type) {
    
    Input input = new Input(name) ;
    
    input.type  = type ;
    input.null_ = true ;
    input.route = -1 ;
    
    return input ;
  }
  
  
  /**
   * Create a new instance for an undefined input.
   * 
   * @param name The input name.
   */
  public static Input createUndefinedInput(String name) {
    
    Input input = new Input(name) ;
    
    input.undefined = true ;
    input.route     = -1 ;
    
    return input ;
  }
  
}
