/*
 * Route.java
 */
package tree;

import java.util.* ;

import parser.* ;


/**
 * A route is a path with F marks at the end.
 * 
 * @author Pierre Salagnac
 */
public class Route {

  private String name ;
  
  private Path path ;
  
  // F marks set
  private List<Integer> marks ;
  
  
  /**
   * Create a new instance.
   * 
   * @param name Name of an associated member. Give the name of the route
   * for a primitive input.
   * @param path The route move.
   */
  public Route(String name,Path path) {
    
    this.name = name ;
    this.path = path ;
    
    marks = new ArrayList<Integer>() ;
  }
  
  
  /**
   * Return the name of the route.
   */
  String getName() {
    return name ;
  }
  
  
  /**
   * Return the path of this route.
   * 
   * @return The corresponding path.
   */
  public Path getPath() {
    return path ;
  }
  
  
  /**
   * Add a F mark at the end of the route.
   * 
   * @param mark The mark id.
   */
  void addMark(int mark) {
    
    Integer integer = Integer.valueOf(mark) ;
    
    if (!marks.contains(integer))
      marks.add(integer) ;
    
  }
  
  
  /**
   * Returns if the specified F mark is at the end of the route.
   * 
   * @param mark The F mark to search.
   * 
   * @return true if the route has the mark.
   */
  public boolean containsMark(int mark) {
    
    Integer integer = Integer.valueOf(mark) ;
    
    return marks.contains(integer) ;
    
  }
  
  
  /**
   * Return a string describing this route.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    buffer.append("name=") ;
    buffer.append(name) ;
    
    buffer.append(",marks={") ;
    for (int i=0;i<marks.size();i++) {
      if (i != 0)
        buffer.append(',') ;
      buffer.append(marks.get(i)) ;
    }
    buffer.append('}') ;
    
    buffer.append(",path=") ;
    buffer.append(path) ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }
  
}
