/*
 * RouteIndexer.java
 */
package tree;

import java.util.* ;


/**
 * Reference all moves in the patterns.
 * 
 * @author Pierre Salagnac
 */
public class RouteIndexer {

  // all the routes
  private List<Route> routes ;
  
  
  /**
   * Create a new instance.
   */
  public RouteIndexer() {
    
    routes = new ArrayList<Route>() ;
    
  }
  
  
  /**
   * Return a new index that can be used with a new route.
   * 
   * @return An available route index.
   */
  int getAvailableIndex() {
    return routes.size() ;
  }
  
  
  /**
   * Add a new route to the list.
   * 
   * @param index The route identifier in the indexer.
   * @param route The route to add. 
   */
  void setRoute(int index,Route route) {
    
    routes.add(route) ;
    
  }
  
  
  /**
   * Return a route.
   * 
   * @param index The route index.
   * 
   * @return The specidied route.
   */
  public Route getRoute(int index) {
    
    return routes.get(index) ;
    
  }
  
  
  /**
   * Return the number of referenced routes, including primitive routes.
   * 
   * @return The number of routes.
   */
  public int getRouteCount() {
    
    return routes.size() ;
    
  }
  
  
  /**
   * Return the index of a primitive route.
   * 
   * @param name The name of the primitive member.
   * 
   * @return The route index, or -1 if the name is not found.
   */
  int getPrimitiveIndex(String name) {
    
    for (int i=0;i<routes.size();i++) {
      
      Route route = routes.get(i) ;
      
      if (route.getName().equals(name))
        return i ;
    }
    
    return -1 ;
  }
  
}
