/*
 * TestRuleBuilder.java
 */
package automata ;

import junit.framework.* ;

import tree.* ;
import parser.* ;


/**
 * Test of automata rules creation. 
 * 
 * @author Pierre Salagnac
 */
public class TestRuleBuilder extends TestCase {

  
  /**
   * Build rules for an unique pattern.
   */
  public void testUnique() throws Exception {
    
    String output  = "output" ;
    String[] roots = new String[]{"x"} ;
    int[] marks    = new int[]{} ;
    
    Input i0 = Input.createNullInput("nullName","nullType") ;
    Input i1 = Input.createUndefinedInput("undefinedName") ;
    Input i2 = Input.createPrimitiveInput("primitiveName","primitiveType",1) ;
    Input i3 = Input.createRoutedInput("routedName","routedType",2) ;
    
    Pattern pattern = new Pattern(output,4,"101",roots) ;
    pattern.setInput(0,i0) ;
    pattern.setInput(1,i1) ;
    pattern.setInput(2,i2) ;
    pattern.setInput(3,i3) ;
    
    RouteIndexer indexer = new RouteIndexer() ;
    ProgramParser program = new MockProgramParser(10,2,"0101","left:0","x:1") ;
    RuleBuilder builder = new RuleBuilder(indexer,program) ;
    
    builder.build(new Pattern[]{pattern}) ;
    Rule[] rules = builder.getRules() ;
    
    for (Rule rule : rules)
      System.out.println(rule) ;
  }
  
  
  /**
   * Build rules for a pattern with no root.
   */
  public void testNoRoot() throws Exception {
    
    String output  = "output" ;
    int[] marks    = new int[]{} ;
    
    Input i0 = Input.createNullInput("nullName","nullType") ;
    Input i1 = Input.createUndefinedInput("undefinedName") ;
    Input i2 = Input.createPrimitiveInput("primitiveName","primitiveType",1) ;
    Input i3 = Input.createRoutedInput("routedName","routedType",2) ;
    
    Pattern pattern = new Pattern(output,4,"0101",null) ;
    pattern.setInput(0,i0) ;
    pattern.setInput(1,i1) ;
    pattern.setInput(2,i2) ;
    pattern.setInput(3,i3) ;
    
    RouteIndexer indexer = new RouteIndexer() ;
    ProgramParser program = new MockProgramParser(10,2,"0101","left:0","x:1") ;
    RuleBuilder builder = new RuleBuilder(indexer,program) ;
    
    builder.build(new Pattern[]{pattern}) ;
    Rule[] rules = builder.getRules() ;
    
    for (Rule rule : rules)
      System.out.println(rule) ;
    
  }
  
}
