/*
 * TestMonaBuilder.java
 */
package mona;

import java.io.* ;

import bdd.* ;
import tree.* ;
import parser.* ;
import automata.* ;

import junit.framework.* ;


/**
 * @author Pierre Salagnac
 */
public class TestMonaBuilder extends TestCase {
  
  
  /**
   * Generates a BDD with an unique rule.
   */
  public void testUnique() throws Exception {
    
    Rule[] rules = {new Rule("0101",Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.ACCEPT_STATE)} ;
    
    TypedefBddBuilder builder = new TypedefBddBuilder(100) ;
    builder.build(rules) ;
    
    Bdd bdd = builder.getBdd() ;
    StateIndexer indexer = builder.getStateIndexer() ;
    MonaBuilder mona = new MonaBuilder(indexer) ;
    
//    mona.writeBdd(bdd,new OutputStreamWriter(System.out)) ;
    
//    FileWriter writer = new FileWriter("../test.gta") ;
//    mona.writeBdd(bdd,writer) ;
//    writer.close() ;
    
  }
  
  /**
   * Tests a whole generation.
   * 
   * The Mona file is written in the standard output.
   */
  public void testAll() throws Exception {
    
    String inputs    = "left:0 right:1 up:2" ;
    String variables = "x:1, y:2" ;
    
    File file = new File(System.getProperty("syntax"),"all") ;
    FileReader reader = new FileReader(file) ;
    
    ProgramParser program = new MockProgramParser(30,10,"0110101",inputs,variables) ;
    program.parse() ;
    
    TypedefParser parser = new TypedefParser(reader) ;
    parser.parse() ;
    reader.close() ;
    
    Typedef[] typedefs = parser.getTypedefs() ;
    
    PatternBuilder patternBuilder = new PatternBuilder(program) ;
    patternBuilder.build(typedefs,true) ;
    
    Pattern[] patterns = patternBuilder.getPatterns() ;
    RouteIndexer routeIndexer = patternBuilder.getRouteIndexer() ;
    
    RuleBuilder ruleBuilder = new RuleBuilder(routeIndexer,program) ;
    ruleBuilder.build(patterns) ;
    Rule[] rules = ruleBuilder.getRules() ;
    
    TypedefBddBuilder bddBuilder = new TypedefBddBuilder(30) ;
    bddBuilder.build(rules) ;
    Bdd bdd = bddBuilder.getBdd() ;
    StateIndexer states = bddBuilder.getStateIndexer() ;
    
    MonaBuilder monaBuilder = new MonaBuilder(states) ;
    //monaBuilder.writeBdd(bdd,new OutputStreamWriter(System.out)) ;
    
    FileWriter writer = new FileWriter("../test.gta") ;
    monaBuilder.writeBdd(bdd,writer) ;
    writer.close() ;
  }
  
  
  /**
   * Tests the generation of a transducer.
   */
  public void testTransducer() throws Exception {
    
    Path move = Path.createPrimitivePath("move") ;
    Route route = new Route("move",move) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","move:0","y:1") ; 
    TransducerBuilder builder = new TransducerBuilder(program) ;
    
    builder.build(route) ;
    TransducerRule[] rules = builder.getRules() ;
    
    TransducerBddBuilder bddBuilder = new TransducerBddBuilder(10) ;
    bddBuilder.build(rules) ;
    StateIndexer states = bddBuilder.getStateIndexer() ;
    Bdd bdd = bddBuilder.getBdd() ;
    
    MonaBuilder mona = new MonaBuilder(states) ;
    
//    mona.writeBdd(bdd,new OutputStreamWriter(System.out)) ;
   
    FileWriter writer = new FileWriter("../transd.gta") ;
    mona.writeBdd(bdd,writer) ;
    writer.close() ;
  }
  
}
