/*
 * MockProgramParser.java
 */
package parser;

import java.io.* ;


/**
 * Simulate a program parser.
 * 
 * @author Pierre Salagnac
 */
public class MockProgramParser extends ProgramParser {
  
  private int nodeWidth ;
  private int descCount ;
  
  private String binary ;
  
  private OrderParser memberParser ;
  private OrderParser variableParser ;
  
  
  /**
   * Create a new instance.
   */
  public MockProgramParser(int nodeWidth,int descCount,String binary,String members,
      String variables)
  throws ParseException {
    
    super((Reader)null) ;
    
    this.nodeWidth = nodeWidth ;
    this.descCount = descCount ;
    this.binary    = binary ;
    
    memberParser   = new OrderParser(new StringReader(members)) ;
    variableParser = new OrderParser(new StringReader(variables)) ;
    
    memberParser.parse() ;
    variableParser.parse(-1) ;
  }
  
  
  /**
   * Do nothing.
   */
  @Override
  public void parse() {
    
  }
  
  
  /**
   * Return the binary code of the specified program line.
   */
  @Override
  public String getLineBinary(int index) {

    if (index == 0)
      return binary ;
    else
      return null ;

  }


  /**
   * Return the maximum with alowed for a tree node.
   */
  @Override
  public int getNodeWidth() {
    return nodeWidth ;
  }


  /**
   * Return the number of F marks that must be put in type 3 nodes.
   */
  @Override
  public int getDescriptorCount() {
    return descCount ;
  }


  /**
   * Return the number of variables declaraed in the program.
   */
  @Override
  public int getVariableCount() {
    return variableParser.getInputCount() ;
  }
  
  
  /**
   * Return the number of member declared in the program.
   */
   @Override
  public int getMemberCount() {
    return memberParser.getInputCount() ;
  }
  
  
  /**
   * Return the index of the specified variable.
   */
  @Override
  public int getVariableIndex(String name) {
    return variableParser.getInputIndex(name) ;
  }
  
  
  /**
   * Return the index of the specified member.
   */
  @Override
  public int getMemberIndex(String name) {
    return memberParser.getInputIndex(name) ;
  }
  
  /**
   * Return the specified variable.
   */
  @Override
  public String getVariable(int index) {
    return variableParser.getInput(index) ;
  }
  
  
  /**
   * Return the specified member name.
   */
  @Override
  public String getMember(int index) {
    return memberParser.getInput(index) ;
  }
  
  /**
   * Return if the specified variable is declared or not.
   */
  @Override
  public boolean containsVariable(String name) {
    return variableParser.containsInput(name) ;
  }


  /**
   * Return if the specified member is declared.
   */
  @Override
  public boolean containsMember(String name) {
    return memberParser.containsInput(name) ;
  }

  
}
