/*
 * TestProgramParser.java
 */
package parser;

import java.io.* ;

import junit.framework.* ;


/**
 * Test parser of the "get_program" function.
 * 
 * @author Pierre Salagnac
 */
public class TestProgramParser extends TestCase {

  
  /**
   * Parse an example.
   */
  public void testParse1() throws Exception {
    
    File file = new File(System.getProperty("program"),"example1.py") ;
    FileReader reader = new FileReader(file) ;
    
    ProgramParser parser = new ProgramParser(reader) ;
    parser.parse() ;
    reader.close();
    
    assertEquals(9,parser.getNodeWidth()) ;
    assertEquals(3,parser.getDescriptorCount()) ;
    assertEquals("00001",parser.getLineBinary(0)) ;
    
    // order
    String memberOrder   = parser.getMemberOrder() ;
    String variableOrder = parser.getVariableOrder() ;
    
    OrderParser memberParser   = new OrderParser(new StringReader(memberOrder)) ;
    OrderParser variableParser = new OrderParser(new StringReader(variableOrder)) ; 
    memberParser.parse() ;
    variableParser.parse() ;
    
    assertEquals(1,memberParser.getInputCount()) ;
    assertEquals(0,memberParser.getInputIndex("s")) ;
    
    assertEquals(3,variableParser.getInputCount()) ;
    assertEquals(1,variableParser.getInputIndex("p")) ;
    assertEquals(2,variableParser.getInputIndex("pp")) ;
    assertEquals(3,variableParser.getInputIndex("tmp1")) ;
  }
  
  
  /**
   * Parse an example.
   */
  public void testParse2() throws Exception {
    
    File file = new File(System.getProperty("program"),"example2.py") ;
    FileReader reader = new FileReader(file) ;
    
    ProgramParser parser = new ProgramParser(reader) ;
    parser.parse() ;
    reader.close() ;
    
    assertEquals(12,parser.getNodeWidth()) ;
    assertEquals(3,parser.getDescriptorCount()) ;
    
    assertEquals("0000001",parser.getLineBinary(0)) ;
    
    // order
    String memberOrder   = parser.getMemberOrder() ;
    String variableOrder = parser.getVariableOrder() ;
    
    OrderParser memberParser   = new OrderParser(new StringReader(memberOrder)) ;
    OrderParser variableParser = new OrderParser(new StringReader(variableOrder)) ; 
    memberParser.parse() ;
    variableParser.parse() ;
    
    assertEquals(2,memberParser.getInputCount()) ;
    assertEquals(0,memberParser.getInputIndex("s")) ;
    assertEquals(1,memberParser.getInputIndex("ss")) ;
    
    assertEquals(3,variableParser.getInputCount()) ;
    assertEquals(1,variableParser.getInputIndex("p")) ;
    assertEquals(2,variableParser.getInputIndex("pp")) ;
    assertEquals(3,variableParser.getInputIndex("tmp1")) ;
  }
  
  
}
