package cril.calife.export.engine.averiles.ARTMC;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import cril.calife.utils.DOMUtil;

/**
 * Cette classe transforme un automate en un automate quivalent ne possdant plus de combinaisons
 * boolennes sur ses gardes
 * 
 * 
 * @author roch.lescuyer
 * 19 septembre 2007
 * 
 */
public class DeployGuards {

	private Element automate;

	private Element[] guards;

	private DeployGuardsUtil nodeBuilder;

	private static String new_place_name = "add";
	private int add_place;

	DeployGuards(Document dom, Element automate, int add_place) {
		this.automate = automate;
		this.add_place = add_place;
		guards = getGuards();
		nodeBuilder = new DeployGuardsUtil(dom);
		run();
	}

	/**
	 * Retourne l'automate transform
	 * @return
	 */
	public Element getAutomate() {
		return automate;
	}

	/**
	 * Retourne le nombre de places qui ont t ajoutes
	 * @return
	 */
	public int getAddPlace() {
		return add_place;
	}

	private void run() {
		while( ! fixPoint() ) {
			for( int i = 0; i < guards.length; i++ ) {
				Element constraint = DOMUtil.getElementByTagName(guards[i], "Constraint")[0];
				if( isDoubleNeg(constraint) ) {
					guards[i].replaceChild(elimDoubleNeg(constraint), constraint);
				}
				if( isNeg(constraint) ) {
					Element main = DOMUtil.getElementByTagName(constraint, "FunT")[0];
					if( ! isNegatomique(main) ) {
						guards[i].replaceChild(elimNeg(constraint), constraint);
					}
				}
				if( isAnd(constraint) ) {
 					Element transition = (Element) guards[i].getParentNode();
					String source = transition.getAttribute("Source");
					String target = transition.getAttribute("Target");
					Element updates = DOMUtil.getElementByTagName(transition, "Updates")[0];
					automate.removeChild(transition);
					//
					Element A = getFirstArg((Element)constraint.cloneNode(true));
					Element B = getSecondArg((Element)constraint.cloneNode(true));
					Element gardeA = nodeBuilder.freshGuard(A);
					Element gardeB = nodeBuilder.freshGuard(B);
					Element gardeNotA = nodeBuilder.freshGuard(not((Element) A.cloneNode(true)));
					Element gardeNotB = nodeBuilder.freshGuard(not((Element) B.cloneNode(true)));
					String id1 = getNewPlaceId(add_place++);
					String id2 = getNewPlaceId(add_place++);
					Element place1 = nodeBuilder.freshPlace(id1);
					Element place2 = nodeBuilder.freshPlace(id2);
					Element transition1 = nodeBuilder.freshTransition(source, id1, gardeA, nodeBuilder.freshUpdates());
					Element transition2 = nodeBuilder.freshTransition(id1, target, gardeB, updates);
					Element transition3 = nodeBuilder.freshTransition(source, id2, gardeNotA, nodeBuilder.freshUpdates());
					Element transition4 = nodeBuilder.freshTransition(id1, id2, gardeNotB, nodeBuilder.freshUpdates());
					//
					Element[] transitions = DOMUtil.getElementByTagName(automate, "Transition");
					for(int j = 0; j < transitions.length; j++)
						if( transitions[j].getAttribute("Source").equals(source) )
								transitions[j].setAttribute("Source", id2);
					//
					automate.appendChild(place1);
					automate.appendChild(place2);
					automate.appendChild(transition1);
					automate.appendChild(transition2);
					automate.appendChild(transition3);
					automate.appendChild(transition4);
					guards = getGuards();
				}
				if( isOr(constraint) ) {
 					Element transition = (Element) guards[i].getParentNode();
					String source = transition.getAttribute("Source");
					String target = transition.getAttribute("Target");
					Element A = getFirstArg((Element)constraint.cloneNode(true));
					Element B = getSecondArg((Element)constraint.cloneNode(true));
					Element gardeA = nodeBuilder.freshGuard(A);
					Element gardeB = nodeBuilder.freshGuard(B);
					Element transitionA = nodeBuilder.freshTransition(source, target, gardeA, nodeBuilder.freshUpdates());
					Element transitionB = nodeBuilder.freshTransition(source, target, gardeB, nodeBuilder.freshUpdates());
					automate.appendChild(transitionA);
					automate.appendChild(transitionB);
					automate.removeChild(transition);
					guards = getGuards();
				}
			}
		}
	}

	/**
	 * Teste l'automate pour savoir si il reste des gardes  dplier
	 * @return
	 */
	private boolean fixPoint() {
		for( int i = 0; i < guards.length; i++) {
			Element[] functions = DOMUtil.getElementByTagName(guards[i], "FunT");
			if( functions.length != 0 ) {
				String main_label = (functions[0]).getAttribute("Label");
				if( main_label.equals("AND") || main_label.equals("OR") )
					return false;
				if( main_label.equals("NEQ") )
					if( !isNegatomique(functions[0]) )
						return false;
			}
		}
		return true;
	}

	/**
	 * Retourne une contrainte portant la mgation de la contrainte passe en argument
	 * @param constraint
	 * @return
	 */
	private Element not(Element constraint) {
		Element main = (Element) constraint.getFirstChild();
		constraint.removeChild(main);
		constraint.appendChild(nodeBuilder.not(main));
		return constraint;
	}

	/**
	 * Teste si une ngation porte sur un atome
	 * @param function
	 * @return
	 */
	private boolean isNegatomique(Element function) {
		
		if (DOMUtil.getElementByTagName(function, "FunT").length != 0)
				{
					String label = (DOMUtil.getElementByTagName(function, "FunT"))[0].getAttribute("Label");
					return ! ( label.equals("AND") || label.equals("OR") || label.equals("NEQ") );
				}
		else
		{
			return true;
		}
	}

	/**
	 * Test si une contrainte est une double ngation
	 * @param constraint
	 * @return
	 */
	private boolean isDoubleNeg(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		if( functions.length < 2 )
			return false;
		else
			return ( (functions[0].getAttribute("Label").equals("NEQ"))
					&& (functions[1].getAttribute("Label").equals("NEQ")) );
	}

	/**
	 * Retourne la contrainte d'une garde sans les doubles ngations
	 * @param constraint
	 * @return
	 */
	private Element elimDoubleNeg(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		constraint.removeChild(functions[0]);
		constraint.appendChild(functions[2]);
		return constraint;
	}

	private boolean isNeg(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		if( functions.length == 0 )
			return false;
		else
			return functions[0].getAttribute("Label").equals("NEQ");
	}

	/**
	 * Retourne une contrainte portant une ngation dans laquelle la ngation est rentre  
	 * l'intrieur
	 * @param constraint
	 * @return
	 */
	private Element elimNeg(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		if( functions[1].getAttribute("Label").equals("AND") ) {
			Node A = functions[1].getFirstChild(),
				B = functions[1].getLastChild();
			functions[1].removeChild(A);
			functions[1].removeChild(B);
			functions[1].setAttribute("Label", "OR");
			functions[1].appendChild(nodeBuilder.not(A));
			functions[1].appendChild(nodeBuilder.not(B));
			constraint.removeChild(constraint.getFirstChild());
			constraint.appendChild(functions[1]);
			return constraint;
		}
		if( functions[1].getAttribute("Label").equals("OR") ) {
			Node A = functions[1].getFirstChild(),
				B = functions[1].getLastChild();
			functions[1].removeChild(A);
			functions[1].removeChild(B);
			functions[1].setAttribute("Label", "AND");
			functions[1].appendChild(nodeBuilder.not(A));
			functions[1].appendChild(nodeBuilder.not(B));
			constraint.removeChild(constraint.getFirstChild());
			constraint.appendChild(functions[1]);
			return constraint;
		}
		return constraint;
	}

	/**
	 * Retourne une contrainte comportant la premire sous-formule de la contrainte passe en argument
	 * @param constraint
	 * @return
	 */
	private Element getFirstArg(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		constraint.replaceChild(functions[0].getFirstChild(), functions[0]);
		return constraint;
	}

	/**
	 * Retourne une contrainte comportant la deuxime sous-formule de la contrainte passe en argument
	 * @param constraint
	 * @return
	 */
	private Element getSecondArg(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		constraint.replaceChild(functions[0].getLastChild(), functions[0]);
		return constraint;
	}

	private boolean isAnd(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		if( functions.length == 0 )
			return false;
		else
			return functions[0].getAttribute("Label").equals("AND");
	}

	private boolean isOr(Element constraint) {
		Element[] functions = DOMUtil.getElementByTagName(constraint, "FunT");
		if( functions.length == 0 )
			return false;
		else
			return functions[0].getAttribute("Label").equals("OR");
	}

	private String getNewPlaceId(int nb) {
		Element[] places = DOMUtil.getElementByTagName(automate, "Place");
		boolean find;
		String name = Integer.toString(nb);
		do {
			find = false;
			name = new_place_name + name;
			for( int i = 0; i < places.length; i++ )
				if( places[i].getAttribute("Id") == name )
					find = true;
		}
		while( find );
		return name;
	}

	private Element[] getGuards() {
		return DOMUtil.getElementByTagName(automate, "Guard");
	}

}
