package cril.calife.export.engine.averiles.ARTMC;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.*;

import cril.calife.export.engine.AutomataBuilder;
import cril.calife.export.interfaces.JavaProjection;
import cril.calife.export.typechecker.TypeChecker;
import cril.calife.locales.Messages;
import cril.calife.utils.DOMUtil;

/**
 * Cette classe est une copie de ARTMCBuilder  la diffrence prs qu'elle est coupe de Calife.
 * Elle permet de crer un excutable qui prend en argument un fichier xml et gnre le fichier
 * program.py correspondant.
 * 
 * @author Roch Lescuyer
 * 
 */
public class Transform {

	%include { dom.tom }

//VAR
	private Document dom;
	private Element automate;

	private XMLAverilesUtil nodeBuilder;

//Variables de ARTMCBuilder
	private String program;
	private String input;
	private int line;

	private int add_place;
	private HashSet places;
	private HashMap place_line;
	private String new_place_name = "add";

	private int length_id;

	// Les informations sur les pointeurs et les slecteurs
	private HashMap pointers;
	private int pointer_id;
	private String tmp_calife1, tmp_calife2;
	private HashMap selectors;
	private int selector_id;

	// Les informations demandes  l'utilisateur
	private String cygwin_path;
	private String typedefs;
	private String exec;
	private boolean array_assign_refused;
	private boolean data_handled;

	// Tableaux des Loop point au sein du programme
	private HashSet loop_point;

	// Ignorer certaines commandes si des instructions IfData sont prsentes
	private boolean ifData;
	private HashMap data_values;

//Variables de l'entre de l'outils
	private int pointer_num;
	private int desc_num;
	private int next_num;
	private String err_line;
	private int restrict_var;
	private int node_width;

	/*private int max_data;*/
	private int data_size;

// END VAR

// Les trois fonctions de l'interface JavaProjection

	private void setInput(String input) {
		try {
			this.dom = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(input);
		} catch( Exception e ) {
			e.printStackTrace();
		}
	}

	/**
	 * Gnre l'entre de l'outils ARTMC
	 * Cette entre est un fichier contenant une fonction python
	 * 
	 * @param String output - le fichier dans lequel l'entre est gnre
	 */
	private void processExport(String output, boolean a, boolean d) {

		array_assign_refused = a;
		data_handled = d;

		// Rcupre le nombre de descripteurs supplmentaires ncessaire en fonction de la configuration initiale
		desc_num = 1;

		// Initialise le constructeur de nouveaux noeuds
		nodeBuilder = new XMLAverilesUtil(dom);

		//Rcupration et typage de l'automate
			// Dans le cas o un block d'automate est pass en argument, c'est le premier composant qui est pris en compte
		automate = DOMUtil.getElementByTagName(dom, "Automata")[0];

		TypeChecker tc = new TypeChecker(automate, AutomataBuilder.buildTypeSystem(dom));
		tc.TypeCheck();

		//Rcupration des variables
		Element[] elts = DOMUtil.getElementByTagName(dom, "Var");

		//Avertissements
			//sur l'usage des tableaux
		for( int i = 0; i < elts.length; i++)
			if( elts[i].getAttribute("Type").equals("ArrayVariables") ) {
				System.err.println(Messages.getString("ARTMCBuilder.array"));
				break;
			}
			//si les donnes sont ignores
		if( !data_handled )
			getWarning("data_ignored");

//Initialisation de ARTMCBuilder
		program = "";
		line = 0;
		add_place = 0;
		places = new HashSet();
		place_line = new HashMap();
			// Initialise le tableau des pointeurs et donne un identifiant entier  chacun
		pointers = new HashMap();
		pointer_id = 1;		// Les pointeurs sont numrots  partir de 1
		for( int i = 0; i < elts.length; i++)
			if( elts[i].getAttribute("Type").equals("PointerVariables") )
				pointers.put(elts[i].getAttribute("Label"), new Integer(pointer_id++) );
			// Initialise les pointeurs temporaires
		initTmpPointer();
		selectors = new HashMap();
		selector_id = 0;	// Les slecteurs sont numrots  partir de 0
		for( int i = 0; i < elts.length; i++)
			if( elts[i].getAttribute("Type").equals("PointerSelectorVariables") )
				selectors.put(elts[i].getAttribute("Label"), new Integer(selector_id++) );

		loop_point = new HashSet();

		ifData = detectIfDataStatement();
		data_values = collectDataValues();
// End Initialisation

		//Dplie l'automate
			//spare les gardes des actions
		separateGuardsInAutomata();
			//Dploie les combinaisons boolennes contenues dans l'automate
		DeployGuards dg = new DeployGuards(dom, automate, add_place);
		automate = dg.getAutomate();
		add_place = dg.getAddPlace();
			//rcrit certaines gardes
		rewriteGuards();

		// Gnre l'entre
			// Gnre le programme
		try {
			ArrayList initial_states = getInitialStates(); //Remarque: Calife impose qu'il n'y ait qu'un tat initial
			// Si l'automate n'a pas d'tat initial, lancer un message d'erreur
			if( initial_states.size() == 0 ) {
				System.err.println(Messages.getString("ARTMCBuilder.initial_missing"));
				return;
			}
			//si l'automate ne contient pas de transitions partant de l'tat initial, le programme 
			//se rsume  une ligne
			if( isFinal((String)initial_states.get(0)) ) {
				program += "\t\t(\"exit\"," + line + ")\n";
			}
			//sinon parser les transitions  partir de l'tat initial
			else
				parseTransition((String)initial_states.get(0));
		} catch( PointerManipulationException pme ) {
			System.err.println(pme.getMsg());
			return;
		}

			// Rcriture des lignes
		length_id = 1 + Integer.toBinaryString(line + 1).length();
		rewriteProgram();
			// Initialise les valeurs de l'entre
		pointer_num = 1 + pointers.size();
		next_num = selectors.size();
		err_line = Integer.toBinaryString(exp(2, length_id) - 1);
		restrict_var = 1;
		if( data_size == 0 )
			data_size = 1;
		node_width = pointer_num + desc_num + 2 + data_size;
			// Gnre la fonction python
		input =
			getComment() +
			"def get_program():\n" +
			"\tprogram=[\n" +
			program +
			"\tnode_width=" + node_width + "\n" +
			"\tpointer_num=" + pointer_num + "\n" +
			"\tdesc_num=" + desc_num + "\n" +
			"\tnext_num=" + next_num + "\n" +
			"\terr_line=\"" + err_line + "\"\n" +
			"\trestrict_var=" + restrict_var + "\n" +
			"\n" +
			"\tenv=(node_width,pointer_num,desc_num,next_num,err_line,restrict_var)\n" +
			"\treturn(program,env)\n";

		// Ecrit l'entre dans un fichier
		try {
			PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(output)));
			out.print(input);
			out.close();
		} catch( IOException e ) {
			e.printStackTrace();
		}

		//Message final
		System.out.println(Messages.getString("ARTMCBuilder.file") + "program.py" + Messages.getString("ARTMCBuilder.generate"));

	}

// END JavaProjection

// FONCTIONS AUXILIAIRES

// Fonctions de bases

	private static int exp(int b, int p) {

		int tmp = 1;
		for( int i = 0; i < p; i++)
			tmp = tmp * b;
		return tmp;

	}

	private static String toBinaryString(int i, int length) {

		String val = Integer.toBinaryString(i);
		int length_tmp = length - val.length();
		String s = "";
		for( int j = 0; j < length_tmp; j++ )
			s += "0";
		return s + val;

	}

	private static boolean present(String val, String s) {

		CharSequence cs = s.subSequence(0, s.length());
		return val.contains(cs);

	}

// Fonctions de manipulations du document XML

	// Un tat est final si aucune flche ne sort de lui
	private boolean isFinal(String name) {

		%match(TNode automate) {
			<_><Transition Source=s Target=t></Transition></_> -> {
				if( `s.equals(name) )
					return false;
			}
		}
		return true;
	}

	private boolean isInitial(String name) {

		%match(TNode automate) {
			<_><Place Id=id Initial=init></Place></_> -> {
				if( `id.equals(name) && `init.equals("true") )
					return true;
			}
		}
		return false;

	}

	// Transfert le statut d'tat initial de from  to
	private void transfertInitial(String from, String to) {

		%match(TNode automate) {
			<_>p@<Place Id=id></Place></_> -> {
				if( `id.equals(from) )
					((Element)`p).setAttribute("Initial", "false");
				if( `id.equals(to) )
					((Element)`p).setAttribute("Initial", "true");
			}
		}
		
	}

	// Retourne l'ensemble des tats initiaux de l'automate.
		// Remarque : Calife impose qu'il n'y en ait qu'un
	private ArrayList getInitialStates() {

		ArrayList al = new ArrayList();
		%match(TNode automate) {
			<_><Place Id=id Initial=init></Place></_> -> {
				if( `init.equals("true") )
					al.add(`id);
			}
		}
		return al;

	}

	// Retourne l'ensemble des transitions qui partent de la place source
	private ArrayList getNextPlaces(String source) {

		ArrayList np = new ArrayList();
		%match(TNode automate) {
			<_>trans@<Transition Source=s Target=t></Transition></_> -> {
				if( `s.equals(source) )
					np.add(`trans);
			}
		}
		return np;

	}

	private void getWarning(String warning) {

		String tmp = "";
		%match(String warning) {
			"pm" -> {
				tmp = Messages.getString("ARTMCBuilder.warning");
			}
			"array_assign" -> {
				tmp = Messages.getString("ARTMCBuilder.array_assign_warning");
			}
			"data_ignored" -> {
				tmp = Messages.getString("ARTMCBuilder.data_ignored_warning");
			}
		}
		System.err.println(tmp);

	}

	private void detectLoopPoint(int line, int next_line) {

		if( next_line < line )
			loop_point.add(next_line);

	}

	private boolean detectIfDataStatement() {

		%match(TNode automate) {
			<_><Transition><Guard><Constraint><FunT>
				<FunT Arg=arg1></FunT><_></_>
			</FunT></Constraint></Guard></Transition></_> -> {
				if( `arg1.equals("PointerVariables:IntegerSelectorVariables") )
					return true;
			}
		}
		return false;

	}

	private HashMap collectDataValues() {

		TreeSet values = new TreeSet();
		%match(TNode automate) {
			<_><Transition><Updates><Update><FunT>
				<FunT Arg=arg1></FunT>c@<Const Type=t2></Const>
			</FunT></Update></Updates></Transition></_> -> {
				if( `arg1.equals("PointerVariables:IntegerSelectorVariables") && `t2.equals("Z") ) {
					values.add(new Integer(Integer.parseInt(`c.getTextContent().trim())));
				}
			}
			<_><Transition><Guard><Constraint><FunT>
				<FunT Arg=arg1></FunT>c@<Const Type=t2></Const>
			</FunT></Constraint></Guard></Transition></_> -> {
				if( `arg1.equals("PointerVariables:IntegerSelectorVariables") && `t2.equals("Z") ) {
					values.add(new Integer(Integer.parseInt(`c.getTextContent().trim())));
				}
			}
			<_><Transition><Guard><Constraint><FunT>
				c@<Const Type=t1></Const><FunT Arg=arg2></FunT>
			</FunT></Constraint></Guard></Transition></_> -> {
				if( `t1.equals("Z") && `arg2.equals("PointerVariables:IntegerSelectorVariables") ) {
					values.add(new Integer(Integer.parseInt(`c.getTextContent().trim())));
				}
			}
			<_><Transition><Guard><Constraint><FunT Label=lab><Const/>
				<FunT><FunT Arg=arg1></FunT>c@<Const Type=t2></Const>
			</FunT></FunT></Constraint></Guard></Transition></_> -> {
				if( `lab.equals("NOT") && `arg1.equals("PointerVariables:IntegerSelectorVariables") && `t2.equals("Z") ) {
					values.add(new Integer(Integer.parseInt(`c.getTextContent().trim())));
				}
			}
			<_><Transition><Guard><Constraint><FunT Label=lab><Const/>
				<FunT>c@<Const Type=t1></Const><FunT Arg=arg2></FunT>
			</FunT></FunT></Constraint></Guard></Transition></_> -> {
				if( `lab.equals("NOT") && `t1.equals("Z") && `arg2.equals("PointerVariables:IntegerSelectorVariables") ) {
					values.add(new Integer(Integer.parseInt(`c.getTextContent().trim())));
				}
			}
		}
		HashMap data = new HashMap();
		String code;
		// si il n'y a pas de donnes, data_size = 0, sinon, c'est la longueur du mot binaire qu'il faut pour les coder
		data_size = (values.size() == 0) ? 0 : Integer.toBinaryString(values.size() - 1).length();
		Iterator e = values.iterator();
		int cpt = 0;
		while( e.hasNext() ) {
			int value = ((Integer)e.next()).intValue();
			code = toBinaryString(cpt, data_size);
			data.put(value, code);
			cpt++;
		}
		return data;

	}

// END FONCTIONS AUXILIAIRES

// REECRITURE DES GARDES

	/**
	 * Dploie une transition.
	 * Si la transition trans part d'un tat A vers un tat B, on ajoute une nouvelle place NEW, on cr
	 * une transition T de NEW vers A, cette transition n'a pas de garde particulire et possde la
	 * mise  jour new_updates, et trans possde la nouvelle garde new_guard.
	 * 
	 * @param Node trans - La transition comprenant une garde  rcrire
	 * @param Node new_updates - La mise  jour correspondant  une ligne de commande  rajouter avant trans
	 * @param Node new_guard - La nouvelle garde pour la transition trans
	 * @return
	 */
	private void deployCond(Node trans, Node new_updates, Node new_guard) {

		String source = ((Element)trans).getAttribute("Source");
		String id = getNewPlaceId(add_place);
		add_place++;
		// Toutes les flches qui arrivaient sur 'source' doivent arriver sur la nouvelle place
		Element[] transitions = DOMUtil.getElementByTagName(automate, "Transition");
		for( int i = 0; i < transitions.length; i++ )
			if( transitions[i].getAttribute("Target").equals(source) )
				transitions[i].setAttribute("Target", id);
		// Ajout d'une place
		Element nNew = nodeBuilder.getFreshPlace(id);
		automate.appendChild(nNew);
		// Cration d'une nouvelle transition, de la nouvelle place vers la source de 'trans'
		nNew = (Element)trans.cloneNode(true);
		nNew.setAttribute("Source", id);
		nNew.setAttribute("Target", source);
		Node old_guard = (DOMUtil.getElementByTagName(nNew, "Guard"))[0];
		Node guard = nodeBuilder.getFreshGuard();
		nNew.replaceChild(guard, old_guard);
		Node old_updates = (DOMUtil.getElementByTagName(nNew, "Updates"))[0];
		nNew.replaceChild(new_updates, old_updates);
		automate.appendChild(nNew);
		// Reconstruction de la transition 'trans'
		old_guard = (DOMUtil.getElementByTagName(trans, "Guard"))[0];
		((Element)trans).replaceChild(new_guard, old_guard);
		// Si la place source tait initiale, il faut transfrer le statut d'tat initial  la place qu'on a rajoute avant elle
		if( isInitial(source) )
			transfertInitial(source, id);

	}

	/**
	 * Parse les gardes de l'automates en vue de rcritures ventuelles
	 * 
	 * @param Node trans - La transition comprenant la garde  rcrire ventuellement
	 * @param Node guard - La garde  rcrire ventuellement
	 */
	private void matchGuard(Node trans, Node guard) {

		%match(TNode guard) {
			<_><Constraint><FunT Label=lab1>
					<FunT Arg=arg Label=lab2>p@<_ Type=t1></_>s@<_ Type=t2></_></FunT>
					a2@<_ Type=t></_>
				</FunT></Constraint></_> -> {
				if( `lab1.equals("EQ") ) {
					String[] args = `arg.split(":");
					if( `lab2.equals("IMPL") && args[1].equals("PointerSelectorVariables") ) {
						String	point = ( `t1.equals(args[0]) ? `p.getTextContent() : `s.getTextContent() ),
								select = ( `t2.equals(args[1]) ? `s.getTextContent() : `p.getTextContent() );
						String typeArg2 = `t;
						%match(String typeArg2) {
							// Rcriture de if( x.next == y )
							"PointerVariables" -> {
								pointers.put(tmp_calife1, new Integer(pointer_id) );
								Node updates = nodeBuilder.getAssignment_x_y_next(tmp_calife1, point, select);
								String y = `a2.getTextContent();
								Node new_guard = nodeBuilder.getNewCond_x_y(tmp_calife1, y);
								deployCond(trans, updates, new_guard);
							}
							// Rcriture de if( x.next == null )
							"null" -> {
								pointers.put(tmp_calife1, new Integer(pointer_id) );
								Node updates = nodeBuilder.getAssignment_x_y_next(tmp_calife1, point, select);
								Node new_guard = nodeBuilder.getNewCond_x_null(tmp_calife1);
								deployCond(trans, updates, new_guard);
							}
						}
					}
				}
			}
			<_><Constraint><FunT Label=lab1>
					a1@<_ Type=t></_>
					<FunT Arg=arg Label=lab2>p@<_ Type=t1></_>s@<_ Type=t2></_></FunT>
				</FunT></Constraint></_> -> {
				if( `lab1.equals("EQ") ) {
					String[] args = `arg.split(":");
					if( `lab2.equals("IMPL") && args[1].equals("PointerSelectorVariables") ) {
						String	point = ( `t1.equals(args[0]) ? `p.getTextContent() : `s.getTextContent() ),
								select = ( `t2.equals(args[1]) ? `s.getTextContent() : `p.getTextContent() );
						String typeArg1 = `t;
						%match(String typeArg1) {
							// Rcriture de if( y == x.next )
							"PointerVariables" -> {
								pointers.put(tmp_calife1, new Integer(pointer_id) );
								Node updates = nodeBuilder.getAssignment_x_y_next(tmp_calife1, point, select);
								String y = `a1.getTextContent();
								Node new_guard = nodeBuilder.getNewCond_x_y(y, tmp_calife1);
								deployCond(trans, updates, new_guard);
							}
							// Rcriture de if( null == x.next )
							"null" -> {
								pointers.put(tmp_calife1, new Integer(pointer_id) );
								Node updates = nodeBuilder.getAssignment_x_y_next(tmp_calife1, point, select);
								Node new_guard = nodeBuilder.getNewCond_x_null(tmp_calife1);
								deployCond(trans, updates, new_guard);
							}
						}
					}
				}
			}
				// Rcriture du cas if( x.next == y.next )
					// Fonctionne comme DeployCond, mais il faut ajouter deux tats en plus cette fois-ci
			<_><Constraint><FunT Label=lab>
					<FunT Arg=arg1 Label=lab1>p1@<_ Type=t11></_>s1@<_ Type=t12></_></FunT>
					<FunT Arg=arg2 Label=lab2>p2@<_ Type=t21></_>s2@<_ Type=t22></_></FunT>
				</FunT></Constraint></_> -> {
				if( `lab.equals("EQ") ) {
					String[] args1 = `arg1.split(":"), args2 = `arg2.split(":");
					if( `lab1.equals("IMPL") && `lab2.equals("IMPL") 
							&& args1[1].equals("PointerSelectorVariables") && args2[1].equals("PointerSelectorVariables") ) {
						String	point1 = ( `t11.equals(args1[0]) ? `p1.getTextContent() : `s1.getTextContent() ),
								select1 = ( `t12.equals(args1[1]) ? `s1.getTextContent() : `p1.getTextContent() ),
								point2 = ( `t21.equals(args2[0]) ? `p2.getTextContent() : `s2.getTextContent() ),
								select2 = ( `t22.equals(args2[1]) ? `s2.getTextContent() : `p2.getTextContent() );
						pointers.put(tmp_calife1, new Integer(pointer_id));
						pointers.put(tmp_calife2, new Integer(pointer_id + 1));
						Node
							assign1 = nodeBuilder.getAssignment_x_y_next(tmp_calife1, point1, select1),
							assign2 = nodeBuilder.getAssignment_x_y_next(tmp_calife2, point2, select2),
							cond = nodeBuilder.getNewCond_x_y(tmp_calife1, tmp_calife2);
						String source = ((Element)trans).getAttribute("Source");
						String id1 = getNewPlaceId(add_place);
						add_place++;
						String id2 = getNewPlaceId(add_place);
						add_place++;
						Element[] transitions = DOMUtil.getElementByTagName(automate, "Transition");
						for( int i = 0; i < transitions.length; i++ )
							if( transitions[i].getAttribute("Target").equals(source) )
								transitions[i].setAttribute("Target", id1);
						Element nNew = nodeBuilder.getFreshPlace(id1);
						automate.appendChild(nNew);
						nNew = nodeBuilder.getFreshPlace(id2);
						automate.appendChild(nNew);
						Node void_guard = nodeBuilder.getFreshGuard();
						nNew = nodeBuilder.getFreshTransition(id1, id2, void_guard, assign1);
						automate.appendChild(nNew);
						void_guard = nodeBuilder.getFreshGuard();
						nNew = nodeBuilder.getFreshTransition(id2, source, void_guard, assign2);
						automate.appendChild(nNew);
						String target = ((Element)trans).getAttribute("Target");
						Element old_updates = (DOMUtil.getElementByTagName(trans, "Updates"))[0];
						nNew = nodeBuilder.getFreshTransition(source, target, cond, old_updates);
						automate.appendChild(nNew);
						automate.removeChild(trans);
						if( isInitial(source) )
							transfertInitial(source, id1);
					}
				}
			}
		}

	}

	private void rewriteGuards() {

		%match(TNode automate) {
			<_>t@<Transition>g@<Guard PrettyPrint=ppg></Guard></Transition></_> -> {
				if( !`ppg.equals("True") ) {
					Node trans = `t, guard = `g;
					matchGuard(trans, guard);
				}
			}
		}

	}

// END REECRITURE DES GARDES


// 	SEPARER LES GARDES DES ACTIONS

	/**
	 * Spare les gardes des actions dans l'automate Averiles
	 * 
	 */
	private void separateGuardsInAutomata() {

		%match(TNode automate) {
			<_>t@<Transition Source=source Target=target>
					g@<Guard PrettyPrint=gpp></Guard>
					u@<Updates PrettyPrint=upp></Updates>
						</Transition></_> -> {
				if( !`gpp.equals("True") ) {

					String new_place_id = getNewPlaceId(add_place);
					add_place++;

					// Ajout d'une place
					Element nNew = nodeBuilder.getFreshPlace(new_place_id);
					automate.appendChild(nNew);

					// Rcupration des anciennes garde et action, et cration des nouvelles
					Node old_guard = `g, old_updates = `u,
						new_guard = nodeBuilder.getFreshGuard(), 
						new_updates = nodeBuilder.getFreshUpdates();

					// Nouvelle transition, de la nouvelle place vers la sortie
					nNew = nodeBuilder.getFreshTransition(new_place_id, `target, new_guard, old_updates);
					automate.appendChild(nNew);

					// Nouvelle transition de l'entre vers la nouvelle place
					nNew = nodeBuilder.getFreshTransition(`source, new_place_id, old_guard, new_updates);
					automate.appendChild(nNew);

					// Effacement de l'ancienne transition
					Node nOld = `t;
					automate.removeChild(nOld);

				}
			}
		}

	}

// END	SEPARER LES GARDES DES ACTIONS


// GENERER LE PROGRAMME A PARTIR DE L'AUTOMATE

	/**
	 * Gnre la partie de la ligne de commande ARTMC correpondant  une condition
	 * Attention, cette portion de code ne gnre pas la ligne de commande entire
	 * Cette dernire est traite par l'algorithme de parcours du graphe afin qu'il traite
	 * toutes les conditions partant d'une place en mme temps
	 * 
	 * @param Node guard - Le noeud XML reprsentant une garde de l'automate Averiles
	 * @return La partie de la ligne de commande ARTMC correspondant  la condition contenue dans la garde
	 */
	private String getCond(Node guard) throws DetectGuardException {

		%match(TNode guard) {
			<_><Constraint><FunT Arg=arg Label=lab>
				a1@<_ Type=t1></_>a2@<_ Type=t2></_></FunT></Constraint></_> -> {
				if( `lab.equals("EQ") ) {
					String arg1 = `t1, arg2 = `t2;
					%match(String arg1, String arg2) {
						"PointerVariables", "PointerVariables" -> {
							Integer p1 = (Integer)pointers.get(`a1.getTextContent());
							Integer p2 = (Integer)pointers.get(`a2.getTextContent());
							return "\"ifx==y\"," + line + "," + p1 + "," + p2;
						}
						"PointerVariables", "null" -> {
							Integer p = (Integer)pointers.get(`a1.getTextContent());
							return "\"ifx==null\"," + line + "," + p;
						}
						"null", "PointerVariables" -> {
							Integer p = (Integer)pointers.get(`a2.getTextContent());
							return "\"ifx==null\"," + line + "," + p;
						}
					}
				}
			}
			<_><Constraint><FunT Label=lab>
					a1@<FunT Arg=arg Label=lab2 Type=t1>b1@<_ Type=s1></_>b2@<_ Type=s2></_></FunT>
					a2@<_ Type=t2></_></FunT></Constraint></_> -> {
				String[] args = `arg.split(":");
				if( `lab.equals("EQ") && `lab2.equals("IMPL") && args[1].equals("IntegerSelectorVariables") ) {
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"lval", "Z" -> {
							if( data_handled ) {
								Integer p = (Integer)pointers.get(`b1.getTextContent());
								int z = Integer.parseInt(`a2.getTextContent().trim());
								String code = (String) data_values.get(z);
								return "\"ifdata\"," + line + "," + p + ",\"" + code + "\"";
							}
						}
					}
				}
			}
			<_><Constraint><FunT Label=lab>
					a1@<_ Type=t1></_>
					a2@<FunT Arg=arg Label=lab2 Type=t2>b1@<_ Type=s1></_>b2@<_ Type=s2></_></FunT>
				</FunT></Constraint></_> -> {
				String[] args = `arg.split(":");
				if( `lab.equals("EQ") && `lab2.equals("IMPL") && args[1].equals("IntegerSelectorVariables") ) {
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"Z", "lval" -> {
							if( data_handled ) {
								Integer p = (Integer)pointers.get(`b1.getTextContent());
								int z = Integer.parseInt(`a1.getTextContent().trim());
								String code = (String) data_values.get(z);
								return "\"ifdata\"," + line + "," + p + ",\"" + code + "\"";
							}
						}
					}
				}
			}
			<_><Constraint>
				<FunT Label=lab0><Const/>
					<FunT Arg=arg Label=lab>
						a1@<_ Type=t1></_>a2@<_ Type=t2></_>
					</FunT></FunT></Constraint></_> -> {
				if( `lab0.equals("NOT") && `lab.equals("EQ") ) {
					String arg1 = `t1, arg2 = `t2;
					%match(String arg1, String arg2) {
						"PointerVariables", "PointerVariables" -> {
							Integer p1 = (Integer)pointers.get(`a1.getTextContent());
							Integer p2 = (Integer)pointers.get(`a2.getTextContent());
							throw new DetectGuardException("\"ifx==y\"," + line + "," + p1 + "," + p2);
						}
						"PointerVariables", "null" -> {
							Integer p = (Integer)pointers.get(`a1.getTextContent());
							throw new DetectGuardException("\"ifx==null\"," + line + "," + p);
						}
						"null", "PointerVariables" -> {
							Integer p = (Integer)pointers.get(`a2.getTextContent());
							throw new DetectGuardException("\"ifx==null\"," + line + "," + p);
						}
					}
				}
			}
			<_><Constraint>
				<FunT Label=lab0><Const/>
					<FunT Label=lab>
					a1@<FunT Arg=arg Label=lab2 Type=t1>b1@<_ Type=s1></_>b2@<_ Type=s2></_></FunT>
					a2@<_ Type=t2></_></FunT></FunT></Constraint></_> -> {
				String[] args = `arg.split(":");
				if( `lab0.equals("NOT") && `lab.equals("EQ") && `lab2.equals("IMPL") && args[1].equals("IntegerSelectorVariables") ) {
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"lval", "Z" -> {
							if( data_handled ) {
								Integer p = (Integer)pointers.get(`b1.getTextContent());
								int z = Integer.parseInt(`a2.getTextContent().trim());
								String code = (String) data_values.get(z);
								throw new DetectGuardException("\"ifdata\"," + line + "," + p + ",\"" + code + "\"");
							}
						}
					}
				}
			}
			<_><Constraint>
				<FunT Label=lab0><Const/>
					<FunT Label=lab>
					a1@<_ Type=t1></_>
					a2@<FunT Arg=arg Label=lab2 Type=t2>b1@<_ Type=s1></_>b2@<_ Type=s2></_></FunT>
				</FunT></FunT></Constraint></_> -> {
				String[] args = `arg.split(":");
				if( `lab0.equals("NOT") && `lab.equals("EQ") && `lab2.equals("IMPL") && args[1].equals("IntegerSelectorVariables") ) {
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"Z", "lval" -> {
							if( data_handled ) {
								Integer p = (Integer)pointers.get(`b1.getTextContent());
								int z = Integer.parseInt(`a1.getTextContent().trim());
								String code = (String) data_values.get(z);
								throw new DetectGuardException("\"ifdata\"," + line + "," + p + ",\"" + code + "\"");
							}
						}
					}
				}
			}
		}
		return "\"if*\"," + line + "";

	}

	/**
	 * Cette fonction renvoit la ligne de commande de l'outils ARTMC correspondant au contenu de
	 * la mise  jour passe en argument.
	 * 
	 *  @param Node updates - Le noeud XML reprsentant une mose  jour dans l'automate Averiles
	 *  @param int next_line - La prochaine ligne de commande  excuter aprs la ligne courante
	 *  @return La ligne de commande ARTMC correspondant  la mise  jour passe en argument 
	 */
	private String getLine(Node updates, int next_line) throws PointerManipulationException {

		%match(TNode updates) {

			<_><Update><FunT Label=lab>a1@<_ Type=t1></_>a2@<_ Type=t2></_></FunT></Update></_> -> {
				if( `lab.equals("EQ") ) {
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"PointerVariables", "null" -> {
							Integer p = (Integer)pointers.get(`a1.getTextContent());
							detectLoopPoint(line, next_line);
							return "\t\t(\"x=null\"," + line + "," + p + "," + next_line + ")\n";
						}
						"PointerVariables", "PointerVariables" -> {
							Integer p1 = (Integer)pointers.get(`a1.getTextContent());
							Integer p2 = (Integer)pointers.get(`a2.getTextContent());
							detectLoopPoint(line, next_line);
							return "\t\t(\"x=y\"," + line + "," + p1 + "," + p2 + "," + next_line + ")\n";
						}
						"PointerVariables", "mallocP" -> {
							Integer p = (Integer)pointers.get(`a1.getTextContent());
							detectLoopPoint(line, next_line);
							return "\t\t(\"new\"," + line + "," + p + "," + next_line + ")\n";
						}
						"PointerVariables", "Z" -> {
							Integer p = (Integer)pointers.get(`a1.getTextContent());
							detectLoopPoint(line, next_line);
							getWarning("pm");
							return "\t\t(\"x=random\"," + line + "," + p + "," + next_line + ")\n";
						}
						"PointerVariables", "IntegerVariables" -> {
							Integer p = (Integer)pointers.get(`a1.getTextContent());
							detectLoopPoint(line, next_line);
							getWarning("pm");
							return "\t\t(\"x=random\"," + line + "," + p + "," + next_line + ")\n";
						}
						"PointerVariables", "ArrayVariables" -> {
							if( array_assign_refused ) {
								throw new PointerManipulationException(Messages.getString("ARTMCBuilder.array_assign_refused"));
							} else {
								Integer p = (Integer)pointers.get(`a1.getTextContent());
								detectLoopPoint(line, next_line);
								getWarning("array_assign");
								return "\t\t(\"x=null\"," + line + "," + p + "," + next_line + ")\n";
							}
						}
					}
				}
			}
			<_><Update><FunT Label=lab>
						a1@<FunT Arg=arg Label=lab2 Type=t1>b1@<_ Type=s1></_>b2@<_ Type=s2></_></FunT>
						a2@<_ Type=t2></_></FunT></Update></_> -> {
				if( `lab.equals("EQ") ) {
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"lval", "null" -> {
							if( `lab2.equals("IMPL") ) {
								String[] args = `arg.split(":");
								if( args[1].equals("PointerSelectorVariables") ) {
									Integer p = (Integer)pointers.get( ( `s1.equals(args[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
									Integer s = (Integer)selectors.get(( `s2.equals(args[1]) ? `b2.getTextContent() : `b1.getTextContent() ));
									detectLoopPoint(line, next_line);
									return "\t\t(\"x.next=null\"," + line + "," + p + "," + s + "," + next_line + ")\n";
								}
								if( args[1].equals("IntegerSelectorVariables") ) {
									throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));									
								}
							}
						}
						"lval", "PointerVariables" -> {
							if( `lab2.equals("IMPL") ) {
								String[] args = `arg.split(":");
								if( args[1].equals("IntegerSelectorVariables") ) {
									throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));
								}
								if( args[1].equals("PointerSelectorVariables") ) {
									Integer p1 = (Integer)pointers.get( ( `s1.equals(args[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
									String select = ( `s2.equals(args[1]) ? `b2.getTextContent() : `b1.getTextContent() );
									Integer s = (Integer)selectors.get(select);
									Integer p2 = (Integer)pointers.get(`a2.getTextContent());
									detectLoopPoint(line, next_line);
									return "\t\t(\"x.next=y\"," + line + "," + p1 + "," + p2 + "," + s + "," + next_line + "," + (desc_num++) + ")\n";
								}
							}
						}
						"lval", "Z" -> {
							if( `lab2.equals("IMPL") ) {
								String[] args = `arg.split(":");
								if( args[1].equals("IntegerSelectorVariables") && data_handled ) {
									Integer p = (Integer)pointers.get( ( `s1.equals(args[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
									int z = Integer.parseInt(`a2.getTextContent().trim());
									String code = (String) data_values.get(z);
									detectLoopPoint(line, next_line);
									return "\t\t(\"setdata\"," + line + "," + p + ",\"" + code + "\"," + next_line + ")\n";
								}
								if( args[1].equals("PointerSelectorVariables") ) {
									Integer p = (Integer)pointers.get( ( `s1.equals(args[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
									Integer s = (Integer)selectors.get( ( `s2.equals(args[1]) ? `b2.getTextContent() : `b1.getTextContent() ) );
									if( !pointers.containsKey(tmp_calife1) )
										pointers.put(tmp_calife1, new Integer(pointer_id) );
									Integer tmp_pt = (Integer)pointers.get(tmp_calife1);
									String tmp = "\t\t(\"x=random\"," + line + "," + tmp_pt + "," + (line + 1) + ")\n";
									line++;
									if( next_line == line )
										next_line++;
									detectLoopPoint(line, next_line);
									tmp += "\t\t(\"x.next=y\"," + line + "," + p + "," + tmp_pt + "," + s + "," + next_line + "," + (desc_num++) + ")\n";
									getWarning("pm");
									return tmp;
								}
							}
						}
						"lval", "ArrayVariables" -> {
							if( `lab2.equals("IMPL") ) {
								String[] args = `arg.split(":");
								if( args[1].equals("PointerSelectorVariables") ) {
									if( array_assign_refused ) {
										throw new PointerManipulationException(Messages.getString("ARTMCBuilder.array_assign_refused"));
									} else {
										Integer p = (Integer)pointers.get( ( `s1.equals(args[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
										Integer s = (Integer)selectors.get( ( `s2.equals(args[1]) ? `b2.getTextContent() : `b1.getTextContent() ) );
										detectLoopPoint(line, next_line);
										getWarning("array_assign");
										return "\t\t(\"x.next=null\"," + line + "," + p + "," + s + "," + next_line + ")\n";
									}
								}
								if( args[1].equals("IntegerSelectorVariables") ) {
									throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));
								}
							}
						}
						"lval", "IntegerVariables" -> {
							if( `lab2.equals("IMPL") ) {
								String[] args = `arg.split(":");
								if( args[1].equals("PointerSelectorVariables") ) {
									Integer p = (Integer)pointers.get( ( `s1.equals(args[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
									Integer s = (Integer)selectors.get( ( `s2.equals(args[1]) ? `b2.getTextContent() : `b1.getTextContent() ) );
									if( !pointers.containsKey(tmp_calife1) )
										pointers.put(tmp_calife1, new Integer(pointer_id) );
									Integer tmp_pt = (Integer)pointers.get(tmp_calife1);
									String tmp = "\t\t(\"x=random\"," + line + "," + tmp_pt + "," + (line + 1) + ")\n";
									line++;
									if( next_line == line )
										next_line++;
									detectLoopPoint(line, next_line);
									tmp += "\t\t(\"x.next=y\"," + line + "," + p + "," + tmp_pt + "," + s + "," + next_line + "," + (desc_num++) + ")\n";
									getWarning("pm");
									return tmp;
								}
								if( args[1].equals("IntegerSelectorVariables") ) {
									if( data_handled && ifData ) {
										throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));
									}
								}
							}
						}
					}
				}
			}
			<_><Update><FunT Label=lab>
						a1@<_ Type=t1></_>
						a2@<FunT Arg=arg Label=lab2 Type=t2>b1@<_ Type=s1></_>b2@<_ Type=s2></_></FunT>
						</FunT></Update></_> -> {
				if( `lab.equals("EQ") ) {
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"PointerVariables", "lval" -> {
							if( `lab2.equals("IMPL") ) {
								String[] args = `arg.split(":");
								if( args[1].equals("IntegerSelectorVariables") ) {
									Integer p = (Integer)pointers.get(`a1.getTextContent());
									detectLoopPoint(line, next_line);
									getWarning("pm");
									return "\t\t(\"x=random\"," + line + "," + p + "," + next_line + ")\n";
								}
								if( args[1].equals("PointerSelectorVariables") ) {
									Integer p1 = (Integer)pointers.get(`a1.getTextContent());
									Integer p2 = (Integer)pointers.get( ( `s1.equals(args[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
									Integer s = (Integer)selectors.get( ( `s2.equals(args[1]) ? `b2.getTextContent() : `b1.getTextContent() ) );
									detectLoopPoint(line, next_line);
									return "\t\t(\"x=y.next\"," + line + "," + p1 + "," + p2 + "," + s + "," + next_line + ")\n";
								}
								if( args[1].equals("ArraySelectorVariables") ) {
									if( array_assign_refused ) {
										throw new PointerManipulationException(Messages.getString("ARTMCBuilder.array_assign_refused"));
									} else {
										Integer p = (Integer)pointers.get(`a1.getTextContent());
										detectLoopPoint(line, next_line);
										getWarning("array_assign");
										return "\t\t(\"x=null\"," + line + "," + p + "," + next_line + ")\n";
									}
								}
							}
							if( `lab2.equals("GET") ) {
								Integer p = (Integer)pointers.get(`a1.getTextContent());
								detectLoopPoint(line, next_line);
								getWarning("pm");
								return "\t\t(\"x=random\"," + line + "," + p + "," + next_line + ")\n";
							}
						}
						"PointerVariables", "expr" -> {
							if( `lab2.equals("PLUS") || `lab2.equals("MOINS") ) {
								Integer p = (Integer)pointers.get(`a1.getTextContent());
								detectLoopPoint(line, next_line);
								getWarning("pm");
								return "\t\t(\"x=random\"," + line + "," + p + "," + next_line + ")\n";
							}
						}
					}
				}
			}
			<_><Update><FunT Label=lab>
						<FunT Arg=arg1 Label=lab1 Type=t1>a1@<_ Type=t11></_>a2@<_ Type=t12></_></FunT>
						<FunT Arg=arg2 Label=lab2 Type=t2>b1@<_ Type=t21></_>b2@<_ Type=t22></_></FunT>
						</FunT></Update></_> -> {
				if( `lab.equals("EQ") ) {
					String args1[] = `arg1.split(":"), args2[] = `arg2.split(":");
					String type1 = `t1, type2 = `t2;
					%match(String type1, String type2) {
						"lval", "lval" -> {
							if( `lab1.equals("IMPL") && `lab2.equals("IMPL") ) {
								if( args1[1].equals("IntegerSelectorVariables") && args2[1].equals("PointerSelectorVariables") ) {
									throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));
								}
								if( args1[1].equals("PointerSelectorVariables") && args2[1].equals("ArraySelectorVariables") ) {
									if( array_assign_refused ) {
										throw new PointerManipulationException(Messages.getString("ARTMCBuilder.array_assign_refused"));
									} else {
										Integer p = (Integer)pointers.get( ( `t11.equals(args1[0]) ? `a1.getTextContent() : `a2.getTextContent() ) );
										Integer s = (Integer)selectors.get( ( `t12.equals(args1[1]) ? `a2.getTextContent() : `a1.getTextContent() ) );
										detectLoopPoint(line, next_line);
										getWarning("array_assign");
										return "\t\t(\"x.next=null\"," + line + "," + p + "," + s + "," + next_line + ")\n";
									}
								}
								if( args1[1].equals("PointerSelectorVariables") && args2[1].equals("IntegerSelectorVariables") ) {
									Integer p = (Integer)pointers.get( ( `t11.equals(args1[0]) ? `a1.getTextContent() : `a2.getTextContent() ) );
									Integer s = (Integer)selectors.get( ( `t12.equals(args1[1]) ? `a2.getTextContent() : `a1.getTextContent() ) );
									if( !pointers.containsKey(tmp_calife1) )
										pointers.put(tmp_calife1, new Integer(pointer_id) );
									Integer tmp_pt = (Integer)pointers.get(tmp_calife1);
									String tmp = "\t\t(\"x=random\"," + line + "," + tmp_pt + "," + (line + 1) + ")\n";
									line++;
									if( next_line == line )
										next_line++;
									detectLoopPoint(line, next_line);
									tmp += "\t\t(\"x.next=y\"," + line + "," + p + "," + tmp_pt + "," + s + "," + next_line + "," + (desc_num++) + ")\n";
									getWarning("pm");
									return tmp;
								}
								if( args1[1].equals("PointerSelectorVariables") && args2[1].equals("PointerSelectorVariables") ) {
									Integer p1 = (Integer)pointers.get( ( `t11.equals(args1[0]) ? `a1.getTextContent() : `a2.getTextContent() ) );
									Integer s1 = (Integer)selectors.get( ( `t12.equals(args1[1]) ? `a2.getTextContent() : `a1.getTextContent() ) );
									Integer p2 = (Integer)pointers.get( ( `t21.equals(args2[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
									Integer s2 = (Integer)selectors.get( ( `t22.equals(args2[1]) ? `b2.getTextContent() : `b1.getTextContent() ) );
									if( !pointers.containsKey(tmp_calife1) )
										pointers.put(tmp_calife1, new Integer(pointer_id) );
									Integer tmp_pt = (Integer)pointers.get(tmp_calife1);
									String tmp = "\t\t(\"x=y.next\"," + line + "," + tmp_pt + "," + p2 + "," + s2 + "," + (line + 1) + ")\n";
									line++;
									if( next_line == line )
										next_line++;
									detectLoopPoint(line, next_line);
									tmp += "\t\t(\"x.next=y\"," + line + "," + p1 + "," + tmp_pt + "," + s1 + "," + next_line + "," + (desc_num++) + ")\n";
									return tmp;
								}
								if( args1[1].equals("IntegerSelectorVariables") && args2[1].equals("ArraySelectorVariables") ) {
									if( data_handled && ifData ) {
										throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));
									}
								}
								if( args1[1].equals("IntegerSelectorVariables") && args2[1].equals("IntegerSelectorVariables") ) {
									// Cas trs particulier : il faut distinguer suivant les diffrentes valeurs des champs data
									// ifdata("000") setdata("000")
									// ifdata("001") setdata("001")
									// ...
									if( data_handled && ifData ) {
										boolean loop = ( next_line != line + 1 );
										int next = next_line;
										next_line = line + 1;
										Integer p1 = (Integer)pointers.get( ( `t11.equals(args1[0]) ? `a1.getTextContent() : `a2.getTextContent() ) );
										Integer p2 = (Integer)pointers.get( ( `t21.equals(args2[0]) ? `b1.getTextContent() : `b2.getTextContent() ) );
										String tmp = "";
										int value;
										String code;
										for( int i = 0; i < (exp(2, data_size) - 1) ; i++ ) {
											code = toBinaryString(i, data_size);
											tmp += "\t\t(\"ifdata\"," + line + "," + p2 + ",\"" + code + "\"," + next_line + "," + (next_line + 1) + ")\n";
											tmp += "\t\t(\"setdata\"," + (line + 1) + "," + p1 + ",\"" + code + "\"," + (next_line + 1) + ")\n";
											line = line + 2;
											next_line = next_line + 2;
										}
										if( !loop )
											next = next_line + 1;
										code = toBinaryString((exp(2, data_size) - 1), data_size);
										tmp += "\t\t(\"ifdata\"," + line + "," + p2 + ",\"" + code + "\"," + next_line + "," + next + ")\n";
										line++;
										tmp += "\t\t(\"setdata\"," + line + "," + p1 + ",\"" + code + "\"," + next + ")\n";
										detectLoopPoint(line, next);
										return tmp;
									}
								}
							}
							if( `lab1.equals("IMPL") &&  `lab2.equals("GET") ) {
								if( args1[1].equals("PointerSelectorVariables") ) {
									Integer p = (Integer)pointers.get( ( `t11.equals(args1[0]) ? `a1.getTextContent() : `a2.getTextContent() ) );
									Integer s = (Integer)selectors.get( ( `t12.equals(args1[1]) ? `a2.getTextContent() : `a1.getTextContent() ) );
									if( !pointers.containsKey(tmp_calife1) )
										pointers.put(tmp_calife1, new Integer(pointer_id) );
									Integer tmp_pt = (Integer)pointers.get(tmp_calife1);
									String tmp = "\t\t(\"x=random\"," + line + "," + tmp_pt + "," + (line + 1) + ")\n";
									line++;
									if( next_line == line )
										next_line++;
									detectLoopPoint(line, next_line);
									tmp += "\t\t(\"x.next=y\"," + line + "," + p + "," + tmp_pt + "," + s + "," + next_line + "," + (desc_num++) + ")\n";
									getWarning("pm");
									return tmp;
								}
								if( args1[1].equals("IntegerSelectorVariables") ) {
									if( data_handled && ifData ) {
										throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));
									}
								}
							}
						}
						"lval", "expr" -> {
							if( `lab1.equals("IMPL") && ( `lab2.equals("PLUS") || `lab2.equals("MOINS") ) ) {
								if( args1[1].equals("PointerSelectorVariables") ) {
									Integer p = (Integer)pointers.get( ( `t11.equals(args1[0]) ? `a1.getTextContent() : `a2.getTextContent() ) );
									Integer s = (Integer)selectors.get( ( `t12.equals(args1[1]) ? `a2.getTextContent() : `a1.getTextContent() ) );
									if( !pointers.containsKey(tmp_calife1) )
										pointers.put(tmp_calife1, new Integer(pointer_id) );
									Integer tmp_pt = (Integer)pointers.get(tmp_calife1);
									String tmp = "\t\t(\"x=random\"," + line + "," + tmp_pt + "," + (line + 1) + ")\n";
									line++;
									if( next_line == line )
										next_line++;
									detectLoopPoint(line, next_line);
									tmp += "\t\t(\"x.next=y\"," + line + "," + p + "," + tmp_pt + "," + s + "," + next_line + "," + (desc_num++) + ")\n";
									getWarning("pm");
									return tmp;
								}
								if( args1[1].equals("IntegerSelectorVariables") ) {
									if( data_handled && ifData ) {
										throw new PointerManipulationException(Messages.getString("ARTMCBuilder.integer_selector_error"));
									}
								}
							}
						}
					}
				}
			}
			<_><Update><FunT Label=lab>a1@<_></_></FunT></Update></_> -> {
				if( `lab.equals("FREE") ) {
					Integer x = (Integer)pointers.get(`a1.getTextContent());
					detectLoopPoint(line, next_line);
					return "\t\t(\"free\"," + line + "," + x + "," + next_line + ")\n";
				}
			}

		}

		detectLoopPoint(line, next_line);
		return "\t\t(\"goto\"," + line + "," + next_line + ")\n";

	}

	/**
	 * Algorithme de parcours du graphe
	 * Gnre rcursivement les lignes du programme au fur et  mesure des transitions rencontres
	 * 
	 * @param String source - la place courante dans l'automate
	 * @return
	 * 
	 */
	private void parseTransition(String source) throws PointerManipulationException {

		%match(TNode automate) {
			<_><Transition Source=s Target=t>
					<Guard PrettyPrint=gpp></Guard>
					u@<Updates PrettyPrint=upp></Updates>
					</Transition></_> -> {
				if( `s.equals(source) && !places.contains(`s) ) {
					place_line.put(`s, new Integer(line));
					places.add(`s);
					if( !(`gpp.equals("True")) ) {
						ArrayList next_places = getNextPlaces(source);
						int nb_targ = next_places.size();
						for(int i = 0; i < nb_targ; i++) {
							Element[] guards = DOMUtil.getElementByTagName((Element)next_places.get(i), "Guard");
							String next_line_then = ((Element)next_places.get(i)).getAttribute("Target");
							int next_line_else = line + 1;
							if( place_line.containsKey(next_line_then) )
								detectLoopPoint(line, ((Integer)place_line.get(next_line_then)).intValue());
							try {
								String cond = getCond(guards[0]);
								program += "\t\t(" + cond + "," + next_line_then + "," + next_line_else + ")\n";
							} catch( DetectGuardException dge ) {
								program += "\t\t(" + dge.getCond() + "," + next_line_else + "," + next_line_then + ")\n";
							}
							line++;
						}
						program += "\t\t(\"exit\"," + line + ")\n";
						line++;
						for(int i = 0; i < nb_targ; i++) {
							String next = ((Element)next_places.get(i)).getAttribute("Target");
							parseTransition(next);
						}
					} else {
						int next_line = places.contains(`t) ? ((Integer)place_line.get(`t)).intValue() : (line + 1);
						try {
							program += getLine(`u, next_line);
						} catch( PointerManipulationException pme ) {
							throw pme;
						}
						line++;
						if( !places.contains(`t) )
							parseTransition(`t);
					}
					if( isFinal(`t) ) {
						program += "\t\t(\"exit\"," + line + ")\n";
						line++;
					}
				}
			}

		}

	}

//	END		GENERER LE PROGRAMME A PARTIR DE L'AUTOMATE


// REECRIRE LE PROGRAMME

	private void rewriteProgram() {

		String tmp, before, after, id, data;
		int index1 = 0, index2;

		//Split du programme ligne par ligne
		String[] prog = program.split("\n");

		//Ajout de "NOABSTR"  la fin des lignes, sauf les loop points
		for( int i = 0; i < prog.length; i++) {
			tmp = prog[i];
			index1 = tmp.indexOf(',');
			index2 = tmp.indexOf(',', index1 + 1);
			if( index2 == -1 )
				index2 = tmp.indexOf(')');
			int line = Integer.valueOf(tmp.substring(index1 + 1, index2)).intValue();
			if( !loop_point.contains(line) ) {
				int index = tmp.lastIndexOf(')');
				prog[i] = tmp.substring(0, index) + ",\"NOABSTR\")";
			}
		}

		//Rcriture des numro de lignes en binaire
		for( int i = 0; i < prog.length; i++) {
			tmp = prog[i];
			index1 = tmp.indexOf(',');
			index2 = tmp.indexOf(',', index1 + 1);
			if( index2 == -1 )
				index2 = tmp.indexOf(')');
			id = toBinaryString((Integer.valueOf(tmp.substring(index1 + 1, index2)) + 1), length_id);
			before = tmp.substring(0, index1 + 1);
			after = tmp.substring(index2);
			prog[i] = ( before + "\"" + id + "\"" + after);
		}

/*
		//Rcriture des donnes en binaire
		for( int i = 0; i < prog.length; i++) {
			tmp = prog[i];
			if( tmp.substring(0, 11).equals("\t\t(\"setdata") || tmp.substring(0, 10).equals("\t\t(\"ifdata") ) {
				int t = 0;
				for( int j = 0; j < 3; j++ ) {
					index1 = tmp.indexOf(",", t);
					t = index1 + 1;
				}
				index2 = tmp.indexOf(",", t);
				before = tmp.substring(0, index1 + 1);
				after = tmp.substring(index2);
				data = toBinaryString(Integer.valueOf(tmp.substring(index1 + 1, index2)), data_size);
				prog[i] = ( before + "\"" + data + "\"" + after );
			}
		}
*/
		//Rcriture des places
		for( int i = 0; i < prog.length; i++) {
			tmp = prog[i];
			if( present(tmp, new_place_name) ) {
				index1 = tmp.indexOf(new_place_name);
				index2 = tmp.indexOf(",", index1);
				if( index2 == -1 )
					index2 = tmp.indexOf(")");
				before = tmp.substring(0, index1);
				after = tmp.substring(index2);
				id = tmp.substring(index1, index2);
				prog[i] = ( before + place_line.get(id) + after );
			}
		}

		//Ajout de virgules  la fin des lignes
		for( int i = 0; i < prog.length; i++) {
			prog[i] += ",";
		}

		//Concatnation des lignes
		program = "";
		for( int i = 0; i < prog.length; i++) {
			program += prog[i] + "\n";
		}

		//Suppression de la dernire virgule
		program = program.substring(0, program.lastIndexOf(',')) + "]\n";

	}

// END	REECRIRE LE PROGRAMME

// GENERER DES COMMENTAIRES EN DEBUT DE FICHIER

	private String iterComment(HashMap variables, String comment) {

		String name, id;
		Set names = variables.keySet();
		Iterator e = names.iterator();
		while( e.hasNext() ) {
			name = (String)e.next();
			id = ((Integer)variables.get(name)).toString();
			comment += name + "=" + id + ", ";
		}
		int index = comment.lastIndexOf(',');
		if( index != -1 )
			comment = comment.substring(0, index) + "\n";
		else
			comment += "\n";

		return comment;

	}

	private String getDataComment() {

		TreeSet ordered_values = new TreeSet();
		Set values = data_values.keySet();
		Iterator e = values.iterator();
		while( e.hasNext() ) {
			ordered_values.add(e.next());
		}

		String comment = "# data values are : ";
		int value;
		String code;
		e = ordered_values.iterator();
		while( e.hasNext() ) {
			value = ((Integer)e.next()).intValue();
			code = (String) data_values.get(value);
			comment += value + "=" + code + ", ";
		}
		int index = comment.lastIndexOf(',');
		if( index != -1 )
			comment = comment.substring(0, index) + "\n";
		else
			comment += "\n";

		return comment;
	}

	private String getComment() {

		// Commentaire dcrivant les pointeurs utiliss
		String ptr_comment = iterComment(pointers, "# pointer variables are : ");

		// Commentaire dcrivant les next pointeurs utiliss
		String next_comment = iterComment(selectors, "# next pointers are : ");

		// Commentaire dcrivant les codes assigns aux champs data
		String data_comment = data_handled ? getDataComment() : "";

		return (ptr_comment + next_comment + data_comment);

	}

// END GENERER LES COMMENTAIRES

	// Attribue un nom  une place nouvellement cre qui n'existe pas encore
	private String getNewPlaceId(int nb) {

		Element[] places = DOMUtil.getElementByTagName(automate, "Place");
		boolean find;
		String name = Integer.toString(nb);
		do {
			find = false;
			name = new_place_name + name;
			for( int i = 0; i < places.length; i++ )
				if( places[i].getAttribute("Id") == name )
					find = true;
		}
		while( find );

		return name;

	}

	//Initialise le nom des pointeurs temporaires qui seront gnrs afin de ne pas entrer
	//en collision avec les noms des pointeurs qui existent dj dans l'automate
	private void initTmpPointer() {

		String name = "";
		boolean find;
		do {
			find = false;
			name = "tmp" + name;
			for( int i = 0; i < pointers.size(); i++ )
				if( pointers.containsKey((name + "1")) || pointers.containsKey((name + "2")) )
					find = true;
		} while( find );

		tmp_calife1 = name + "1";
		tmp_calife2 = name + "2";

	}

	public static void main(String[] args) {

		if( args.length < 3 ) {
			System.out.println("Syntax : java -jar transform.jar file array data\n" +
					"\tfile = the name of the xml file which contains the automata\n" +
					"\tarray = 1 if array assignements are not allowed, 0 if not\n" +
					"\tdata = 1 if data are handled, 0 if not\n");
			return;
		}
		Transform t = new Transform();
		t.setInput(args[0]);
		boolean a = ( Integer.parseInt(args[1]) == 1 ) ? true : false ;
		boolean d = ( Integer.parseInt(args[2]) == 1 ) ? true : false ;
		t.processExport("program.py", a, d);

	}

}
