/* Copyright (C) 2004  Bertrand Tavernier (CRIL Technology)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

package cril.calife.export.typechecker;

/**
 * @author bertrand.tavernier
 *
 * Version DOMTOM
 * Ajout de backtracking pour les fonctions polymorphes
 */

import org.w3c.dom.*;
import java.util.*;
import cril.calife.utils.DOMUtil;
import cril.calife.export.typechecker.TypeCaster;

public class TypeChecker {

	%include{dom.tom} //Mapping DOM -> Tom Term

	private Element n;
	private ArrayList others;
	private Enumeration backtrack;
	private TypeSystem ts;

	/**
	 * Cre un typechecker pour le document dom
	 * Ce document doit tre correctement typage aux vues du 
	 * systme de types intgr au noeud <Modle>
	 * @param dom Automate  typer
	 */
	
	 public TypeChecker(Element n, TypeSystem ts) {
		this.n=n; 
		this.ts=ts;
		this.CheckFun();
		this.TypeLeaf(n);
	 }

	/**
	 * Type toutes les contraintes rencontres dans l'AST XML
	 * @return Automate correctement typ
	 */
	public boolean TypeBound() {
		return TypeConstraint(n);
	}


	/**
	 * Type toutes les contraintes rencontres dans l'AST XML
	 * @return Automate correctement typ
	 */
	public boolean TypeCheck() {
		boolean result=CheckInitial();
		ArrayList allDecl = DOMUtil.getElementsByTagNames(n,
							new String[]{"Constraint","Update","Activite"});
		for (int it=0;it<allDecl.size();it++) {
			result &= TypeConstraint((Element)allDecl.get(it));
		}
		return result;
	}

	/**
	 * Vrifie la prsence d'un tat initial dans le composant
	 * @return prsence d'un tat initial
	 */
	public boolean CheckInitial() {
		ArrayList allLoc = DOMUtil.copyNodeList(n.getElementsByTagName("Place"));
		%match(TNodeList allLoc) {
			(_*,ElementNode[attrList=(_*,AttributeNode(_,"Initial","true"),_*)],_*) -> {return true;}
		}
		System.err.println("Erreur>> L'automate n'a pas d'tat initial !");
		return false;
	}

	/**
	* Type toutes les feuilles (de tag VAR) du document en fonction
	* des lements dclars dans le contexte
	* return 
	**/
	public void TypeLeaf(Element n) {
		ArrayList allVar = DOMUtil.copyNodeList(n.getElementsByTagName("VAR"));
		ArrayList allDecl = ts.getTerms();
		Document dom = n.getOwnerDocument();;
		for (int it=0;it<allVar.size();it++) {
			Node nVar = (Node)allVar.get(it);
	
			search : %match(TNode nVar,TNodeList allDecl) {
				<_>#TEXT(used)</_>,
				(_*,c @ <(Var|Param) Label=used></(Var|Param)>,_*) -> {
					String tag = `c.getNodeName().substring(0,1)
								+(`c.getParentNode().getNodeName().equals("Local")?"Loc":"Ext");
					Node nNew = `ElementNode(dom,tag
								,concTNode(AttributeNode(dom,"Type","true",((Element)c).getAttribute("Type")))
								,concTNode(TextNode(dom,used)));
					nVar.getParentNode().replaceChild(nNew,nVar);
					break search;
				}
				<_>#TEXT(used)</_>,
				(_*, c @ <ConstType Match=match></ConstType>,_*) -> {
					if (`used.matches(`match)) {
						Node nNew = `ElementNode(dom,"Const"
								,concTNode(AttributeNode(dom,"Type","true",TypeCaster.findSortOf(ts,used)))
								,concTNode(TextNode(dom,used)));
						nVar.getParentNode().replaceChild(nNew,nVar);
						break search;
					}
				}
				<_>#TEXT(used)</_>,_ -> {
					System.err.println("Erreur>> Symbole "+`used+" non dclar !");
				}
			}
		}
	}

	/**
	* Vrifie la bonne dclaration de toutes les fonctions utilises
	* et initialise les signatures en ("Undef")(":Undef")*
	* return
	**/
	private void CheckFun() {
		ArrayList allfun = DOMUtil.copyNodeList(n.getElementsByTagName("FunT"));
		ArrayList alldec = ts.getFunctions();		
		for (int it=0;it<allfun.size();it++) {
			Element efun = (Element)allfun.get(it);
			search : %match(TNode efun,TNodeList alldec) {
				<_ Label=l></_>,
				(_*,d @ <_ Label=l></_>,_*) -> {
					int i_arg = ((Element)`d).getAttribute("Arg").split(";")[0].split(":").length;
					String str_arg="";
					for (int kt=0;kt<i_arg;kt++) {
						str_arg+="Undef"; 
						if (kt!=(i_arg-1))
							str_arg+=":"; 
					}
					if (!((Element)efun).getAttribute("Label").equals("Localize"))
						((Element)efun).setAttribute("Arg",str_arg);	
					break search;
				}
				<_ Label=l></_>,_ -> {
					System.err.println("Erreur>> Fonction "+`l+" non dclare !");
				}
			}
		}
	}

	/**
	 * Type la contrainte reprsente par un AST de racine e_const
	 * Celle-ci doit auparavant tre initialise  l'aide des mthodes
	 * CheckFun() et TypeLeaf()
	 * @param e_const Racine de l'AST de contrainte
	 * @return Contrainte correctement type
	 */
	public boolean TypeConstraint(Element e_const) {
		boolean correctlyTyped = false;
		ArrayList leaf = DOMUtil.getElementsByTagNames(e_const,
								new String[]{"Const","VLoc","VExt","PLoc","PExt"});
		ArrayList backtrackList = new ArrayList();
		String constType = e_const.getNodeName();
		for (int it=0;it<leaf.size();it++) {
			others = new ArrayList(); // Sauvegarde de l'tat
			for (int jt=it+1;jt<leaf.size();jt++)
				others.add(leaf.get(jt));
			propagateType((Element)leaf.get(it),backtrackList);
		}
		
		if (TypeCaster.isCastAllowed(ts,e_const.getAttribute("Type"),constType)>=0)
			return true;
		//BackTracking
		for (int it=backtrackList.size()-1;it>=0;it--)
			if (((BackTrack)backtrackList.get(it)).backTrack(e_const))
				return true;
		System.err.println("Problme de Typage : ");
		DOMUtil.print(e_const,true);
		return false;
	}

	/**
	 * Propage le type de la feuille e_cur vers la racine.
	 * La liste backtracklist contient les retours en arrire
	 * possible pour typer correctement l'arbre (@see BackTrack) 
	 * @param e_cur Noeud  propager
	 * @param backtrackList Liste des backtracks possibles
	 */
	private void propagateType(Element e_cur,ArrayList backtrackList) {		
		Element e_parent = (Element)e_cur.getParentNode();
		%match(TNode e_cur,TNode e_parent) {
			<(Constraint|Update|Activite|Bound)></(Constraint|Update|Activite|Bound)>,_ -> {	return;	}
			<_ Type=cur_type></_>,
			<(Constraint|Update|Activite|Bound|OUT)></(Constraint|Update|Activite|Bound|OUT)> -> {
				e_parent.setAttribute("Arg",`cur_type); 
				e_parent.setAttribute("Type",`cur_type);
				propagateType(e_parent,backtrackList);
				return;
			}
			<_ Type="Undef"></_>,_ -> { 
				e_parent.setAttribute("Type","Undef"); 
				propagateType(e_parent,backtrackList);
				return;
			}
			<_ Type=cur_type></_>,<_ Arg=str_arg></_> -> { 
				int pos = DOMUtil.getNodePosition(e_cur);
				String[] args = `str_arg.split(":");
							
				if(pos<args.length)
					args[pos]=`cur_type;
				
				String newarg = ""; 
				for (int jt=0;jt<args.length;jt++)
					newarg += args[jt]+":"; 
	
				e_parent.setAttribute("Arg",newarg.substring(0,newarg.length()-1)); 
				e_parent.setAttribute("Type",TypeFun(e_parent,backtrackList));
				propagateType(e_parent,backtrackList);
				return;	
			}
		}
	}


	/**
	 * Tente de typer la fonction reprsente par efun
	 * Le typage ne peut tre ralis que lorsque la signature
	 * est complte. Cette mthode alimente la liste des backtrack
	 * lorsque d'une mme fonction peut conduire  plusieurs signatures.
	 * Remarque : Lorsque plusieurs signatures sont possibles pour 
	 * un mme type, le type choisi minimise le nombre de casts utiliss
	 * @param eFun Noeud associ  la fonction  typer
	 * @param backtrackList Liste des backtracks
	 * @return Premier Type correct rencontr
	 */
	private String TypeFun(Element eFun,ArrayList backtrackList) {
		String result="Undef";
		String str_f = eFun.getAttribute("Label"); 
		String str_arg = eFun.getAttribute("Arg"); 	

		Element e_fun = ts.getFunction(str_f);	
		if (e_fun==null) //Fonction non dclare !
			return "Undef"; 
			
		String[] args = e_fun.getAttribute("Arg").split(";"); 
		String[] types = e_fun.getAttribute("Type").split(";"); 
		int curcast=-1;
		Hashtable ht_possibleback = new Hashtable();
		for (int it=0;it<args.length;it++) {
			curcast = TypeCaster.getCastLength(ts,str_arg,args[it]);
			if (curcast>=0) { //La signature est trouve !
				BackTrack bk = (BackTrack)ht_possibleback.get(types[it]);
				if (bk==null) //Premire occurence de la fonction
					ht_possibleback.put(types[it], new BackTrack(eFun,types[it],curcast));
				else //La fonction est dj trouve, on minimise les casts !
				  if (curcast<bk.getCastLength())
						bk.setValue(types[it],curcast);
			}
		}
		Enumeration backs = ht_possibleback.elements();
		if (backs.hasMoreElements()) { //Enregistrement des backtracks
			result = ((BackTrack)backs.nextElement()).getType();
			while(backs.hasMoreElements())
				backtrackList.add(backs.nextElement());
		}
		return result;
	}

	/**
	 * Classe associe au backtracking sur les types de fonction
	 * @author bertrand.tavernier
	 * Cette classe sauvegarde l'tat et peut rcursivement revenir
	 *  des points de choix diffrents
	 */
	class BackTrack {
		private ArrayList state;
		private String str_type;
		private Element efun;
		private int castlength=-1;

		public void setValue(String str_type, int castlength){
			this.str_type=str_type;
			this.castlength = castlength;
		}

		public String getType() { return str_type; }
		public int getCastLength() { return castlength;	}

		/**
		 * Excute le backtracking associ  l'tat sauvegarder.
		 * Peut potentiellement excuter de nouveaux backtrackings
		 * attention  la complexit exponentielle sur le nombre
		 * de fonctions polymorphes
		 * @param e_const Contrainte  typer
		 * @return typage russi ?
		 */
		public boolean backTrack(Element e_const) {
			efun.setAttribute("Type",str_type);
			ArrayList backtrackList = new ArrayList();
			String constType = e_const.getNodeName();
			propagateType(efun,backtrackList);
			
			for (int it=0;it<state.size();it++) {
				others = new ArrayList();
				for (int jt=it+1;jt<state.size();jt++)
					others.add(state.get(jt));
				propagateType((Element)state.get(it),backtrackList);
			}
			//Ce backtrack a t'il rsolu le pb de typage ?
			if (TypeCaster.isCastAllowed(ts,e_const.getAttribute("Type"),constType)>=0)			
				return true;
			//Non ! -> BackTracking recursif
			for (int it=backtrackList.size()-1;it>=0;it--) {
				if (((BackTrack)backtrackList.get(it)).backTrack(e_const))
					return true;
			}
			return false;
		}
		
		public BackTrack(Element efun, String str_type,int castlength) {
			this.str_type=str_type;
			this.state = (ArrayList)others.clone();
			this.castlength = castlength;
			this.efun=efun;
		}		
	}
}
