/* Copyright (C) 2003  Bertrand Tavernier (CRIL Technology)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

package cril.calife.utils;

/**
 * @author bertrand.tavernier
 *
 */

import java.io.File;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;


import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
//import javax.xml.xpath.XPath;
//import javax.xml.xpath.XPathConstants;
//import javax.xml.xpath.XPathExpressionException;
//import javax.xml.xpath.XPathFactory;


import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


public class DOMUtil {
	
	private static DTDEntityResolver resolver = null;
	
	/**
	 * Supprime le noeud pass en paramtre de son document DOM
	 * @param Node n
	 */

	public static void deleteNode(Node n) {
		try {
			n.getParentNode().removeChild(n);
		} catch (NullPointerException npx) {
		}
	}

	/**
	 * Clone un document DOM
	 * @param Document DOM
	 * @return
	 */

	public static Document cloneDocument(Document DOM) {
		Document d_res = DOMUtil.newDocument();
		d_res.appendChild(d_res.importNode(DOM.getDocumentElement(), true));
		return d_res;
	}

	/**
	 * Retourne le premier fils du noeud npere nomm str_childname
	 * @param npere
	 * @param str_childname
	 * @return
	 */

	public static Element getChildNamed(Node npere, String str_childname) {
		NodeList allChild = npere.getChildNodes();
		for (int it = 0; it < allChild.getLength(); it++) {
			if (allChild.item(it).getNodeName().equals(str_childname))
				return (Element) allChild.item(it);
		}
		return null;
	}
	
	/**
	 * Retourne la liste des fils nomms str_childname
	 * @param npere Noeud pere
	 * @param str_childname Nom ddes fils  retourner
	 * @return
	 */
	
	public static Element[] getChildsNamed(Node npere, String str_childname) {
		NodeList allChild = npere.getChildNodes();
		ArrayList result = new ArrayList();
		for (int it = 0; it < allChild.getLength(); it++) {
			if (allChild.item(it).getNodeName().equals(str_childname))
				result.add(allChild.item(it));
		}
		Element[] res = new Element[result.size()];
		for (int it=0;it<result.size();it++)
		    res[it]=(Element) result.get(it);
		return res;
	}
	
	/**
	 * Cre un nouveau document sans racine
	 * @return
	 */

	public static Document newDocument() {
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document d_tmp = builder.newDocument();
			return d_tmp;
		} catch (Exception e) {}
		return null;
	}

	/**
	 * Cre un nouveau document DOM de racine "root"
	 * @param String root racine du document
	 * @return
	 */

	public static Document newDocument(String root) {
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document d_tmp = builder.newDocument();
			d_tmp.appendChild(d_tmp.createElement(root));
			return d_tmp;
		} catch (Exception e) {}
		return null;
	}

	public static void addResolver(String str_url, String str_path) {
		if (resolver==null) {
			resolver = new DTDEntityResolver();
		}
		resolver.addResolver(str_url,str_path);
	}
	
	/**
	 * Parse le fichier XML f 
	 * @param File f
	 * @return
	 */

	public static Document parseDocument(File f,boolean validate) {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		if (validate) {
			factory.setNamespaceAware(true);
			factory.setValidating(true);
			factory.setIgnoringElementContentWhitespace(true);
		}
		
		try {
			DocumentBuilder builder = factory.newDocumentBuilder();
			
			if (validate) {
				builder.setErrorHandler(
					new XMLErrorManager(f.getName()));
				addResolver(
						"http://calife.criltechnology.com/xml/component.dtd",
						"/cril/calife/utils/dtd/component.dtd");
				addResolver(
						"http://calife.criltechnology.com/xml/model.dtd",
						"/cril/calife/utils/dtd/model.dtd");
				addResolver(
						"http://calife.criltechnology.com/xml/trace.dtd",
						"/cril/calife/utils/dtd/trace.dtd");
				
				builder.setEntityResolver(resolver);
			}
			return builder.parse(f);
		} catch (Exception x) {x.printStackTrace();}
		return null;
	}

	/**
	 * Parse le Document XML filename
	 * @param String filename
	 * @return
	 */

	public static Document parseDocument(String filename,boolean validate) {
		return parseDocument(new File(filename),validate);
	}
	/*
	 * System.setProperty("javax.xml.transform.TransformerFactory",
                     "net.sf.joost.trax.TransformerFactoryImpl");

        try {

            TransformerFactory factory = TransformerFactory.newInstance();
            Transformer transformer =
                factory.newTransformer(new StreamSource(stxId));

            transformer.transform(
                new SAXSource(new SAXTest(xmlId), new InputSource()),
                new StreamResult(System.out));

        } catch (Exception e) {
            System.err.println(e.toString());
        }
	 */
	
	public static String applySTX(String xmlFile, String stxFile) {
	    String tprop = System.getProperty("javax.xml.transform.TransformerFactory");
	    
	    System.setProperty("javax.xml.transform.TransformerFactory",
        	"net.sf.joost.trax.TransformerFactoryImpl");

	    
	    System.setProperty("javax.xml.transform.TransformerFactory",tprop);
	    return null;
	}
	
	public static String applyXSL(Document DOM, InputStream xslFile) {
		try { 
			Transformer transform =
				TransformerFactory.newInstance().newTransformer(
					new StreamSource(xslFile));
			StringWriter res = new StringWriter();
			StreamResult result = new StreamResult(res);
			transform.transform(new DOMSource(DOM), result);
			return res.toString();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public static String applyXSL(Document DOM, String xslFile) {
		try { 
			Transformer transform =
				TransformerFactory.newInstance().newTransformer(
					new StreamSource(xslFile));
			StringWriter res = new StringWriter();
			StreamResult result = new StreamResult(res);
			transform.transform(new DOMSource(DOM), result);
			return res.toString();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Applique la transformation xsl xslFilename au Document DOM
	 * @param Document DOM
	 * @param String xslFilename
	 * @return booleen d'tat
	 */

	public static boolean applyXSL(
		Document DOM,
		String xslFile,
		String resultFile) {
		try { 
			File f_result = new File(resultFile);
			f_result.getParentFile().mkdirs(); //Cration du rpertoire de sortie
			/*System.setProperty("javax.xml.transform.TransformerFactory",
														 "jd.xml.xslt.trax.TransformerFactoryImpl");
			*/
			Transformer transform =
				TransformerFactory.newInstance().newTransformer(
					new StreamSource(xslFile));
			StreamResult result = new StreamResult(new File(resultFile));
			transform.transform(new DOMSource(DOM), result);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}

	/**
	 * Sauvegarde l'automate au format DOM pass en paramtre
	 * le nom du fichier est dduit de l'attribut Label du DOM.
	 * La dtd est celle d'un composant.
	 * @param Document DOM
	 * @param String str_path rpertoire de travail
	 * @param String str_ext extention  ajouter  @Label
	 * @return boolean bOK
	 */

	public static boolean saveDocument(
		Document DOM,
		String str_path,
		String str_ext) {
			return saveDocument(DOM,str_path
							+ NtoE(DOM.getDocumentElement()).getAttribute("Label") //$NON-NLS-1$
							+ str_ext);
	}
	
	/**
	 * Sauvegarde l'automate au format DOM pass en paramtre
	 * le nom du fichier pass en paramtre.
	 * La dtd est celle d'un composant.
	 * @param DOM
	 * @param nomEtChemin
	 * @return
	 */
	
	public static boolean saveDocument(
		Document DOM,
		String nomEtChemin) {
		return saveDocument(DOM
				,nomEtChemin
				,"http://calife.criltechnology.com/xml/component.dtd"
				,true);
	}

	/**
	 * Sauvegarde le document DOM vers le fichier nomEtChemin en lui
	 * appliquant la dtd stt_dtd.
	 * @param DOM Document source
	 * @param nomEtChemin Chemin vers le fichier xml de sauvegarde
	 * @param str_dtd Nom de la dtd  appliquer
	 * @param indent indentation du fichier de sortie
	 * @return tat de la procedure de sauvegarde
	 */
	
	public static boolean saveDocument(
			Document DOM,
			String nomEtChemin,
			String str_dtd,
			boolean indent) {
			try {
				if (DOM.getDocumentElement().getAttribute("Modif").equals("True"))
					DOMUtil.flagModif(DOM, false);
				Transformer transform = TransformerFactory.newInstance().newTransformer();

				transform.setOutputProperty(OutputKeys.ENCODING, "ISO-8859-1"); //$NON-NLS-1$
				if (indent) {
					transform.setOutputProperty(OutputKeys.INDENT, "yes"); //$NON-NLS-1$
					transform.setOutputProperty(
						"{http\u003a//xml.apache.org/xslt}indent-amount", //$NON-NLS-1$
						"4"); //$NON-NLS-1$
				}
				if (str_dtd!=null && !str_dtd.equals(""))
					transform.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,str_dtd); //$NON-NLS-1$

				StreamResult result =
					new StreamResult(
						new File(nomEtChemin));
				transform.transform(new DOMSource(DOM), result);
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
			return true;
		}

/*	public static Element[] getElementByXPath(Node n,String query) {
		try {
			XPath p = XPathFactory.newInstance().newXPath();
			NodeList nodes = (NodeList)p.evaluate(query
					,n,XPathConstants.NODESET);
			
			Element[] result = new Element[nodes.getLength()];
			for (int it=0;it<result.length;it++)
				result[it]= (Element)nodes.item(it);
			
			return result;
		} catch (XPathExpressionException e) {
			e.printStackTrace();
		}
		return null;
	}
	*/
	
	public static Element[] getElementByTagName(Document dom, String tag) {
		NodeList nodes = dom.getElementsByTagName(tag);

		Element[] result = new Element[nodes.getLength()];
		for (int it=0;it<result.length;it++)
			result[it]= (Element)nodes.item(it);
		
		return result;
	}

	public static Element[] getElementByTagName(Node n, String tag) {
		NodeList nodes = ((Element)n).getElementsByTagName(tag);

		Element[] result = new Element[nodes.getLength()];
		for (int it=0;it<result.length;it++)
			result[it]= (Element)nodes.item(it);
		
		return result;
	}
	
	/**
	 * Renvoie dans une Arraylist les noeuds de tag tags 
	 * @param dom document source
	 * @param tags liste des tags  rechercher
	 * @return liste des noeuds de tag tags
	 */
	
	public static ArrayList getElementsByTagNames(Document dom,String[] tags) {
		ArrayList res = new ArrayList();
		
		for (int it=0;it<tags.length;it++)
			res.addAll(copyNodeList(dom.getElementsByTagName(tags[it])));
		
		return res;
	}
	
	/**
	 * Retourne tous les noeuds dont les tags sont passs dans le tableau de String
	 * @param nSource Origine de la recherche
	 * @param tags Tableau des tags
	 * @return Collection des noeuds
	 */
	
	public static ArrayList getElementsByTagNames(Node nSource,String[] tags) {
		ArrayList res = new ArrayList();
		
		for (int it=0;it<tags.length;it++)
			res.addAll(copyNodeList(NtoE(nSource).getElementsByTagName(tags[it])));
		
		return res;
	}

	/**
	 * Retourne dans un ArrayList tous les fils de type ElementNode
	 * @param nSource
	 * @return
	 */
	
	public static ArrayList getElementsChilds(Node nSource) {
		ArrayList res = new ArrayList();
		NodeList childs = nSource.getChildNodes();
		
		for (int it=0;it<childs.getLength();it++)
			if (childs.item(it).getNodeType()==Node.ELEMENT_NODE)
				res.add(childs.item(it));
		
		return res;
	}
	
	/**
	 * Identique  Node.getElementsByTagName mais retourne un ArrayList
	 * @param nSource
	 * @param tag
	 * @return
	 */
	
	public static ArrayList getElementsByTagName(Node nSource,String tag) {	
		return copyNodeList(NtoE(nSource).getElementsByTagName(tag));
	}
	
	/**
	 * Recopie une NodeList dans un Vector
	 * cette mthode est  utiliser pour viter le bug du bucheron
	 * @param NodeList source
	 * @return
	 */

	public static ArrayList copyNodeList(NodeList source) {
		ArrayList result = new ArrayList();
		for (int it = 0; it < source.getLength(); it++)
			result.add(source.item(it));
		return result;
	}

	public static Element[] convertNodeList(NodeList source) {
	    Element[] res = new Element[source.getLength()];
	    
	    for (int it=0;it<source.getLength();it++)
	        res[it]=(Element)source.item(it);
	    
	    return res;
	}
	
	/**
	 * Cast d'un Node en Element
	 * @param Node n
	 * @return
	 */

	public static Element NtoE(Node n) {
		return (Element) n;
	}

	/**
	 * Supprime tous les noeuds fils du noeud n pass en paramtre
	 * @param Node n
	 */

	public static void removeAllChild(Node n) {
		ArrayList AllChild = copyNodeList(NtoE(n).getChildNodes());
		for (int it = 0; it < AllChild.size(); it++)
			n.removeChild((Node) AllChild.get(it));
	}

	/**
	 * Modifie l'attribut Modif du document DOM en "True" ou "False"
	 * @param Document DOM
	 * @param boolean bModif
	 */

	public static void flagModif(Document DOM, boolean bModif) {
		NtoE(DOM.getDocumentElement()).setAttribute(
			"Modif", //$NON-NLS-1$
			bModif ? "True" : "False"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Renvoie l'tat du document (Modifi ou non)
	 * @param Document DOM
	 * @return 
	 */

	public static boolean isDocumentModified(Document DOM) {
		return NtoE(DOM.getDocumentElement()).getAttribute("Modif").equals("True"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Renvoie la position du noeud pass en paramtre, par rapport  son pre
	 * @param Node nfils
	 * @return int position
	 */

	public static int getNodePosition(Node nfils) {
		int result = -1;
		NodeList allFils = nfils.getParentNode().getChildNodes();

		for (int it = 0; it < allFils.getLength(); it++) {
			if (allFils.item(it).getNodeType() == Node.ELEMENT_NODE)
				result++;
			if (allFils.item(it) == nfils)
				break;
		}

		return result;
	}
	
	/**
	 * Affiche le xml associ  un noeud n
	 * @param n noeud  afficher
	 * @param indent indentation on/off
	 */
	
	public static void print(Node n, boolean indent) {
	    
	    StringWriter wout = new StringWriter();  
	    PrintWriter out = new PrintWriter(wout,true); 
	    print(n,indent,0,out);
	    System.out.println(wout.toString());
	}
	
	/**
	 * Convertit en String le noeud n
	 * @param n noeud  convertir
	 * @param indent indentation on/off
	 * @return contenu textuel du xml associ  n
	 */
	
	public static String toString(Node n, boolean indent) {
	    StringWriter wout = new StringWriter();  
	    PrintWriter out = new PrintWriter(wout,true); 
	    print(n,indent,0,out);
	    return wout.toString();
	}
	
	private static void print(Node n,boolean indent, int depth,PrintWriter out) {
	    switch(n.getNodeType()) {
	    	case Node.ELEMENT_NODE : {
	    	    String str_label = n.getNodeName();
	    	    if (indent)
	    	        for (int it=0;it<depth;it++)
	    	            out.print("    ");
	    	    out.print("<"+str_label);
	    	    NamedNodeMap attr = n.getAttributes(); 
	    	    for (int it=0;it<attr.getLength();it++)
	    	        print(attr.item(it),indent,depth,out);
	    	    if (n.getChildNodes().getLength()==0) {
	    	        out.print("/>");
	    	        if (indent) out.println();
	    	    } else if(n.getChildNodes().getLength()==1 && n.getChildNodes().item(0).getNodeType()==Node.TEXT_NODE) {
	    	        out.print(">");
	    	        print(n.getChildNodes().item(0),false,depth+1,out);
	    	        out.print("</"+str_label+">");
	    	        if (indent) out.println();
	    	    } else {
		    	    out.print(">");
		    	    if (indent) out.println();
		    	    
		    	    NodeList nChilds = n.getChildNodes();
		    	    for (int it=0;it<nChilds.getLength();it++)
		    	        print(nChilds.item(it),indent,depth+1,out);	    	    
		    	    
		    	    if (indent)
		    	        for (int it=0;it<depth;it++)
		    	            out.print("    ");
		    	    out.print("</"+str_label+">");
		    	    if (indent) out.println();
	    	    } 
	    	    break;
	    	}
	    	
	    	case Node.TEXT_NODE : {
	    	    if (indent)
	    	        for (int it=0;it<depth;it++)
	    	            out.print("    ");
	    	    out.print(n.getNodeValue());
	    	    if (indent) out.println();
	    	    break;
	    	}
	    	case Node.ATTRIBUTE_NODE : {
	    	    out.print(" "+n.getNodeName()+"=\""+n.getNodeValue()+"\"");
	    	    break;
	    	}
	    	case Node.DOCUMENT_NODE : {
	    	    print(((Document)n).getDocumentElement(),indent,depth,out);
	    	    break;
	    	}
	    }
	}
	
}
