/* Copyright (C) 2003  Bertrand Tavernier (CRIL Technology)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

package cril.calife.utils;

import java.io.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import cril.calife.utils.parseur.config.ConfigParser;
import cril.calife.utils.parseur.xmlconfig.XMLConfigParser;

/**
 * @author bertrand.tavernier
 * 14 mai 2003
 */

public class FileUtil {

	/**
	 * Supprime rcursivement le rpertoire pass en paramtre
	 * renvoit un booleen d'tat sur l'opration
	 * @param String sdir
	 * @return
	 */

	public static boolean deleteDirectory(String sdir) {
		return deleteDirectory(new File(sdir));
	}

	/**
	 * Supprime rcursivement le rpertoire pass en paramtre
	 * renvoit un booleen d'tat sur l'opration
	 * @param File fdir
	 * @return
	 */

	public static boolean deleteDirectory(File fdir) {
		boolean result = true;
		File[] allFile = fdir.listFiles();
		if (allFile!=null)
			for (int it=0;it<allFile.length;it++) {
				if (allFile[it].isDirectory()) {
					result = result & deleteDirectory(allFile[it]);
				} else {
					result = result & allFile[it].delete();
				}
			}
		result = result & fdir.delete();
		return result;
	}

	/**
	 * Cre un Document AST ini vierge pour le fichier ini fSource
	 * @param fSource Fichier ini cible
	 * @return AST ini
	 */
	public static Document initIniAST(File fSource) {
	    Document dom = DOMUtil.newDocument("Config");
	    dom.getDocumentElement().setAttribute("Filename",fSource.getAbsolutePath());
	    return dom;
	}
	
	/**
	 * Construit l'arbre de syntaxe abstraite du fichier ini fSource
	 * @param fSource Fichier ini  la windows
	 * @return Document XML reprsentant l'AST
	 */

	public static Document buildIniAST(File fSource) {
	    try {
            XMLConfigParser xcf = new XMLConfigParser(fSource);
            return xcf.buildAST();
        } catch (Exception e) {
            return initIniAST(fSource);
        }
	}

	private static void incrementIniLine(Element eSect,int inc) {
	    int lastLine = -1;
	    Element[] allKey;
	    while (eSect.getNextSibling()!=null) {
            if (eSect.getNextSibling().getNodeType()==Node.ELEMENT_NODE) {
                
                eSect = (Element) eSect.getNextSibling();
	            lastLine = Integer.parseInt(eSect.getAttribute("Line"));
	            eSect.setAttribute("Line",Integer.toString(lastLine+inc));
	            
	            allKey = DOMUtil.convertNodeList(eSect.getChildNodes());
		        for (int jt=0;jt<allKey.length;jt++) {
		            lastLine = Integer.parseInt(allKey[jt].getAttribute("Line"));
		            allKey[jt].setAttribute("Line",Integer.toString(lastLine+inc));
		        }	            
            }
        }    
	}
	
	public static void removePrivateProfileSection(File fSource, String section) {
	    Document ast = buildIniAST(fSource);
	    Element[] allSect = DOMUtil.getChildsNamed(ast.getDocumentElement(),"Section");
	    int firstLine = -1, lastLine = -1;
	    Element eSect;
	    Element[] allKey;
	    for (int it=0;it<allSect.length;it++) {
	        if (allSect[it].getAttribute("Label").equals(section)) {
	            eSect = allSect[it];	            
	            if (it!=allSect.length-1) {
	                firstLine = Integer.parseInt(allSect[it].getAttribute("Line"));
	                lastLine = Integer.parseInt(allSect[it+1].getAttribute("Line"));
	                int diff = lastLine-firstLine;
	                incrementIniLine(eSect,-diff);
	                                    
	            }
	            allSect[it].getParentNode().removeChild(allSect[it]);	     
	            break;
	        }                
	    }
	    saveIniAST(ast);	    
	}
	
	/**
	 * Sauvegarde l'AST XML d'un fichier ini vers le fichier ini dcrit par
	 * l'attribut Filename
	 * @param dom Document XML correspondant  un AST de fichier ini
	 */
	
	public static void saveIniAST(Document dom) {
	    String str_file = dom.getDocumentElement().getAttribute("Filename");
	    String str_ini="";
	    int line = 0;
	    int tline;
	    Element[] sections = DOMUtil.convertNodeList(dom.getDocumentElement().getChildNodes());
	    Element[] keys;
	    String result = "";
	    for (int it=0;it<sections.length;it++) {
	        tline = Integer.parseInt(sections[it].getAttribute("Line"));
	        for (line++;line<tline;line++) {
	            result+=("\n");
	        }
	        
	        if (sections[it].getNodeName().equals("Comment")) {
	            result+=(sections[it].getAttribute("Value"))+"\n";
	        } else if (!sections[it].getAttribute("Label").equals("")) {
	            result+=("["+sections[it].getAttribute("Label")+"]\n");
	            keys = DOMUtil.convertNodeList(sections[it].getChildNodes());
	            
	            for (int jt=0;jt<keys.length;jt++) {
	    	        tline = Integer.parseInt(keys[jt].getAttribute("Line"));
	    	        for (line++;line<tline;line++) {
	    	            result+=("\n");
	    	        }
	    	        if (keys[jt].getNodeName().equals("Comment")) {
	    	            result+=(keys[jt].getAttribute("Value"))+"\n";
	    	        } else if (!(keys[jt].getAttribute("Key").equals("")||keys[jt].getAttribute("Value").equals(""))) {
	    	            result+=keys[jt].getAttribute("Key")+"="+keys[jt].getAttribute("Value")+"\n";
	    	        }
	    	    }	         
	        }
	    }
	    FileUtil.setContent(str_file,result);
	}
	
	/**
	 * Positionne la cl str_key de la section str_champ  la valeur str_value 
	 * dans le fichier ini fSource.
	 * Si le fichier n'existe pas, il est initialis.
	 * Si la section n'existe pas, elle est cre
	 * SI la cl n'existe pas, elle est cre.
	 * @param fSource Fichier ini
	 * @param str_champ Section
	 * @param str_key Cl
	 * @param str_value Valeur
	 */
	
	public static void setPrivateProfilString(File fSource,String str_champ,String str_key,String str_value) {
	    Document ast = buildIniAST(fSource);
	    Element[] allSect = DOMUtil.getChildsNamed(ast.getDocumentElement(),"Section");
	    Element[] allKey = null;
	    Element eSect = null, eChamp = null;
	    int lastLine = -1;
	    for (int it=0;it<allSect.length;it++) {
	        if (allSect[it].getAttribute("Label").equals(str_champ)) {
	            eSect = allSect[it];
	            break;
	        }
	        lastLine = Integer.parseInt(allSect[it].getAttribute("Line"));
	        allKey = DOMUtil.convertNodeList(allSect[it].getChildNodes());
	        for (int jt=0;jt<allKey.length;jt++)
	            lastLine = Integer.parseInt(allKey[jt].getAttribute("Line"));
	    }
	    
	    if (eSect == null) { // Creation d'une nouvelle section  la fin du document
	        eSect = ast.createElement("Section");
	        eSect.setAttribute("Label",str_champ);
	        eSect.setAttribute("Line",Integer.toString(lastLine+2));
	        ast.getDocumentElement().appendChild(eSect);
	    }
	    
	    allKey = DOMUtil.getChildsNamed(eSect,"Champ");
	    for (int it=0;it<allKey.length;it++)
	        if (allKey[it].getAttribute("Key").equals(str_key)) {
	            eChamp = allKey[it];
	            break;
	        }
	    
	    if (eChamp == null) { // Creation d'une nouvelle cl 
	        eChamp = ast.createElement("Champ");
	        eChamp.setAttribute("Key",str_key);
	        
	        // On recherche la dernire ligne de la section courrante
	        lastLine = Integer.parseInt(eSect.getAttribute("Line"));
	        allKey = DOMUtil.convertNodeList(eSect.getChildNodes());
	        for (int jt=0;jt<allKey.length;jt++)
	            lastLine = Integer.parseInt(allKey[jt].getAttribute("Line"));
	        eChamp.setAttribute("Line",Integer.toString(lastLine+1));
	        eSect.appendChild(eChamp);
	        // On incrmente la valeur de l'attribut Line pour tous les sections et champs suivantes 
	        incrementIniLine(eSect,1);
	    }

	    //On positionne la valeur sur la bonne cl    
	    eChamp.setAttribute("Value",str_value);
	    saveIniAST(ast);  
	}
	
	/**
	 * Lecture d'un paramtre d'un fichier ini windows
	 * Renvoit null lorsque la cl n'est pas trouve
	 * @param File fSource fichier ini
	 * @param String str_champ nom du champ
	 * @param String str_key nom de la cl
	 * @return String value
	 */
	
	public static String getPrivateProfilString(File fSource,String str_champ,String str_key) {
		try {
			ConfigParser p = new ConfigParser(fSource);
			return p.getPrivateProfilString(str_champ,str_key).trim();
		} catch (FileNotFoundException fx) {
		}
		return null;
	}

	/**
	 * Copie le fichier de nom str_Source vers str_Copy
	 * renvoit un booleen sur l'tat de la copie
	 * @param String str_Source nom du fichier source
	 * @param String str_Copy nom de la destination
	 * @return 
	 */

	public static boolean Copy(String str_Source, String str_Copy) {
		return FileUtil.Copy(new File(str_Source), new File(str_Copy));
	}
	
	/**
	 * Copie le fichier fSource sur fCopy
	 * @param File fSource
	 * @param File fCopy
	 * @return
	 */
	
	public static boolean Copy(File fSource, File fCopy) {
		try {
			int c = 0;
			FileInputStream in = new FileInputStream(fSource);
			FileOutputStream out = new FileOutputStream(fCopy);
			while ((c = in.read()) != -1)
				out.write(c);
			in.close();
			out.close();
		} catch (IOException iox) {
			iox.printStackTrace();
		}
		return true;
	}

	/**
	 * Renvoit le contenu du fichier nomm filename
	 * @param String filename
	 * @return
	 */

	public static String getContent(String filename) {
		return FileUtil.getContent(new File(filename));
	}

	/**
	 * Crer un fichier nomm filename et contenant content
	 * @param String filename
	 * @param String content
	 */
	
	public static void setContent(String filename,String content) {
			FileUtil.setContent(new File(filename),content);
	}
	
	/**
	 * Renvoit le contenu du fichier texte f
	 * @param File f
	 * @return
	 */
		
	public static String getContent(File f) {
		String result = "";
		try {
			BufferedReader finput =
				new BufferedReader(new InputStreamReader(new FileInputStream(f)));
			String ligne;
			int i = 0;
			while ((ligne = finput.readLine()) != null) {
				if (i > 0)
					result += "\n";
				result += ligne;
				i++;
			}
			finput.close();
		} catch (Exception e) { return null;}
		return result;
	}

	/**
	 * Cre le fichier f contentant le texte content
	 * @param File f
	 * @param String content
	 */

	public static void setContent(File f, String content) {
		try {
			if (f.getParentFile()!=null)
				if (!f.getParentFile().exists()) {
					f.getParentFile().mkdirs();
					System.out.println("Creation du rpertoire de sortie...");
				}
			PrintWriter pw = new PrintWriter(new FileOutputStream(f));
			pw.print(content);
			pw.flush();
			pw.close();
		} catch (Exception e) {e.printStackTrace();}
	}

}
