/*
 *  Client			: CEA/DAM                                                         
 *  Projet			: cril.calife.utils 
 * 					  Logiciel danalyse dInteractions Neutron - Silicium                                                
 *  Module			: ImagePrinter.java                
 *  Emetteur		:
 *      CRIL TECHNOLOGY - Agence IDF                       
 *      9-11, rue Jeanne Braconnier                                       
 *      Le Newton                                                         
 *      92360 Meudon-la-fort                                             
 *
 * Created on 4 mai 2004
 */

package cril.calife.utils;

/**
 * @author bertrand.tavernier
 *
 */

import java.awt.*;
import java.awt.image.*;
import java.awt.print.*;

import javax.print.PrintService;

/** This class prints an image without the need for a GUI component. 
  * It doesn't use MediaTracker to load images. 
  */
public class ImagePrinter implements Printable, ImageObserver {

   protected Image awtImage = null;
   protected boolean printFlag = true;
   protected int imageWidth, imageHeight;
   protected Point printLoc = new Point(0,0);
   protected boolean printdialog = false;
   protected PrintService pserv = null;
   
   public ImagePrinter(boolean dialog) {
   	this.printdialog = dialog;
   }
   
   public void setPrintLocation(Point d) {
       printLoc = d;
   }

   public Point getPrintLocation() {
      return printLoc;
   }

   public void loadAndPrint(String filename){
      awtImage = Toolkit.getDefaultToolkit().getImage(filename);
      imageWidth = awtImage.getWidth(this);
      imageHeight = awtImage.getHeight(this);
      if((imageWidth <0 )|| (imageHeight <0)) printFlag = false;
      else print();
   }

   public void Print(Image img) {
   	  awtImage = img;
   	  imageWidth = awtImage.getWidth(this);
   	  imageHeight = awtImage.getHeight(this);
   	  if((imageWidth <0 )|| (imageHeight <0)) printFlag = false;
   	  else print();
   }
   
   protected void print() {
      PrinterJob pj = PrinterJob.getPrinterJob();
      pj.setPrintable(this);
      if (printdialog && pserv==null) {
      	pj.printDialog();
      	pserv = pj.getPrintService();
      } else if (printdialog) {
      	try {
			pj.setPrintService(pserv);
		} catch (PrinterException e1) {
			e1.printStackTrace();
		}
      }
      try{
          pj.print();
      }catch(Exception e){System.out.println(e);}
   }


   public int print(Graphics g, PageFormat f, int pageIndex){
       if(pageIndex >= 1) return Printable.NO_SUCH_PAGE;
       Graphics2D g2d = (Graphics2D) g;
       g2d.translate(f.getImageableX(), f.getImageableY());
       double dw = awtImage.getWidth(this)+100;
       double dh = awtImage.getHeight(this)+100;

       if (f.getImageableWidth()<dw || f.getHeight()<dh) {
       		double dscale = Math.min(f.getImageableWidth()/dw,f.getImageableHeight()/dh);
       		g2d.scale(dscale,dscale);
       }
       
       if(awtImage != null){
          printImage(g2d, awtImage);
          return Printable.PAGE_EXISTS;
       } else  return Printable.NO_SUCH_PAGE;
    }

    public void printImage(Graphics2D g2d, Image image){
       if((image == null)|| (g2d == null)) return;
       int x = printLoc.x;
       int y = printLoc.y;
       int wid = imageWidth;
       int ht = imageHeight;
       g2d.drawImage(image,x,y,wid,ht, this);
    }

    public boolean imageUpdate(Image img,
                             int infoflags,
                             int x,
                             int y,
                             int width,
                             int height){

       if((infoflags & ImageObserver.ERROR) != 0){
          System.out.println("ERROR in image loading or drawing");
          return false;
       }

       if((infoflags & (ImageObserver.FRAMEBITS | ImageObserver.ALLBITS))!= 0) {
          if(printFlag == false) {
             imageWidth = width;
             imageHeight = height;
             print();
             printFlag = true;
          }
         return false;
       }
       return true;
   }
}
