/* Copyright (C) 2003  Bertrand Tavernier (CRIL Technology)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

package cril.calife.utils;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.AbstractButton;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.LookAndFeel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.plaf.FontUIResource;


/**
 * @author bertrand.tavernier
 * 5 sept. 2003
 */
public class SwingUtil {

	/**
	 * Renvoit le point permettant de centrer le composant en fonction 
	 * de sa taille et de la rsolution 
	 * @param Component jcomp
	 * @return Point
	 */
	
	public static Point getCentrerDim(Component jcomp) {
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Dimension frameSize = jcomp.getSize();
		
		return new Point(
		(screenSize.width - frameSize.width) / 2,
		(screenSize.height - frameSize.height) / 2);		
	}
	
	/**
	 * Crer une boite de dialogue avec JProgressBar indtermine
	 * @param parent Frame appelante
	 * @param str_title Titre de la boite de dialogue
	 * @param str_message Message affich
	 * @return
	 */
	
	public static JDialog createWaitWindow(Component parent, String str_title, String str_message) {
		JProgressBar jpg_bar = new JProgressBar(JProgressBar.HORIZONTAL);
		jpg_bar.setString(str_message);
		jpg_bar.setStringPainted(true);
		jpg_bar.setIndeterminate(true);
		JOptionPane waitpane =
		new JOptionPane(jpg_bar, JOptionPane.INFORMATION_MESSAGE);
		Object[] options = { "Annuler" }; 
		waitpane.setOptions(options);
		return waitpane.createDialog(parent, str_title);
	}

	/**
	 * Modifie la font par dfaut de tous les composants
	 * @param f Font 
	 */
	
	public static void setUIFont(Font f){
		FontUIResource fui = new FontUIResource(f);	
		java.util.Enumeration keys = UIManager.getDefaults().keys();
		while (keys.hasMoreElements()) {
		  Object key = keys.nextElement();
		  Object value = UIManager.get (key);
		  
		  if (value instanceof javax.swing.plaf.FontUIResource) {
		    UIManager.put (key, fui);
		  }
		  }
		}    
	
	/**
	 * Cre un JMenu contenant comme Items tous les LookAndFeel installs
	 * Chaque sous Item contient un ActionListener permettant d'appliquer
	 * le LookAndFeel. Le JMenu est cr avec l'icne ic
	 * @param src Fentre contenant le JMenu
	 * @param ic Icne du JMenu
	 * @return Menu "Look'n feels"
	 */
	
	public static JMenu buildLookAndFeelJMenu(final JFrame src, Icon ic) {
	    JMenu res = buildLookAndFeelJMenu(src);
	    res.setIcon(ic);
	    return res;
	}
	
	/**
	 * Recherche tous les jars contenant des Look'n Feel contenus dans le
	 * rpertoire fdir, les installe automatiquement et les ajoute au classpath !
	 * @param fdir Rpertoire contenant les Look'nFeels
	 */
	
	public static void autoInstallLookAndFeel(File fdir) {
	    ClassFinder cf = new ClassFinder(LookAndFeel.class);
		Object[] allc = cf.getClasses(fdir).toArray();
		String strc,strl;
		for (int it=0;it<allc.length;it++) {
		    strc = (String)allc[it];
		    strl = strc ;
		    if (strl.lastIndexOf(".")>=0)
		        strl = strl.substring(strl.lastIndexOf(".")+1);
		    if (strl.length()>12)
		        strl = strl.substring(0,strl.length()-11);
		    UIManager.installLookAndFeel(strl
		            ,strc);
		}	    
	}
	
	/**
	 * Cette mthode recherche automatiquement tous les look'n feels du
	 * classdpath et les installe.
	 *
	 */
	
	public static void autoInstallLookAndFeel() {
	    ClassFinder cf = new ClassFinder(LookAndFeel.class);
		Object[] allc = cf.getClasses().toArray();
		String strc,strl;
		for (int it=0;it<allc.length;it++) {
		    strc = (String)allc[it];
		    strl = strc ;
		    if (strl.lastIndexOf(".")>=0)
		        strl = strl.substring(strl.lastIndexOf(".")+1);
		    if (strl.length()>12)
		        strl = strl.substring(0,strl.length()-11);
		    
		    UIManager.installLookAndFeel(strl
		            ,strc);
		}
	}
	
	/**
	 * Cre un JMenu contenant comme Items tous les LookAndFeel installs
	 * Chaque sous Item contient un ActionListener permettant d'appliquer
	 * le LookAndFeel.
	 * @param src Fentre hebergeant le JMenu
	 * @return Menu "Look'n feels"
	 */
	
	public static JMenu buildLookAndFeelJMenu(final JFrame src) {
	    JMenu jm_lnf = new JMenu("Look'n feels"); //$NON-NLS-1$
	    
		UIManager.LookAndFeelInfo[] lnf =  UIManager.getInstalledLookAndFeels();
		ButtonGroup bg = new ButtonGroup();
		String lnfclass = UIManager.getLookAndFeel().getClass().getName();

		for (int it=0;it<lnf.length;it++) {
		    JRadioButtonMenuItem jmi_tmp = new JRadioButtonMenuItem(lnf[it].getName());

		    if (lnf[it].getClassName().equals(lnfclass))
		        jmi_tmp.setSelected(true);
		    jmi_tmp.setActionCommand(lnf[it].getClassName());
		    bg.add(jmi_tmp);
		    jmi_tmp.addActionListener(new ActionListener() {
		        	public void actionPerformed(ActionEvent e) {
		        	    try {
		                    UIManager.setLookAndFeel(((AbstractButton) e.getSource()
		                            ).getActionCommand());
		                    SwingUtilities.updateComponentTreeUI(src);
		                    if(src instanceof LookNFeelSaver)
		                        ((LookNFeelSaver)src).saveLookNFeel(
		                                ((AbstractButton) e.getSource()
		                                	).getText());
		                    src.validate();
		                } catch (Exception x) {
		                }
		        	}
		        });
		    jm_lnf.add(jmi_tmp);
		}
		return jm_lnf;
	}
}
